import Ext from "../../../tr-grid-util/es6/Ext.js";
import EventDispatcher from "../grid/event/EventDispatcher.js";

/** @description Trigger when data within the table has been changed
 * @event DataCache#dataChanged
 * @property {boolean} globalChange Indicates a big change. User should expect all data has been change
 * @property {string} type Type of changes. Possible values are "inserted", "removed", "updated", and undefined
 * @property {string} rid Row Id
 * @property {Object.<string, *>} rowData Column values of the changed row in JSON object format
 */

/** @description Trigger before DataCache#dataChanged. Perform any data update during the event will NOT cause more dataChanged events
 * @event DataCache#dataComposed
 */

/** Class for caching data and firing event when data is changed
 * @constructor
 * @extends {EventDispatcher}
 */
let DataCache = function () {
	this._rows = {};

	this._addEvent("dataComposed");
	this._addEvent("dataChanged");
};
Ext.inherits(DataCache, EventDispatcher);

/** @protected
 * @ignore
 * @type {!Object.<string, Object>}
 */
DataCache.prototype._rows;

/** @protected
 * @ignore
 * @type {boolean}
 */
DataCache.prototype._composing = false;

//#region Public
/**
 * @public
 */
DataCache.prototype.dispose = function () {
	this.unlistenAll();
	this.clearAllData(true);
};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.setSubscriptions = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 * @return {Object} Always returns null
 */
DataCache.prototype.getSubscriptions = function () { return null; };

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.addSubscription = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.removeSubscription = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.startAllSubscriptions = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.stopAllSubscriptions = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 * @return {Object} Always returns null
 */
DataCache.prototype.getSubscription = function () { return null; };

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 * @return {string} Always returns empty string
 */
DataCache.prototype.getPrimaryRic = function () { return ""; };

/** @public
 * @param {boolean=} opt_suppressEvent
 * @fires DataCache#dataChanged
 */
DataCache.prototype.clearAllData = function (opt_suppressEvent) {
	this._rows = {};

	if (!opt_suppressEvent) {
		this._dispatchDataChange({"globalChange": true});
	}
};

/**
 * @public
 * @param {Array.<string>|string} colId
 * @param {boolean=} opt_suppressEvent
 * @fires DataCache#dataChanged
 */
DataCache.prototype.clearColumnData = function (colId, opt_suppressEvent) {
	let cids = (typeof colId === "string") ? [/** @type {string} */(colId)] : /** @type {Array.<string>} */(colId);

	for (let i = cids.length; --i >= 0;) {
		let cid = cids[i];

		if (cid) {
			for (let rid in this._rows) {
				delete this._rows[rid][cid];
			}
		}
	}

	if (!opt_suppressEvent) {
		this._dispatchDataChange({"globalChange": true});
	}
};

/**
 * @public
 * @param {string} rid
 * @param {string} cid
 * @return {*}
 */
DataCache.prototype.getData = function (rid, cid) {
	let row = this.getRowData(rid);

	return (row) ? row[cid] : null;
};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 * @return {Object.<string, Array.<string>>} Always returns null
 */
DataCache.prototype.getAllRics = function () { return null; };

/**
 * @public
 * @return {Array.<string>}
 */
DataCache.prototype.getAllRowIds = function () {
	let rids = [];

	for (let rid in this._rows) {
		rids.push(rid);
	}

	return rids;
};

/**
 * @public
 * @param {string} rid ID of a row to be checked for its presence.
 * @return {boolean} `true` if row was found, `false` if row was not found.
 */
DataCache.prototype.hasRowId = function (rid) {
	return !!this._rows[rid];
};

/**
 * @public
 * @param {string} cid
 * @return {!Array.<*>}
 */
DataCache.prototype.getColumnData = function (cid) {
	let fieldRows = [];

	for (let rid in this._rows) {
		fieldRows.push(this._rows[rid][cid]);
	}

	return fieldRows;
};

/**
 * @public
 * @param {string} rid
 * @return {Object.<string, *>}
 */
DataCache.prototype.getRowData = function (rid) {
	return this._rows[rid] || null;
};

/**
 * @public
 * @param {Array.<string>} rids
 * @param {number=} opt_from
 * @param {number=} opt_to EXCLUSIVE
 * @return {!Array.<Object>}
 */
DataCache.prototype.getMultipleRowData = function (rids, opt_from, opt_to) {
	let len = rids.length;

	if (opt_to == null) {
		opt_to = len;
	}

	if (opt_to > len) {
		opt_to = len;
	}

	if (!(opt_from > 0)) {
		opt_from = 0;
	}

	let out_ary = new Array(len);
	let rowMap = this._rows;

	for (let i = opt_from; i < opt_to; ++i) {
		out_ary[i] = rowMap[rids[i]] || null;
	}

	return out_ary;
};

/** Return a new copy of row data
 * @public
 * @return {!Array.<Object.<string, *>>}
 */
DataCache.prototype.getAllRowData = function () {
	let rows = [];
	let rowMap = this._rows;

	for (let rid in rowMap) {
		rows.push(rowMap[rid]);
	}

	return rows;
};

/**
 * @public
 * @param {string} rid Row Id
 * @param {string} cid Column Id
 * @param {*} value
 * @return {boolean} Return true if there is any change, and false otherwise
 * @fires DataCache#dataChanged
 */
DataCache.prototype.setData = function (rid, cid, value) { // Data changed event is always dispatched
	let values = {};

	values[cid] = value;

	return this.setRowData(rid, values);
};

/**
 * @public
 * @ignore
 * @param {string} rid
 */
DataCache.prototype.hasDataCloudData = function (rid) {};

/**
 * @public
 * @param {string} rid Row Id
 * @param {Object.<string, *>=} values Object where the key is column Id and value is the data to be set
 * @param {Object=} opt_eventArg Additional arguments to be attached to the event fired
 * @return {boolean} Return true if there is any change, and false otherwise
 * @fires DataCache#dataChanged
 */
DataCache.prototype.setRowData = function (rid, values, opt_eventArg) { // Data changed event is always dispatched
	let row = this.getRowData(rid);
	let dirty = false;
	let changes = null;

	if (values) {
		if (!row) {
			row = values; // WARNING: Dangerous use of external reference. User may modify data without using DataCache's API

			this._rows[rid] = row;
			dirty = true;

		} else {
			for (let cid in values) {
				row[cid] = values[cid];
				dirty = true;
			}

			changes = values;
		}
	} else {
		if (row) {
			delete this._rows[rid];
			dirty = true;
		}
	}

	if (dirty && !this._composing) {
		let e = opt_eventArg || {};

		e["rid"] = rid;

		if (changes) {
			e["changes"] = changes;
		}

		e["rowData"] = row;

		this._dispatchDataChange(e);
	}

	return dirty;
};

/**
 * @public
 * @param {string} rid
 * @return {Object}
 */
DataCache.prototype.cloneRowData = function (rid) {
	let fields = this._rows[rid];

	if (!fields) {
		return null;
	}

	let values = {};

	for (let key in fields) {
		if (key.indexOf("FREE_TEXT") === 0) { // TODO: This is not generic and will cause unintentional error. Also, it posts some performance penalty
			continue; // Skip free text column
		}

		values[key] = fields[key];
	}

	return values;
};

/**
 * @protected
 * @ignore
 * @param {string} fromRid
 * @param {string} toRid
 * @return {boolean}
 */
DataCache.prototype._replaceRowId = function(fromRid, toRid) {
	let rows = this._rows;
	if(rows[fromRid]) {
		rows[toRid] = rows[fromRid];
		delete rows[fromRid];
		return true;
	}
	return false;
};
/**
 * @public
 * @ignore
 * @param {Object} ridPair
 */
DataCache.prototype.replaceRowIds = function(ridPair) {
	if(typeof ridPair === "object") {
		for(let oldRid in ridPair) {
			this._replaceRowId(oldRid, ridPair[oldRid]);
		}
	}
};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.addStaticFields = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.removeStaticFields = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.resetStaticFields = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.getStaticFields = function () {};

/**
 * Construct a JSON object that is compatible with console.table()
 * @public
 * @param {Object=} options
 * @return {!Array.<Object>} Return a JSON object that is compatible with console.table()
 * @example
 * let options = {
 *   "colNames": Array.<string>,
 *   "rowCount": number,
 *   "startIndex": number,
 *   "getter": function(Object, string):*,
 *   "withRowIds": boolean
 * };
 */
DataCache.prototype.dump = function (options) {
	return DataCache.constructTable(this.getAllRowData(), options, Object.keys(this._rows));
};

/**
 * Print current data and structure to console panel
 * @public
 * @param {Object=} opt_options
 * @example
 * let options = {
 *   "colNames": Array.<string>,
 *   "rowCount": number,
 *   "startIndex": number,
 *   "getter": function(Object, string):*,
 *   "withRowIds": boolean
 * };
 */
DataCache.prototype.log = function (opt_options) {
	console.table(this.dump(opt_options));
};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.setDataCloudSettings = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.getDataCloudFields = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.addDataCloudFields = function () {};

/** Deprecated. Built-in Data Service is deprecated due to the lack of flexibility.
 * @public
 * @ignore
 */
DataCache.prototype.removeDataCloudField = function () {};
//#endregion Public

/**
 * @protected
 * @ignore
 * @param {!Object} e
 * @fires DataTable#dataComposed
 * @fires DataTable#dataChanged
 */
DataCache.prototype._dispatchDataChange = function (e) {
	if (this._composing) {
		return;
	}

	this._composing = true;
	this._dispatch("dataComposed", e);
	this._composing = false;
	this._dispatch("dataChanged", e);
};

/**
 * @private
 * @param {string} propertyName
 * @param {Object} rowData
 * @param {string} colName
 * @return {*}
 */
DataCache._defaultPropertyGetter = function (propertyName, rowData, colName) {
	let d = rowData[colName];

	return d ? d[propertyName] : null;
};

/**
 * @private
 * @param {Object} rowData
 * @param {string} colName
 * @return {*}
 */
DataCache._defaultGetter = function (rowData, colName) {
	return rowData[colName];
};


/** Construct a JSON object that is compatible with console.table()
 * @public
 * @ignore
 * @param {!Array.<Object>} dataset
 * @param {Object=} opt_options { "rowCount": number,  "colNames": Array.<string>,  "getter": function(Object, string):*, startIndex: number, "withRowIds": boolean }
 * @param {Array.<string>=} opt_rowIds Optional row ids corresponding to the given dataset
 * @return {!Array.<Object>} Return a JSON object that is compatible with console.table()
 */
DataCache.constructTable = function (dataset, opt_options, opt_rowIds) {
	let startIndex = 0;
	let rLen = dataset.length;
	let rowLim = rLen;
	let colMap = null;
	let rids = opt_rowIds || null;
	let getter = DataCache._defaultGetter;
	let transformation = false;

	if (opt_options) {
		if (opt_options["rowCount"]) {
			rowLim = opt_options["rowCount"];
			transformation = true;
		}

		if (opt_options["colNames"]) {
			colMap = {};

			let len = opt_options["colNames"].length;

			for (let i = 0; i < len; ++i) {
				colMap[opt_options["colNames"][i]] = true;
			}

			transformation = true;
		}

		if (opt_options["getter"]) {
			if (typeof opt_options["getter"] == "string") {
				getter = DataCache._defaultPropertyGetter.bind(null, opt_options["getter"]);
			} else { // Function
				getter = opt_options["getter"];
			}

			transformation = true;
		}

		if (opt_options["startIndex"]) {
			startIndex = /** @type{number} */(opt_options["startIndex"]);
			transformation = true;
		}

		if (opt_options["withRowIds"] === false) {
			rids = null;
		}
	}

	if (!transformation && !rids) {
		return dataset; // The dataset is already compatible with console.table
	}

	let rows = [];

	for (let r = startIndex; r < rLen; ++r) {
		if (rows.length >= rowLim) {
			break;
		}

		let rowData = dataset[r];
		let columns = (rids) ? {"rid": rids[r]} : {};

		if (rowData) { // Header and footer don't have any row data
			let colName;

			if (colMap) {
				for (colName in colMap) {
					if (rowData[colName] != null) {
						columns[colName] = getter(rowData, colName);
					}
				}
			} else {
				for (colName in rowData) {
					columns[colName] = getter(rowData, colName);
				}
			}
		}

		rows.push(columns);
	}

	return rows;
};


export default DataCache;
export { DataCache };
