import Ext from "../../../tr-grid-util/es6/Ext.js";
import GroupDefinitions from "../../../tr-grid-util/es6/GroupDefinitions.js"; // eslint-disable-line
import { isEmptyObject } from "../../../tr-grid-util/es6/Util.js";
import ElementWrapper from "./components/ElementWrapper.js";
import ILayoutGrid from "./ILayoutGrid.js"; // eslint-disable-line
import LayoutGrid from "./LayoutGrid.js";
import DataTable from "../data/DataTable.js";
import DataView from "../data/DataView.js";
import Util from "./util/util.js";
import SectionSettings from "./util/SectionSettings.js";
import Virtualizer from "./util/Virtualizer.js";
import Conflator from "./util/Conflator.js";
import TrackLayout from "./util/TrackLayout.js";
import HScrollbar from "./components/HScrollbar.js";
import VScrollbar from "./components/VScrollbar.js";
import VirtualizedLayoutGrid from "./VirtualizedLayoutGrid.js";

//#region Events
/** @event Core#sectionAdded */
/** @event Core#columnAdded */
/** @event Core#columnRemoved */
/** @event Core#columnVisibilityChanged */
/** @event Core#columnAlignmentChanged */
/** @event Core#widthChanged */
/** @event Core#dataSourceChanged */
/** @event Core#rowExpansionBinding
* @description Trigger after column data binding. Fires for all rows within the view.
* @property {Object.<string, *>} originalRowData
* @property {string} originalRowId The row that rowExpansion belongs to
* @property {boolean} rowExpansion
* @property {string} rowId
* @property {number} rowIndex
*/
/** @event Core#beforeContentBinding */
/** @event Core#postSectionDataBinding */
/** @event Core#firstRendered */
/** @event Core#afterContentBinding */
/** @event Core#rowHighlighted */

/** @event Core#rowAdded
* @ignore
*/
/** @event Core#postDataSourceChanged
* @ignore
*/
/** @event Core#preSectionRender
* @ignore
*/
/** @event Core#preSectionDataBinding
* @ignore
*/
/** @event Core#preForcedUpdate
* @ignore
*/
/** @event Core#tabNavigation
* @ignore
*/
//#endregion Events

/** @private
* @param {number} a
* @param {number} b
* @return {number}
*/
let ascNumberSorter = function (a, b) {
	return a - b;
};
/** @private
* @return {!Element}
*/
let _createHiddenInput = function () {
	let hiddenInput = document.createElement("input");
	let styleObj = hiddenInput.style;
	styleObj.position = "absolute";
	styleObj.width = styleObj.height = styleObj.padding = styleObj.border = "0";
	hiddenInput.value = "0";
	hiddenInput.role = "grid";
	hiddenInput.readOnly = true;
	return hiddenInput;
};
/** @private
* @param {Object} e
* @return {boolean}
*/
let _isTabCommand = function (e) {
	if(e.keyCode === 9) {
		return !e.ctrlKey && !e.altKey && !e.metaKey;
	}
	return false;
};
/** @private
* @param {Element} elem
* @return {Element}
*/
let _getActiveElement = function (elem) {
	if(elem) {
		if(elem.getRootNode) {
			let rootNode = elem.getRootNode(); // Get uncomposed root node
			if(rootNode && rootNode !== elem) { // The root node could be the element itself, if it is not attached to the DOM tree
				return rootNode.activeElement || null;
			}
		} else { // Older browser does not support getRootNode
			return document.activeElement;
		}
	}
	return null;
};

/** @constructor
* @param {Element=} opt_initializer this can be either element id (string) or DOM element.
* @extends {ElementWrapper}
*/
let Core = function (opt_initializer) {
	if (!Core._sectionType) { // Initialize static variables
		Core._sectionType = {};
		Core._sectionType["title"] = true;
		Core._sectionType["header"] = true;
		Core._sectionType["content"] = true;
		Core._sectionType["footer"] = true;
	}

	let _t = this;

	// Initialize method binding
	_t.getColumnId = _t.getColumnId.bind(_t);
	_t.getColumnIndex = _t.getColumnIndex.bind(_t);

	_t._onMouseMove = _t._onMouseMove.bind(_t);
	_t._onRowHightlighted = _t._onRowHightlighted.bind(_t);
	_t._onGridClicked = _t._onGridClicked.bind(_t);
	_t._onKeyDown = _t._onKeyDown.bind(_t);

	_t._onWindowResize = _t._onWindowResize.bind(_t);
	_t._onSectionDataChanged = _t._onSectionDataChanged.bind(_t);
	_t._onRowCountChanged = _t._onRowCountChanged.bind(_t);
	_t._onRowHeightChanged = _t._onRowHeightChanged.bind(_t);
	_t._onRowAvailable = _t._onRowAvailable.bind(_t);

	_t._onVScroll = _t._onVScroll.bind(_t);
	_t._onHScroll = _t._onHScroll.bind(_t);
	_t._onSyncVScroll = _t._onSyncVScroll.bind(_t);
	_t._onSyncHScroll = _t._onSyncHScroll.bind(_t);
	_t.updateLayout = _t.updateLayout.bind(_t);
	_t._onRowRefresh = _t._onRowRefresh.bind(_t);
	_t._onVScrollEnabled = _t._onVScrollEnabled.bind(_t);
	_t._onVScrollDisabled = _t._onVScrollDisabled.bind(_t);

	_t._onRowInViewChanged =  _t._onRowInViewChanged.bind(_t);
	_t._onColInViewChanged =  _t._onColInViewChanged.bind(_t);

	_t._updateVScrollbar = _t._updateVScrollbar.bind(_t);
	_t.updateColumnBounds = _t.updateColumnBounds.bind(_t);
	_t._dispatchColumnPositionChanged = _t._dispatchColumnPositionChanged.bind(_t);
	_t._dispatchRowPositionChanged = _t._dispatchRowPositionChanged.bind(_t);
	_t._requestScrollbarUpdate = _t._requestScrollbarUpdate.bind(_t);

	// Text nodes are unintentionally getting in the tag.
	if(opt_initializer) { // Any node other than element node is not allowed within the tag.
		for(let i = opt_initializer.childNodes.length; --i >= 0;) {
			if(!Util.isElement(opt_initializer.childNodes[i])) {
				opt_initializer.removeChild(opt_initializer.childNodes[i]);
			}
		}
	}

	// Initialize members
	_t._element = opt_initializer ? opt_initializer : document.createElement("div");
	_t._element.classList.add("tr-grid");

	_t._layoutX = new TrackLayout();
	_t._layoutY = new TrackLayout();
	_t._layoutX.setDefaultLaneSize(100); //Default column width
	_t._layoutY.setDefaultLaneSize(32); //Default row height

	_t._settings = [];
	_t._plugins = {};
	_t._columnDefs = [];
	_t._sectionStarts = [];
	_t._rowHeightException = "title";

	_t._rowVirtualizer = new Virtualizer(_t._layoutY);
	_t._rowVirtualizer.deactivate();
	_t._rowVirtualizer.setViewSize(360); // Minimum view size

	_t._colVirtualizer = new Virtualizer(_t._layoutX);
	_t._colVirtualizer.enable(false);
	_t._colVirtualizer.deactivate();
	_t._colVirtualizer.setViewSize(400); // Minimum view size

	// Initialize vertical scrollbar
	_t._vscrollbar = new VScrollbar();
	_t._vscrollbar.disable();
	_t._vscrollbar.setParent(_t.getParent() || _t.getElement());

	_t._vscrollbar.listen("scroll", _t._onVScroll);
	_t._vscrollbar.listen("layoutChanged", _t._onVScroll);
	_t._vscrollbar.listen("activated", _t.updateLayout); // TODO: remove column virtualization logic
	_t._vscrollbar.listen("deactivated", _t.updateLayout); // TODO: remove column virtualization logic
	_t._vscrollbar.listen("thicknessChanged", _t.updateLayout); // TODO: remove column virtualization logic
	_t._vscrollbar.listen("enabled", _t._onVScrollEnabled);
	_t._vscrollbar.listen("disabled", _t._onVScrollDisabled);

	// Initialize horizontal scrollbars
	_t._hscrollbar = new HScrollbar();
	_t._hscrollbar.disable();
	_t._hscrollbar.setParent(_t.getParent() || _t.getElement());

	_t._hscrollbar.listen("scroll", _t._onHScroll);
	_t._hscrollbar.listen("layoutChanged", _t._onHScroll);
	_t._hscrollbar.listen("activated", _t.updateLayout);
	_t._hscrollbar.listen("deactivated", _t.updateLayout);

	// cross-reference scrollbars
	_t._hscrollbar.setOtherScrollbar(_t._vscrollbar);
	_t._vscrollbar.setOtherScrollbar(_t._hscrollbar);


	_t._element.addEventListener("keydown", _t._onKeyDown);
	if (Util.isMobile || Util.isTouchDevice) {
		_t._element.addEventListener("touchmove", _t._onMouseMove, false);
	} else {
		_t._element.addEventListener("mousemove", _t._onMouseMove, false);
	}

	if(Util.isSafari){
		_t._element.addEventListener("click", _t._onGridClicked);
	}

	window.addEventListener("resize", _t._onWindowResize, false); // Should be unlistened after destroyed
	_t._rowVirtualizer.listen("indexChanged", _t._onRowInViewChanged);
	_t._colVirtualizer.listen("indexChanged", _t._onColInViewChanged);
	_t._rowHeightConflator = new Conflator(_t._onRowHeightChanged, 50);
	_t._vScrollbarConflator = new Conflator(_t._updateVScrollbar, 200);
	_t._columnBoundConflator = new Conflator(_t.updateColumnBounds, 10);
	_t._columnPositionConflator = new Conflator(_t._dispatchColumnPositionChanged, 10);
	_t._rowPositionConflator = new Conflator(_t._dispatchRowPositionChanged, 10);

	_t._firstHiddenInput = _createHiddenInput();
	_t._firstHiddenInput.className = "first-input";
	_t._lastHiddenInput = _createHiddenInput();
	_t._lastHiddenInput.className = "last-input";
	_t._element.insertBefore(_t._firstHiddenInput, _t._element.firstChild);
	_t._element.appendChild(_t._lastHiddenInput);

	// Initialize events for external users
	_t._addEvents(
		"sectionAdded",
		"columnAdded",
		"columnRemoved",
		"columnMoved",
		"columnVisibilityChanged",
		"columnAlignmentChanged",
		"widthChanged",
		"dataSourceChanged",
		"postDataSourceChanged",
		"preSectionRender",
		"postSectionRender",
		"beforeContentBinding",
		"preSectionDataBinding",
		"postSectionDataBinding",
		"firstRendered",
		"afterContentBinding",
		"rowExpansionBinding",
		"rowHighlighted",
		"preForcedUpdate",
		"rowAdded",
		"rowRemoved",
		"columnPositionChanged",
		"rowPositionChanged",
		"beforeColumnBoundUpdate",
		"beforeBatchOperation",
		"afterBatchOperation",
		"pinningChanged",
		"tabNavigation"
	);

	// For debugging in advanced optimization mode
	let map = Core["map"];
	if(!map) {
		map = {};
		Core["map"] = map;
	}

	_t._element["_control"] = _t;
	let id = _t._element.id || _t._element.name;
	if(!id || map[id]) {
		id = "_grid" + Core._runningGridId;
	}
	_t["_id"] = id;
	map[id] = _t;
	Core._runningGridId++;

	// Ensure all affected plugins are loaded prior zoom plugin
	// use as entity to trigger updateLayout once zoom is changed
	Object.defineProperty(_t, "zoomFactor", {
		set: function(factor) {
			if(!factor || factor < 0) {
				return;  // Cannot have negative factor or factor of zero
			}
			if(this._zoomFactor !== factor) {
				this._zoomFactor = factor;
				// Panes and view port are affected and need to be updated
				_t._updateLayout(); // Super slow. Do not execute this so often
			}
		},
		get: function() {
			return this._zoomFactor;
		}
	});
};
Ext.inherits(Core, ElementWrapper);

/** @typedef {number|string|ILayoutGrid|SectionSettings} Core~SectionReference
* @description A section in core grid can be refered by the following object <br>
* `{@link ILayoutGrid}` : A pointer to Section instance <br>
* `{@link SectionSettings}` : A pointer to SectionSettings instance <br>
* `number` : Section index <br>
* `string` : Section type (e.g. "content", "title", "header", "footer"), or Section id (element id)
*/
Core.SectionReference;

/** @typedef {Object} Core~MouseInfo
* @description Object returned from Core#getRelativePosition method
* @property {boolean} hit - Indicates whether the given position is on the grid
* @property {number} x - Horizontal coordinate that is relative to the top-left of the grid (Zero means left-most)
* @property {number} y - Vertical coordinate that is relative to the top-left of the grid (Zero means top-most)
* @property {number} colIndex
* @property {number} sectionIndex
* @property {number} rowIndex
* @property {string} sectionType
* @property {SectionSettings} sectionSettings
* @property {ILayoutGrid} section
* @property {Element} cellElement
* @property {DataView} dataSource
*/
Core.MouseInfo;

/** @typedef {Object} Core~BatchInfo
* @private
* @property {string=} reset //set columns
* @property {string=} insertion //add cols
* @property {string=} removal //remove cols
* @property {string=} moving //reorder
*/
Core.BatchInfo;

/** @typedef {Core.MouseInfo|ElementWrapper|Element} Core~CellReference
* @description A section in core grid can be refered by the following object <br>
* `{Core.MouseInfo}` : Object with valid x, y coordinates and section index <br>
* `{@link ElementWrapper}` : a pointer to ElementWrapper instance <br>
* `Element` : cell element
*/
Core.CellReference;

/** @typedef {Object} Core~ColumnOptions
* @description Column options that can be provided to {@link Core#insertColumn}
* @property {number=} width Column width
* @property {number=} minWidth Minimum column width
* @property {number=} minSize Alias to `minWidth`
* @property {number=} size Alias to `width`
* @property {boolean=} visible Column visibility
* @property {boolean=} scalable Column scalability
* @property {Object.<string, string>=} styles Key/value pair object specifying CSS styles
* @property {Object.<string, string>=} classes Key/value pair object specifying CSS classes
* @property {string=} dataColumnName Data column for sorting
* @property {Function=} renderingHandler - Deprecated
* @property {Function=} dataBindingHandler Column binding function
* @property {string=} columnData For attaching context (user data)
* @property {boolean=} stationary=false If enabled, the column order cannot be changed (i.e., this column and any column to its left cannot be moved)
* @property {boolean=} leftPinned=false If enabled, the column will not be part of the scrollable area and is pinned to the left side
* @property {boolean=} rightPinned=false If enabled, the column will not be part of the scrollable area and is pinned to the right side
* @property {TrackLayout=} layout - Deprecated
*/
Core.ColumnOptions;

/** For runtime debugging!
* @type {number}
* @private
*/
Core._runningGridId = 0; // Static variable
/** For identifying column
* @type {number}
* @private
*/
Core._runningColumnId = 0; // Static variable

/** @type {!Object.<string, boolean>}
* @private
* @const
*/
Core._sectionType; // Static variable

/** @type {!Object.<string, *>}
* @private
*/
Core.prototype._plugins;
/** @type {DataView}
* @private
*/
Core.prototype._dataSource = null;

/** @type {!TrackLayout}
* @private
*/
Core.prototype._layoutX;
/** @type {!TrackLayout}
* @private
*/
Core.prototype._layoutY;
/** @type {!VScrollbar}
* @private
*/
Core.prototype._vscrollbar; // vertical scrollbar is non-nullable
/** @type {!HScrollbar}
* @private
*/
Core.prototype._hscrollbar; // horizontal scrollbar is non-nullable


/** reference of top frozen section. It's a temporaly use between function _freezeTopSectionAt and _onSectionCountChanged
* @type {string}
* @private
*/
Core.prototype._tempFrozenSectionRef = "";
/** first index of section that vertical scroll placed, negative value mean no vertical scroll
* @type {number}
* @private
*/
Core.prototype._startVScrollbarIndex = -1;
/** @type {boolean}
* @private
*/
Core.prototype._hScrollbarEnabled = true;
/** @type {number}
* @private
*/
Core.prototype._pinnedLeftColumnCount = 0; // Cached value, that should always the same as the actual value in the UI
/** @type {number}
* @private
*/
Core.prototype._pinnedRightColumnCount = 0; // Cached value, that should always the same as the actual value in the UI
/** number of footer that being freeze right now
* @type {number}
* @private
*/
Core.prototype._frozenFooterCount = 0;
/** If true, the number of frozen top sections will not be changed automatically
* @type {boolean}
* @private
*/
Core.prototype._fixFrozenTopSections = false;
/** If true, the number of frozen bottom sections will not be changed automatically
* @type {boolean}
* @private
*/
Core.prototype._fixFrozenBottomSection = false;
/** cumulative row indices
* @type {Array.<number>}
* @private
*/
Core.prototype._sectionStarts;

/** @type {!Array.<SectionSettings>}
* @private
*/
Core.prototype._settings;
/** @type {SectionSettings}
* @private
*/
Core.prototype._firstSettings;
/** @type {SectionSettings}
* @private
*/
Core.prototype._lastSettings;

/** @type {boolean}
* @private
*/
Core.prototype._fitContentWidth = false;

/** @type {!Array.<Object.<string, *>>}
* @private
*/
Core.prototype._columnDefs;

/** @type {!Virtualizer}
* @private
*/
Core.prototype._rowVirtualizer;
/** @type {!Virtualizer}
* @private
*/
Core.prototype._colVirtualizer;

/** when set datasource without loading DataSourceTreePlugin, this datasource is used for every section.
* @type {boolean}
* @private
*/
Core.prototype._autoSetDataSource = true;

/** @type {Function}
* @private
*/
Core.prototype._renderingHandler = null;
/** @type {Function}
* @private
*/
Core.prototype._dataBindingHandler = null;

/** @type {boolean}
* @private
*/
Core.prototype._frozenLayout = false;
/** @type {boolean}
* @private
*/
Core.prototype._noBinding = false;
/** @type {boolean}
* @private
*/
Core.prototype._rowHeightSync = true;
/** @type {boolean}
* @private
*/
Core.prototype._rowHighlighting = false;
/** @type {boolean}
* @private
*/
Core.prototype._dispatchingDataChanged = false;

/** @type {string}
* @private
*/
Core.prototype._rowHeightException = "";
/** @type {number}
* @private
*/
Core.prototype._rowScrollingStep = 0;
/** @type {number}
* @private
*/
Core.prototype._zoomFactor = 1;
/** @type {number}
* @private
*/
Core.prototype._reservedSpace = 0;

/** @type {string}
* @private
*/
Core.prototype._hidingMethod = "";
/** @type {Conflator}
* @private
*/
Core.prototype._rowHeightConflator = null;
/** @type {Conflator}
* @private
*/
Core.prototype._vScrollbarConflator = null;
/** @type {Conflator}
* @private
*/
Core.prototype._columnBoundConflator = null;
/** @type {number}
* @private
*/
Core.prototype._rowRefreshTimer = 0;
/** @type {boolean}
* @private
*/
Core.prototype._layoutUpdating = false;
/** A hidden input that allows grid to receive keyboard input and focus
* @type {!Element}
* @private
*/
Core.prototype._firstHiddenInput;
/** A hidden input that allows grid to receive keyboard input and focus
* @type {!Element}
* @private
*/
Core.prototype._lastHiddenInput;
/** @type {number}
* @private
*/
Core.prototype._bottomPadding = 6;
/** @type {ILayoutGrid}
* @private
*/
Core.prototype._bottomPaddingSect = null;
/** @type {boolean}
* @private
*/
Core.prototype._preserveProportion = false;
/** @type {boolean}
* @private
*/
Core.prototype._preserveGridSize = false;
/** @type {number}
* @private
*/
Core.prototype._rowHeightTimerId = 0;
/** @type {GroupDefinitions}
* @private
*/
Core.prototype._groupDefs = null;
/** @type {BatchInfo}
* @private
*/
Core.prototype._batches = null;
/** @type {boolean}
* @private
*/
Core.prototype._firstRendered = false;
/** @type {boolean}
* @private
*/
Core.prototype._hasPendingRowChange = false;
//#region Public Methods

/**
* @public
* @return {string}
*/
Core.getVersion = function () {
	return "5.2.3";
};
/** {@link ElementWrapper#dispose}
* @override
*/
Core.prototype.dispose = function () {
	let internalId = this["_id"];
	let map = Core["map"];
	if (map[internalId]) {
		map[internalId] = null;
	} else {
		console.log("WARNING: Core grid is disposed twice");
		return;
	}
	this._disposed = true;
	if(this._rowRefreshTimer) {
		clearTimeout(this._rowRefreshTimer);
		this._rowRefreshTimer = 0;
	}

	// Clear all events
	this.unlistenAll();
	window.removeEventListener("resize", this._onWindowResize, false);

	for (let i = this._columnDefs.length; --i >= 0; ) {
		let colDef = this._getColumnDef(i);
		colDef["renderingHandler"] = colDef["dataBindingHandler"] = null;
	}
	for (let key in this._plugins) {
		let plugin = this._plugins[key];
		if (plugin["unload"]) {
			plugin["unload"](this);
		}
	}

	// Clear all Elements
	this.removeAllSections(); // Remove all Sections and LayoutGrids
	let mainScrolbar = this._vscrollbar._mainScrollbar;
	if(mainScrolbar) {
		mainScrolbar.unlisten("scroll", this._onSyncVScroll);
		this._vscrollbar._mainScrollbar = null;
	}
	mainScrolbar = this._hscrollbar._mainScrollbar;
	if(mainScrolbar) {
		mainScrolbar.unlisten("scroll", this._onSyncHScroll);
		this._hscrollbar._mainScrollbar = null;
	}

	this._vscrollbar.dispose();
	this._hscrollbar.dispose();
	this._rowHeightConflator.dispose();
	this._vScrollbarConflator.dispose();
	this._columnBoundConflator.dispose();
	this._columnPositionConflator.dispose();
	this._rowPositionConflator.dispose();
	this._layoutX.dispose();
	this._layoutY.dispose();


	// Clean Top node
	let elem = this._element;
	if (elem) {
		if (elem["_control"]) {
			delete elem["_control"];
		}
		elem.removeChild(this._firstHiddenInput);
		elem.removeChild(this._lastHiddenInput);
	}
	this._dispose();

	this._columnDefs.length = 0;
	delete this._plugins;
};
/** Indicates if the object has been disposed or not
* @public
* @return {boolean}
*/
Core.prototype.isDisposed = function () {
	return this._disposed;
};

/** {@link ElementWrapper#setParent}
* @override
*/
Core.prototype.setParent = function (obj, opt_prepend) {
	this._setParent(obj, opt_prepend);
	this._onParentChange();
};
/** {@link ElementWrapper#insertBefore}
* @override
*/
Core.prototype.insertBefore = function (obj) {
	this._insertBefore(obj);
	this._onParentChange();
};
/** @private */
Core.prototype._onParentChange = function () {
	let newParent = this.getParent() || this.getElement();
	if(!this._vscrollbar._mainScrollbar) { // HACK
		this._vscrollbar.setParent(newParent);
	}
	this._hscrollbar.setParent(newParent);

	this._updateLayout();
};

/** @public
* @param {Object=} gridOptions
* @return {!Object}
*/
Core.prototype.getConfigObject = function (gridOptions) {
	let obj = gridOptions || {};

	let columnCount = this.getColumnCount();
	if (!obj["columns"]) {
		obj["columns"] = [];
	}

	for (let colIndex = 0; colIndex < columnCount; colIndex++) {
		let column = obj["columns"][colIndex];
		if (!column) {
			column = obj["columns"][colIndex] = {};
		}

		let columnDef = this._getColumnDef(colIndex);
		if(columnDef["userId"]) {
			column["id"] = columnDef["userId"];
		}
		if(columnDef["field"]) {
			column["field"] = columnDef["field"];
		}
		if (columnDef["styles"]) {
			column["styles"] = columnDef["styles"];
		}
		if (columnDef["stationary"]) {
			column["stationary"] = columnDef["stationary"];
		}
		if (columnDef["leftPinned"]) {
			column["leftPinned"] = columnDef["leftPinned"];
		}
		if (columnDef["rightPinned"]) {
			column["rightPinned"] = columnDef["rightPinned"];
		}
		if (columnDef["backgroundColor"]) {
			column["backgroundColor"] = columnDef["backgroundColor"];
		}
	}

	// It will be overwrite in rt-grid or atlas-blotter
	obj["rowVirtualization"] = this.isRowVirtualizationEnabled();
	obj["columnVirtualization"] = this.isColumnVirtualizationEnabled();
	obj["rowHighlighting"] = this._rowHighlighting;

	if(this._fixFrozenTopSections) {
		obj["topFreezingCount"] = this._startVScrollbarIndex >= 0 ? (this._startVScrollbarIndex + 1) : false;
	}
	if(this._fixFrozenBottomSection) {
		obj["bottomFreezingCount"] = this._frozenFooterCount;
	}


	if(this._rowScrollingStep) {
		obj["stepScroll"] = (this._rowScrollingStep === 1) ? true : this._rowScrollingStep;
	}

	let val = this._vscrollbar.getAutoHide();
	if(val !== true) {
		// TODO: check this._hscrollbar and this._vscrollbar can be difference config
		obj["autoHideScrollbar"] = val; // this._hscrollbar has the same settings
	}

	let wheelSpeed = this._vscrollbar.getMouseWheelSpeed();
	if(wheelSpeed) {
		obj["linearWheelScrolling"] = wheelSpeed;
	}

	if(this.hasClass("no-borders")) {
		obj["borders"] = false;
	}
	if(this.hasClass("no-gridlines")) {
		obj["gridlines"] = false;
	}
	if(this.hasClass("no-vertical-grid-line")) {
		obj["verticalLines"] = false;
	}
	if(this.hasClass("no-horizontal-grid-line")) {
		obj["horizontalLines"] = false;
	}
	if(this.hasClass("no-content-vertical-line")) {
		obj["contentVerticalLines"] = false;
	}
	if(this._bottomPaddingSect) {
		obj["contentBottomPadding"] = this._bottomPadding;
	}
	if(this._reservedSpace > 0) {
		obj["contentRightPadding"] = this._reservedSpace;
	}

	let conflationRate = this.getDataConflationRate();
	if (conflationRate) {
		obj["dataConflationRate"] = conflationRate;
	}

	for(let key in this._plugins) {
		let pluginInstance = this._plugins[key];
		if(typeof pluginInstance["getConfigObject"] === "function") {
			pluginInstance["getConfigObject"](obj);
		}
	}

	return obj;
};

/** {@link ElementWrapper#fillParentWidth}
* @override
*/
Core.prototype.fillParentWidth = function(pct) {
	if(this._fitContentWidth) { return false; }
	if(this._layoutX.isFixedSize()) { return false; }

	// uses getBoundingClientRect() instead of clientWidth to prevents
	// fraction error occured on Chrome (Windows 10) which make clientWidth
	// exceeds expected value by 1px and will cause incorrect scrollbar active frag
	let el = this._element;
	let px = (el.getBoundingClientRect().width | 0) - el.clientLeft;
	let changes = this._layoutX.setTrackSize(px - this._reservedSpace);
	return changes !== null;
};
/** {@link ElementWrapper#show}
* @override
*/
Core.prototype.show = function (opt_shown) {
	this._element.style.display = (opt_shown !== false) ? "" : "none";
	if (opt_shown !== false) {
		this.updateLayout();
	}
};
/** @public
* @return {number}
*/
Core.prototype.getContentWidth = function() {
	return this._layoutX.getTrackSize();
};

/** @public
* @return {number}
*/
Core.prototype.getSectionCount = function() {
	return this._settings.length;
};

/** @private */
Core.prototype._updateFrozenTopSections = function () {
	if (!this._fixFrozenTopSections) {
		let topCount = this._getTopSectionCount();
		this._freezeTopSectionAt(topCount - 1);
	}
};

/** @private */
Core.prototype._updateFrozenBottomSections = function () {
	if (!this._fixFrozenBottomSection) {
		let bottomCount = this._getBottomSectionCount();
		this._setFrozenFooter(bottomCount);
	}
};

/** @private
* @return {number}
*/
Core.prototype._getTopSectionCount = function () {
	let sectionSettings = this._settings;
	let length = sectionSettings.length;
	let count = 0;
	for (let i = 0; i < length; i++) {
		let type = sectionSettings[i].getType();
		if (type === "header" || type === "title") {
			count++;
		} else {
			break;
		}
	}
	return count;
};

/** @private
* @return {number}
*/
Core.prototype._getBottomSectionCount = function () {
	let sectionSettings = this._settings;
	let length = sectionSettings.length;
	let count = 0;
	for (let i = length - 1; i >= 0; i--) {
		let type = sectionSettings[i].getType();
		if (type === "footer") {
			count++;
		} else {
			break;
		}
	}
	return count;
};

/** @public
* @param {string=} opt_type
* @param {string=} opt_sectionName
* @return {ILayoutGrid|null}
*/
Core.prototype.addSection = function (opt_type, opt_sectionName) {
	return this.addSectionAt(this.getSectionCount(), opt_type, opt_sectionName);
};
/** @public
* @param {number} at
* @param {string=} opt_type
* @param {string=} opt_sectionName
* @return {ILayoutGrid|null}
* @fires Core#sectionAdded
*/
Core.prototype.addSectionAt = function (at, opt_type, opt_sectionName) {
	let sectionCount = this._settings.length;
	if (at < 0) { at = 0; }
	if (at > sectionCount) { at = sectionCount; }

	let sectType = this._toSectionType(opt_type);
	let isBottomPadding = (opt_sectionName === "bottomPadding");
	let sectionSettings  = this._newSection(
		isBottomPadding ? "footer" : sectType, // Always use footer as a setup for bottom padding section
		opt_sectionName
	);
	if(!sectionSettings){ return null; }

	let section = sectionSettings.getSection();
	if(isBottomPadding) {
		sectionSettings.setAutoSyncRowCount(false); // Important
		this._bottomPaddingSect = section;
	}

	if (at >= sectionCount) {
		this._putToLast(section);
		this._settings.push(sectionSettings);
	} else {
		this._settings.splice(at, 0, sectionSettings);
		section.insertBefore(this._settings[at + 1].getSection());
	}

	this._updateSectionIndices(at);
	if (this._autoSetDataSource) {
		// This could trigger data binding and column virtualization
		sectionSettings.setDataSource(this._dataSource);
	}

	sectType = sectionSettings.getType();
	if (sectType === "footer") {
		this._updateFrozenBottomSections();
	} else if (sectType === "header" || sectType === "title") {
		this._updateFrozenTopSections();
	}

	this._onColumnCountChanged(); // When a new section is added, new columns are also added
	this._onSectionCountChanged(true);

	if(this._colVirtualizer.isEnabled()) {
		let activations = this._getColActivationList();
		let colCount = this.getColumnCount();
		sectionSettings.activateColumns(activations, 0, colCount - 1);
	}

	if (this._hasListener("sectionAdded")) {
		let e = sectionSettings.extendEventArg();
		this._dispatch("sectionAdded", e);
	}

	return section;
};
/** @public
* @param {Core.SectionReference} sectionRef
* @return {ILayoutGrid}
*/
Core.prototype.removeSection = function(sectionRef) {
	let section = this.getSection(sectionRef);
	if(section !== null) {
		return this.removeSectionAt(section.getIndex());
	}
	return null;
};
/** @public
* @param {number} at
* @return {ILayoutGrid}
*/
Core.prototype.removeSectionAt = function (at) {
	let sectionCount = this._settings.length;
	if (at >= 0 && at < sectionCount) {
		let sectionSettings = this._settings[at];
		let removedIndex = sectionSettings.getIndex();
		let section = sectionSettings.getSection();
		let sectionType = sectionSettings.getType();

		section.setRowCount(0); // Row count changed
		this._settings.splice(at, 1);

		this._updateSectionIndices(removedIndex);

		if (sectionType === "footer") {
			this._updateFrozenBottomSections();
			this._onColumnCountChanged(); // Update footer count in hScrollbar
		} else if (sectionType === "header" || sectionType === "title") {
			this._updateFrozenTopSections();
		}

		this._onSectionCountChanged();  // Update vScrollbar

		sectionSettings.dispose();
		return section;
	}
	return null;
};
/** @public
* @return {ILayoutGrid}
*/
Core.prototype.removeLastSection = function() {
	return this.removeSectionAt(this._settings.length - 1);
};
/** @public */
Core.prototype.removeAllSections = function () {
	let len = this._settings.length;
	for (let i = 0; i < len; ++i) {
		let sectionSettings = this._settings[i];
		sectionSettings.dispose();
	}
	this._bottomPaddingSect = null;
	this._firstSettings = this._lastSettings = null;
	this._settings.length = 0;
	this._updateSectionIndices(0);
	this._onSectionCountChanged();
};
/** @public
* @param {boolean|number} paddingSize=6 Add special section at the bottom. If falsy value is specified, the section is removed.
*/
Core.prototype.setBottomPaddingSection = function (paddingSize) {
	let sect = this._bottomPaddingSect;
	if(paddingSize) {
		if(typeof paddingSize === "number") {
			this._bottomPadding = paddingSize;
		}
		if(!sect) {
			sect = this.addSection("content", "bottomPadding");
		}
		if(sect) {
			sect.setDefaultRowHeight(this._bottomPadding);
			if(!sect.getRowCount()) {
				sect.addRow(1);
			}
		}
	} else if(sect) {
		this.removeSection(sect);
	}
};

/** @private
* @return {boolean} Returns true if there is any change
*/
Core.prototype._updatePaddingBottomSect = function () {
	let bps = this._bottomPaddingSect;
	if(bps) { // Always move bottom padding sect to the last
		let sectCount = this._settings.length;
		let last = sectCount - 1;
		if(last > 0) {
			let lastSettings = this._settings[last];
			if(lastSettings.getSection() !== bps) {
				let bpss = null;
				for(let i = last; --i >= 0;) {
					let settings = this._settings[i];
					if(settings.getSection() === bps) {
						bpss = settings;
						break;
					}
				}
				if(i >= 0) {
					this._putToLast(bps);
					this._settings.splice(i, 1);
					this._settings.push(bpss);
					// Blend styling with previous one
					bpss.setType(lastSettings.getType() === "footer" ? "footer" : "content");
					return true;
				} else {
					this._bottomPaddingSect = null; // The settings and section must have been removed
				}
			} else {
				let prevSettings = this._settings[last - 1];
				if(prevSettings) {
					// Blend styling with previous one
					lastSettings.setType(prevSettings.getType() === "footer" ? "footer" : "content");
				}
			}
		}
	}
	return false;
};
/** @public
* @ignore
* @param {number} from INCLUSIVE
* @param {number} dest INCLUSIVE
*/
Core.prototype.moveSection = function (from, dest) {
	let sectionCount = this.getSectionCount();
	if (from < 0) { from = 0; }
	else if (from >= sectionCount) { return; }
	if (dest < 0) { dest = 0; }
	else if (dest >= sectionCount) { dest = sectionCount - 1; }
	if (from === dest) { return; }

	let fromSettings = this._settings[from];
	let destSettings = this._settings[dest];
	Util.moveArrayItem(this._settings, from, dest);

	if (destSettings) {
		fromSettings.getSection().insertBefore(destSettings.getSection());
	} else {
		this._putToLast(fromSettings.getSection());
	}

	this._updateSectionIndices((from <= dest) ? from : dest);
	this._syncRowHeights();
	this._onSectionCountChanged();
	if (!this._frozenLayout) {
		this._rowVirtualizer.update(true); // Force firing of row activation
	}
};

/** @public
* @param {Core.SectionReference} sectionRef
* @return {ILayoutGrid}
*/
Core.prototype.getSection = function (sectionRef) {
	let settings = this.getSectionSettings(sectionRef);
	return (settings !== null) ? settings.getSection() : null;

};
/** @public
* @param {Core.SectionReference} sectionRef
* @return {SectionSettings}
*/
Core.prototype.getSectionSettings = function (sectionRef) {
	if(Util.isString(sectionRef)) {
		if(Core._sectionType[/** @type{string} */(sectionRef)]) {
			return this._getNextSection(1, null, /** @type{string} */(sectionRef));
		}
		for(let i = this._settings.length; --i >= 0;) {
			if(this._settings[i].getId() === sectionRef) {
				return this._settings[i];
			}
		}
	} else {
		if(sectionRef != null) {
			let num = -1;
			if(sectionRef.getIndex != null) {
				num = /** @type{number} */(sectionRef.getIndex());
			} else if(Util.isNumber(sectionRef)) {
				num = /** @type{number} */(sectionRef);
			}
			if(num >= 0 && num < this._settings.length) {
				return this._settings[num];
			}
		}
	}
	return null;
};

/** @public
* @param {string=} opt_type
* @return {!Array.<ILayoutGrid>}
*/
Core.prototype.getAllSections = function (opt_type) {
	if(opt_type == null) { return this._getAllSections(); }

	let list = this.getAllSectionSettings(opt_type);
	let ary = /** @type{!Array.<ILayoutGrid>} */(list);
	for (let i = ary.length; --i >= 0; ) {
		ary[i] = list[i].getSection();
	}
	return ary;
};
/** @public
* @param {string=} opt_type
* @return {!Array.<SectionSettings>}
*/
Core.prototype.getAllSectionSettings = function (opt_type) {
	if(!opt_type) {
		return this._settings; // WARNING: Return internal private member
	}

	let len = this._settings.length;
	let ary = [];
	opt_type = this._toSectionType(opt_type);
	for (let i = 0; i < len; ++i) {
		let sectionSettings = this._settings[i];
		if(opt_type === sectionSettings.getType()) {
			ary.push(sectionSettings);
		}
	}
	return ary;
};

/** @public
* @param {(MouseEvent|Core.SectionReference)=} identifier This can be MouseEvent, index, or section id or pointer reference
* @return {number} Return negative value, if the section does not exist
*/
Core.prototype.getSectionIndex = function(identifier) {
	if(Util.isNumber(identifier)) {
		return /** @type{number} */(identifier);
	}

	if(identifier instanceof MouseEvent) {
		let mPos = Util.getRelativePosition(/** @type{MouseEvent} */(identifier), this._element);
		let obj = {};
		this._hitTestY(mPos["y"], obj); // TODO: This will not work with zoom feature
		return obj["sectionIndex"];
	}

	let section = this.getSection(/** @type{string} */(identifier));
	if(section == null) { return -1; }

	return section.getIndex();
};
/** @public
* @param {Core.SectionReference} sectionRef
* @return {number} Return zero, if identifier does not exist
*/
Core.prototype.getSectionTop = function (sectionRef) {
	let section = this.getSection(sectionRef);
	if (section == null) { return 0; }
	let top = this._layoutY.getLaneStart(section.getRowOffset());

	if (this._startVScrollbarIndex >= 0 && section.getIndex() >= this._startVScrollbarIndex) {
		top -= this._vscrollbar.getScrollTop();
	}
	return top;
};

/** @public
* @param {(MouseEvent|Core.SectionReference)=} identifier This can be MouseEvent, index, or section id or pointer reference <br>
* If not specified, the last item is returned
* @param {string=} opt_type The previous section with the specified type. Default is any type
* @return {ILayoutGrid}
*/
Core.prototype.getPreviousSection = function (identifier, opt_type) {
	if(identifier == null) { identifier = this._settings.length; }
	let s = this._getNextSection(-1, identifier, opt_type);
	return (s !== null) ? s.getSection() : null;
};
/** @public
* @param {(MouseEvent|Core.SectionReference)=} identifier This can be MouseEvent, index, or section id or pointer reference <br>
* If not specified, the first item is returned
* @param {string=} opt_type The next section with the specified type. Default is any type
* @return {ILayoutGrid}
*/
Core.prototype.getNextSection = function (identifier, opt_type) {
	let s = this._getNextSection(1, identifier, opt_type);
	return (s !== null) ? s.getSection() : null;
};
/** @public
* @return {ILayoutGrid}
*/
Core.prototype.getLastSection = function () {
	return this.getSection(this._settings.length - 1);
};

/** @public
* @param {Core.SectionReference} sectionRef
* @param {number} colIndex
* @param {number} rowIndex
* @param {boolean=} opt_includeSpan Default is true
* @return {ElementWrapper}
*/
Core.prototype.getCell = function (sectionRef, colIndex, rowIndex, opt_includeSpan) {
	let s = this.getSection(sectionRef);
	return (s) ? s.getCell(colIndex, rowIndex, opt_includeSpan) : null;
};
/** Get information about the cell, inclusing column index, row index, and section that the cell resides
* @public
* @param {Core.CellReference} cellRef Reference to cell element. This can be something inside the cell as well.
* @return {Core.MouseInfo} Information about the given cell
* @see {@link Core#getRelativePosition}
* @example
* core.getCellInfo({"colIndex": 0, "rowIndex": 0}); // Get cell information about the top left cell in the content section
* core.getCellInfo(document.createElement("some_element"); // Get cell information for cell that contains the element with id "some_element"
* core.getCellInfo(mouseEventArg); // Get cell information from the given MouseEvent argument (this is the same as getRelativePosition() method)
* let cellInfo = core.getCellInfo({"colIndex": 1, "rowIndex": 1});
* cellInfo.colIndex = cellInfo.colIndex + 1; // Get next cell to the right
* core.getCellInfo(cellInfo);
*/
Core.prototype.getCellInfo = function (cellRef) {
	if(!cellRef) { return null; }
	let colIndex = /** @type{number} */(cellRef["colIndex"]);
	let rowIndex = /** @type{number} */(cellRef["rowIndex"]);
	let sectionName = /** @type{string} */(cellRef["sectionName"]);
	let mouseInfo = null;
	if(colIndex == null || rowIndex == null) { // Core.MouseInfo Non Mouse Info
		mouseInfo = this.getRelativePosition(/** @type{ElementWrapper|Element} */(cellRef));
		if(!mouseInfo["hit"]) {
			mouseInfo = null;
		}
	} else {
		if(!sectionName) {
			sectionName = "content";
		}
		let ss = this.getSectionSettings(sectionName);
		if(ss) {
			let s = ss.getSection();
			// WARNING: getCellInfo does not required cell to exist in the DOM (i.e., virtualized or hidden cells are included)
			if(colIndex >= 0 && colIndex < s.getColumnCount() &&
				rowIndex >= 0 && rowIndex < s.getRowCount()) {
				mouseInfo = {
					"colIndex": colIndex,
					"rowIndex": rowIndex
				};
				ss.extendEventArg(mouseInfo);
			}
		}
	}
	return mouseInfo;
};
/** Get cell information of the next cell in the same row. If there is no next cell in the row, `cell` parameter is set to null
* @public
* @param {Core.CellReference} cellRef Reference to cell element. This can be something inside the cell as well.
* @param {number=} step=1
* @return {Core.MouseInfo} Information about the next cell in the row. Returns null if the given parameter is invalid
*/
Core.prototype.getNextCellInRow = function (cellRef, step) {
	return this._getNextCellInRow(cellRef, (step != null) ? step : 1);
};
/** Get cell information of the previous cell in the same row. If there is no previous cell in the row, `cell` parameter is set to null
* @public
* @param {Core.CellReference} cellRef Reference to cell element. This can be something inside the cell as well.
* @param {number=} step=1
* @return {Core.MouseInfo} Information about the previous cell in the row. Returns null if the given parameter is invalid
*/
Core.prototype.getPrevCellInRow = function (cellRef, step) {
	return this._getNextCellInRow(cellRef, (step != null) ? -step : -1);
};
/** @private
* @param {Core.CellReference} cellRef
* @param {number} step
* @return {Core.MouseInfo}
*/
Core.prototype._getNextCellInRow = function (cellRef, step) {
	let mouseInfo = this.getCellInfo(cellRef);
	if(mouseInfo) {
		let colIndex = /** @type{number} */(mouseInfo["colIndex"]) + step;
		let rowIndex = /** @type{number} */(mouseInfo["rowIndex"]);
		let section = /** @type{ILayoutGrid} */(mouseInfo["section"]);

		mouseInfo["colIndex"] = colIndex;
		mouseInfo["cell"] = section.getCell(colIndex, rowIndex);
	}
	return mouseInfo;
};
/** Get cell information of the next cell. If bottom right has been reached, `cell` parameter is set to null
* @public
* @param {Core.CellReference} cellRef Reference to cell element. This can be something inside the cell as well.
* @return {Core.MouseInfo} Information about the next cell from left to right and top to bottom. Returns null if the given parameter is invalid
*/
Core.prototype.getNextCell = function (cellRef) {
	let mouseInfo = this.getNextCellInRow(cellRef);
	if(mouseInfo) {
		let cell = /** @type{Core.Cell} */(mouseInfo["cell"]);
		if(!cell) { // End of the row
			let section = /** @type{ILayoutGrid} */(mouseInfo["section"]);
			let colIndex = mouseInfo["colIndex"] = 0;
			let rowIndex = /** @type{number} */(mouseInfo["rowIndex"]);

			mouseInfo["rowIndex"] = ++rowIndex;
			mouseInfo["cell"] = section.getCell(colIndex, rowIndex);
		}
	}
	return mouseInfo;
};
/** Get cell information of the previous cell. If top left cell has been reached, `cell` parameter is set to null
* @public
* @param {Core.CellReference} cellRef Reference to cell element. This can be something inside the cell as well.
* @return {Core.MouseInfo} Information about the previous cell in the row. Returns null if the given parameter is invalid
*/
Core.prototype.getPrevCell = function (cellRef) {
	let mouseInfo = this.getPrevCellInRow(cellRef);
	if(mouseInfo) {
		let cell = /** @type{Core.Cell} */(mouseInfo["cell"]);
		if(!cell) { // End of the row
			let section = /** @type{ILayoutGrid} */(mouseInfo["section"]);
			let colIndex = mouseInfo["colIndex"] = section.getColumnCount() - 1;
			let rowIndex = /** @type{number} */(mouseInfo["rowIndex"]);

			mouseInfo["rowIndex"] = --rowIndex;
			mouseInfo["cell"] = section.getCell(colIndex, rowIndex);
		}
	}
	return mouseInfo;
};

/** @public
* @return {number}
*/
Core.prototype.getColumnCount = function () {
	return this._layoutX.getLaneCount();
};
/** Column count can be independently set regardless of row count and section count
* @public
* @param {number} num
* @fires Core#columnAdded
* @fires Core#preSectionRender
* @fires Core#columnRender
* @fires Core#beforeContentBinding
* @fires Core#preSectionDataBinding
* @fires Core#columnDataBinding
* @fires Core#postSectionDataBinding
* @fires Core#afterContentBinding
*/
Core.prototype.setColumnCount = function(num) {
	let colCount = this._layoutX.getLaneCount();
	if(num >= colCount) {
		this._layoutX.setLaneCount(num);
		for (let i = this._settings.length; --i >= 0; ) {
			let section = this._settings[i].getSection();
			section.setColumnCount(num); // Column may not be added to the document due to virtualization
		}

		this._onColumnCountChanged();
		this._syncLayoutToColumns(colCount, num);
		this._dispatchColumnAddedEvent(colCount, num - colCount, false);
	} else {
		this._removeColumn(colCount - num);
	}
};

/** @public
* @param {number} index
* @param {Core.ColumnOptions=} jsonObj Column definition object
* @fires Core#columnAdded
* @fires Core#preSectionRender
* @fires Core#columnRender
* @fires Core#beforeContentBinding
* @fires Core#preSectionDataBinding
* @fires Core#columnDataBinding
* @fires Core#postSectionDataBinding
* @fires Core#afterContentBinding
*/
Core.prototype.insertColumn = function (index, jsonObj) {
	let prevCount = this.getColumnCount();
	if (!(index < prevCount)) { index = prevCount; }

	this._layoutX.insertLane(index);
	for (let i = this._settings.length; --i >= 0; ) {
		let section = this._settings[i].getSection();
		section.insertColumn(index);
	}

	this._columnDefs.splice(index, 0, null); // TODO: Always initialize column def
	this._deserializeColumn(index, jsonObj); // Set leftPinned and other properties

	this._onColumnCountChanged();
	this._syncLayoutToColumns(index, index + 1); // _updateScrollbarWidth should be called

	this._dispatchColumnAddedEvent(index, 1, (index < prevCount), jsonObj);
};

/** @public
* @param {number} at
* @param {(number|Core.ColumnOptions|Array.<Core.ColumnOptions>)=} cols Column definition object or column count
*/
Core.prototype.insertColumns = function (at, cols) {
	let prevCount = this.getColumnCount();
	if (!(at < prevCount)) { at = prevCount; }

	let count = 0;
	let colDefs;
	if (Array.isArray(cols)) {
		count = cols.length;
		if(count === 1) {
			this.insertColumn(at, cols[0]);
			return;
		}

		colDefs = cols;
	} else if (typeof cols === "number") {
		count = cols;
		if(count === 1) {
			this.insertColumn(at, null);
			return;
		}
	} else {
		this.insertColumn(at, cols);
		return;
	}
	if (count <= 0) { return; }

	let prevState = this.freezeLayout();
	let args = [at, 0].concat(new Array(count));
	Array.prototype.splice.apply(this._columnDefs, args);

	for (let i = 0; i < count; i++) {
		let index = at + i;
		this._layoutX.insertLane(index);
		for (let s = this._settings.length; --s >= 0; ) {
			let section = this._settings[s].getSection();
			section.insertColumn(index);
		}
		let colDef = colDefs && colDefs[i];
		this._deserializeColumn(index, colDef);
	}

	this._dispatchColumnAddedEvent(at, count, (at < prevCount), colDefs);

	this.freezeLayout(prevState); // call _onColumnCountChanged and _syncLayoutToColumns internally
};

/** @public
* @param {number} index
* @return {Object} Return column definition object
* @see {@link Core#removeAllColumns}
* @fires Core#widthChanged
* @fires Core#columnRemoved
*/
Core.prototype.removeColumnAt = function (index) {
	let colCount = this.getColumnCount();
	if(index < 0 || index >= colCount) { return null; }

	let vEnd = -1;
	if(this._colVirtualizer.isVirtualizable()) {
		vEnd = this._colVirtualizer.getLastIndexInView();
	}

	let layoutSettings = this._layoutX.removeLaneAt(index);
	let sectionCount = this._settings.length;
	for (let i = 0; i < sectionCount; ++i) {
		this._settings[i].getSection().removeColumnAt(index);
	}

	let leftPinnedCount = this._pinnedLeftColumnCount;
	let rightPinnedCount = this._pinnedRightColumnCount;

	let colDef = this._columnDefs.splice(index, 1)[0]; // width and styles

	this._onColumnCountChanged();
	 // TODO: Column count changed may cause scroll pane to changed, and thus _updateScrollbarWidth should be included in the event handler
	if (index < leftPinnedCount) {
		this._updateScrollbarWidth(true, true);
	} else if (index >= (colCount - rightPinnedCount)) {
		this._updateScrollbarWidth(true, true);
	}
	this._syncLayoutToColumns(index, colCount - 1, true); // WARNiNG: _updateScrollbarWidth may be called again here

	if(!colDef) {
		colDef = {};
	}
	colDef["layout"] = layoutSettings;

	if (this._hasListener("columnRemoved")) {
		let e = {};
		e["atTheMiddle"] = true;
		e["colIndex"] = index;
		e["columns"] = "deprecated";
		e["columnData"] = colDef["columnData"];
		e["colId"] = colDef["id"] || "";
		this._dispatchColumnEvent("columnRemoved", e);
	}

	// Last index in view here might be different from before moving column if columns have different width.
	if(index <= vEnd && vEnd === this._colVirtualizer.getLastIndexInView()) {
		this._colVirtualizer.update(); // Since columns are shifted to the left, we need to activate the columns to the right
	}

	return colDef;
};
/** @public
* @see {@link Core#removeColumnAt}
* @fires Core#widthChanged
* @fires Core#columnRemoved
*/
Core.prototype.removeAllColumns = function () {
	if (this.getColumnCount() > 0) {
		this._pinnedLeftColumnCount = 0;
		this._pinnedRightColumnCount = 0;
		this.setColumnCount(0);
	}
};

/** @public
* @return {number}
*/
Core.prototype.getStationaryColumnIndex = function () {
	let colCount = this.getColumnCount();
	for (let i = colCount - 1; i >= 0; i--) {
		if (this._getColumnDef(i)["stationary"]) {
			return i;
		}
	}
	return -1;
};

/** Moving columns from or to steady area marked by stationary flag is not allowed. This method is used for checking ability of move.
* @public
* @param {number|Array.<number>} fromColIndex index or array of indices of the column(s) to be moved.
* @param {number} toColIndex target column index
* @return {boolean} Return true if the moving is doable, and false otherwise
*/
Core.prototype.isColumnMoveAllowed = function (fromColIndex, toColIndex) {
	if (!Array.isArray(fromColIndex)) {
		fromColIndex = [fromColIndex];
	}

	if (fromColIndex.length === 0 || toColIndex == null) {
		return false;
	}

	let stationaryIndex = this.getStationaryColumnIndex();

	if (stationaryIndex > -1) {
		if (toColIndex <= stationaryIndex) {
			return false;
		}

		for (let i = 0; i < fromColIndex.length; i++) {
			if (fromColIndex[i] <= stationaryIndex) {
				return false;
			}
		}
	}

	return true;
};

/** This method is used for checking movability of column.
* @public
* @param {number} colIndex Index of the column.
* @return {boolean} Return true if the column moving is movable, and false otherwise
*/
Core.prototype.isColumnMovable = function (colIndex) {
	let stationaryIndex = this.getStationaryColumnIndex();

	if (stationaryIndex > -1) {
		if (colIndex <= stationaryIndex) {
			return false;
		}
	}

	return true;
};

/** @public
* @param {number|Array.<number>} fromCol INCLUSIVE index/indices of the column to be moved
* @param {number} destCol INCLUSIVE
* @return {boolean} Return true if there is any change, and false otherwise
* @fires Core#columnMoved
*/
Core.prototype.moveColumn = function (fromCol, destCol) {
	if(typeof fromCol === "number") {
		return this._moveColumn(fromCol, destCol);
	}
	if(!Array.isArray(fromCol)) {
		return false;
	}

	// Validate user inputs
	let i;
	let len = fromCol.length;
	let colIndices = [];
	let colIndex = 0;
	let colCount = this.getColumnCount();
	let dict = {};
	for(i = 0; i < len; ++i) {
		colIndex = fromCol[i];
		if(colIndex >= 0 && colIndex < colCount) { // Filter out invalid index
			if(dict[colIndex] == null) { // Remove duplication
				dict[colIndex] = 1;
				colIndices.push(colIndex);
			}
		}
	}
	if (destCol < 0) {
		destCol = 0;
	} else if (destCol >= colCount) {
		destCol = colCount - 1;
	}

	// Skip unneccessary cases
	len = colIndices.length;
	if(!len) {
		return false;
	}
	if(len === 1) {
		return this._moveColumn(colIndices[0], destCol);
	}
	if(!this.isColumnMoveAllowed(colIndices, destCol)) {
		return false;
	}

	colIndices.sort(ascNumberSorter);
	if(dict[destCol]) { // Destination falls in the same place as source columns
		// Check if source columns are in continuous sequence
		let continuousSequence = true;
		colIndex = colIndices[0];
		for(i = 1; i < len; ++i) {
			if(++colIndex !== colIndices[i]) {
				continuousSequence = false;
				break;
			}
		}
		if(continuousSequence) {
			return false; // No moving operation is required
		}
	}
	let prevState = this.freezeLayout(true);

	// Start moving multiple items
	let rhsCount = 0;
	let srcIndex, destIndex;
	// Perform move operation on the right hand side
	for(i = 0; i < len; ++i) {
		srcIndex = colIndices[i];
		if(srcIndex >= destCol) {
			destIndex = destCol + rhsCount;
			this._moveColumn(srcIndex, destIndex);
			++rhsCount;
		}
	}
	// Perform move operation on the left hand side
	let lhsCount = len - rhsCount;
	destIndex = rhsCount ? destCol - 1 : destCol;
	for(i = 0; i < lhsCount; ++i) {
		srcIndex = colIndices[i] - i;
		this._moveColumn(srcIndex, destIndex);
	}

	this.freezeLayout(prevState);
	return true;
};
/** @private
* @param {number} fromCol INCLUSIVE
* @param {number} destCol INCLUSIVE
* @return {boolean} Return true if there is any change, and false otherwise
* @fires Core#columnMoved
*/
Core.prototype._moveColumn = function (fromCol, destCol) {
	let colCount = this.getColumnCount();
	if (fromCol < 0) { fromCol = 0; }
	else if (fromCol >= colCount) { fromCol = colCount - 1; }
	if (destCol < 0) { destCol = 0; }
	else if (destCol >= colCount) { destCol = colCount - 1; }
	if (fromCol === destCol) { return false; }

	if (!this.isColumnMoveAllowed(fromCol, destCol)) {
		return false;
	}

	let vEnd = -1;
	if(this._colVirtualizer.isVirtualizable()) {
		vEnd = this._colVirtualizer.getLastIndexInView();
	}

	let colId = this.getColumnId(fromCol);
	this._layoutX.moveLane(fromCol, destCol);
	let sectionCount = this._settings.length;
	for (let i = 0; i < sectionCount; ++i) {
		let sectionSettings = this._settings[i];
		let section = sectionSettings.getSection();
		section.moveColumn(fromCol, destCol);
	}

	let colDef = this._columnDefs[fromCol];
	let leftPinnedCount = this._countPinnedLeftColumns();
	let rightPinnedCount = this._countPinnedRightColumns();
	// let minColumn = (fromCol < destCol) ? fromCol : destCol;
	Util.moveArrayItem(this._columnDefs, fromCol, destCol);

	if(this._hScrollbarEnabled) {
		if(fromCol < leftPinnedCount) {
			// Left pinned column

			if(destCol >= (colCount - rightPinnedCount)) {
				// moved into right frozen zone
				this.freezeColumn(leftPinnedCount - 2, rightPinnedCount + 1);
			} else if(destCol >= leftPinnedCount) {
				// moved out
				this.freezeColumn(leftPinnedCount - 2, rightPinnedCount);
			}
		} else if(colDef["rightPinned"]) {
			// Right pinned column

			if(destCol < leftPinnedCount) {
				// moved into left frozen zone
				this.freezeColumn(leftPinnedCount, rightPinnedCount - 1);
			} else if(destCol < (colCount - rightPinnedCount)) {
				// moved out
				this.freezeColumn(leftPinnedCount - 1, rightPinnedCount - 1);
			}
		} else {
			// unpinned column

			if(destCol < leftPinnedCount) {
				// moved into left frozen zone
				this.freezeColumn(leftPinnedCount, rightPinnedCount);
			} else if(destCol >= (colCount - rightPinnedCount)) {
				// moved into right frozen zone
				this.freezeColumn(leftPinnedCount - 1, rightPinnedCount + 1);
			}
		}
	}

	// The deactivated column may be moved directly or indirectly to the current view. It's necessary to make all columns in view activated.
	if(!this._frozenLayout) {
		if(this._colVirtualizer.isEnabled()) {
			let vBegin = this._colVirtualizer.getFirstIndexInView();
			if(!((fromCol < vBegin && destCol < vBegin) || (fromCol > vEnd && destCol > vEnd))) { // Columns does not move between hidden columns
				this._activateColumns(vBegin, vEnd, vBegin, vEnd); // To confirm that all columns in view are activated
			}
		}
	}

	// no need to invoke because moving column does not change column-width
	// this._syncLayoutToColumns(minColumn, this.getColumnCount());

	if (this._hasListener("columnMoved")) {
		let e = {};
		e["fromColIndex"] = fromCol;
		e["toColIndex"] = destCol;
		e["colId"] = colId; // TODO: Id may not needed
		this._dispatchColumnEvent("columnMoved", e); // add remove move
	}

	// Last index in view here might be different from before moving column if columns have different width.
	if(vEnd === this._colVirtualizer.getLastIndexInView()) {
		if (fromCol <= vEnd && destCol > vEnd) { // Columns are shifted to the left, we need to activate the columns to the right
			this._colVirtualizer.update();
		} else if (destCol <= vEnd && fromCol > vEnd) { // Move virtualized column into view, need to rerender destination column
			this._colVirtualizer.update();
		}
	}
	this.updateColumnBounds();
	this._updateColumnSeparators();
	return true;
};

/** If source column is not found, no operation is performed. If destination column is not found, the source column will be moved to the last position.<br>
* Note: this method behaves slightly different from moveColumn method in that it always put source column at the position before the specified destination column, while moveColumn method will put column at exactly at the destination index.
* @public
* @param {number|string} srcCol Column Id or index
* @param {(number|string)=} destCol Column Id or index of the destination
* @return {boolean} Return true if there is any change, and false otherwise
* @see {@link Core#moveColumn}
* @example
* grid.moveColumnById(3, 1); // Move column 3 to position before column 1
* grid.moveColumnById(0, 2); // Move column 0 to position before column 2 (column index 1)
* grid.moveColumnById(0, 1); // Nothing is moved
* grid.moveColumnById("sourceColumnId", "anotherId");
* grid.moveColumnById("sourceColumnId", ""); // move to the last position
*/
Core.prototype.moveColumnById = function (srcCol, destCol) {
	let colCount = this.getColumnCount();
	let srcIndex = this.getColumnIndex(srcCol);
	if(srcIndex < 0 || srcIndex >= colCount) {
		return false;
	}
	let destIndex = destCol != null ? this.getColumnIndex(destCol) : -1;
	if(destIndex < 0) {
		destIndex = colCount;
	}
	return this._moveColumnByIndex(srcIndex, destIndex);
};
/** Move column without verification for better performance
* @private
* @param {number} srcIndex Column index
* @param {number} destIndex Column index of the destination
* @return {boolean} Return true if there is any change, and false otherwise
*/
Core.prototype._moveColumnByIndex = function (srcIndex, destIndex) {
	if(srcIndex < destIndex) { // Ensure that the source column is put in front of the destination index
		--destIndex;
	}
	if(srcIndex === destIndex) {
		return false;
	}
	return this.moveColumn(srcIndex, destIndex);
};
/** @public
* @param {number|string|Array.<number|string>} colRefs List of column index or column id to be moved
* @param {(number|string)=} destCol Destination position where the moved columns will be placed BEFORE the specified position. This can be column id or index
* @return {boolean} Return true if there is any change, and false otherwise
*/
Core.prototype.reorderColumns = function (colRefs, destCol) {
	let destId = "";
	if(typeof destCol === "number") {
		destId = this.getColumnId(destCol);
	} else if(destCol) {
		destId = destCol;
	}

	this.startBatch("move");

	let dirty = 0;
	if(Array.isArray(colRefs)) {
		let srcLen = colRefs.length;
		if(srcLen > 1) {
			let colIds = this.getColumnIds();
			let colCount = colIds.length;
			let srcIds = [];
			let invalidDest = false;
			let i, srcId, srcIdx;
			for(i = 0; i < srcLen; ++i) {
				let colRef = colRefs[i];
				if(typeof colRef === "number") {
					srcIdx = colRef;
					srcId = colIds[colRef] || "";
				} else {
					srcId = colRef;
					srcIdx = colIds.indexOf(srcId);
				}
				if(srcId && srcIdx >= 0) {
					srcIds.push(srcId);
					if(destId === srcId) {
						invalidDest = true; // Destination must not exist in source columns
					}
				}
			}

			let destIdx;
			srcLen = srcIds.length;
			if(invalidDest) { // Find the next valid destination where it is not contained in the source columns
				destIdx = this.getColumnIndex(destId);
				if(destIdx >= 0) {
					while(++destIdx < colCount) {
						destId = colIds[destIdx];
						if(srcIds.indexOf(destId) < 0) {
							break;
						}
					}
				}
				if(destIdx < 0 || destIdx >= colCount) {
					destId = "";
				}
			}

			for(i = srcLen; --i >= 0;) {
				srcId = srcIds[i]; // Only valid source columns are left at this point
				srcIdx = this.getColumnIndex(srcId);
				destIdx = this.getColumnIndex(destId);
				if(destIdx < 0) { // Insert to the back when id is not found
					destIdx = colCount;
				}
				dirty |= this._moveColumnByIndex(srcIdx, destIdx);
				destId = srcId;
			}
		} else {
			dirty = this.moveColumnById(colRefs[0], destId);
		}
	} else if(colRefs != null) {
		// colRefs will be a number or string
		dirty = this.moveColumnById(colRefs, destId);
	}
	this.stopBatch("move");
	return dirty ? true : false;
};

/** @public
* @ignore
* @return {!TrackLayout}
*/
Core.prototype.getColumnLayout = function () {
	return this._layoutX;
};

/** @public
* @ignore
* @fires Core#widthChanged
*/
Core.prototype.updateColumnLayout = function () {
	this._syncLayoutToColumns(0, this.getColumnCount());
};

/** @private
* @param {number} index
* @param {Object=} jsonObj
*/
Core.prototype._deserializeColumn = function (index, jsonObj) {
	if (!jsonObj) { return; }

	let layout = jsonObj["layout"] || jsonObj;
	let val = jsonObj["width"];
	if (val != null) { layout["size"] = val; }
	val = jsonObj["scalable"];
	if (val != null) { layout["scalable"] = val; }
	val = jsonObj["minWidth"];
	if (val != null) { layout["minSize"] = val; }
	if (layout["scalable"] == null) {
		layout["scalable"] = (layout["size"] == null);
	}

	this._layoutX.deserializeLane(index, layout);

	let name;
	let styles = jsonObj["styles"];
	for (name in styles) {
		this.setColumnStyle(index, name, styles[name]);
	}
	let classes = jsonObj["classes"];
	for (name in classes) {
		this.enableColumnClass(index, name, classes[name]);
	}
	// TODO: Load specific styles for each column type (e.g. "content", "title", "header")

	let colDef = this._getColumnDef(index);
	let colId = jsonObj["id"];
	if(colId && typeof colId === "string") {
		colDef["id"] = colId; // WARNING: We do not guarantee uniqueness of user id
		colDef["userId"] = colId;
	}
	let field = jsonObj["field"];
	if(field && typeof field === "string") {
		colDef["field"] = field;
	}

	let value = jsonObj["dataColumnName"]; // Deprecated
	if (value != null) {
		colDef["dataColumnName"] = value;
	}

	value = jsonObj["stationary"];
	if (value != null) {
		colDef["stationary"] = value ? true : false;
	}
	value = jsonObj["leftPinned"];
	if (value != null) {
		colDef["leftPinned"] = value ? true : false;
	}
	value = jsonObj["rightPinned"];
	if (value != null) {
		colDef["rightPinned"] = value ? true : false;
	}
	value = jsonObj["backgroundColor"];
	if (value != null) {
		this.setColumnBackgroundColor(index, value);
	}

	this.setColumnRenderingHandler(index, jsonObj["renderingHandler"]);
	this.setColumnDataBindingHandler(index, jsonObj["dataBindingHandler"]);

	value = jsonObj["columnData"];
	if(value) {
		this.setColumnData(index, value);
	}

	let formatter = jsonObj["binding"] || jsonObj["formatter"]; // support both composite and real-time grid, binding have piority first
	if(Array.isArray(formatter)) {
		formatter = formatter[0]; // Support only one formatter in the array
	}
	if(formatter && typeof formatter !== "function") {
		let formatterInitializer = formatter["init"];
		if(typeof formatterInitializer === "function") {
			formatterInitializer(jsonObj, this); // Pass on userJson object and core grid instance
		}
	}
};

/** Add specified number of rows to the last section. This method is intended for fast layout generation. <br>
* If column count is zero, one column will be added automatically. <br>
* If there is no section, a new section is created automatically. <br>
* No row is added to the data source.
* @public
* @param {number=} opt_num Default is one row
* @fires Core#preSectionRender
* @fires Core#columnRender
* @fires Core#beforeContentBinding
* @fires Core#preSectionDataBinding
* @fires Core#columnDataBinding
* @fires Core#postSectionDataBinding
* @fires Core#afterContentBinding
*/
Core.prototype.addRow = function (opt_num) {
	if (opt_num == null) { opt_num = 1; }
	else if (opt_num <= 0) { return; }

	if (this.getColumnCount() <= 0) {
		this.setColumnCount(1); // WARNING: Auto column adding
	}

	let section = this.getLastSection();
	if (section == null) {
		section = this.addSection(); // WARNING: Auto section adding
	}

	section.addRow(opt_num);
};
/** Remove specified number of rows from the last section. <br>
* Note that this will NOT remove row in the DataTable, as well as the section. <br>
* Use removeRowData to remove the supporting data row from the data source
* @public
* @param {number=} opt_num Default is one row
*/
Core.prototype.removeRow = function (opt_num) {
	let section = this.getLastSection();
	if (section == null) { return; }

	if (opt_num == null) { opt_num = 1; }
	section.removeRow(opt_num);
};
/** @public
* @return {number}
*/
Core.prototype.getRowCount = function () {
	return this._layoutY.getLaneCount();
};
/** The specified section and columns must exist prior to calling this method
* @public
* @param {Core.SectionReference} sectionRef
* @param {number} num
*/
Core.prototype.setSectionRowCount = function(sectionRef, num) {
	let section = this.getSection(sectionRef);
	if (section == null) { return; }
	section.setRowCount(num);
};
/** @public
* @param {number} sectionIndex
* @param {number} rowIndex relative to the given section
* @return {number} rowIndex relative to the first section (index as a whole)
*/
Core.prototype.getRowIndex = function(sectionIndex, rowIndex) {
	return this._sectionStarts[sectionIndex] + rowIndex;
};
/** @private
* @function
*/
Core.prototype.getRelativeIndex = Util._deprecatedFunction;

/** WARNING: This only removes all rows in each section, but the sections and columns remain intact
* @public
*/
Core.prototype.removeAllRows = function() {
	let len = this._settings.length;
	for(let i = 0; i < len; ++i) {
		this._settings[i].getSection().setRowCount(0);
		this._sectionStarts[i] = 0;
	}
	this._sectionStarts[len] = 0;
};
/** @public
* @return {number}
*/
Core.prototype.getDefaultRowHeight = function () {
	return this._layoutY.getDefaultLaneSize();
};
/** This default row height only applies to header and content section. <br>
* Use the second flag to apply the change to title section too. <br>
* To set default row height for a specific section, please call setDefaultRowHeight directly from the section
* @public
* @param {number} val
* @param {boolean=} opt_includeTitle
*/
Core.prototype.setDefaultRowHeight = function (val, opt_includeTitle) {
	if (this._layoutY.setDefaultLaneSize(val)) {
		this._rowHeightSync = false;
		let defaultHeight = this._layoutY.getDefaultLaneSize();
		this._rowHeightException = (opt_includeTitle) ? "" : "title";
		let sectCount = this._settings.length;
		for (let i = 0; i < sectCount; ++i) {
			let sectionSettings = this._settings[i];
			if (
				sectionSettings.getType() !== this._rowHeightException &&
				sectionSettings.getId() !== "bottomPadding"
			) {
				sectionSettings.getSection().setDefaultRowHeight(defaultHeight); // An event will be fired for each row
			}
		}
		this._syncRowHeights();
		this._rowHeightSync = true;
		this.setRowScrollingStep(this._rowScrollingStep);
		this._updateColumnSeparators();
	}
};

/** @public
* @return {DataView} dataSource
*/
Core.prototype.getDataSource = function () {
	return this._dataSource;
};

/** set Data Conflation for better performance
* @public
* @param {number} milliSecond
*/
Core.prototype.setDataConflationRate = function (milliSecond) {
	this._dataSource.setConflationRate(milliSecond);
};

/**
* @public
* @return {number}
*/
Core.prototype.getDataConflationRate = function () {
	return this._dataSource.getConflationRate();
};

/** @public
* @param {DataTable|DataView} dataSource
* @fires Core#dataSourceChanged
* @fires Core#postDataSourceChanged
* @return {DataView} Returns the root data view
*/
Core.prototype.setDataSource = function (dataSource) {
	if (dataSource) {
		if (dataSource instanceof DataTable) { // Sometimes, user put in a wrong parameter
			dataSource = new DataView(dataSource);
		}
	}

	let e = {};
	e["prevDataSource"] = this._dataSource;
	this._dataSource = dataSource || null;
	e["dataSource"] = this._dataSource;

	this._dispatch("dataSourceChanged", e);

	if (this._autoSetDataSource) {
		let len = this._settings.length;
		for (let i = 0; i < len; ++i) {
			this._settings[i].setDataSource(dataSource);
		}
	}

	this._dispatch("postDataSourceChanged", e);

	return this._dataSource;
};

/** @private
* @function
*/
Core.prototype.setColumnTemplate = Util._deprecatedFunction;
/** @private
* @function
*/
Core.prototype.getColumnTemplate = Util._deprecatedFunction;

/** Deprecated. Use setColumnRenderingHandler instead
* @public
* @function
*/
Core.prototype.getColumnRenderEvent = Util._deprecatedFunction;
/** Deprecated. Use setColumnDataBindingHandler instead
* @public
* @function
* @param {number} c
* @return {Object}
* @see {@link Core#setColumnDataBindingHandler}
*/
Core.prototype.getColumnDataBindingEvent = function(c) {
	console.log("getColumnDataBindingEvent() is deprecated. Use setColumnDataBindingHandler() instead");
	let t = this;
	return {
		"listen": function(f) {
			t.setColumnDataBindingHandler(c, f);
		},
		"unlisten": function() {
			t.setColumnDataBindingHandler(c, null);
		}
	};
};


/** @public
* @param {number} colIndex
* @param {string} style
* @param {string} value
* @param {string=} opt_type Apply the style only to the specified row type. Default is applying to all rows
*/
Core.prototype.setColumnStyle = function (colIndex, style, value, opt_type) {
	if (!style) { return; }
	if (!Util.isNumber(colIndex)) {
		colIndex = 0;
		console.log("WARNING: the first parameter for setColumnStyle must be column index");
	}

	let styles;
	if (opt_type == null) {
		styles = this._getNestedColumnDef(colIndex, "styles");
	} else {
		opt_type = this._toSectionType(opt_type);
		styles = this._getNestedColumnDef(colIndex, opt_type, "styles");
	}
	styles[style] = value;

	for (let i = this._settings.length; --i >= 0; ) {
		let sectionSettings = this._settings[i];
		if (opt_type == null || opt_type === sectionSettings.getType()) {
			let section = sectionSettings.getSection();
			section.setColumnStyle(colIndex, style, value);
		}
	}
};

/** @public
* @param {number} colIndex
* @return {string}
*/
Core.prototype.getColumnBackgroundColor = function(colIndex) {
	let colDef = this._getColumnDef(colIndex);
	return colDef["backgroundColor"] || "";
};
/** @public
* @param {number} colIndex
* @param {string=} color
*/
Core.prototype.setColumnBackgroundColor = function(colIndex, color) {
	if(colIndex == null || typeof colIndex !== "number") {
		return;
	}

	let colDef = this._getColumnDef(colIndex);
	if(colDef["backgroundColor"] === color) {
		return;
	}

	colDef["backgroundColor"] = color;
	color = color != null ? color : "";

	let sectionCount = this._settings.length;
	for(let i = 0; i < sectionCount; ++i) {
		let settings = this._settings[i];
		let section = settings.getSection();
		if(!section) {
			continue;
		}

		if(settings.getType() !== "content") {
			this._applyCellsBackgroundColor(section, colIndex, color);
		} else {
			let column = section.getColumn(colIndex);
			column.setStyle("backgroundColor", color);
		}
	}
};
/** @private
* @param {ILayoutGrid} section
* @param {number} colIndex
* @param {string} color
*/
Core.prototype._applyCellsBackgroundColor = function(section, colIndex, color) {
	let rowCount = section.getRowCount();
	for(let r = 0; r < rowCount; r++) {
		let cellSpan = section.getCellColSpan(colIndex, r);
		if(cellSpan > 1) { continue; }
		let cell = section.getCell(colIndex, r);
		if(cell) {
			cell.setStyle("backgroundColor", color);
		}
	}
};
/** @public
* @ignore
* @param {ILayoutGrid} section
*/
Core.prototype.applyColumnsBackgroundColor = function(section) {
	if(!section) {
		return;
	}

	let colCount = section.getColumnCount();
	for(let c = 0; c < colCount; c++) {
		let colDef = this._getColumnDef(c);
		let color = colDef["backgroundColor"] || "";
		this._applyCellsBackgroundColor(section, c, color);
	}
};

/** @public
* @param {number} colIndex
* @param {string} clsName
* @param {boolean=} enabled
* @param {string=} opt_type Apply the class only to the specified row type. Default is applying to all rows
* @return {boolean} Return true, if there is any change
*/
Core.prototype.enableColumnClass = function (colIndex, clsName, enabled, opt_type) {
	if (typeof colIndex !== "number") {
		colIndex = 0;
	}

	let classes;
	if (opt_type == null) {
		classes = this._getNestedColumnDef(colIndex, "classes");
	} else {
		opt_type = this._toSectionType(opt_type);
		classes = this._getNestedColumnDef(colIndex, opt_type, "classes");
	}

	let dirty = false;
	enabled = (enabled !== false);
	if (enabled) {
		if(!classes[clsName]) {
			dirty = classes[clsName] = true;
		}
	} else if (classes[clsName]) {
		delete classes[clsName];
		dirty = true;
	}

	if(dirty) {
		for (let i = this._settings.length; --i >= 0; ) {
			let sectionSettings = this._settings[i];
			if (opt_type == null || opt_type === sectionSettings.getType()) {
				sectionSettings.getSection().enableColumnClass(colIndex, clsName, enabled);
			}
		}
	}
	return dirty;
};
/** @public
* @param {number} colIndex
* @param {string} clsName
* @param {string=} opt_type Row type
* @return {boolean}
*/
Core.prototype.hasColumnClass = function (colIndex, clsName, opt_type) {
	let classes;
	if (opt_type == null) {
		classes = this._getNestedColumnDef(colIndex, "classes");
	} else {
		opt_type = this._toSectionType(opt_type);
		classes = this._getNestedColumnDef(colIndex, opt_type, "classes");
	}
	return (classes && classes[clsName]) ? true : false;
};

/** @public
* @param {number} colIndex
* @param {string=} alignment Available options are "center"|"right"|"left"|"c"|"r"|"l". Use empty string (falsy value) to reset back to default aliganment.
*/
Core.prototype.setColumnAlignment = function(colIndex, alignment) {
	let ch = alignment ? alignment.charAt(0).toLowerCase() : "";

	let dirty = 0;
	dirty |= this.enableColumnClass(colIndex, "tr-align-left", ch === "l");
	dirty |= this.enableColumnClass(colIndex, "tr-align-right", ch === "r");
	dirty |= this.enableColumnClass(colIndex, "tr-align-center", ch === "c");

	if(dirty) {
		this._dispatch("columnAlignmentChanged", {
			"colIndex": colIndex,
			"alignment": ch
		});
	}
};
/** @public
* @param {number} colIndex
* @return {string} Possible returned values are "default"|"left"|"center"|"right"
*/
Core.prototype.getColumnAlignment = function(colIndex) {
	if(this.hasColumnClass(colIndex, "tr-align-right")) {
		return "right";
	} else if(this.hasColumnClass(colIndex, "tr-align-center")) {
		return "center";
	} else if(this.hasColumnClass(colIndex, "tr-align-left")) {
		return "left";
	}
	return "default";
};


/** @public
* @param {number} colIndex
* @return {number}
*/
Core.prototype.getColumnLeft = function (colIndex) {
	let colDef = this._columnDefs[colIndex];
	if(colDef && colDef["rightPinned"]) {
		let gridWidth = this.getElement().clientWidth;
		let contentWidth = this._layoutX.getTrackSize();
		let colLeft = contentWidth < gridWidth ? contentWidth : gridWidth;
		let lastIndex = this.getColumnCount() - 1;
		for(let i = lastIndex; i >= colIndex; --i) {
			colLeft -= this._layoutX.getLaneSize(i);
		}
		return colLeft;
	} else if (this._hScrollbarEnabled && colIndex >= this._pinnedLeftColumnCount) {
		return this._layoutX.getLaneStart(colIndex) - this._hscrollbar.getScrollLeft();
	} else {
		return this._layoutX.getLaneStart(colIndex);
	}
};
/** @public
* @param {number} colIndex
* @return {number}
*/
Core.prototype.getColumnRight = function (colIndex) {
	return this.getColumnLeft(colIndex) + this._layoutX.getLaneSize(colIndex);
};
/** @public
* @param {number} colIndex
* @return {number}
*/
Core.prototype.getColumnWidth = function (colIndex) {
	return this._layoutX.getLaneSize(colIndex);
};
/** @public
* @param {number} colIndex
* @return {number}
*/
Core.prototype.getColumnCustomLaneSize = function (colIndex) {
	return this._layoutX.getCustomLaneSize(colIndex);
};
/** @private
* @param {Array.<number>} indices
* @param {Array.<number>} widths
* @return {Array.<number>}
*/
Core.prototype._calculateColumnWidths = function (indices, widths) {
	if(!this._preserveGridSize) {
		return widths;
	}

	let i, newWidth = 0, settingCols = {};
	let len = indices.length;
	let calWidths = new Array(len);
	for (i = 0; i < len; i++) {
		settingCols[indices[i]] = true;
		calWidths[i] = widths[i];
		newWidth += widths[i];
	}
	let fixedSize = 0, scalableCount = 0;
	let colCount = this.getColumnCount();
	let layoutX = this._layoutX;

	for(i = 0; i < colCount; i++) {
		if (settingCols[i]) { continue; }
		if (layoutX.isLaneVisible(i)) { // If column is visible
			if (!layoutX.getLaneScalability(i)) {
				fixedSize += layoutX.getLaneSize(i);
			} else {
				fixedSize += layoutX.getMinimumLaneSize(i);
				scalableCount++;
			}
		}
	}
	let gridSize = this.getContentWidth();
	let excess = newWidth + fixedSize - gridSize;

	if (scalableCount < 1 || excess > 0) {
		layoutX.resetLaneToDefaultSize();
		layoutX.setTrackSize(gridSize - this._reservedSpace);
		fixedSize = 0;
		for(i = 0; i < colCount; i++) {
			if (settingCols[i]) { continue; }
			if (layoutX.isLaneVisible(i)) { // If column is visible
				if (!layoutX.getLaneScalability(i)) {
					fixedSize += layoutX.getLaneSize(i);
				} else {
					fixedSize += layoutX.getMinimumLaneSize(i);
				}
			}
		}
		excess = newWidth + fixedSize - gridSize;
		if (excess > 0) {
			excess = excess / len;
			for (i = 0; i < len; i++) {
				calWidths[i] -= excess;
			}
		}
	}
	return calWidths;
};
/** @public
* @param {number} colIndex
* @param {number} px
* @param {boolean=} opt_scalable If not defined false value is used instead
* @return {boolean} True if there is any changed
* @fires Core#widthChanged
* @example
* core.setColumnWidth(0, 100); // The first column will have a fixed width
* core.setColumnWidth(1, 100, true); // The second column will have a percentage width where 100 is the weight comparing with other percentage column
*/
Core.prototype.setColumnWidth = function (colIndex, px, opt_scalable) {
	let calWidths = this._calculateColumnWidths([colIndex], [px]);
	let dirty = this._setColumnWidth(colIndex, calWidths[0], opt_scalable);
	if (dirty) {
		this._syncLayoutToColumns(colIndex, colIndex + 1);
	}
	return dirty;
};
/** Specifies multiple column widths all at once.
* @public
* @param {Array.<number>} columnWidths in pixel. If the specified value is undefined or null, nothing is performed on that column.
* @param {(boolean|Array.<boolean>)=} opt_proportion This will make columns scale in proportion specified by the given width
* @return {boolean} True if there is any changed
* @fires Core#widthChanged
* @example
* core.setColumnWidths([1, 2, 1, 1], true); // The first four columns have percentage width with the second column is twice as big as the others
* core.setColumnWidths([1, 200, 1, 150], [true, false, true]); // The first column make column scale, seconds make column pixel, third make column scale and four make column pixel
*/
Core.prototype.setColumnWidths = function (columnWidths, opt_proportion) {
	let dirty = 0;
	let indices = [], widths = [], propotions = [], i;
	for(i = columnWidths.length; --i >= 0;) {
		let px = columnWidths[i];
		if(px != null) {
			indices.push(i);
			widths.push(px);
		}
		let proportionMode = Array.isArray(opt_proportion) ? opt_proportion[i] : opt_proportion;
		propotions.push(proportionMode);
	}
	let calWidths = this._calculateColumnWidths(indices, widths);
	for(i = 0; i < calWidths.length; i++) {
		dirty |= this._setColumnWidth(indices[i], calWidths[i], propotions[i] );
	}

	if (dirty) {
		this._syncLayoutToColumns(0, columnWidths.length);
	}
	return dirty ? true : false;
};
/** Use {@link Core#hideColumn} to hide the column
* @public
* @param {number} colIndex
* @return {boolean}
*/
Core.prototype.isColumnVisible = function (colIndex) {
	return this._layoutX.isLaneVisible(colIndex);
};
/** Note that the hidden columns continue to exist and occupy the index. So there is no index shifted after this function call.
* @public
* @param {number} colIndex
* @param {boolean=} hidden If the value is true or undefined, hide the specified column. Otherwise, show the column
* @fires Core#widthChanged
* @fires Core#columnVisibilityChanged
* @see {@link Core#isColumnVisible}
* @see {@link Core#hideColumns}
* @see {@link Core#showAllColumns}
*/
Core.prototype.hideColumn = function (colIndex, hidden) {
	let visibility = hidden === false;
	this.setColumnVisibility(colIndex, visibility, 0);
};
/** Hide multiple columns at once, improving the performance. Note that the hidden columns continue to exist and occupy the index. So there is no index shifted after this function call.
* @public
* @param {Array.<number>} colIndices
* @param {boolean=} hidden If the value is true or undefined, hide the specified columns. Otherwise, show the columns
* @fires Core#widthChanged
* @fires Core#columnVisibilityChanged
* @see {@link Core#hideColumn}
* @see {@link Core#showAllColumns}
*/
Core.prototype.hideColumns = function (colIndices, hidden) {
	let len = Array.isArray(colIndices) ? colIndices.length : 0;
	let changeIndices = [];
	let colCount = this.getColumnCount();
	hidden = hidden !== false;
	for (let i = 0; i < len; ++i) {
		let colIndex = colIndices[i];
		if (colIndex >= 0 && colIndex < colCount) {
			if (this._layoutX.hideLane(colIndex, hidden, 0)) {
				changeIndices.push(colIndex);
			}
		}
	}
	if (changeIndices.length) {
		this._syncLayoutToColumns();
		this.requestRowRefresh();

		if (this._hasListener("columnVisibilityChanged")) {
			let e = {};
			e["colIndices"] = changeIndices;
			e["hidden"] = hidden;
			this._dispatch("columnVisibilityChanged", e);
		}
	}
};
/** @public
* @fires Core#widthChanged
* @fires Core#columnVisibilityChanged
* @see {@link Core#hideColumn}
* @see {@link Core#hideColumns}
*/
Core.prototype.showAllColumns = function () {
	let changeIndices = [];
	let colCount = this.getColumnCount();
	for (let c = 0; c < colCount; ++c) {
		if (this._layoutX.hideLane(c, false, 0)) {
			changeIndices.push(c);
		}
	}
	if (changeIndices.length) {
		this._syncLayoutToColumns();
		this.requestRowRefresh();

		if (this._hasListener("columnVisibilityChanged")) {
			let e = {};
			e["colIndices"] = changeIndices;
			e["hidden"] = false;
			this._dispatch("columnVisibilityChanged", e);
		}
	}
};
/** @public
* @param {number} colIndex
* @return {number}
*/
Core.prototype.getMinimumColumnWidth = function (colIndex) {
	return this._layoutX.getMinimumLaneSize(colIndex);
};
/** @public
* @param {number} colIndex
* @param {number} width
* @fires Core#widthChanged
*/
Core.prototype.setMinimumColumnWidth = function (colIndex, width) {
	if (this._layoutX.setMinimumLaneSize(colIndex, width)) {
		this._syncLayoutToColumns(colIndex, colIndex + 1);
	}
};
/** @public
* @param {Array.<number|null>} minWidths
* @param {boolean=} proportion This will make the column expand in proportion of its minimum width when there is enough space.
* @fires Core#widthChanged
* @return {boolean} True if there is any change
*/
Core.prototype.setMinimumColumnWidths = function (minWidths, proportion) {
	let dirty = 0;
	let first = NaN;
	let last = NaN;
	for (let c = minWidths.length; --c >= 0; ) {
		let minWidth = minWidths[c];
		if (minWidth != null) {
			if(!last) { last = c; }
			first = c;
			dirty |= this._layoutX.setMinimumLaneSize(c, minWidth, proportion);
		}
	}
	if (dirty) {
		this._syncLayoutToColumns(first, last + 1);
	}
	return dirty ? true : false;
};
/** @public
* @param {number} colIndex
* @return {boolean}
*/
Core.prototype.getColumnScalability = function (colIndex) {
	return this._layoutX.getLaneScalability(colIndex);
};
/** @public
* @param {number|boolean} val column index or default column scalability
* @param {boolean=} opt_scalable
* @return {boolean}
* @fires Core#widthChanged
*/
Core.prototype.setColumnScalability = function (val, opt_scalable) {
	let dirty = false;
	if (Util.isNumber(val)) {
		opt_scalable = (opt_scalable !== false);
		dirty = this._layoutX.setLaneScalability(/** @type{number} */(val), opt_scalable);
		if (dirty) {
			this._syncLayoutToColumns(/** @type{number} */(val), val + 1);
		}
	} else {
		this.setDefaultColumnScalable(/** @type{boolean} */(val));
	}
	return dirty;
};
/** @public
* @param {boolean} scalable
*/
Core.prototype.setDefaultColumnScalable = function (scalable) {
	this._layoutX.setDefaultScalability(!!scalable);
};
/** @public
* @return {boolean}
*/
Core.prototype.resetColumnWidths = function () {
	let dirty = this._layoutX.resetLaneToDefaultSize();
	if (dirty) {
  	this._syncLayoutToColumns(0, this.getColumnCount());
	}
	return dirty;
};
/** @public
* @param {number} colIndex
* @param {number} width
*/
Core.prototype.setColumnDefaultWidth = function (colIndex, width) {
	this._layoutX.setLaneDefaultSize(colIndex, width);
};
/** @public
* @param {boolean} preserve=false
*/
Core.prototype.preserveProportion = function (preserve) {
	this._preserveProportion = preserve === true;
};
/** @public
* @return {boolean} Proportion preserved
*/
Core.prototype.isProportionPreserved = function () {
	return this._preserveProportion;
};
/** @public
* @param {boolean} preserve=false
*/
Core.prototype.preserveGridSize = function (preserve) {
	this._preserveGridSize = preserve === true;
};
/** @public
* @param {number} colIndex
* @return {!Object} Column data object
*/
Core.prototype.getColumnData = function (colIndex) {
	return this._getColumnDef(colIndex)["columnData"];
};
/** @public
* @deprecated
* @ignore
* @param {number} colIndex
* @param {Object} userData
* @return {Object}
*/
Core.prototype.setColumnData = function (colIndex, userData) {
	this._getColumnDef(colIndex)["columnData"] = userData || {};
	return userData;
};
/** @public
* @function
* @deprecated
* @ignore
* @param {number} colIndex
* @return {!Object}
*/
Core.prototype.newColumnData = Core.prototype.getColumnData;
/** @public
* @param {number} colIndex
* @param {boolean=} alwaysRender
*/
Core.prototype.setAlwaysRenderColumn = function(colIndex, alwaysRender) {
	this.getColumnData(colIndex)["alwaysRender"] = (alwaysRender !== false);
};
/** @private
* @param {number} colIndex
* @return {!boolean}
*/
Core.prototype._isAlwaysRenderColumn = function (colIndex) {
	let colData = this.getColumnData(colIndex);
	return colData["alwaysRender"] === true;
};

/** @public
* @return {boolean}
*/
Core.prototype.getFitContentWidth = function () {
	return this._fitContentWidth;
};
/** @public
* @param {boolean} bool
*/
Core.prototype.setFitContentWidth = function (bool) {
	if (this._fitContentWidth !== bool) {
		this._fitContentWidth = bool;
		this.enableClass("fit-width", this._fitContentWidth);
	}
};

/** A plugin can be loaded only once
* @public
* @param {*} pluginRef Plugin/extension instance. Plugin or extension name (string) is deprecated in favor of html import syntax
* @param {Object=} opt_options Option for initialize the given plugin
* @return {*} Return the plugin that has been successfully loaded. Returns null if the plugin is not loaded
* @example
* core.loadPlugin(new InCellEditingExtension());
*/
Core.prototype.loadPlugin = function (pluginRef, opt_options) {
	let pluginInstance = null;
	if (typeof pluginRef === "string") {
		let trNamespace = window["tr"];
		let gridNamespace = trNamespace ? trNamespace["grid"] : null;
		if(!gridNamespace) {
			return null;
		}
		if (gridNamespace[pluginRef]) {
			pluginInstance = new gridNamespace[pluginRef]();
		} else if (gridNamespace[pluginRef + "Plugin"] ) {
			pluginInstance = new gridNamespace[pluginRef + "Plugin"]();
		} else {
			return null;
		}
	} else {
		pluginInstance = pluginRef;
	}

	if (!pluginInstance || !pluginInstance["getName"]) {
		return null;
	}

	let pluginName = pluginInstance["getName"]();
	if(!pluginName) {
		return null;
	}

	let prevInstance = this._plugins[pluginName];
	if (prevInstance) {
		if(prevInstance === pluginInstance) {
			return prevInstance; // The same instance is loaded twice
		}
		console.log("WARNING: Two of the same " + pluginName + "s cannot be loaded on a single grid. The existing plugin is replaced.");
		this.unloadPlugin(pluginName);
	}

	if (pluginInstance["initialize"]) {
		pluginInstance["initialize"](this, opt_options);
	}
	this._plugins[pluginName] = pluginInstance;

	return pluginInstance;
};
/** @public
* @param {*} pluginRef Plugin name (string) or instance
* @return {*|null}
*/
Core.prototype.unloadPlugin = function (pluginRef) {
	let pluginInstance;
	if (pluginRef) {
		let pluginName = "";
		if (typeof pluginRef === "string") {
			pluginName = pluginRef;
		} else if (pluginRef["getName"]) {
			pluginName = pluginRef["getName"]();
		}

		pluginInstance = this._plugins[pluginName];
		if(!pluginInstance) {
			pluginName += "Plugin";
			pluginInstance = this._plugins[pluginName];
		}

		if (pluginInstance) {
			if (pluginInstance["unload"]) {
				pluginInstance["unload"](this);
			}

			delete this._plugins[pluginName];
		}
	}
	return pluginInstance || null;
};
/** Return null if the specified plugin has not been loaded
* @public
* @param {string} pluginName
* @return {*|null} plugin Return null if the specified plugin has not been loaded
*/
Core.prototype.getPlugin = function (pluginName) {
	let p = this._plugins[pluginName];
	if (!p) {
		p = this._plugins[pluginName + "Plugin"];
	}
	return p || null;
};
/** Get a list of all plugin instances
* @public
* @return {!Array}
*/
Core.prototype.getPlugins = function () {
	let list = [];
	for (let key in this._plugins) {
		list.push(this._plugins[key]);
	}
	return list;
};
/** Get a list of existing plugin names
* @public
* @return {!Array.<string>}
*/
Core.prototype.getPluginList = function () {
	let list = [];
	for (let key in this._plugins) {
		list.push(key);
	}
	return list;
};

/** Vertical scrollbar will automatically appear for the rest of the sections, when container of Grid does NOT have enough space for Grid's content.
* @public
* @param {(string|number|null)=} opt_frozenSection the last section that will be fixed in place. <br>
* Negative value means applying the scrollbar to all sections. <br>
* null value means disabling the feature.
*/
Core.prototype.freezeSection = function (opt_frozenSection) {
	this._fixFrozenTopSections = true;
	this._freezeTopSectionAt(opt_frozenSection);
};
/** @public
*/
Core.prototype.resetSectionFreezing = function () {
	this._fixFrozenTopSections = false;
	this._updateFrozenTopSections();
};

/** Vertical scrollbar will automatically appear for the rest of the sections, when container of Grid does NOT have enough space for Grid's content.
* @private
* @param {(string|number|null)=} opt_frozenSection the last section that will be fixed in place. <br>
* Negative value means applying the scrollbar to all sections. <br>
* null value means disabling the feature.
*/
Core.prototype._freezeTopSectionAt = function (opt_frozenSection) {
	if (typeof opt_frozenSection == "number") {
		let newIndex = opt_frozenSection + 1;
		if (newIndex === this._startVScrollbarIndex) return; // same position then do nothing
		this._startVScrollbarIndex = newIndex;
	} else {
		this._startVScrollbarIndex = -1;
		if (typeof opt_frozenSection == "string") {
			this._tempFrozenSectionRef = opt_frozenSection;
		}
	}
	this._onSectionCountChanged();
};

/** @public
* @param {number} frozenCount the amount of sections that will be fixed in place (from bottom up).
*/
Core.prototype.freezeFooter = function (frozenCount) {
	this._fixFrozenBottomSection = true;
	if (this._frozenFooterCount === frozenCount) return;
	this._setFrozenFooter(frozenCount);
	this._onSectionCountChanged();
};
/** @private
* @param {number} frozenCount
*/
Core.prototype._setFrozenFooter = function (frozenCount) {
	if (this._frozenFooterCount === frozenCount) return;
	this._frozenFooterCount = frozenCount;
	this._vscrollbar.setFooterCount(frozenCount); // Vertical scrollbar needs to know number of body sections for framing (excluding header and footer sections)
	this._hscrollbar.setFooterCount(frozenCount); // Horizontal scrollbar needs this for determining the scrollbar track width
};

/** @public
* @return {number} Total footer count
*/
Core.prototype.getFooterCount = function () {
	return this._frozenFooterCount;
};

/** Left columns will be frozen and horizontal scrollbar will automatically appear for the rest columns. This will happens only when the container of grid does NOT have enough space for the grid's content.
* @public
* @param {(number|null)=} frozenColIndex Last column index to be fixed/pinned on the left side. <br>
* Use negative value (-1) make horizontal scrollbar bar cover the first column. <br>
* Use null value to disable the scrollbar
* @param {number=} numRightColumn Number of columns on the right side to be fixed/pinned on the right side
* @example
* core.freezeColumn(0); // The first column is pinned to the left side
* core.freezeColumn(1); // The first two columns are pinned to the left side
* core.freezeColumn(-1); // Grid's horizontal scrollbar will cover entire are of the grid.
* core.freezeColumn(null); // Turn off grid's horizontal scrollbar
* core.freezeColumn(-1, 1); // The last column is pinned to the right side
* core.freezeColumn(-1, 2); // The last two columns are pinned to the right side
* core.freezeColumn(0, 2); // Grid's horizontal scrollbar will affect only the middle part
*/
Core.prototype.freezeColumn = function (frozenColIndex, numRightColumn) {
	let i, colDef;
	let colCount = this.getColumnCount();
	let leftPinnedCount = 0;
	let rightPinnedCount = 0;
	let dirty = false;
	if (frozenColIndex || frozenColIndex === 0) {
		this._hScrollbarEnabled = true;
		leftPinnedCount = (frozenColIndex >= 0) ? frozenColIndex + 1 : 0;
		if (this._pinnedLeftColumnCount !== leftPinnedCount) {
			dirty = true;
			this._pinnedLeftColumnCount = leftPinnedCount; // This variable is used for caching
		}

		for (i = 0; i < colCount; ++i) {
			colDef = this._getColumnDef(i);
			colDef["leftPinned"] = (--leftPinnedCount >= 0);
		}
	} else {
		this._hScrollbarEnabled = false;
	}
	if(numRightColumn != null) {
		rightPinnedCount = numRightColumn > 0 ? numRightColumn : 0;
		if (this._pinnedRightColumnCount !== rightPinnedCount) {
			dirty = true;
			this._pinnedRightColumnCount = rightPinnedCount; // This variable is used for caching
		}

		for (i = colCount; --i >= 0;) {
			colDef = this._getColumnDef(i);
			colDef["rightPinned"] = (--rightPinnedCount >= 0);
		}
	}

	this._onColumnCountChanged(); // Activate horizontal scrollbar and column virtualization
	this._updateScrollbarWidth(true, true);
	this._updateColumnSeparators();

	if (dirty) {
		if (!this._isEventDispatching("pinningChanged")) {
			this._dispatch("pinningChanged", {});
		}
	}
};

/** @private
* @return {number} Array of number of left pinned column count and right pinned column count
*/
Core.prototype._countPinnedLeftColumns = function () {
	let leftPinnedCount = 0;
	if(this._hScrollbarEnabled) {
		for (let i = this._columnDefs.length; --i >= 0;) {
			let colDef = this._columnDefs[i];
			if(colDef) {
				if(colDef["leftPinned"]) {
					leftPinnedCount = i + 1; // Use the first right most pinned column
					break;
				}
			}
		}
	}
	this._pinnedLeftColumnCount = leftPinnedCount; // This variable is used as a cache
	return leftPinnedCount;
};
/** @private
* @return {number} numRightColumn Number of columns on the right side to be fixed/pinned on the right side
*/
Core.prototype._countPinnedRightColumns = function () {
	let rightPinnedCount = 0;
	if(this._hScrollbarEnabled) {
		let colCount = this._columnDefs.length;
		let i;
		for(i = 0; i < colCount; ++i) {
			let colDef = this._columnDefs[i];
			if(colDef && colDef["rightPinned"]) {
				break;
			}
		}
		rightPinnedCount = colCount - i;
	}
	this._pinnedRightColumnCount = rightPinnedCount;
	return rightPinnedCount;
};

/** @public
* @param {number} colIndex
* @return {boolean} return weather colIndex in pinned or not
*/
Core.prototype.isPinnedColumn = function (colIndex) {
	return colIndex < this.getFrozenColumnCount() || colIndex >= this.getFirstPinnedRightIndex();
};
/** Returns number of pinned left columns.
* @public
* @return {number}
*/
Core.prototype.getFrozenColumnCount = function() {
	return this._hScrollbarEnabled ? this._pinnedLeftColumnCount : 0;
};
/** Returns number of pinned left columns. This is an alias to getFrozenColumnCount
* @public
* @function
* @return {number}
*/
Core.prototype.getPinnedLeftColumnCount = Core.prototype.getFrozenColumnCount;
/** Returns number of pinned right columns.
* @public
* @return {number}
*/
Core.prototype.getPinnedRightColumnCount = function() {
	return this._pinnedRightColumnCount;
};
/** @public
* @return {number} If there is no pinned right column, column count is returned
*/
Core.prototype.getFirstPinnedRightIndex = function () {
	return this.getColumnCount() - this._pinnedRightColumnCount;
};

/** Sometimes there is a change in layout without Grid noticing it (e.g tab pane is switched, or parent is hidden from view). <br>
* Hence, the layout must be manually and explicitly update by calling this method.
* @public
* @fires Core#widthChanged
*/
Core.prototype.updateLayout = function () {
	if(this._disposed || this._layoutUpdating) { return; }
	this._layoutUpdating = true;

	this._dispatch("preForcedUpdate", {});
	this._updateLayout();

	if(this._colVirtualizer.isEnabled()) {
		this._activateColumns();
	}

	this._layoutUpdating = false;
};

/** @public
* @param {boolean} updateHeight
* @param {boolean} updateWidth
*/
Core.prototype.updateScrollbars = function(updateHeight, updateWidth) {
	if (updateWidth) {
		this._updateScrollbarWidth(true, false);
	}

	if (updateHeight) {
		this._updateScrollbarHeight(true, false);
	}
};

/** Stop grid from updating layout to improve performance during multiple layout changes. Unfreeze the layout once all changes have been made.
* @public
* @param {boolean=} opt_bool
* @return {boolean} Previous freezing state (Used for restoring original state)
* @fires Core#widthChanged
* @see {@link Core#isLayoutFrozen}
*/
Core.prototype.freezeLayout = function (opt_bool) {
	opt_bool = opt_bool !== false;
	let prevState = this._frozenLayout;
	if (prevState !== opt_bool) {
		this._frozenLayout = opt_bool;

		let stp = this.getPlugin("SortableTitlePlugin");
		if(this._frozenLayout) {
			if(stp) {
				stp.freezeIndicator(true);
			}
		}

		if (!this._frozenLayout) {
			// Width has not yet changed so we need to disable it first
			this._disableEvent("widthChanged");
			this._onSectionCountChanged();
			this._onColumnCountChanged(); // Update horizontal scrollbar
			if(this._colVirtualizer.isEnabled()) {
				this._activateColumns();
			}
		}

		this._rowHeightSync = false;
		let s, section;
		let sectionCount = this._settings.length;
		for (s = 0; s < sectionCount; s++) {
			section = this._settings[s].getSection();
			section.setFrozenLayout(this._frozenLayout);
		}

		if (!this._frozenLayout) {
			this._syncLayoutToColumns(0, this.getColumnCount()); // Column layout should be synchronized after sections are unfrozen
			// Layout has been changed dispatch it
			this._disableEvent("widthChanged", false);
			this._dispatch("widthChanged", {});

			let viewChanged = this._syncRowHeights();
			if(!viewChanged) { // Always update virtualizer
				this._rowVirtualizer.update(true); // Force section activation
			}
			if(stp) {
				stp.freezeIndicator(false);
			}
		}
		this._rowHeightSync = true;
	}
	return prevState;
};
/** Deprecated. Please use freezeLayout instead
* @public
* @param {boolean=} opt_bool
* @fires Core#widthChanged
* @see {@link Core#freezeLayout}
*/
Core.prototype.unfreezeLayout = function (opt_bool) {
	this.freezeLayout(opt_bool === false);
};
/** @public
* @return {boolean}
*/
Core.prototype.isLayoutFrozen = function () {
	return this._frozenLayout;
};

/** Deprecated
* @public
* @ignore
* @function
* @param {boolean=} opt_shown
* @see {@link Core#toggleBorders}
*/
Core.prototype.showBorders = Util._deprecatedFunction;
/** Deprecated
* @public
* @ignore
* @function
* @param {boolean=} opt_hidden
* @see {@link Core#toggleBorders}
*/
Core.prototype.hideBorders = Util._deprecatedFunction;
/** Toggle borders around core grid element. WARNING: You will need to manually call updateLayout() method, if you call this method at run-time
* @public
* @param {boolean=} forcedVal If value is true, show border lines, otherwise hide them
*/
Core.prototype.toggleBorders = function(forcedVal) {
	this.toggleClass("no-borders", forcedVal != null ? !forcedVal : null);
};
/** Toggle both horizontal and vertical gridlines. This applies to ONLY content sections.
* @public
* @param {boolean=} forcedVal If value is true, show gridlines, otherwise hide them
*/
Core.prototype.toggleGridlines = function(forcedVal) {
	this.toggleClass("no-gridlines", forcedVal != null ? !forcedVal : null);
};
/** Toggle vertical gridlines. This applies to all sections.
* @public
* @param {boolean=} forcedVal If value is true, show vertical lines, otherwise hide them
*/
Core.prototype.toggleVerticalLines = function(forcedVal) {
	this.toggleClass("no-vertical-grid-line", forcedVal != null ? !forcedVal : null);
};
/** Toggle horizontal gridlines. This applies to all sections.
* @public
* @param {boolean=} forcedVal If value is true, show horizontal lines, otherwise hide them
*/
Core.prototype.toggleHorizontalLines = function(forcedVal) {
	this.toggleClass("no-horizontal-grid-line", forcedVal != null ? !forcedVal : null);
};
/** Toggle vertical gridlines. This applies to ONLY content sections.
* @public
* @param {boolean=} forcedVal If value is true, show horizontal lines, otherwise hide them
*/
Core.prototype.toggleContentVerticalLines = function(forcedVal) {
	this.toggleClass("no-content-vertical-line", forcedVal != null ? !forcedVal : null);
};


/** @public
* @return {Array.<string>}
*/
Core.prototype.getDataColumnMap = function () {
	let dataMap = [];
	for (let i = this.getColumnCount(); --i >= 0; ) {
		dataMap[i] = this.getDataColumnName(i);
	}
	return dataMap;
};
/** For indirectional one to one data binding
* @public
* @param {Array.<string>=} opt_dataMap index in the array will be matched with the display column
*/
Core.prototype.setDataColumnMap = function (opt_dataMap) {
	let i, colDef;
	if (opt_dataMap) {
		for (i = opt_dataMap.length; --i >= 0; ) {
			colDef = this._getColumnDef(i);
			colDef["dataColumnName"] = opt_dataMap[i];
		}
	} else {
		for (i = this._columnDefs.length; --i >= 0; ) {
			colDef = this._columnDefs[i];
			if (colDef && colDef["dataColumnName"] != null) {
				colDef["dataColumnName"] = null;
			}
		}
	}
};
/** @public
* @param {number} colIndex
* @return {string}
*/
Core.prototype.getDataColumnName = function (colIndex) {
	let colDef = this._getColumnDef(colIndex);
	let dataColumnName = colDef["dataColumnName"];
	return dataColumnName || "";
};
/** @public
* @param {number} colIndex
* @param {(string|null)=} opt_dataColumnName Use string for normal mapping, null to disable, and undefined for default mapping
*/
Core.prototype.setDataColumnName = function (colIndex, opt_dataColumnName) {
	this._getColumnDef(colIndex)["dataColumnName"] = opt_dataColumnName;
};

/** If the given column has no rendering handler, this default handler will be executed instead
* @public
* @param {Function} handler that takes json object as a first parameter
* @see {Core#setColumnRenderingHandler}
*/
Core.prototype.setDefaultRenderingHandler = function (handler) {
	this._renderingHandler = (typeof handler === "function") ? handler : null;
};
/** If the given column has no data binding handler, this handler will be executed instead
* @public
* @param {Function} handler that takes json object as a first parameter
* @see {Core#setColumnDataBindingHandler}
*/
Core.prototype.setDefaultDataBindingHandler = function (handler) {
	this._dataBindingHandler = (typeof handler === "function") ? handler : null;
};
/** @public
* @param {number} c
* @param {*} func
* @see {Core#setDefaultRenderingHandler}
*/
Core.prototype.setColumnRenderingHandler = function(c, func) {
	this._getColumnDef(c)["renderingHandler"] = (typeof func === "function") ? func : null;
};
/** @public
* @param {number} c
* @param {*} func
* @see {Core#setDefaultDataBindingHandler}
*/
Core.prototype.setColumnDataBindingHandler = function(c, func) {
	this._getColumnDef(c)["dataBindingHandler"] = (typeof func === "function") ? func : null;
};
/** @public
* @param {number} c
* @return {Function} func
*/
Core.prototype.getColumnRenderingHandler = function(c) {
	return this._getColumnDef(c)["renderingHandler"] || null;
};
/** @public
* @param {number} c
* @return {Function} func
*/
Core.prototype.getColumnDataBindingHandler = function(c) {
	return this._getColumnDef(c)["dataBindingHandler"] || null;
};

/** Rows will be highlighted when mouse is hoved over rows. This will works only with content section
* @public
* @param {boolean=} opt_enabled
*/
Core.prototype.enableRowHighlighting = function (opt_enabled) {
	if (opt_enabled == null) { opt_enabled = true; }

	if (this._rowHighlighting !== opt_enabled) {
		this._rowHighlighting = opt_enabled;
		let contentSections = this.getAllSections("content");
		for (let i = contentSections.length; --i >= 0; ) {
			contentSections[i].enableRowHighlighting(this._rowHighlighting);
		}
	}
};

/** Get target from the given object for hit testing
* @private
* @param {!ElementWrapper|Element|Event|MouseEvent} obj Element, Browser's Event object, Browser's Mouse Event object are all valid
* @return {Element}
*/
Core.prototype._getTarget = function (obj) {
	let targetEl = /** @type{Element} */(obj["target"]);
	if(targetEl) { // The Given object is an Event object
		let path = null;
		if(!this._element.contains(targetEl)) {
			if(obj["composedPath"]) { // Handle the case where grid is used in the Shadow Dom
				path = /** @type{Array.<Element>} */(obj["composedPath"]());
			} else { // 'path' property is an alternative to composedPath, but it is a non-standard property
				path = /** @type{Array.<Element>} */(obj["path"]);
			}
		}
		if(path) {
			return path[0] || null;
		} else {
			return targetEl;
		}
	} else {
		if(obj["getElement"]) { // The Given object is implementing ElementWrapper
			return /** @type{Element} */(obj["getElement"]());
		} else if(Util.isElement(obj)){ // The Given object is just an element
			return /** @type{Element} */(obj);
		}
	}
	return null;
};
/** Return all information, from Grid and its internal components, related to the given position.
* @public
* @param {!ElementWrapper|Element|Event|MouseEvent} obj Element, Browser's Event object, Browser's Mouse Event object are all valid
* @param {Object=} context An object that stores returned values
* @return {!Core.MouseInfo}
*/
Core.prototype.getRelativePosition = function (obj, context) {
	if(!obj) {
		return {};
	}

	let ret_obj = context || {};
	let targetEl = this._getTarget(obj);

	// Calculate object's x and y coordinates that are relative to the top left of this grid
	if(obj["target"]) { // The Given object is an Event object
		Util.getRelativePosition(/** @type{Event} */(obj), this.getElement(), ret_obj);
	} else if(targetEl) {
		Util.getRelativePosition(targetEl, this.getElement(), ret_obj);
	}

	// The coordinates must be scaled by the zoom factor
	if (this._zoomFactor !== 1) {
		ret_obj["x"] /= this._zoomFactor;
		ret_obj["y"] /= this._zoomFactor;
	}

	ret_obj["hit"] = false;
	ret_obj["grid"] = this;
	ret_obj["clientX"] = obj["clientX"];
	ret_obj["clientY"] = obj["clientY"];
	ret_obj["zoomFactor"] = this._zoomFactor;

	// Find cell element
	let cellElement = Util.closestElement(targetEl, "cell");
	if(!cellElement) {
		return ret_obj;
	}
	ret_obj["cellElement"] = cellElement;

	// Find column element
	let colElement = Util.closestElement(cellElement, "column");
	ret_obj["columnElement"] = colElement;

	// Find section element
	let sectionElement = null;
	if (colElement) {
		// Column element may not be in the document (the column could be hidden)
		sectionElement = Util.closestElement(colElement, "tr-lg");
	} else {
		// Floating cell does have the column
		sectionElement = Util.closestElement(cellElement, "tr-lg");
	}

	if(!sectionElement) {
		return ret_obj;
	}
	ret_obj["sectionName"] = sectionElement["id"] || Util.closestElement(sectionElement, "tr-vlg")["id"];

	let sectionSettings = this.getSectionSettings(ret_obj["sectionName"]);
	if (!sectionSettings) {
		return ret_obj; // Cannot find a section from the given id
	}
	sectionSettings.extendEventArg(ret_obj);

	// Find column index
	let section = sectionSettings.getSection();
	let colIndex = section.getColumnIndex(colElement || cellElement);
	if (colIndex < 0) {
		return ret_obj; // The given columnElement does not resolve in any column index
	}
	ret_obj["colIndex"] = colIndex;

	// Find row index
	if (this._hitTestY(-1, ret_obj) === false) { // hitTestY requires only sectionSettings, colIndex, and cellElement
		return ret_obj;
	}

	if(cellElement["stretched"]) {
		ret_obj["cell"] = section.getStretchedCell(colIndex, ret_obj["rowIndex"]);
	} else {
		ret_obj["cell"] = section.getCell(colIndex, ret_obj["rowIndex"]);
	}
	ret_obj["hit"] = true;
	return ret_obj;
};

/** @public
* @return {!ElementWrapper}
*/
Core.prototype.getVScrollbar = function () {
	return this._vscrollbar;
};
/** @public
* @return {!ElementWrapper}
*/
Core.prototype.getHScrollbar = function () {
	return this._hscrollbar;
};
/** @public
* @param {boolean=} opt_hide True if autohide (default) or false to keep scrollbars permanently
*/
Core.prototype.autoHideScrollbars = function(opt_hide){
	opt_hide = opt_hide != null ? opt_hide : true;
	this._vscrollbar.setAutoHide(opt_hide);
	this._hscrollbar.setAutoHide(opt_hide);
};
/** @public
* @ignore
* @return {number}
*/
Core.prototype.getVScrollStartIndex = function () {
	return this._startVScrollbarIndex;
};
/** @public
* @ignore
* @return {number}
*/
Core.prototype.getHScrollStartIndex = function () {
	return this._hScrollbarEnabled ? this._pinnedLeftColumnCount : -1;
};
/** @public
* @param {number} colIndex
* @param {boolean=} leftOfView Default is false. If true, always make to specified column at the left most of the view.
* @return {boolean} Return true, if there is any change
*/
Core.prototype.scrollToColumn = function (colIndex, leftOfView) {
	if(!this._updateScrollbar(this._hscrollbar)) {
		return false;
	}

	let colCount = this._layoutX.getLaneCount();
	if (colIndex <= 0) {
		colIndex = 0;
	} else if (colIndex >= colCount) {
		colIndex = colCount - 1;
	}
	if(this.isPinnedColumn(colIndex)) {
		return false;
	}

	let widthOffset = this._hScrollbarEnabled ? this._layoutX.getLaneStart(this._pinnedLeftColumnCount) : 0;
	let scrollLeft = this._hscrollbar.getScrollLeft();
	let viewLeft = scrollLeft + widthOffset;
	let viewWidth = this._hscrollbar.getWidth();
	let viewTopIndex = this._layoutX.hitTest(viewLeft);
	let laneLeft = this._layoutX.getLaneStart(colIndex);
	let laneSize = this._layoutX.getLaneSize(colIndex);

	if (leftOfView || laneSize >= viewWidth || colIndex <= viewTopIndex) {
		let toLeftOfView = laneLeft - widthOffset;
		if(scrollLeft !== toLeftOfView) {
			this._hscrollbar.setScrollLeft(toLeftOfView);
			return true;
		}
	} else {
		let viewRight = viewLeft + viewWidth;
		let viewRightIndex = this._layoutX.hitTest(viewRight - 0.1);
		if (colIndex >= viewRightIndex) {
			let laneRight = laneLeft + laneSize;
			let toEndOfView = scrollLeft + laneRight - viewRight; // Fit end of column into the view

			this._hscrollbar.setScrollLeft(toEndOfView);
			return true;
		}
	}
	return false;
};
/** @public
* @override
* @return {number} pixels
*/
Core.prototype.getScrollLeft = function () {
	return this._updateScrollbar(this._hscrollbar) ? this._hscrollbar.getScrollLeft() : 0;
};
/** @public
* @param {number} pixels
*/
Core.prototype.setScrollLeft = function (pixels) {
	if(!this._updateScrollbar(this._hscrollbar)) {
		return;
	}

	this._hscrollbar.setScrollLeft(pixels);
};
/** @public
* @param {number} pixels
*/
Core.prototype.scrollRight = function (pixels) {
	this.setScrollLeft(this._hscrollbar.getScrollLeft() + pixels);
};
/** @public
 * @returns {number}
 */
Core.prototype.getScrollWidth = function () {
	if(!this._updateScrollbar(this._hscrollbar)) {
		return 0;
	}
	return this._hscrollbar.getContentWidth();
};
/** Return the scroll value (in pixels) to make specified row visible in the view
* @public
* @param {Core.SectionReference} sectionRef Specify section that the given rowIndex is related to. Use null for the first scrollable section (first non-frozen section)
* @param {number} rowIndex
* @param {boolean=} topOfView=false. If true, Make to specified row at the top of the view.
* @returns {number|null} The scroll value (in pixels)
*/
Core.prototype.getYScrollVal = function (sectionRef, rowIndex, topOfView) {
	if(!this._updateScrollbar(this._vscrollbar)) {
		return null;
	}
	let section = null;
	if(sectionRef) {
		section = this.getSection(sectionRef);
	}

	let rowIndexOffset = (section) ? section.getRowOffset() : this._sectionStarts[this._startVScrollbarIndex];
	let heightOffset = this._layoutY.getLaneStart(this._startVScrollbarIndex);

	let rowCount = this._layoutY.getLaneCount();
	if (rowIndex <= 0) { rowIndex = 0; }
	else if (rowIndex >= rowCount) { rowIndex = rowCount - 1; }

	if (topOfView) {
		return this._layoutY.getLaneStart(rowIndex + rowIndexOffset) - heightOffset;
	}

	let bufferSize = 2; // for more appealing space
	let viewInfo = this.getVerticalViewInfo();
	let firstFullRow = viewInfo.firstFullRow; // TODO: Make it work in zooming mode
	if(rowIndex < firstFullRow) { // Scroll up, as the target row is outside of view
		let targetIndex = rowIndex - bufferSize; // Have some spaces at the top for more appealing visual
		if(targetIndex < 0) {
			targetIndex = 0;
		}

		return this._layoutY.getLaneStart(targetIndex + rowIndexOffset) - heightOffset;
	}

	let lastFullRow = viewInfo.lastFullRow;
	if (rowIndex > lastFullRow) { // Scroll down, as the target row is outside of view
		let targetScroll = this._layoutY.getLaneStart(rowIndex + rowIndexOffset + bufferSize + 1);

		return targetScroll - viewInfo.viewHeight - heightOffset;
	}

	return null;
};
/** Scroll up or down to make specified row visible in the view
* @public
* @param {Core.SectionReference} sectionRef Specify section that the given rowIndex is related to. Use null for the first scrollable section (first non-frozen section)
* @param {number} rowIndex
* @param {boolean=} topOfView=false. If true, Make to specified row at the top of the view.
*/
Core.prototype.scrollToRow = function (sectionRef, rowIndex, topOfView) {
	let scrollVal = this.getYScrollVal(sectionRef, rowIndex, topOfView);
	if(scrollVal !== null) {
		this._vscrollbar.setScrollTop(scrollVal);
	}
};
/** @public
* @return {Object}
*/
Core.prototype.getVerticalViewInfo = function() {
	let rowIndexOffset = this._sectionStarts[this._startVScrollbarIndex];
	let heightOffset = this._layoutY.getLaneStart(rowIndexOffset);
	let viewHeight = this._vscrollbar.getHeight();
	let viewTop = this._vscrollbar.getScrollTop() + heightOffset;
	let viewBottom = viewTop + viewHeight;

	let topRowIndex = this._layoutY.hitTest(viewTop) - rowIndexOffset;
	let bottomRowIndex = this._layoutY.hitTest(viewBottom - 0.1);
	if(bottomRowIndex < 0) { // view is larger than existing row count
		bottomRowIndex = this._layoutY.getLaneCount() - 1;
	}
	bottomRowIndex -= rowIndexOffset;

	let laneTop = this._layoutY.getLaneStart(topRowIndex + rowIndexOffset);
	let laneBottom = this._layoutY.getLaneEnd(bottomRowIndex + rowIndexOffset);

	return {
		topRowIndex: topRowIndex,
		firstFullRow: (laneTop < viewTop) ? topRowIndex + 1 : topRowIndex,
		lastFullRow: (laneBottom > viewBottom) ? bottomRowIndex - 1 : bottomRowIndex,
		bottomRowIndex: bottomRowIndex,
		viewHeight: viewHeight,
		viewTop: viewTop,
		viewBottom: viewBottom
	};
};
/** @public
* @return {Object} Returns null if vscrollbar does not exists
*/
Core.prototype.getVScrollView = function () {
	if(!this._updateScrollbar(this._vscrollbar)) {
		return null;
	}

	let rowCount = this._layoutY.getLaneCount();
	if (rowCount === 0) { return null; }

	let rowIndexOffset = this._sectionStarts[this._startVScrollbarIndex];
	let heightOffset = this._layoutY.getLaneStart(rowIndexOffset);
	let viewTop = this._vscrollbar.getScrollTop() + heightOffset;
	let viewHeight = this._vscrollbar.getHeight();
	let viewBottom = viewTop + viewHeight;
	let topRowIndex = this._layoutY.hitTest(viewTop);
	let bottomRowIndex = this._layoutY.hitTest(viewBottom - 0.1);

	let laneTop = this._layoutY.getLaneStart(topRowIndex + rowIndexOffset);
	let laneBottom = this._layoutY.getLaneEnd(bottomRowIndex + rowIndexOffset);

	let prevAnchor = viewTop - viewHeight;
	let prevPage = (prevAnchor <= heightOffset) ? rowIndexOffset : this._layoutY.hitTest(prevAnchor) + 1;
	let nextAnchor = viewBottom + viewHeight;
	let trackSize = this._layoutY.getTrackSize();
	let nextPage = (nextAnchor >= trackSize) ? rowCount - 1 : this._layoutY.hitTest(nextAnchor - 0.1) - 1;

	topRowIndex -= rowIndexOffset;
	bottomRowIndex -= rowIndexOffset;
	let obj = {};
	obj["viewTop"] = viewTop - heightOffset;
	obj["viewBottom"] = viewBottom - heightOffset;
	obj["topRowIndex"] = topRowIndex;
	obj["bottomRowIndex"] = bottomRowIndex;
	obj["firstFullRow"] = (laneTop < viewTop) ? topRowIndex + 1 : topRowIndex;
	obj["lastFullRow"] = (laneBottom > viewBottom) ? bottomRowIndex - 1 : bottomRowIndex;
	obj["prevPageRowIndex"] = prevPage - rowIndexOffset;
	obj["nextPageRowIndex"] = nextPage - rowIndexOffset;
	return obj;
};
/** @public
* @override
* @return {number} pixels
*/
Core.prototype.getScrollTop = function () {
	return this._updateScrollbar(this._vscrollbar) ? this._vscrollbar.getScrollTop() : 0;
};
/** @public
* @param {number} pixels
*/
Core.prototype.setScrollTop = function (pixels) {
	if(!this._updateScrollbar(this._vscrollbar)) {
		return;
	}

	this._vscrollbar.setScrollTop(pixels);
};
/** @public
* @param {number} pixels
*/
Core.prototype.scrollDown = function (pixels) {
	this.setScrollTop(this._vscrollbar.getScrollTop() + pixels);
};
/** @public
 * @returns {number}
 */
Core.prototype.getScrollHeight = function () {
	if(!this._updateScrollbar(this._vscrollbar)) {
		return 0;
	}
	return this._vscrollbar.getContentHeight();
};
/** @public
*/
Core.prototype.restoreScrollbars = function () {
	this._vscrollbar.restoreTrackPosition();
};

/** @public
* @ignore
* @param {boolean} bool
*/
Core.prototype.enableAutoSetDataSource = function (bool) {
	this._autoSetDataSource = bool;
};

/** @public
* @return {!Virtualizer}
*/
Core.prototype.getRowVirtualizer = function () {
	return this._rowVirtualizer;
};


/** @public
* @return {!Virtualizer}
*/
Core.prototype.getColumnVirtualizer = function () {
	return this._colVirtualizer;
};

/** No runtime support
* @public
* @param {boolean} opt_enabled
*/
Core.prototype.enableRowVirtualization = function (opt_enabled) {
	this._rowVirtualizer.enable(opt_enabled);
	if (this._rowVirtualizer.isEnabled()) {
		if (this._vscrollbar.isEnabled()) {
			this._rowVirtualizer.activate();
		} else {
			this._rowVirtualizer.deactivate();
		}
	} else {
		this._rowVirtualizer.deactivate();
	}
};

/** @public
* @return {boolean}
*/
Core.prototype.isRowVirtualizationEnabled = function () {
	return this._rowVirtualizer.isEnabled();
};

/** No runtime support
* @public
* @param {boolean=} opt_enabled
*/
Core.prototype.enableColumnVirtualization = function (opt_enabled) {
	this._colVirtualizer.enable(opt_enabled);

	let colVir = this._colVirtualizer.isEnabled();
	for (let i = this._settings.length; --i >= 0; ) {
		let sectionSettings = this._settings[i];
		sectionSettings.enableColumnVirtualization(colVir);
	}
};

/** @public
* @return {boolean}
*/
Core.prototype.isColumnVirtualizationEnabled = function () {
	return this._colVirtualizer.isEnabled();
};

/** Scrolling step will be multiple of the default row height in pixels. For example, if default row height is 32 pixel and row scrolling step is 2, <br>
* then the actual scrolling step for vertical scrollbar will be (32 * 2) = 64 pixels
* @public
* @param {number} rowCount Number of rows for a single scrolling step. Use zero to disable this feature.
*/
Core.prototype.setRowScrollingStep = function (rowCount) {
	this._rowScrollingStep = (rowCount > 0) ? rowCount : 0;
	this._vscrollbar.setScrollingStep(this._layoutY.getDefaultLaneSize() * this._rowScrollingStep);
};

/** Allows other elements such as background to accept wheel event and scroll grid's vertical scrollbar
* @public
* @return {function(Event)}
*/
Core.prototype.getVMouseWheelHandler = function () {
	return this._vscrollbar.getDefaultMouseWheelLogic();
};
/** Synchronize two vertical scrollbars of two grid, by hiding its scrollbar and using the one from the given grid
* @public
* @param {Core} subGrid
*/
Core.prototype.synchronizeVScrollbar = function (subGrid) {
	subGrid.unlisten("mousemove", subGrid._onMouseMove);
	subGrid.listen("mousemove", this._onMouseMove);

	let vscrollbar = subGrid.getVScrollbar();
	vscrollbar._mainScrollbar = this._vscrollbar; // HACK

	vscrollbar.setStyle("visibility", "hidden");
	vscrollbar.setStyle("pointerEvents", "none");
	vscrollbar.disableMouseWheel(); // Disable sub-grid wheel behavior
	vscrollbar.attachToExternalElement(this._vscrollbar.getParent()); // MouseWheel event is still available on the main

	// TODO: Check if we need to re-append this._vscrollbar to move it to the front over other grid elements.
	this._vscrollbar.listen("scroll", subGrid._onSyncVScroll);
};

/** Synchronize two horizontal scrollbars of two grid, by hiding its scrollbar and using the one from the given grid
* @public
* @param {Core} subGrid
*/
Core.prototype.synchronizeHScrollbar = function (subGrid) {
	subGrid.unlisten("mousemove", subGrid._onMouseMove);
	subGrid.listen("mousemove", this._onMouseMove);

	let hscrollbar = subGrid.getHScrollbar();
	hscrollbar._mainScrollbar = this._hscrollbar; // HACK

	hscrollbar.setStyle("visibility", "hidden");
	hscrollbar.setStyle("pointerEvents", "none");
	hscrollbar.disableMouseWheel(); // Disable sub-grid wheel behavior
	hscrollbar.attachToExternalElement(this._hscrollbar.getParent()); // MouseWheel event is still available on the main

	// TODO: Check if we need to re-append this._hscrollbar to move it to the front over other grid elements.
	this._hscrollbar.listen("scroll", subGrid._onSyncHScroll);
};

/** Fires data binding event without actual change in the data source. <br>
* This will force visual elements to be re-rendered with the latest data in the data source.
* @public
* @param {Core.SectionReference=} sectionRef If not defined, "content" is used
* @param {number=} fromRowIndex INCLUSIVE If the value is undefined, the first row index will be used
* @param {number=} lastRowIndex INCLUSIVE If the value is undefined, the last row index will be used
* @param {Object=} userParam Addtional parameters to be fired with the event
* @fires Core#beforeContentBinding
* @fires Core#preSectionDataBinding
* @fires Core#columnDataBinding
* @fires Core#postSectionDataBinding
* @fires Core#afterContentBinding
*/
Core.prototype.updateRowData = function (sectionRef, fromRowIndex, lastRowIndex, userParam) {
	let ss = this.getSectionSettings(sectionRef || "content");
	if (ss) {
		ss.updateRowData(fromRowIndex, lastRowIndex, userParam);
	}
};
/** Re-render all sections (headers and contents) in grid. Layout will not be affected by this method
* @public
*/
Core.prototype.rerender = function () {
	let sectionCount = this._settings.length;
	for (let s = 0; s < sectionCount; s++) { // for each section
		this._settings[s].rerender();
	}
};
/** Set a timer to call updateRowData only once to avoid performance issue due to multiple call of updateRowData()
* @public
*/
Core.prototype.requestRowRefresh = function() {
	if(!this._rowRefreshTimer) {
		this._rowRefreshTimer = setTimeout(this._onRowRefresh, 100);
	}
};
/** Set a timer to call updateScrollbarHeight only once to avoid performance issue due to multiple call of _updateScrollbarHeight()
* @private
*/
Core.prototype._requestScrollbarUpdate = function() {
	this._updateScrollbarHeight(true, true);
	this._rowHeightTimerId = 0;
};

/** prevent bind data process
* @public
* @param {boolean=} disabled
*/
Core.prototype.deactivateRendering = function(disabled) {
	disabled = disabled !== false;

	// if option value not changed then do nothing
	if (disabled == this._noBinding) return;

	this._noBinding = disabled;
	let contentSection = this.getSectionSettings("content");
	if (!contentSection) return;

	contentSection.disableDataBinding(disabled);

	if (!disabled) {
		this.requestRowRefresh();
	}
};

/** @private */
Core.prototype._onRowRefresh = function() {
	this._rowRefreshTimer = 0;
	this.updateRowData();
};

/** @public
* @ignore
* @param {boolean=} opt_bool
*/
Core.prototype.disableLazyRendering = function(opt_bool) {
	console.log("WARNING: This function is deprecated.");
};

/** @param {Core.SectionReference} sectionRef
* @param {number|string} colRef Either column index or data column name
* @param {number|string} rowRef Row Index of the display section or row id of the data table
* @return {*} Returns undefined if not success
*/
Core.prototype.getData = function (sectionRef, colRef, rowRef) {
	let sectionSettings = this.getSectionSettings(sectionRef);
	if (!sectionSettings) { return null; }

	let colName = this._toDataColumnName(colRef);

	return sectionSettings.getData(colName, rowRef);
};
/** Do not used this method too often. This method is slow. It is always better to set data directly to the data table
* @param {Core.SectionReference} sectionRef
* @param {number|string} colRef Either column index or data column name
* @param {number|string} rowRef Row Index of the display section or row id of the data table
* @param {*} data
* @return {boolean} True for successful data setting
*/
Core.prototype.setData = function (sectionRef, colRef, rowRef, data) {
	let sectionSettings = this.getSectionSettings(sectionRef);
	if (sectionSettings && sectionSettings.isDataBindable()) {
		let colName = this._toDataColumnName(colRef);
		return sectionSettings.setData(colName, rowRef, data);
	}
	return false;
};

/** Do not used this method too often. This method is slow. It is always better to set data directly to the data table
* @param {Core.SectionReference} sectionRef
* @param {number|string} rowRef Row Index of the display section or row id of the data table
* @return {boolean} True for successful data setting
*/
Core.prototype.removeRowData = function (sectionRef, rowRef) {
	let sectionSettings = this.getSectionSettings(sectionRef);
	if (sectionSettings && sectionSettings.isDataBindable()) {
		return sectionSettings.removeRowData(rowRef);
	}
	return false;
};

/** @param {string} method */
Core.prototype.setHidingMethod = function (method) {
	if(!method) { method = ""; }
	if(this._hidingMethod !== method) {
		this._hidingMethod = method;
		let sections = this.getAllSections();
		for(let i = sections.length; --i >= 0;) {
			sections[i].setHidingMethod(method);
		}
	}
};

/** @public
* @ignore
* @return {!TrackLayout}
*/
Core.prototype.getHorizontalLayout = function () {
	return this._layoutX;
};
/** @public
* @ignore
* @param {number} colIndex
* @param {boolean} bool
* @param {number} flag
* @return {boolean}
*/
Core.prototype.setColumnVisibility = function (colIndex, bool, flag) {
	if (this._layoutX.hideLane(colIndex, !bool, flag)) {
		this._syncLayoutToColumns(colIndex, colIndex + 1);

		if (bool) { // If column turns back to visible, update ui. TODO: This will cause a rendering issue for column grouping
			this.requestRowRefresh();
		}
		if (this._hasListener("columnVisibilityChanged")) {
			let e = {};
			e["colIndex"] = colIndex;
			e["hidden"] = !bool;
			this._dispatch("columnVisibilityChanged", e);
		}
		return true;
	}
	return false;
};
/** Get visibility state for the specified flag. This is for internal use only
* @public
* @ignore
* @param {number} colIndex
* @param {number=} flag
* @return {boolean}
*/
Core.prototype.getColumnVisibility = function (colIndex, flag) {
	return this._layoutX.getLaneVisibilityBit(colIndex, flag);
};

/** @public
* @param {number} size
* @return {boolean}
*/
Core.prototype.reserveRightSpace = function (size) {
	if(this._reservedSpace !== size) {
		this._reservedSpace = size;
		for(let i = this._settings.length; --i >= 0;) {
			this._settings[i].getSection().reserveRightSpace(size);
		}
		this._onColumnCountChanged(); // Activate horizontal scrollbar and column virtualization
		this._updateScrollbarWidth(true, true);
		return true;
	}
	return false;
};

/** Get the hidden input. This input allows grid to receive keyboard input
* @public
* @ignore
* @param {boolean} firstInput
* @return {!Element}
*/
Core.prototype.getHiddenInput = function (firstInput) {
	return firstInput ? this._firstHiddenInput : this._lastHiddenInput;
};

/** Focus grid element without bringing grid into window's view. This is useful when grid is very wide or tall, since window can be scrolled to focused element by default in some browsers.
* @public
* @see {@link https://developer.mozilla.org/en/docs/Web/API/HTMLElement/focus}
* @see {@link http://help.dottoro.com/ljqmdirr.php}
*/
Core.prototype.focus = function () {
	let elem = this._lastHiddenInput;
	let activeElem = _getActiveElement(elem);
	if(elem && elem !== activeElem) {
		let x = window.pageXOffset;
		let y = window.pageYOffset;
		if(elem["setActive"]) { // Non-standard IE specific method
			try { // MON-909
				elem["setActive"]();
			} catch (error) {
				elem["focus"]({"preventScroll": true});
			}
		} else {
			elem["focus"]({"preventScroll": true});
		}
		if(y !== window.pageYOffset || x !== window.pageXOffset) {
			window.scrollTo(x, y); // Restore scrolling
		}
	}
};

/** @public
* @return {boolean}
*/
Core.prototype.isBinding = function() {
	return this._dispatchingDataChanged;
};

/**
 * Hide the specified rows by row id
 * @param {number | string} rowId rowId of the row to hide
 * @param {boolean=} hidden=true If the value is false, the specified rows will be shown.
 * @public
 */
Core.prototype.hideRow = function (rowId, hidden) {
	this._dataSource.hideRow(rowId, hidden);
};
/**
 * Hide the specified rows by rowIds
 * @param {number | string | Array.<number | string>} rowIds rowId(s) of the rows to hide
 * @param {boolean=} hidden=true If the value is false, the specified rows will be shown.
 * @public
 */
Core.prototype.hideRows = function (rowIds, hidden) {
	if (!Array.isArray(rowIds)) {
		this._dataSource.hideRow(/** @type{number | string} */(rowIds), hidden);
	} else {
		this._dataSource.hideRows(rowIds, hidden);
	}
};

/**
 * Show the specified rows by rowIds. To show all rows, use unhideAllRows method for better performance.
 * @param {number | string | Array.<number | string>} rowIds rowId(s) of the rows to show
 * @param {boolean=} shown=true If the value is false, the specified rows will be hidden.
 * @public
 */
Core.prototype.showRows = function (rowIds, shown) {
	if (!Array.isArray(rowIds)) {
		this._dataSource.showRows([rowIds], shown);
	} else {
		this._dataSource.showRows(rowIds, shown);
	}
};

/**
 * Unhide all hidden rows in the Grid
 * @public
 */
Core.prototype.unhideAllRows = function () {
	this._dataSource.unhideAllRows();
};

/**
 * @public
 * @return {boolean}
 */
Core.prototype.hasHiddenRow = function () {
	return this._dataSource.hasHiddenRow();
};

/** @public
 * @ignore
 * @param {Array=} columns
 */
Core.prototype.initSimpleTable = function (columns) {
	let colCount = 0;
	if(!columns && !this.getColumnCount()) {
		columns = ["Default"]; // Row cannot be added unless there is at least 1 column
	}

	if(columns) {
		colCount = columns.length;
		this.setColumnCount(colCount);
		this.setDataColumnMap(columns);
		for(let c = 0; c < colCount; ++c) {
			this.setColumnField(c, columns[c]);
		}
	}

	let titleSect = this.addSection("title");
	titleSect.addRow(1);

	if(!this._renderingHandler) { // Render title if there is no default rendering
		for(let c = 0; c < colCount; ++c) {
			titleSect.setCellContent(c, 0, columns[c]);
		}
	}

	this.addSection("content");
	let titleSectionSetting = this.getSectionSettings("title");
	titleSectionSetting.disableDataBinding();

	let contentSectionSetting = this.getSectionSettings("content");
	contentSectionSetting.setAutoSyncRowCount(true);
	contentSectionSetting.setParent(titleSectionSetting);

	if(!this._dataBindingHandler) {
		this.setDefaultDataBindingHandler(Core._defaultDataBinding);
	}
};
/** @private
 * @param {Object} e
 */
Core._defaultDataBinding = function(e) {
	e.cell.setContent(e.dataValue);
};

/** @public
* @param {number} colIndex
* @param {boolean=} selected
*/
Core.prototype.selectColumn = function (colIndex, selected) {
	for (let i = this._settings.length; --i >= 0; ) {
		this._settings[i].getSection().selectColumn(colIndex, selected);
	}
	this.updateColumnBounds();
};
/** @public
* @param {number} colIndex
* @return {boolean}
*/
Core.prototype.isSelectedColumn = function (colIndex) {
	let first = this._settings[0] || null;
	if(first) {
		return first.getSection().isSelectedColumn(colIndex);
	}
	return false;
};

/** @public
*/
Core.prototype.updateColumnBounds = function () {
	if(this._columnBoundConflator.conflate()) {
		return;
	}

	let sectCount = this._settings.length;
	if(!sectCount) {
		return;
	}

	// Collecting column selection and selection ranges
	let selRanges = [];
	let pair = null;
	let colCount = this.getColumnCount();
	let selIndices = [];
	let i;
	for(i = 0; i < colCount; i++) {
		if(this.isSelectedColumn(i)) {
			selIndices.push(i);
			if(!pair) {
				pair = [i, -1];
			}
		} else if(pair) {
			pair[1] = i - 1;
			selRanges.push(pair);
			pair = null;
		}
	}
	if(pair) {
		pair[1] = colCount - 1;
		selRanges.push(pair);
		pair = null;
	}

	let arg = {
		"selectedColumns": selIndices,
		"selectionRanges": selRanges
	};
	this._dispatch("beforeColumnBoundUpdate", arg);

	// Calculate position from ranges
	let rangeCount = selRanges.length;
	let posAry = [];
	let noBorderAry = [];
	let topSectionSettings = this._settings[0];
	let section = topSectionSettings.getSection();
	for(i = 0; i < rangeCount; ++i) {
		pair = selRanges[i];
		let positions = [0, 0];
		let noBorders = [false, false];
		section.calculateColumnBounds(pair[0], pair[1], positions, noBorders);
		if(positions[0] < positions[1]) {
			posAry.push(positions);
			noBorderAry.push(noBorders);
		}
	}

	// Render column bounds
	let topPx = 0;
	let topBoundIdx = -1;
	if(arg["topBoundRowIndex"] != null) {
		topBoundIdx = +arg["topBoundRowIndex"];
	}
	if(topBoundIdx >= 0 && topSectionSettings.getType() === "title") {
		topPx = this._layoutY.getLaneStart(topBoundIdx);
	}
	section.updateColumnBounds(posAry, noBorderAry, topPx);
	for(i = 1; i < sectCount; i++) {
		section = this._settings[i].getSection();
		section.updateColumnBounds(posAry, noBorderAry);
	}
};
/* @private
*/
Core.prototype._updateColumnSeparators = function() {
	let sectCount = this._settings.length;
	if(!sectCount) {
		return;
	}

	for(let i = 0; i < sectCount; i++) {
		let section = this._settings[i].getSection();
		if (section) {
			section.updateColumnSeparators();
		}
	}
};

/** @public
* @param {number} startColIndex INCLUSIVE
* @param {number} endColIndex INCLUSIVE
* @return {Object}
*/
Core.prototype.getColumnRect = function (startColIndex, endColIndex) {
	let gridWidth = this.getElement().clientWidth;
	let scrollViewSize = gridWidth;
	let hscrollbar = this._hscrollbar;
	if (hscrollbar.isActive()) {
		scrollViewSize = hscrollbar.getTrackSize();
	}

	let left = 0;
	let width = 0;
	if (this._settings.length) {
		let section = this._settings[0].getSection();
		let positions = [0, 0];
		section.calculateColumnBounds(startColIndex, endColIndex, positions, []);
		left = positions[0];
		width = positions[1] - left;
	}
	let rect = {
		top: 0, // TODO: Support column grouping
		left: left,
		width: width,
		height: this._layoutY.getTrackSize(),
		leftPinnedIndex: this._pinnedLeftColumnCount - 1,
		rightPinnedIndex: this.getFirstPinnedRightIndex(),
		scrollViewSize: scrollViewSize,
		gridWidth: gridWidth
	};
	return rect;
};

/** @public
* @param {number} startRowIndex INCLUSIVE
* @param {number} endRowIndex INCLUSIVE
* @return {Object}
*/
Core.prototype.getRowRect = function (startRowIndex, endRowIndex) {
	let contentWidth = this._layoutX.getTrackSize();
	let gridElem = this.getElement();
	let gridWidth = gridElem.clientWidth;
	contentWidth = contentWidth < gridWidth ? contentWidth : gridWidth;

	let section = this.getSection("content");
	let rowOffset = section ? section.getRowOffset() : 1;
	let layoutY = this._layoutY;
	let sectionOffsetPx = layoutY.getLaneStart(rowOffset);

	let scrollTop = 0;
	let scrollViewSize = section.getHeight();
	let vscrollbar = this._vscrollbar;
	if (vscrollbar.isActive()) {
		scrollTop = vscrollbar.getScrollTop();
		scrollViewSize = vscrollbar.getTrackSize();
	}
	let scrollViewBottom = scrollViewSize + sectionOffsetPx;

	let topPx = layoutY.getLaneStart(startRowIndex + rowOffset) - scrollTop;
	let bottomPx = layoutY.getLaneEnd(endRowIndex + rowOffset) - scrollTop;
	if (topPx < sectionOffsetPx) {
		topPx = sectionOffsetPx;
	} else if (topPx > scrollViewBottom) {
		topPx = scrollViewBottom;
	}
	if (bottomPx < sectionOffsetPx) {
		bottomPx = sectionOffsetPx;
	} else if (bottomPx > scrollViewBottom) {
		bottomPx = scrollViewBottom;
	}

	// Change origin point from grid to section
	topPx -= sectionOffsetPx;
	bottomPx -= sectionOffsetPx;

	let rect = {
		top: topPx,
		left: 0,
		width: contentWidth,
		height: bottomPx - topPx,
		scrollViewSize: scrollViewSize,
		scrollViewTop: sectionOffsetPx,
		scrollViewBottom: scrollViewBottom
	};
	return rect;
};
/** @public
* @param {Object} configObj
* @return {!Object}
*/
Core.prototype.normalizeConfig = function (configObj) {
	if(!configObj) {
		configObj = {};
	}
	let columns = configObj["columns"];
	let colCount = columns ? columns.length : 0;
	let i, column;

	let freezingIndex = configObj["columnFreezing"];
	if(freezingIndex == null) {
		freezingIndex = configObj["freezeColumn"];
	}
	let pinnedRightColumns = configObj["pinnedRightColumns"] || configObj["frozenRightColumns"];
	if (pinnedRightColumns) {
		delete configObj["pinnedRightColumns"];
		delete configObj["frozenRightColumns"];

		configObj["columnVirtualRendering"] = false;
		configObj["columnVirtualization"] = false;

		for(i = colCount; --i >= 0;) {
			if(--pinnedRightColumns >= 0) {
				column = columns[i];
				if(column) {
					column["rightPinned"] = true;
				}
			}
		}
	}
	if(freezingIndex != null) {
		delete configObj["columnFreezing"];
		delete configObj["freezeColumn"];

		if(freezingIndex >= 0) {
			if(columns) {
				column = columns[freezingIndex];
				if(column) {
					column["leftPinned"] = true;
				}
			}
		}
	}

	return configObj;
};
/** @public
* @param {GroupDefinitions} groupDefs
*/
Core.prototype.setColumnGrouping = function (groupDefs) {
	this._groupDefs = groupDefs || null;
};
/** @public
* @param {string|number} colRef
* @return {string} Return group id of immediate group parent. Return empty string, if no parent is found.
*/
Core.prototype.getColumnGroupParentId = function (colRef) {
	if(this._groupDefs) {
		let colId = (typeof colRef === "number") ? this.getColumnId(colRef) : colRef;
		return this._groupDefs.getParentId(colId);
	}
	return "";
};

/** @public
* @param {string} groupId
* @return {Array.<string>} Return array of column id. Return null if the specified group has no child
*/
Core.prototype.getColumnGroupChildIds = function (groupId) {
	if(this._groupDefs) {
		let ary = this._groupDefs.getLeafDescendants(groupId);
		if(ary && ary.length > 0) {
			return ary;
		}
	}
	return null;
};

/** @public
* @description Get a list of objects with column id and column index in sorted order
* @param {Array.<string>} colIds
* @param {Object=} columnMap
* @return {!Array.<string>} Return column array with corresponding order to UI
* @example
* core.getValidColumnList(["c1","c2","c5"]); // Get list of valid columns
* core.getValidColumnList(["c1","c2","c5"],{ "c2":true, "c5":true }); // Get list of valid columns from specific mapping
*/
Core.prototype.getValidColumnList = function (colIds, columnMap) {
	let colList = [];
	if(!colIds){
		return colList;
	}
	if(!columnMap){
		columnMap = this.createColumnMap(colIds);
	}
	let colCount = this.getColumnCount();
	for(let c = 0; c < colCount; ++c) {
		let id = this._getColumnDef(c)["id"];
		if(columnMap[id] != null){
			colList.push({"index": c, "id": id});
		}
	}
	return colList;
};

/** @public
* @description Create mapping object from an array of strings or column indices
* @param {Array.<string|number>=} colRefs
* @return {!Object} Column mapping object
*/
Core.prototype.createColumnMap = function (colRefs) {
	if(!colRefs){
		colRefs = this.getColumnIds();
	}
	let mappingObj = {};
	let count = colRefs.length;
	for(let i = 0; i < count; i++){
		let colId = colRefs[i];
		if(typeof colId ===  "number"){
			colId = this.getColumnId(colId);
		}
		if(colId){
			mappingObj[colId] = true;
		}
	}
	return mappingObj;
};

/** @public
* @param {string} batchType
* @return {boolean}
* @fires Core#beforeBatchOperation
*/
Core.prototype.startBatch = function (batchType) {
	if(!batchType){
		return false;
	}
	if(!this._batches){
		this._batches = {};
	}
	this._batches[batchType] = batchType;
	// The "columnVisibilityChanged" event is blocked while the "reset" batch operation is in progress, so this event will not be fired until the batch operation succeeds.
	if(batchType === "reset") {
		this._disableEvent("columnVisibilityChanged", true);
	}
	this._dispatch("beforeBatchOperation", { batches: this._batches, batchType: batchType });
	return true;
};
/** @public
* @param {string} batchType
* @return {boolean}
* @fires Core#afterBatchOperation
*/
Core.prototype.stopBatch = function (batchType) {
	if(!batchType){
		return false;
	}
	this._dispatch("afterBatchOperation", { batches: this._batches, batchType: batchType });
	if(batchType === "reset") {
		this._disableEvent("columnVisibilityChanged", false);
	}

	delete this._batches[batchType];
	if(isEmptyObject(this._batches)){
		this._batches = null;
	}
	return true;
};
//#endregion Public Methods

//#region Private Methods
/** @private
* @param {string} type
* @param {!Object} eventArg
*/
Core.prototype._dispatchColumnEvent = function (type, eventArg) {
	let batches = this._batches;
	if(batches){
		eventArg["batches"] = batches;
	}
	this._dispatch(type, eventArg);
};
/** @private
*/
Core.prototype._dispatchColumnPositionChanged = function () {
	if(this._columnPositionConflator.conflate()) {
		return;
	}

	this._dispatch("columnPositionChanged", { sender: this });
};

/** @private
*/
Core.prototype._dispatchRowPositionChanged = function () {
	if(this._rowPositionConflator.conflate()) {
		return;
	}

	this._dispatch("rowPositionChanged", { sender: this });
};
/** @private
* @return {!Array.<ILayoutGrid>}
*/
Core.prototype._getAllSections = function () {
	let sections = [];
	let sectionCount = this._settings.length;
	for (let i = 0; i < sectionCount; i++) {
		sections.push(this._settings[i].getSection());
	}
	return sections;
};
/** @private
* @param {number} increment
* @param {(MouseEvent|Core.SectionReference)=} identifier This can be MouseEvent, index, or element id or pointer reference
* @param {string=} opt_type The next section with the specified type. Default is any type
* @return {SectionSettings}
*/
Core.prototype._getNextSection = function (increment, identifier, opt_type) {
	let nextAt = this.getSectionIndex(identifier) + increment;
	if(opt_type == null) { return this.getSectionSettings(nextAt); }

	opt_type = this._toSectionType(opt_type);
	let count = this._settings.length;
	while (nextAt >= 0 && nextAt < count) {
		let sectionSettings = this._settings[nextAt];
		if (opt_type === sectionSettings.getType()) {
			return /**@type{SectionSettings}*/(sectionSettings);
		}
		nextAt += increment;
	}
	return null;
};

/** @private
* @param {*} val
* @return {string}
*/
Core.prototype._toSectionType = function (val) {
	if(Core._sectionType[/** @type{string} */(val)]) { return /** @type{string} */(val); }

	if (Util.isString(val)) {
		let lowerCaseStr = val.charAt(0).toLowerCase();
		switch (lowerCaseStr) {
		case "t": return "title";
		case "h": return "header";
		case "s": return "header";
		case "f": return "footer";
		}
	}

	return "content";
};

/** @private
* @param {string=} opt_type Possible value are "content", "title", "header", and "footer"
* @param {string=} sectionName
* @return {SectionSettings|null}
*/
Core.prototype._newSection = function (opt_type, sectionName) {
	if(sectionName) {
		for(let i = this._settings.length; --i >= 0;) {
			if (this._settings[i].getId() === sectionName) { // check if section name isn't already used
				return null;
			}
		}
	}

	opt_type = this._toSectionType(opt_type);
	let sectionOpt = {
		"horizontalScrollbar": this._hscrollbar
	};
	let section = (opt_type === "content") ? new VirtualizedLayoutGrid(sectionOpt) : new LayoutGrid(sectionOpt);
	section._setContext(this);

	let sectionSettings = new SectionSettings(section);
	sectionSettings.setId(sectionName);
	sectionSettings.setType(opt_type);

	sectionSettings.listen("dataChanged", this._onSectionDataChanged);
	sectionSettings.listen("rowAvailable", this._onRowAvailable);

	let rowHighlighting = (opt_type === "content") ? this._rowHighlighting : false;
	section.enableRowHighlighting(rowHighlighting);
	section.enableShading(false); // No alternate shading
	section.setHidingMethod(this._hidingMethod);

	sectionSettings.enableColumnVirtualization(this._colVirtualizer.isEnabled());
	section.setColumnCount(this._layoutX.getLaneCount());
	section.reserveRightSpace(this._reservedSpace);
	// TODO: Add any column customization
	this._syncLayoutToSectionColumns(section);
	this._syncStyleToSectionColumns(section, opt_type);

	if (opt_type !== this._rowHeightException) { section.setDefaultRowHeight(this._layoutY.getDefaultLaneSize()); }
	section.listen("rowHighlighted", this._onRowHightlighted);
	section.listen("rowCountChanged", this._onRowCountChanged);
	section.listen("rowHeightChanged", this._onRowHeightChanged);

	return sectionSettings;
};

/** @private
* @param {ILayoutGrid} section
*/
Core.prototype._putToLast = function(section) {
	let sectionCount = this._settings.length;
	if (sectionCount === 0) {
		section.insertBefore(this._lastHiddenInput);
	} else {
		let lastGrid = this.getLastSection();
		let nextSibling = lastGrid.getElement().nextSibling;
		if (nextSibling !== null) {
			section.insertBefore(nextSibling);
		} else {
			section.insertBefore(this._lastHiddenInput);
		}
	}
};

/** @private
* @param {ILayoutGrid} section
* @param {string} sectionType
*/
Core.prototype._syncStyleToSectionColumns = function (section, sectionType) {
	let name;
	for (let i = this._columnDefs.length; --i >= 0; ) {
		let colDef = this._getColumnDef(i);
		let styles = colDef["styles"];
		for (name in styles) {
			section.setColumnStyle(i, name, styles[name]);
		}
		let classes = colDef["classes"];
		for (name in classes) {
			section.enableColumnClass(i, name, true);
		}

		let sectionDef = colDef[sectionType];
		if(sectionDef) {
			styles = sectionDef["styles"];
			for (name in styles) {
				section.setColumnStyle(i, name, styles[name]);
			}
			classes = sectionDef["classes"];
			for (name in classes) {
				section.enableColumnClass(i, name, true);
			}
		}
	}
};
/** @private
* @return {boolean}
*/
Core.prototype._syncRowHeights = function () {
	let sectionCount = this._settings.length;
	for (let s = 0; s < sectionCount; ++s) {
		this._syncSectionRowHeights(this._settings[s].getSection());
	}
	if (!this._frozenLayout) {
		return this._updateScrollbarHeight(true, true);
	}
	return false;
};
/** @private
* @param {ILayoutGrid} section
*/
Core.prototype._syncSectionRowHeights = function (section) {
	this._layoutY.copyFrom(
		section.getRowLayout(),
		this._sectionStarts[section.getIndex()]
	);
};

/** @private
* @param {number} colIndex
* @param {number} px If this is negative
* @param {boolean=} opt_scalable Default is false
* @return {boolean} True if there is any changed
*/
Core.prototype._setColumnWidth = function (colIndex, px, opt_scalable) {
	let dirty = false;
	let layoutX = this._layoutX;
	if(opt_scalable != null) {
		dirty = layoutX.setLaneScalability(colIndex, opt_scalable ? true : false);
	} else {
		let scalable = layoutX.getLaneScalability(colIndex);
		if(this._preserveProportion) {
			if(scalable) {
				//calculate the proportion
				let colCount = this.getColumnCount();
				let totalWeight = 0, totalWidth = 0;
				for (let j = 0; j < colCount; ++j) {
					if(!layoutX.getLaneScalability(j)) { continue; }
					let colWidth = layoutX.getLaneSize(j);
					let colWeight = layoutX.getCustomLaneSize(j);
					totalWeight += colWeight;
					totalWidth += colWidth;
				}
				let oldWeight = layoutX.getCustomLaneSize(colIndex);
				px = ((px * totalWeight) - (px * oldWeight)) / (totalWidth - px);
				if(!px || px < 0) {
					px = totalWeight; // In case of new width exceeds total available width, set new weight as total weight
				}
			}
		} else {
			dirty = layoutX.setLaneScalability(colIndex, false);
		}
	}
	dirty = layoutX.setLaneSize(colIndex, px) || dirty;
	return dirty;
};

/** @private
* @param {number} at
* @param {number} count
* @param {boolean=} atTheMiddle
* @param {*=} ctx
*/
Core.prototype._dispatchColumnAddedEvent = function (at, count, atTheMiddle, ctx) {
	let e = {};
	if (this._hasListener("columnAdded")) {
		e["atTheMiddle"] = atTheMiddle;
		if(count === 1) {
			e["colIndex"] = at;
			e["context"] = ctx;
			this._dispatchColumnEvent("columnAdded", e);
		} else {
			let ary = Array.isArray(ctx) ? ctx : [];
			for (let i = 0; i < count; ++i) {
				e["colIndex"] = at + i;
				e["context"] = ary[i];
				this._dispatchColumnEvent("columnAdded", e);
			}
		}
	}

	if(this._frozenLayout){
		e["noDataUpdate"] = true;
	}

	this._dispatchColumnRenderEvent(e,
		at, at + count,
		0, this._settings.length,
		NaN, NaN);
};

/** @private
* @param {!Object} e
* @param {number} fromC
* @param {number} toC EXCLUSIVE
* @param {number} fromS
* @param {number} toS EXCLUSIVE
* @param {number} fromR If NaN or negative value is given. The entire section will be updated
* @param {number} toR EXCLUSIVE
*/
Core.prototype._dispatchColumnRenderEvent = function (e, fromC, toC, fromS, toS, fromR, toR) {
	if (!(fromS < toS)) { // Column is added before section
		return;
	}

	e["sender"] = this; // Make sure that the sender is this grid
	e["fromColIndex"] = fromC;
	e["toColIndex"] = toC;

	let s;
	if (this._hasListener("preSectionRender")) {
		for (s = fromS; s < toS; ++s) {
			this._settings[s].extendRenderEventArg(e, fromR, toR);
			this._dispatch("preSectionRender", e);
		}
	}

	// Dispatch column render event
	for (let c = fromC; c < toC; ++c) { // For each column
		let handler = this.getColumnRenderingHandler(c) || this._renderingHandler;
		if (handler) {
			e["colIndex"] = c;
			e["dataColumnName"] = this.getDataColumnName(c);

			for (s = fromS; s < toS; ++s) { // For each section
				let sectionSettings = this._settings[s];
				sectionSettings.extendRenderEventArg(e, fromR, toR);

				let section = sectionSettings.getSection();
				let startR = e["fromRowIndex"];
				let destR = e["toRowIndex"];
				for (let r = startR; r < destR; ++r) { // For each row
					e["rowIndex"] = r;
					e["cell"] = section["getCell"](c, r, false); // Accessing cell by using bracket allows extenal object to mock Section
					handler(e);
				}
			}
		}
	}

	if (this._hasListener("postSectionRender")) {
		for (s = fromS; s < toS; ++s) {
			this._settings[s].extendRenderEventArg(e, fromR, toR);
			this._dispatch("postSectionRender", e);
		}
	}

	// Dispatch data binding event once after rendering for populating content of newly rendered cells
	if (!e["noDataUpdate"] && !this._noBinding) {
		for (s = fromS; s < toS; ++s) { // For each sectionSettings
			this.updateRowData(this._settings[s], fromR, toR - 1);
		}
	}
};
/** @private
* @param {Object} e
*/
Core.prototype._dispatchRowExpansionBinding = function (e) {
	if (!this._hasListener("rowExpansionBinding")) {
		return;
	}

	let dataView = /** @type{DataView} */(e["dataSource"]);
	if (!dataView || !dataView.getDataSource()) {
		return;
	}

	let fromR = /** @type{number} */(e["fromRowIndex"]);
	let toR = /** @type{number} */(e["toRowIndex"]);
	let rids = dataView.getVisibleRowIds(true);
	let section = /** @type{ILayoutGrid} */(e["section"]);

	e["rowData"] = null;
	e["cell"] = null;
	let lastRowIndexChanged = false;
	for (let r = toR - 1; r >= fromR; --r) {
		let rowId = dataView.getRowId(r);
		if(!lastRowIndexChanged) {
			let reCount = dataView.getRowExpansionCount(rowId);
			if(reCount > 0){
				r = r + reCount + 1; // include parent row
				lastRowIndexChanged = true;
				continue;
			}
		}

		let ctxRow = section.getContextRow(r);
		if(!ctxRow) {
			continue; // The row may have been pushed out of view due to row count or height changed during the binding
		}
		let parentId = dataView.getExpansionParentByRowId(rowId);
		if(parentId) { // dispatch to render row expansion
			e["originalRowData"] = dataView.getRowData(parentId);
			e["originalRowId"] = parentId;
			e["rowExpansion"] = ctxRow["rowExpansion"] = true;
		} else if(ctxRow["rowExpansion"] ){ // dispose cases
			e["originalRowData"] = null;
			e["originalRowId"] = null;
			e["rowExpansion"] = ctxRow["rowExpansion"] = false;
		} else { // normal row no need to dispatch
			continue;
		}
		e["rowId"] = rids[r];
		e["rowIndex"] = r;
		this._dispatch("rowExpansionBinding", e);
	}
	e["originalRowData"] = null;
	e["originalRowId"] = null;
	e["rowExpansion"] = false;
};

/** @private
* @param {number} num
*/
Core.prototype._removeColumn = function (num) {  // TODO: change the logic to use removeColumnAt
	let colCount = this._layoutX.getLaneCount();
	if (num > colCount) { num = colCount; }
	if (num <= 0) { return; }

	let newCount = colCount - num;
	this._layoutX.setLaneCount(newCount);

	for (let s = this._settings.length; --s >= 0; ) {
		let section = this._settings[s].getSection();
		section.setColumnCount(newCount);
	}

	let removedCols = this._columnDefs.slice(newCount); // Shallow cloning
	this._columnDefs.length = newCount;

	this._onColumnCountChanged();
	this._syncLayoutToColumns(newCount, 0, true); // Virtualizer's view bound will not be affected from column

	if (this._hasListener("columnRemoved")) {
		let e = {};
		for (let c = colCount; --c >= newCount; ) {
			let colDef = removedCols[c - newCount];
			e["colIndex"] = c;
			e["columnData"] = colDef ? colDef["columnData"] : null;
			this._dispatchColumnEvent("columnRemoved", e);
		}
	}
};

/** `section` variable in this scope can either be instance of LayoutGrid or SectionWriter (from tr-grid-printer).
 * Need to force calling public methods (section["isColumnActive"] and section["getCell"]) so it works while being minified.
 * @private
* @param {!Object} e
*/
Core.prototype._onSectionDataChanged = function (e) {
	if(this._dispatchingDataChanged) { return; } // Prevent infinite loop

	let colCount = this._layoutX.getLaneCount();
	this._dispatchingDataChanged = colCount ? true : false;
	if(!colCount) {
		return;
	}

	let fromR = /** @type{number} */(e["fromRowIndex"]);
	let toR = /** @type{number} */(e["toRowIndex"]);
	let section = /** @type{ILayoutGrid} */(e["section"]);
	let dataView = /** @type{DataView} */(e["dataSource"]);
	let hasDataView = (dataView && dataView.getDataSource()) ? 1 : 0;
	let rids, rowDataCollection;

	section._startBindingSession(true);

	if(hasDataView) {
		rids = dataView.getVisibleRowIds(true);
		rowDataCollection = dataView.getMultipleRowData(rids, fromR, toR);
		e["dataRows"] = rowDataCollection;
	}

	let isContentSection = e["sectionType"] === "content";
	if(isContentSection){
		this._dispatch("beforeContentBinding", e);
	}
	this._dispatch("preSectionDataBinding", e);

	let dataMap = this.getDataColumnMap();
	for (let c = 0; c < colCount; ++c) {
		if(section["isColumnActive"](c) && (this.isColumnVisible(c) || this._isAlwaysRenderColumn(c))) {
			let handler = this.getColumnDataBindingHandler(c) || this._dataBindingHandler;
			if (handler) {
				let cid = dataMap[c];
				e["dataColumnName"] = cid;
				e["colIndex"] = c;

				for (let r = fromR; r < toR; ++r) {
					let rowData = hasDataView ? rowDataCollection[r] : null;
					if(rowData) {
						e["rowData"] = rowData;
						e["rowId"] = rids[r];
						e["dataValue"] = rowData ? rowData[cid] : null;
						e["rowIndex"] = r;
						e["cell"] = section["getCell"](c, r, false); // Accessing cell by using bracket allows extenal object to mock Section

						handler(e);
					}
				}
			}
		}
	}

	this._dispatch("postSectionDataBinding", e);
	this._dispatchRowExpansionBinding(e);

	if(isContentSection){
		if(!this._firstRendered) {
			this._dispatch("firstRendered", e);
			this._firstRendered = true;
		}
		this._dispatch("afterContentBinding", e);
	}
	section._startBindingSession(false);
	this._dispatchingDataChanged = false;
};

/** @public
* @ignore
*/
Core.prototype.resetInternalState = function() {
	this._firstRendered = false;
};

/** @private
* @param {VScrollbar|HScrollbar} scrollbar
* @returns {boolean}
*/
Core.prototype._updateScrollbar = function (scrollbar) {
	if(scrollbar) {
		if(scrollbar.isEnabled()) { scrollbar.update(); }
		return scrollbar.isActive();
	}
	return false;
};
/** @private
* @param {Object} e
*/
Core.prototype._onVScroll = function (e) {
	this._rowVirtualizer.setViewOffset(this._vscrollbar.getScrollTop());  // Trigger virtualization event
	this._dispatchRowPositionChanged();
};
/** @private
* @param {Object} e
*/
Core.prototype._onHScroll = function (e) {
	let scrollVal = this._hscrollbar.getScrollLeft();
	this._colVirtualizer.setViewOffset(scrollVal); // Trigger virtualization event
	this.updateColumnBounds();
	this._dispatchColumnPositionChanged();
};
/** @private
*/
Core.prototype._onVScrollEnabled = function () {
	this._updateScrollbarHeight(true, false);
	if(this._rowVirtualizer.isEnabled()) {
		this._rowVirtualizer.activate();
	}
	if(!this._vscrollbar.getAutoHide()) {
		this._updateVScrollbar();
	}
};
/** @private
*/
Core.prototype._onVScrollDisabled = function () {
	if(this._rowVirtualizer.isEnabled()) {
		this._rowVirtualizer.deactivate();
	}
};
/** @private
* @param {Object} e
*/
Core.prototype._onRowInViewChanged = function (e) { // Triggered from virtualizer
	let fi = e["firstIndex"];
	let ei = e["lastIndex"] + 1; // End index
	let sectionCount = this._settings.length;
	let updateRange = [0, 0];

	for (let s = 0; s < sectionCount; ++s) { // For each section
		let sectionSettings = this._settings[s];
		let section = sectionSettings.getSection();
		if(section && !section.getFrozenLayout()) { // Section may have been disposed or frozen
			section.activateRows(fi, ei, updateRange);
			this.updateRowData(sectionSettings,
				updateRange[0],
				updateRange[1] - 1
			); // The updateRange are relative to the section
		}
	}
};
/** @private
* @param {Object} e
*/
Core.prototype._onColInViewChanged = function (e) {
	if(!this.getColumnCount()) {
		return; // No column for activation and no cell to be rendered
	}
	let pfi = e["prevFirstIndex"];
	let pli = e["prevLastIndex"]; // INCLUSIVE
	let fi = e["firstIndex"];
	let li = e["lastIndex"]; // INCLUSIVE

	let first = pfi < fi ? pfi : fi;
	let last = pli > li ? pli : li; // INCLUSIVE

	this._activateColumns(fi, li, first, last);
	let sectionCount = this.getSectionCount();
	for (let s = 0; s < sectionCount; ++s) { // For each section
		let sectionSettings = this._settings[s];
		this.updateRowData(sectionSettings); // It will be render all rows in view
	}
};
/** @private
* @param {number=} fi First index in view
* @param {number=} li Last index in view (inclusive)
* @param {number=} first Starting index of range
* @param {number=} last Ending index of range (inclusive)
*/
Core.prototype._activateColumns = function(fi, li, first, last) {
	if (first == null) {
		first = 0;
	}

	if (last == null) {
		last = this.getColumnCount() - 1;
	}

	let activations = this._getColActivationList(fi, li, first, last);
	let sectionCount = this.getSectionCount();
	for(let s = sectionCount; --s >= 0;) {
		let sectionSettings = this._settings[s];
		sectionSettings.activateColumns(activations, first, last);
	}
};
/** @private
* @param {number=} fi First index in view
* @param {number=} li Last index in view (inclusive)
* @param {number=} first Starting index of range
* @param {number=} last Ending index of range (inclusive)
* @return {Array.<boolean>}
*/
Core.prototype._getColActivationList = function (fi, li, first, last) {
	if (first == null) { first = 0; }

	let colCount = this.getColumnCount();
	if (last == null) { last = colCount - 1; }

	if (fi == null) {
		fi = this._colVirtualizer.getFirstIndexInView();
	}

	if (li == null) {
		li = this._colVirtualizer.getLastIndexInView();
	}

	let leftPinnedIndex = this.getHScrollStartIndex();
	let rightPinnedIndex = this.getFirstPinnedRightIndex();
	let activations = new Array(last + 1);
	for (let c = first; c <= last; ++c) {
		activations[c] = (c < leftPinnedIndex) ||
			(c >= fi && c <= li) ||
			(c >= rightPinnedIndex);
	}
	return activations;
};
/** @public
* @ignore
* @param {number} colIndex
* @return {!Object}
*/
Core.prototype._getColumnDef = function (colIndex) {
	if (colIndex >= 0) {
		let def = this._columnDefs[colIndex];
		if (!def) {
			def = this._columnDefs[colIndex] = {
				"id": "" + Core._runningColumnId++, // Unique identifier
				"columnData": {}
			};
		}
		return def;
	}
	return {
		"columnData": {}
	};
};
/** Retrieve 1 or 2 level of nested object definition
* @private
* @param {number} colIndex
* @param {string=} firstLvl
* @param {string=} secondLvl
* @return {!Object}
*/
Core.prototype._getNestedColumnDef = function (colIndex, firstLvl, secondLvl) {
	let def = this._getColumnDef(colIndex);
	if(firstLvl) {
		let parentDef = def;
		def = parentDef[firstLvl];
		if(!def) {
			def = parentDef[firstLvl] = {};
		}
		if(secondLvl) {
			parentDef = def;
			def = parentDef[secondLvl];
			if(!def) {
				def = parentDef[secondLvl] = {};
			}
		}
	}
	return def;
};

/** @public
* @param {number} colIndex
* @return {string} Return empty string if the specified column does not exist
*/
Core.prototype.getColumnId = function (colIndex) {
	if(colIndex >= 0 && colIndex < this.getColumnCount()) {
		return this._getColumnDef(colIndex)["id"] || "";
	}
	return "";
};
/** @deprecated
* @public
* @function
* @ignore
* @param {number} colIndex
* @return {string}
*/
Core.prototype._getColumnId = Core.prototype.getColumnId;
/** @public
* @return {!Array.<string>} Return all column ids from existing column
*/
Core.prototype.getColumnIds = function () {
	let colCount = this.getColumnCount();
	let ary = new Array(colCount);
	for(let c = 0; c < colCount; ++c) {
		ary[c] = this._getColumnDef(c)["id"] || "";
	}
	return ary;
};

/** @public
* @param {number} colIndex
* @param {string} field
*/
Core.prototype.setColumnField = function (colIndex, field) {
	this._getColumnDef(colIndex)["field"] = field || "";
};
/** @public
* @param {number} colIndex
* @return {string} Return empty string if the specified column does not exist
*/
Core.prototype.getColumnField = function (colIndex) {
	if(colIndex >= 0 && colIndex < this.getColumnCount()) {
		return this._getColumnDef(colIndex)["field"] || "";
	}
	return "";
};
/** @public
* @return {!Array.<string>} Return all column fields from existing columns
*/
Core.prototype.getColumnFields = function () {
	let colCount = this.getColumnCount();
	let ary = new Array(colCount);
	for(let c = 0; c < colCount; ++c) {
		ary[c] = this._getColumnDef(c)["field"] || "";
	}
	return ary;
};

/** Get column index by column id or column field
* @public
* @param {string|number} colRef Column id or data column name
* @return {number} Return negative value if there is no match
*/
Core.prototype.getColumnIndex = function (colRef) {
	if(typeof colRef === "number") {
		return colRef;
	} else if(colRef) {
		let str = colRef;
		let indexByField = -1;
		let colCount = this.getColumnCount();
		for(let c = 0; c < colCount; ++c) {
			let colDef = this._getColumnDef(c);
			if(str === colDef["id"]){
				return c;
			}
			if(str === colDef["field"]) { // In case colId and field are the same, use colId first and field as a fallback
				if(indexByField < 0) {
					indexByField = c;
				}
			}
		}
		return indexByField;
	}
	return -1;
};
/** Any invalid column reference will be excluded from the output array
* @public
* @param {Array.<number|string>} colRefs
* @return {!Array.<number>} Return negative value if there is no match
*/
Core.prototype.getColumnIndices = function (colRefs) {
	let ary = [];
	let colCount = this.getColumnCount();
	let inputAry = Array.isArray(colRefs) ? colRefs : [colRefs];
	let len = inputAry.length;
	// Verify user input
	for(let i = 0; i < len; ++i) {
		let colIndex = this.getColumnIndex(inputAry[i]);
		if(colIndex >= 0 && colIndex < colCount) {
			ary.push(colIndex); // WARNING: We have not check for duplication
		}
	}
	return ary;
};

/** Get a map from column id and field to column index. This should be useful when finding multiple column indices.
* @public
* @return {!Object.<string, number>}
*/
Core.prototype.getColumnIndexMap = function () {
	let colCount = this.getColumnCount();
	let obj = {};
	let str = "";
	for(let c = 0; c < colCount; ++c) {
		let colDef = this._getColumnDef(c);

		str = colDef["field"];
		if(str) {
			obj[str] = c;
		}

		str = colDef["id"];
		if(str) {
			obj[str] = c;
		}
	}
	return obj;
};


/** @private */
Core.prototype._onWindowResize = function() {
	this._updateLayout();
	this._rowVirtualizer.setViewOffset(this._vscrollbar.getScrollTop());
};
/** @private */
Core.prototype._onMouseMove = function () {
	this._vscrollbar.flash();
	this._hscrollbar.flash();
};
/** @private */
Core.prototype._onGridClicked = function () {
	let selection = window.getSelection();
	if(!selection.toString()){
		if(!this._element.contains(_getActiveElement(this._element))){
			this.focus();
		}
	}
};

/** @private
* @param {Object} e
*/
Core.prototype._onKeyDown = function (e) {
	if(!_isTabCommand(e)) {
		return;
	}
	let activeElement = _getActiveElement(this._element);
	let onTheEdge = false;
	if(this._firstHiddenInput === activeElement) {
		onTheEdge = -1;
	} else if(this._lastHiddenInput === activeElement) {
		onTheEdge = 1;
	}

	this._dispatch("tabNavigation", {
		"activeElement": activeElement,
		"firstHiddenInput": this._firstHiddenInput,
		"lastHiddenInput": this._lastHiddenInput,
		"onTheEdge": onTheEdge,
		"shiftKey": e.shiftKey,
		"event": e
	});

	if(onTheEdge && !e.defaultPrevented) {
		if(onTheEdge > 0 && e.shiftKey) {
			this._firstHiddenInput.focus(); // jump to the top and move out of grid
		} else if(onTheEdge < 0 && !e.shiftKey) {
			this._lastHiddenInput.focus(); // skip to the end and move out of grid
		}
	}
};

/** @private
* @param {Object} e
*/
Core.prototype._onRowHightlighted = function (e) {
	if (!this._hasListener("rowHighlighted")) { return; }

	let section = e["sender"];
	let settings = this.getSectionSettings(section.getId());
	if (settings == null) { return; }

	settings.extendEventArg(e);

	let dataSource = settings.getDataSource();
	if (dataSource !== null) {
		e["rowId"] = dataSource.getRowId(/** @type{number} */(e["rowIndex"]));
		e["dataRow"] = dataSource.getRowData(e["rowId"]);
	}
	this._dispatch("rowHighlighted", e);
};
/** @private
* @param {Object} e
*/
Core.prototype._onRowCountChanged = function (e) {
	if(this._disposed) {
		return; // There is no need to do anything after grid is disposed
	}
	let section = /** @type {ILayoutGrid} */(e["sender"]);
	let prevRowCount = e["prevCount"];
	let newRowCount = e["rowCount"];
	let sectionIndex = section.getIndex();
	let offset = this._updateSectionIndices(sectionIndex);
	let i = 0;

	if(prevRowCount > newRowCount) { // case row removed
		this._layoutY.removeLaneAt(offset, prevRowCount - newRowCount);
		if(!this._layoutY.isHomogeneous()) {
			this._syncSectionRowHeights(section);
		}
	} else { // case row added
		let defaultHeight = this._layoutY.getDefaultLaneSize();
		let customHeight = section.getDefaultRowHeight();

		this._layoutY.insertLane(offset + prevRowCount, newRowCount - prevRowCount);
		if (customHeight !== defaultHeight) {
			for (i = prevRowCount; i < newRowCount; ++i) {
				this._layoutY.setLaneSize(offset + i, customHeight);
			}
		}
	}

	let noBinding = e["noBinding"];
	let paneChanged = (sectionIndex <= this._startVScrollbarIndex) || (sectionIndex >= this._getFooterStartIndex());
	let forceUpdate = this._frozenFooterCount > 0 ? true : false; // Prevent from footer section flashing
	let viewChanged = this._updateScrollbarHeight(paneChanged, forceUpdate, noBinding); // Virtualization is triggered
	if(noBinding){
		this._hasPendingRowChange = true;
	}
	if(!this._frozenLayout && !viewChanged) { // Grid must activate newly created section if no event is dispatched from the row virtualizer
		if(this._rowVirtualizer.isVirtualizable()) {
			section.activateRows(this._rowVirtualizer.getFirstIndexInView(),
				this._rowVirtualizer.getLastIndexInView());
		} else {
			section.activateRows();
		}
	}
	this._hasPendingRowChange = false;

	if(!forceUpdate) {
		this._updateVScrollbar(); // Asynchronous
	}

	this._updateColumnSeparators();
	if(prevRowCount < newRowCount) {
		// Apply column background color
		let settings = this.getSectionSettings(section);
		if(settings.getType() !== "content") {
			this.applyColumnsBackgroundColor(section);
		}
		this._dispatch("rowAdded", e);
	} else if(prevRowCount > newRowCount) {
		this._dispatch("rowRemoved", e);
	}
};

/** @private
* @param {Object} e
*/
Core.prototype._onRowHeightChanged = function (e) {
	if (!this._rowHeightSync) { return; }
	if (this._disposed) { return; } // There is no need to sync anything after disposition
	if(this._rowHeightConflator.conflate(e)) {
		return;
	}

	let evts = this._rowHeightConflator.popAllData();

	let len = evts.length;
	let minSectionIndex = NaN;

	// TODO: Set row height one by one is not good
	for(let j = 0; j < len; ++j) {
		let evt = evts[j];
		let section = /** @type {ILayoutGrid} */(evt["section"]);
		if(!this._containsSection(section)) {
			continue; // The section is no longer contained in this grid
		}

		let sectIndex = section.getIndex();
		let relIndex = evt["rowIndex"];
		let rowIndex = this.getRowIndex(sectIndex, relIndex);

		let changeLength = evt["changeLength"];
		for (let i = 0; i < changeLength; ++i) {
			if(this._layoutY.setLaneSize(rowIndex + i, section.getRowHeight(relIndex + i))) {
				if(!(sectIndex >= minSectionIndex)) {
					minSectionIndex = sectIndex;
				}
			}
		}
	}

	if(this._rowHeightConflator._needScrollbarUpdate) {
		this._rowHeightConflator._needScrollbarUpdate = false;
		this._updateScrollbarHeight(true, true);
	} else if(minSectionIndex >= 0) {
		if(this._dispatchingDataChanged){
			if(!this._rowHeightTimerId){
				this._rowHeightTimerId = setTimeout(this._requestScrollbarUpdate, 0);
			}
		} else {
			this._updateScrollbarHeight(minSectionIndex < this._startVScrollbarIndex,
				minSectionIndex >= this._startVScrollbarIndex);
		}
	}
	this._updateColumnSeparators();
	this._dispatchRowPositionChanged();
};
/** @private
* @param {ILayoutGrid} s
* @returns {boolean}
*/
Core.prototype._containsSection = function (s) {
	for(let i = this._settings.length; --i >= 0;) {
		if(this._settings[i].getSection() === s) {
			return true;
		}
	}
	return false;
};

/** @private
* @param {!Object} e
*/
Core.prototype._onRowAvailable = function (e) {
	if(this._hasPendingRowChange){
		return;
	}
	let sectionIndex = /** @type{number} */(e["sectionIndex"]);
	this._dispatchColumnRenderEvent(e,
		0, this.getColumnCount(),
		sectionIndex, sectionIndex + 1,
		e["prevRowIndex"], e["rowIndex"]);
};

/** @private
* @param {boolean=} opt_suppressLayout
*/
Core.prototype._onSectionCountChanged = function (opt_suppressLayout) {
	if(this._frozenLayout) { return; }

	if (this._tempFrozenSectionRef) {
		let startVScrollbarSection = this.getNextSection(this._tempFrozenSectionRef);
		if (startVScrollbarSection !== null) {
			this._startVScrollbarIndex = startVScrollbarSection.getIndex();
			this._tempFrozenSectionRef = "";
		}
	}

	if (this._startVScrollbarIndex >= 0 && this._startVScrollbarIndex < this._getFooterStartIndex()) {
		// TODO: should enable scrollbar inside DataSourceTreePlugin instead
		this._vscrollbar.enable();

		// Reinsert sections
		this._vscrollbar.setScrollContent(this, this._getAllSections(), this._startVScrollbarIndex);
		this._element.appendChild(this._lastHiddenInput); // Ensure that the hidden input is always at the last position

		if(!opt_suppressLayout) {
			this._updateScrollbarHeight(true, true);
		}
		this._syncLayoutToColumns();
	} else {
		this._vscrollbar.disable();
	}
};
/** @private */
Core.prototype._onColumnCountChanged = function () {
	if(this._frozenLayout || this._disposed) { return; }

	// TODO: Optimize these methods during initialization
	let pinnedLeft = this._countPinnedLeftColumns();
	let pinnedRight = this._countPinnedRightColumns();

	this.updateColumnBounds();
	this._updateColumnSeparators();

	if (this._hScrollbarEnabled && pinnedLeft + pinnedRight < this.getColumnCount()) {
		this._hscrollbar.enable();
		this._colVirtualizer.activate();

		this._hscrollbar.setScrollContent(
			this,
			this._getAllSections(),
			pinnedLeft,
			pinnedRight
		);
	} else { // All columns are pinned and thus no column is scrollable
		this._hscrollbar.disable();
		this._colVirtualizer.deactivate();

		pinnedLeft = pinnedRight = 0;
	}

	this._element.classList.toggle("right-pinning", pinnedRight > 0 ? true : false);
	this._dispatchColumnPositionChanged();
};

/** @private
* @param {Object} e
*/
Core.prototype._onSyncVScroll = function (e) {
	this._vscrollbar.setScrollTop(e.scrollTop);
};
/** @private
* @param {Object} e
*/
Core.prototype._onSyncHScroll = function (e) {
	this._hscrollbar.setScrollLeft(e.scrollLeft);
};
/** @private
* @return {number} index of footer section
*/
Core.prototype._getFooterStartIndex = function () {
	return this.getSectionCount() - this._frozenFooterCount;
};
/** @private
* @return {number} pixel without zoom factor
*/
Core.prototype._getVScrollStart = function () {
	return this._getSectionSize(0, this._startVScrollbarIndex);
};
/** @private
* @return {number} pixel without zoom factor
*/
Core.prototype._getFooterSize = function () {
	return (this._frozenFooterCount > 0) ? this._getSectionSize(this._getFooterStartIndex()) : 0;
};
/** @private
* @param {number=} startIndex
* @param {number=} count
* @return {number} pixel without zoom factor
*/
Core.prototype._getSectionSize = function (startIndex, count) {
	if(!startIndex) {
		startIndex = 0;
	} else if(startIndex < 0) {
		return 0;
	}
	if(count == null) {
		count = this.getSectionCount() - startIndex;
	}
	let startPos = this._layoutY.getLaneStart(
		this._sectionStarts[startIndex]
	);

	let endIndex = startIndex + count;
	let endPos = count > 0 ? this._layoutY.getLaneStart(
		this._sectionStarts[endIndex]
	) : 0;
	return endPos - startPos;
};

/** @private
* @param {boolean} paneChanged
* @param {boolean} contentChanged
* @param {boolean=} noBinding
* @return {boolean}
*/
Core.prototype._updateScrollbarHeight = function (paneChanged, contentChanged, noBinding) {
	if(!this._vscrollbar.isEnabled()) {
		return false;
	}

	let forceUpdateLayout = contentChanged;
	let recalculation = paneChanged && this._startVScrollbarIndex >= 0;
	let hasRowHeightChanged = this._rowHeightConflator.isConflating();
	if(hasRowHeightChanged) {
		this._rowHeightConflator._needScrollbarUpdate = true;

		// Should not force layout-update while rowHeightConflator is executing
		// because states have not updated and will actives scrollbar unintentionally.
		forceUpdateLayout = false;
	}

	if (recalculation && !hasRowHeightChanged) {
		// Scroll Frame
		let gridElem = this.getElement();
		if(!gridElem) {
			return false;
		}

		let gridParent = this.getParent();
		let scrollFrame = this._vscrollbar.getParent();
		if(scrollFrame) {
			if(scrollFrame !== gridParent) { // If vscrollbar is attached to an external element
				// Assume that there is always a horizontal scrollbar host underneath
				if(scrollFrame.firstElementChild) {
					scrollFrame = scrollFrame.firstElementChild;
				}
			}
		} else {
			scrollFrame = gridElem;
			this._vscrollbar.setParent(scrollFrame);
		}

		let sectionStart = this._getVScrollStart();
		let footerSize = this._getFooterSize();

		if(scrollFrame !== gridElem) {
			sectionStart += gridElem.clientTop;
		}

		// Scroll Track and Pane
		this._vscrollbar.resetPaneSize(); // scrollFrame could have a dynamic height that rely on pane's content.
		let containerHeight = scrollFrame.clientHeight - (sectionStart + footerSize) * this._zoomFactor;

		// Pane size needs to be bigger when zooming out to keep the same view size
		this._vscrollbar.setTop(sectionStart * this._zoomFactor);
		this._vscrollbar.setTrackSize(containerHeight, containerHeight / this._zoomFactor);

		// This calculation is for finding a appropriate dynamic buffer size
		// because in the large content number, buffer need to be bigger.
		// TODO: move buffer calculation to virtualizer
		let defaultSize = this._layoutY.getDefaultLaneSize();
		let viewSize = defaultSize * 12; // Display at least 12 rows

		if (containerHeight > viewSize) {
			viewSize = (Math.ceil(containerHeight / defaultSize) + 4) * defaultSize;
		}

		// HACK: Due to fixed layout size we need to scale view size instead of content size, when zooming
		// TODO: Check if zoom factor is used for virtualization correctly
		this._rowVirtualizer.setViewSize(viewSize / this._zoomFactor);

		let offsetRowCount = this._sectionStarts[this._startVScrollbarIndex];
		let footerCount = this.getFooterCount();
		let footerRowCount = 0;
		if(footerCount) {
			let sectionCount = this.getSectionCount();
			footerRowCount = this._sectionStarts[sectionCount] - this._sectionStarts[sectionCount - footerCount];
		}
		this._rowVirtualizer.setViewBounds(offsetRowCount, footerRowCount);
	} else {
		this._rowVirtualizer.validateVirtualization(); // Content height may be changed
	}

	let viewChanged = false;
	if(!noBinding) {
		viewChanged = this._rowVirtualizer.update(); // May fire onRowInViewChanged
	}

	this._vscrollbar.invalidate(forceUpdateLayout); // Update visual for vertical scrollbar

	return viewChanged;
};
/** @private
* @param {boolean} paneChanged
* @param {boolean} contentChanged
*/
Core.prototype._updateScrollbarWidth = function (paneChanged, contentChanged) {
	if (this._frozenLayout) {
		return;
	}

	if (paneChanged && this._hScrollbarEnabled) {
		// Scroll Frame
		let gridElem = this.getElement();
		let scrollFrame = this._hscrollbar.getParent();
		if (!scrollFrame) {
			scrollFrame = this.getParent() || gridElem;
			this._hscrollbar.setParent(scrollFrame);
		}

		// Frozen Area
		let leftFrozenArea = this._layoutX.getLaneStart(this._pinnedLeftColumnCount);

		let colCount = this._layoutX.getLaneCount();
		let rightFrozenArea = 0;
		let rightPinnedCount = this._countPinnedRightColumns(); // TODO: Check if we need to do a recount here
		if (rightPinnedCount) {
			rightFrozenArea = this._layoutX.getLaneStart(colCount) -
				this._layoutX.getLaneStart(colCount - rightPinnedCount);
		}

		// Scroll Track and Pane
		let gridClientWidth = (scrollFrame !== gridElem) ? gridElem.clientLeft : 0;
		let containerWidth = scrollFrame.clientWidth -
				(leftFrozenArea + gridClientWidth + rightFrozenArea) * this._zoomFactor;

		if (rightPinnedCount > 0) {
			containerWidth = containerWidth - this._reservedSpace;
		}

		this._hscrollbar.setPaneLeft(leftFrozenArea);
		this._hscrollbar.setLeft((leftFrozenArea + gridClientWidth) * this._zoomFactor);
		this._hscrollbar.setTrackSize(containerWidth, containerWidth / this._zoomFactor);

		let defaultSize = this._layoutX.getDefaultLaneSize();
		let viewSize = defaultSize * 4; // Display at least 4 column
		if (containerWidth > viewSize) {
			viewSize = containerWidth;
		}
		this._colVirtualizer.setViewSize(viewSize / this._zoomFactor);
		this._colVirtualizer.setViewBounds(this._pinnedLeftColumnCount, rightPinnedCount);
	} else {
		this._colVirtualizer.validateVirtualization(); // Content width may be changed
	}

	this._colVirtualizer.update();

	this._hscrollbar.invalidate(contentChanged);  // Update content width
};

/** @private
* @ignore
* @param {number} relativeY y position that is relative to the top left this object.
* @param {!Object} returnedObj
* @return {boolean}
*/
Core.prototype._hitTestY = function (relativeY, returnedObj) {
	let sectionSettings;
	if (returnedObj["sectionSettings"]) {
		sectionSettings = returnedObj["sectionSettings"];
	} else {
		let rowIndex;
		if (returnedObj["rowIndex"] == 'number') {
			rowIndex = returnedObj['rowIndex'];
		} else {
			rowIndex = this._layoutY.hitTest(relativeY);
		}
		let validRowIndex = (rowIndex >= 0) ? rowIndex : ~rowIndex; // TODO: Check if this is necessary
		let sectionIndex = Util.upperBound(this._sectionStarts, validRowIndex) - 1;
		returnedObj["sectionIndex"] = sectionIndex;
		sectionSettings = this._settings[sectionIndex];
		if(!sectionSettings) { // Cannot find SectionSettings
			return false;
		}
		sectionSettings.extendEventArg(returnedObj);
	}

	let section = sectionSettings.getSection();
	if(returnedObj["colIndex"] != null && returnedObj["cellElement"]) {
		returnedObj["rowIndex"] = section.getCellIndex(returnedObj["colIndex"], returnedObj["cellElement"]);
	} else {
		returnedObj["rowIndex"] = rowIndex + section.getRowOffset();
	}

	return true;
};

/** @private
* @param {number=} from
* @param {number=} to
* @param {boolean=} opt_forceDispatching
* @fires Core#widthChanged
*/
Core.prototype._syncLayoutToColumns = function (from, to, opt_forceDispatching) {
	if (this._frozenLayout) { return; }
	let colCount = this._layoutX.getLaneCount();
	let forceUpdate = this.fillParentWidth(100); // Resize all columns if there is a scaling column
	if (forceUpdate || (from == null) || (to == null)) {
		from = 0;
		to = colCount;
		forceUpdate = true;
	}

	let sectCount = this._settings.length;
	if (sectCount <= 0) { return; }

	let dirty = false;
	if(from < to) {
		for (let i = 0; i < sectCount; ++i) {
			dirty = this._syncLayoutToSectionColumns(this._settings[i].getSection(), from, to) || dirty;
		}
	}

	// TODO: Check if "to" should be greater than or equal to first pinnied right index
	let paneChanged = forceUpdate || (from < this.getHScrollStartIndex()) || (to > this.getFirstPinnedRightIndex());
	this._updateScrollbarWidth(paneChanged, true /* contentChanged */);
	this.updateColumnBounds();
	this._updateColumnSeparators();
	this._dispatchColumnPositionChanged();

	if (dirty || opt_forceDispatching) {
		this._dispatch("widthChanged", {
			"from": from,
			"to": to,
			"colChanged": from
		});
	}
};
/** @private
* @param {ILayoutGrid} section
* @param {number=} opt_from
* @param {number=} opt_to
* @return {boolean} True if there is any change
*/
Core.prototype._syncLayoutToSectionColumns = function (section, opt_from, opt_to) {
	if (opt_from == null) { opt_from = 0; }
	if (opt_to == null) { opt_to = this._layoutX.getLaneCount(); }

	let dirty = 0;
	for (let i = opt_from; i < opt_to; ++i) {
		dirty |= section._setColumnWidth(i, this._layoutX.getLaneSize(i));
	}
	if(dirty && !this._frozenLayout) {
		section._updateColumnLayout(); // TODO: This could be further improved by identifying first dirty index
		return true;
	}
	return false;
};

/** @private
* @param {number|string} colRef
* @return {string}
*/
Core.prototype._toDataColumnName = function (colRef) {
	if (Util.isNumber(colRef)) {
		return this.getDataColumnName(/** @type{number} */(colRef));
	}
	return /** @type{string} */(colRef);
};

/** @private
* @param {number} from
* @return {number} offset value at the given index
*/
Core.prototype._updateSectionIndices = function (from) {
	if(this._updatePaddingBottomSect()) { // Bottom padding must always be at the last section
		--from; // The section has been moved, so we shift the position required for update
	}

	if (from < 0) { from = 0; }

	// Adjust Index offsets for hittest
	let sectionCount = this._settings.length;
	let first = this._settings[0] || null;
	if (first !== this._firstSettings) {
		if (this._firstSettings) { this._firstSettings.getSection().removeClass("first"); }
		this._firstSettings = first;
		if (this._firstSettings) { this._firstSettings.getSection().addClass("first"); }
	}
	let last = this._settings[sectionCount - 1] || null;
	if (last !== this._lastSettings) {
		if (this._lastSettings) { this._lastSettings.getSection().removeClass("last"); }
		this._lastSettings = last;
		if (this._lastSettings) { this._lastSettings.getSection().addClass("last"); }
	}

	this._sectionStarts.length = sectionCount + 1;
	this._sectionStarts[0] = 0;
	for (let i = from; i < sectionCount; ++i) {
		let section = this._settings[i].getSection();
		section.setIndex(i);
		section.setRowOffset(this._sectionStarts[i]);
		this._sectionStarts[i + 1] = this._sectionStarts[i] + section.getRowCount();
	}

	return this._sectionStarts[from];
};

/** @private */
Core.prototype._updateLayout = function () {
	let element =  this.getElement();
	if(this._disposed || !element.offsetParent) { return; }

	this._syncLayoutToColumns(); // Update only if need

	this._updateScrollbarHeight(true, true);
	this._updateScrollbarWidth(true, true); // WARNING this may be redundant with _syncLayoutToColumns

	let sectionCount = this.getSectionCount();
	for (let s = 0; s < sectionCount; ++s) {
		let section = this._settings[s].getSection();
		section.updateLayout(); // Notify section about forced recalculation of the layout
	}
	this._updateColumnSeparators();
};

/** @private */
Core.prototype._updateVScrollbar = function() {
	if (this._disposed || !this._vscrollbar.isEnabled()) { return; } // There is no need to sync anything after disposition
	if(this._vScrollbarConflator.conflate()) {
		return;
	}
	this._vscrollbar.update();
};

//#endregion Private Methods

Core._proto = Core.prototype;

export default Core;
export { Core };
