import CellSpan from "./CellSpan.js";

/** @constructor */
let CellSpans = function () {
	this._spans = [];
	this._occupiedMap = {};
};

/** @public
* @param {number} at
* @return {boolean} Return true if there is any change
*/
CellSpans.prototype.removeColumn = function (at) { // Use when a column is removed
	let dirty = false;
	if(this.removeSpanByColIndex(at)) {
		dirty = true;
	}
	if(this.shiftColumn(at, -1)) {
		dirty = true;
	}
	return dirty;
};
/** @public
* @param {boolean=} bool
*/
CellSpans.prototype.freezeMapping = function (bool) {
	this._frozenMapping = bool !== false;
	if(!this._frozenMapping) {
		this._mapCellSpans();
	}
};


/** @private
*/
CellSpans.prototype._mapCellSpans = function () {
	let cellSpans = this._spans;
	this._spans = [];
	this._occupiedMap = {};
	let i, cellSpan;
	for(i = cellSpans.length; --i >= 0;) {
		cellSpan = cellSpans[i];
		if(cellSpan.indexX >= 0) {
			cellSpan.register(this._spans, this._occupiedMap);
		}
	}
};


/** @public
* @param {number} from
* @param {number} amount This can be negative number
* @return {boolean} Return true if there is any change
*/
CellSpans.prototype.shiftColumn = function (from, amount) { // Use when a column is inserted or removed
	if(this._spans.length <= 0 || amount === 0) { return false; }

	let dirty = false;
	for (let i = this._spans.length; --i >= 0;) {
		let cellSpan = this._spans[i];
		if(cellSpan.indexX >= from) {
			dirty = true;
			cellSpan.indexX += amount;
		}
	}
	if(!dirty || this._frozenMapping) { return false; }

	// Re-registers all cellspans. Some of the cellspans may be lost due to shifting
	this._mapCellSpans();
	return true;
};
/** @public
 * @param {number} indexX
 * @param {number} indexY
* @return {CellSpan}
*/
CellSpans.prototype.removeSpan = function (indexX, indexY) {
	if(this._spans.length <= 0) { return null; }

	let key = indexX + "," + indexY;
	let cellSpan = this._occupiedMap[key];
	if(!cellSpan || cellSpan.isOccupiedIndices(indexX, indexY)) { return null; }

	cellSpan.unregister(this._spans, this._occupiedMap);
	return cellSpan;
};
/** @public
* @param {number} at
* @param {Array.<CellSpan>=} opt_ary Array that contains removed CellSpan.
* @return {Array.<CellSpan>} Return null if no CellSpan is removed
*/
CellSpans.prototype.removeSpanByColIndex = function (at, opt_ary) {
	let len = this._spans.length;
	if(len <= 0) { return null; }

	let cellSpan;
	let removedCellSpans = opt_ary || [];
	let count = 0;
	let i;
	for (i = 0; i < len; ++i) {
		cellSpan = this._spans[i];
		if(cellSpan && cellSpan.indexX === at) {
			removedCellSpans.push(cellSpan);
			count++;
		}
	}
	if(count <= 0) { return null; }

	len = removedCellSpans.length;
	for(i = 0; i < count; ++i) {
		removedCellSpans[len - i - 1].unregister(this._spans, this._occupiedMap);
	}
	return removedCellSpans;
};
/** @public
* @param {number} at
* @param {Array.<CellSpan>=} opt_ary Array that contains removed CellSpan.
* @return {Array.<CellSpan>} Return null if no CellSpan is removed
*/
CellSpans.prototype.removeSpanStartedFromRowIndex = function (at, opt_ary) {
	let len = this._spans.length;
	if(len <= 0) { return null; }

	let cellSpan;
	let removedCellSpans = opt_ary || [];
	let count = 0;
	let i;
	for (i = 0; i < len; ++i) {
		cellSpan = this._spans[i];
		if(cellSpan && cellSpan.indexY >= at) {
			removedCellSpans.push(cellSpan);
			count++;
		}
	}
	if(count <= 0) { return null; }

	len = removedCellSpans.length;
	for(i = 0; i < count; ++i) {
		removedCellSpans[len - i - 1].unregister(this._spans, this._occupiedMap);
	}
	return removedCellSpans;
};

/** @public
* @return {Array.<CellSpan>}
*/
CellSpans.prototype.clearAllSpans = function () {
	if(this._spans.length <= 0) { return null; }

	let allCellSpans = this._spans;
	this._spans = [];
	this._occupiedMap = {};
	return allCellSpans;
};
/** @public
* @return {boolean}
*/
CellSpans.prototype.hasSpan = function () {
	return (this._spans.length  > 0);
};

/** @public
* @param {number} indexX
* @param {number} indexY
* @return {CellSpan|null}
*/
CellSpans.prototype.getSpan = function (indexX, indexY) {
	let cellSpan = this.getOccupyingSpan(indexX, indexY);
	if(cellSpan && cellSpan.isSpanIndices(indexX, indexY)) {
		return cellSpan;
	}
	return null;
};
/** @public
* @param {number} indexX
* @param {number} indexY
* @return {CellSpan|null}
*/
CellSpans.prototype.getOccupyingSpan = function (indexX, indexY) {
	if(this._spans.length <= 0) { return null; }
	return this._occupiedMap[indexX + "," + indexY] || null;
};
/** Shallow copy is used for better performance. The returned cellspans are still referenced.
* @public
* @return {!Array.<CellSpan>}
*/
CellSpans.prototype.getAllSpans = function () {
	return this._spans;
};
/** @description Returns zero or negative value, if the cell is spanned over (i.e. occupied by the spanning cell). <br>
* Negative value is distance from the spanning cell (i.e. -1 means one cell from the spanning cell and -2 means two cells from the spanning cell) <br>
* Return one, if the cell has no span (normal cell). <br>
* Return value greater than one, if the cell is spanning (getting expanded)
* @public
* @param {number} indexX
* @param {number} indexY
* @return {number}
*/
CellSpans.prototype.getColSpan = function (indexX, indexY) {
	let cellSpan = this._occupiedMap[indexX + "," + indexY];
	if(cellSpan == null) {
		return 1;
	}
	if (cellSpan.isOccupiedIndices(indexX, indexY)) {
		return cellSpan.indexX - indexX;
	}
	return cellSpan.colSpan;
};
/** @description Returns zero or negative value, if the cell is spanned over (i.e. occupied by the spanning cell). <br>
* Negative value is distance from the spanning cell (i.e. -1 means one cell from the spanning cell and -2 means two cells from the spanning cell) <br>
* Return one, if the cell has no span (normal cell). <br>
* Return value greater than one, if the cell is spanning (getting expanded)
* @public
* @param {number} indexX
* @param {number} indexY
* @return {number}
*/
CellSpans.prototype.getRowSpan = function (indexX, indexY) {
	let cellSpan = this._occupiedMap[indexX + "," + indexY];
	if(cellSpan == null) {
		return 1;
	}
	if (cellSpan.isOccupiedIndices(indexX, indexY)) {
		return cellSpan.indexY - indexY;
	}
	return cellSpan.rowSpan;
};

/** @public
* @param {number} indexX
* @param {number} indexY
* @param {number} spanVal
* @return {CellSpan} Return effected CellSpan definition
*/
CellSpans.prototype.setColSpan = function (indexX, indexY, spanVal) {
	return this._updateSpan(indexX, indexY, true, spanVal);
};
/** @public
* @param {number} indexX
* @param {number} indexY
* @param {number} spanVal
* @return {CellSpan} Return effected CellSpan definition
*/
CellSpans.prototype.setRowSpan = function (indexX, indexY, spanVal) {
	return this._updateSpan(indexX, indexY, false, spanVal);
};

/** @private
* @param {number} c0
* @param {number} r0
* @param {boolean} isColumn
* @param {number} spanVal
* @return {CellSpan} Return effected CellSpan definition
*/
CellSpans.prototype._updateSpan = function (c0, r0, isColumn, spanVal) {
	if (spanVal < 1) { spanVal = 1; }
	let cellSpan = this.getSpan(c0, r0);
	if(cellSpan) { cellSpan.retVal = 0; }

	let c1 = c0;
	let r1 = r0;
	let cs = (cellSpan) ? cellSpan.colSpan : 1;
	let rs = (cellSpan) ? cellSpan.rowSpan : 1;
	let c2, r2, target, key;
	let isExpanding = true;
	if (isColumn) {
		if (cs === spanVal) { return cellSpan; }
		if (cs > spanVal) {
			c1 = c0 + spanVal;
			c2 = c0 + cs;
			isExpanding = false;
		} else {
			target = this._occupiedMap[c0 + "," + r0];
			if (target && target !== cellSpan) {
				return cellSpan;
			}
			c1 = c0 + cs;
			c2 = c0 + spanVal;
		}
		r2 = r0 + rs;
	} else {
		if (rs === spanVal) { return cellSpan; }
		if (rs > spanVal) {
			r1 = r0 + spanVal;
			r2 = r0 + rs;
			isExpanding = false;
		} else {
			target = this._occupiedMap[c0 + "," + r0];
			if (target && target !== cellSpan) {
				return cellSpan;
			}
			r1 = r0 + rs;
			r2 = r0 + spanVal;
		}
		c2 = c0 + cs;
	}

	// Collecting affected cells
	let c, r;
	if (isExpanding) { // Check all affected cells for occupying
		let keys = [];
		for (c = c1; c < c2; ++c) {
			for (r = r1; r < r2; ++r) {
				key = c + "," + r;
				if (this._occupiedMap[key]) { //Invalid input is given
					return cellSpan;
				}
				keys.push(key);
			}
		}

		if (cellSpan == null) {
			cellSpan = new CellSpan(c0, r0);
			key = c0 + "," + r0;
			this._spans.push(cellSpan);
			this._occupiedMap[key] = cellSpan;
		}

		for (c = keys.length; --c >= 0; ) { // Update occupied cell map
			this._occupiedMap[keys[c]] = cellSpan;
		}
	} else {
		let noLongerSpan = (isColumn) ? (spanVal + rs <= 2) : (spanVal + cs <= 2);
		if (noLongerSpan) {
			key = c0 + "," + r0;
			let at = this._spans.indexOf(cellSpan);
			this._spans.splice(at, 1);
			this._occupiedMap[key] = null;
		}

		for (c = c1; c < c2; ++c) { // Update occupied cell map
			for (r = r1; r < r2; ++r) {
				this._occupiedMap[c + "," + r] = null;
			}
		}
	}

	if (isColumn) {
		cellSpan.prevColSpan = cellSpan.colSpan;
		cellSpan.colSpan = spanVal;
	} else {
		cellSpan.prevRowSpan = cellSpan.rowSpan;
		cellSpan.rowSpan = spanVal;
	}
	cellSpan.retVal = isExpanding ? 1 : -1;
	return cellSpan;
};

/** A list contains all CellSpan definitions
* @private
* @type {!Array.<CellSpan>}
*/
CellSpans.prototype._spans;
/** Map of occupied cells using string as key : _occupiedMap[colIndex + "," +rowIndex]
* @private
* @type {!Object.<string, CellSpan>}
*/
CellSpans.prototype._occupiedMap;

/**
* @private
* @type {boolean}
*/
CellSpans.prototype._frozenMapping = false;

export default CellSpans;
export { CellSpans };

