import Util from "../util/util.js";
import TrackLayout from "../util/TrackLayout.js"; // eslint-disable-line
import ILayoutGrid from "../ILayoutGrid.js"; // eslint-disable-line
import Cell from "./Cell.js";


/** Update cell's top and height
* @private
* @param {Cell} cell
* @param {TrackLayout} trackY
* @param {number} rowIndex
*/
let _updateYAxis = function (cell, trackY, rowIndex) {
	cell.setTop(trackY.getLaneStart(rowIndex));
	cell.setHeight(trackY.getLaneSize(rowIndex));
};
/** Update cell's left and width
* @private
* @param {Cell} cell
* @param {TrackLayout} trackX
* @param {number} colIndex
* @param {number} width
*/
let _updateXAxis = function (cell, trackX, colIndex, width) {
	if(colIndex) {
		let leftStart = trackX.getLaneStart(colIndex);
		cell.setLeft(leftStart);
		cell.setWidth(width - leftStart);
	} else {
		cell.setStyle("left", "");
		cell.setWidth(width); // Cell width is always modified (colSpan styling will be reset)
	}
};


/** @constructor
* @ignore
* @param {!TrackLayout} trackX
* @param {!TrackLayout} trackY
* @param {!ILayoutGrid} section
*/
let StretchedCells = function (trackX, trackY, section) {
	this._trackX = trackX;
	this._trackY = trackY;
	this._section = section;
	this._cells = [];
};
/** @type {!TrackLayout}
* @private
*/
StretchedCells.prototype._trackX;
/** @type {!TrackLayout}
* @private
*/
StretchedCells.prototype._trackY;
/** @type {!ILayoutGrid}
* @private
*/
StretchedCells.prototype._section;
/** @type {!Array.<Cell>}
* @private
*/
StretchedCells.prototype._cells;
/** @type {number}
* @private
*/
StretchedCells.prototype._activeCount = 0;
/** @type {number}
* @private
*/
StretchedCells.prototype._stretchSize = 0;

/** @public
*/
StretchedCells.prototype.dispose = function () {
	let cells = this._cells;
	if(this._activeCount) {
		this._activeCount = 0;
		for(let i = cells.length; --i >= 0;) {
			let cell = cells[i];
			if(cell) {
				cell.dispose();
			}
		}
	}
	cells.length = 0;
};

/** @public
* @ignore
* @param {number} num
*/
StretchedCells.prototype.setStretchSize = function(num) {
	this._stretchSize = num;
};
/** @public
* @ignore
* @return {number}
*/
StretchedCells.prototype.getStretchSize = function() {
	return this._stretchSize;
};

/** @public
* @ignore
* @param {number} count
*/
StretchedCells.prototype.setCellCount = function (count) {
	let cells = this._cells;
	let len = cells.length;
	if(len < count) {
		cells.length = count;
		while(len < count) {
			cells[len++] = null;
		}
	} else if(len > count) {
		while(--len >= count) {
			let cell = cells[len];
			if(cell) {
				cell.dispose();
			}
		}
		cells.length = count;
	}
};
/** @public
* @ignore
* @param {number} rowIndex
* @return {Cell}
*/
StretchedCells.prototype.getCell = function (rowIndex) {
	let cell = this._cells[rowIndex];
	if(cell && cell["getParent"]()) {
		return cell;
	}
	return null;
};
/** @public
* @ignore
* @return {Array.<Cell>}
*/
StretchedCells.prototype.getCells = function () {
	let outAry = [];
	if(this._activeCount) {
		let cells = this._cells;
		let rowCount = cells.length;
		for(let r = 0; r < rowCount; ++r) {
			let cell = cells[r];
			if(cell && cell.getParent()) {
				outAry.push(cell);
			}
		}
		this._activeCount = outAry.length;
	}
	return outAry;
};
/** @public
* @ignore
* @param {Cell|Element} cellRef
* @return {number}
*/
StretchedCells.prototype.getColumnIndex = function (cellRef) {
	if(cellRef) {
		if(cellRef["getElement"]) {
			let cellElement = cellRef["getElement"]();
			if(cellElement) {
				return cellElement._colIndex != null ? cellElement._colIndex : -1;
			}
		} else {
			return cellRef._colIndex != null ? cellRef._colIndex : -1;
		}
	}
	return -1;
};
/** @public
* @ignore
* @param {Cell|Element} cellRef
* @return {number}
*/
StretchedCells.prototype.getRowIndex = function (cellRef) {
	if(cellRef) {
		if(cellRef["getElement"]) {
			return cellRef["getParent"]() ? this._cells.indexOf(cellRef) : -1;
		} else {
			let cells = this._cells;
			let rowCount = cells.length;
			for(let r = 0; r < rowCount; ++r) {
				let cell = cells[r];
				if(cell && cell["getParent"]() && cell["getElement"]() === cellRef) {
					return r;
				}
			}
		}
	}
	return -1;
};

/** @public
* @ignore
* @param {number} rowIndex
* @param {number=} colIndex Column index indicates the position of the cell. If not defined, the cell will always position at the left most
* @return {Cell}
*/
StretchedCells.prototype.stretchCell = function (rowIndex, colIndex) {
	// Increase the active count regardless of redundancy. This acts as a dirty flag. It does not need to be accurate
	++this._activeCount;
	if(!colIndex || colIndex < 0) {
		colIndex = 0;
	}

	let cell = this._cells[rowIndex] || null;
	if(!cell) {
		cell = new Cell(null, this._section);
		cell.addClass("tr-stretched");
		cell["stretched"] = true;
		this._cells[rowIndex] = cell;
	}
	let cellElem = cell["getElement"]();
	cellElem["stretched"] = true;
	if(cellElem._colIndex !== colIndex) {
		cellElem._colIndex = colIndex;
		if(colIndex > 0) {
			cellElem.classList.add("right-stretching");
		} else {
			cellElem.classList.remove("right-stretching");
		}
	}
	if(cellElem._rowIndex !== rowIndex) {
		cellElem._rowIndex = rowIndex;
	}

	let width = this._stretchSize || this._trackX.getTrackSize();
	_updateXAxis(cell, this._trackX, colIndex, width);
	_updateYAxis(cell, this._trackY, rowIndex);

	return cell;
};
/** @public
* @ignore
* @param {number} rowIndex
* @return {Cell}
*/
StretchedCells.prototype.unstretchCell = function(rowIndex) {
	let cell = this._cells[rowIndex] || null;
	if(cell) {
		cell.setParent(null);
	}
	return cell;
};
/** @public
*/
StretchedCells.prototype.unstretchAllCells = function () {
	if(this._activeCount) {
		this._activeCount = 0;
		let cells = this._cells;
		for(let i = cells.length; --i >= 0;) {
			let cell = cells[i];
			if(cell) {
				cell.setParent(null);
			}
		}
	}
};

/** Update cell's left, top, width and height for all active cells
* @public
* @ignore
*/
StretchedCells.prototype.updateCells = function() {
	if(this._activeCount) {
		let activeCount = 0;
		let cells = this._cells;
		let rowCount = cells.length;
		let trackX = this._trackX;
		let trackY = this._trackY;
		let width = this._stretchSize || trackX.getTrackSize();
		for(let r = 0; r < rowCount; ++r) {
			let cell = cells[r];
			if(cell && cell.getParent()) {
				++activeCount;
				_updateXAxis(cell, trackX, cell.getElement()._colIndex, width);
				_updateYAxis(cell, trackY, r);
			}
		}
		this._activeCount = activeCount;
	}
};
/** Update cell's left, top, width and height for all active cells
* @public
* @ignore
*/
StretchedCells.prototype.updateXAxis = function() {
	if(this._activeCount) {
		let activeCount = 0;
		let cells = this._cells;
		let rowCount = cells.length;
		let trackX = this._trackX;
		let width = this._stretchSize || trackX.getTrackSize();
		for(let r = 0; r < rowCount; ++r) {
			let cell = cells[r];
			if(cell && cell.getParent()) {
				++activeCount;
				_updateXAxis(cell, trackX, cell.getElement()._colIndex, width);
			}
		}
		this._activeCount = activeCount;
	}
};
/** Update cell's top and height for all active cells
* @public
* @ignore
*/
StretchedCells.prototype.updateYAxis = function () {
	if(this._activeCount) {
		let activeCount = 0;
		let cells = this._cells;
		let rowCount = cells.length;
		let trackY = this._trackY;
		for(let r = 0; r < rowCount; ++r) {
			let cell = cells[r];
			if(cell && cell.getParent()) {
				++activeCount;
				_updateYAxis(cell, trackY, r);
			}
		}
		this._activeCount = activeCount;
	}
};

/** @public
* @ignore
* @param {number} numCells Index of the first item to be moved to the top
*/
StretchedCells.prototype.moveCellsToTop = function (numCells) {
	if(this._activeCount) {
		this._cells = Util.cycleArray(this._cells, numCells);
		this.updateYAxis(); // Reposition cells due to moving row index
	}
};

export default StretchedCells;
export { StretchedCells };
