import EventListeners from "./EventListeners.js";

/** Abstract base class that provides event management methods for derived class
 * @constructor
 */
let EventDispatcher = function () {
	/* throw("Cannot instantiate an abstract class"); */
	/*
	this._addEvent("customEvent"); // Add this line in your derived class, if you want to add custom event to your class object
	*/
};
/** @protected
* @ignore
* @type {Object.<string, !EventListeners>}
*/
EventDispatcher.prototype._customEvents = null;

//#region Public
/**
* @public
* @param {string} type Any event provided by derived class
* @param {Function} handler Callback method
* @param {number=} opt_priority The higher the priority, the sooner the listener get executed. Undefined priority listener will be executed last
* @example
* let grid = new Core();
* grid.listen("click", function(e) { console.log("Grid is clicked"); });
*/
EventDispatcher.prototype.listen = function (type, handler, opt_priority) {
	if (this._customEvents && this._customEvents[type]) {
		this._customEvents[type].listen(handler, opt_priority);
	}
};
/** Removes an event listener which was added with listen(). The same function handler must be sent in, or else it will not be removed
* @public
* @param {string} type The name of the event without the 'on' prefix.
* @param {Function} handler The listener function to remove.
*/
EventDispatcher.prototype.unlisten = function (type, handler) {
	if(this._customEvents && this._customEvents[type]) {
		this._customEvents[type].unlisten(handler);
	}
};
/** Removes all listeners from an object. If no type is specified, it will remove all listeners that have been registered. <br>
* You can also optionally remove listeners of a particular type.
* @public
* @param {string=} type Type of event to, default is all types.
*/
EventDispatcher.prototype.unlistenAll = function (type) {
	if(!type) {
		for (let key in this._customEvents) {
			this._customEvents[key].unlistenAll();
		}
	} else if(this._customEvents && this._customEvents[type]) {
		this._customEvents[type].unlistenAll();
	}
};
//#endregion Public

//#region Protected
/** Adds custom event
* @protected
* @ignore
* @param {string} type
*/
EventDispatcher.prototype._addEvent = function (type) {
	if(!type) { return; }

	if(!this._customEvents) {
		this._customEvents = {};
	}
	let e = this._customEvents[type];
	if(!e) {
		this._customEvents[type] = new EventListeners(this);
	}
};
/** Adds multiple events at once
* @protected
* @ignore
* @param {...string} types
*/
EventDispatcher.prototype._addEvents = function (types) {
	let ary = Array.isArray(types) ? types : arguments;
	let len = ary.length;
	for(let i = 0; i < len; ++i) {
		this._addEvent(ary[i]);
	}
};
/** Check if the custom event is disptaching as dispatching for custom event is syncronous operations. Native events will always return false. This is useful to stop infinite loop.
* @protected
* @ignore
* @param {string} type
* @return {boolean} Returns false if the event is not dispatching
*/
EventDispatcher.prototype._isEventDispatching = function (type) {
	let evt = this._customEvents[type];
	if(evt) {
		return evt.isDispatching();
	}
	return false;
};
/** Disable custom event from operating. This does NOT work with native events.
* @protected
* @ignore
* @param {string} type
* @param {boolean=} disabled
*/
EventDispatcher.prototype._disableEvent = function (type, disabled) {
	let evt = this._customEvents[type];
	if(evt) {
		evt.disable(disabled);
	}
};
/** This is used in conjunction with _dispatch to avoid creating event argument without actual listener
* @protected
* @ignore
* @param {string} type
* @return {boolean}
*/
EventDispatcher.prototype._hasListener = function (type) {
	if(this._customEvents && this._customEvents[type]) {
		return this._customEvents[type].hasListener();
	}
	return false;
};
/** Only custom event that is added by _addEvent() can be dispatched
* @protected
* @ignore
* @param {string} type
* @param {!Object} eventArg
*/
EventDispatcher.prototype._dispatch = function (type, eventArg) {
	this._customEvents[type].dispatch(eventArg); // Intentional no error checking. You must add custom event first
};

EventDispatcher._proto = EventDispatcher.prototype;

export default EventDispatcher;
export { EventDispatcher };
