import Util from "../util/util.js";

/** @constructor
* @param {*|null=} opt_sender The "sender" reference added automatically to event argument when dispatching event
*/
let EventListeners = function (opt_sender) {
	this.dispatch = this.dispatch.bind(this);

	this._listeners = [];
	this.setSender(opt_sender);
};

//#region Public methods
/** @public
* @param {Function} listener should be already bound by the bind method
* @param {number=} opt_priority The higher the priority, the sooner the listener get executed. Undefined priority listener will be executed last
*/
EventListeners.prototype.listen = function (listener, opt_priority) {
	if (listener && Util.isFunction(listener)) {
		if(this._listeners.indexOf(listener) < 0) {
			if(opt_priority == null) {
				this._listeners.push(listener);
			} else {
				listener._priority = opt_priority;
				let at = Util.lowerBound(this._listeners, listener, 0, this._listeners.length, EventListeners.listenerComparer);
				this._listeners.splice(at, 0, listener);
			}
		}
	}
};
/** @public
* @param {Function} listener
*/
EventListeners.prototype.unlisten = function (listener) {
	if (listener) {
		let index = this._listeners.indexOf(listener);
		if (index >= 0) {
			this._listeners.splice(index, 1);
		}
	}
};
/** @public
*/
EventListeners.prototype.unlistenAll = function () {
	this._listeners.length = 0;
};

/** @public
* @return {number}
*/
EventListeners.prototype.getListenerCount = function () {
	return this._listeners.length;
};
/** @public
* @return {boolean}
*/
EventListeners.prototype.hasListener = function () {
	return this._listeners.length > 0;
};

/** @public
* @param {*} eventArg
*/
EventListeners.prototype.dispatch = function (eventArg) {
	if (!this._enabled || this._listeners.length <= 0) { return; }

	if (this._dispatching) {
		window.console.log("WARNING: There is an error in one of your binding methods.");
	}
	this._dispatching = true;

	if (eventArg == null) { eventArg = {}; }

	if (this._sender !== null) {
		eventArg["sender"] = this._sender;
	}

	if (this._extender !== null) {
		this._extender(/** @type{Object} */(eventArg));
	}

	for (let i = 0; i < this._listeners.length; i++) {
		this._listeners[i](eventArg);
	}

	if(this._sender !== null) {
		eventArg["sender"] = null; // Clear ref variable to prevent memory leak
	}
	this._dispatching = false;
};
/** @public
* @return {boolean}
*/
EventListeners.prototype.isDispatching = function () {
	return this._dispatching;
};

/** @public
* @return {boolean}
*/
EventListeners.prototype.isEnabled = function () {
	return this._enabled;
};
/** @public
* @param {boolean=} opt_enabled
*/
EventListeners.prototype.enable = function (opt_enabled) {
	this._enabled = (opt_enabled !== false);
};
/** @public
* @param {boolean=} opt_disabled
*/
EventListeners.prototype.disable = function (opt_disabled) {
	this._enabled = (opt_disabled === false);
};

/** Specifies "sender" reference to be added to the given event argument
* @public
* @param {*|null} sender
*/
EventListeners.prototype.setSender = function (sender) {
	this._sender = sender || null;
};
/** Specifies a handler to be executed before every dispatching
* @public
* @param {Function=} extender
*/
EventListeners.prototype.setEventArgExtender = function (extender) {
	this._extender = (extender != null) ? extender : null;
};
//#endregion Public methods

//#region Private
EventListeners.listenerComparer = function(a, b) {
	if(b._priority == null) { return true; }
	return a._priority >= b._priority; // Descending order
};

/** @private
 * @type {Array.<function(*)>}
 */
EventListeners.prototype._listeners = null;

/** @private
 * @type {*|null}
 */
EventListeners.prototype._sender = null;
/** @private
 * @type {function(Object)|null}
 */
EventListeners.prototype._extender = null;

/** @private
 * @type {boolean}
 */
EventListeners.prototype._enabled = true;
/** @private
 * @type {boolean}
 */
EventListeners.prototype._dispatching = false;
//#endregion  Private

EventListeners._proto = EventListeners.prototype; // To allow compression of prototype keyword

export default EventListeners;
export { EventListeners };
