/** @private
* @function
* @param {number} idx
* @param {number} limit
* @return {number}
*/
let _validateIndex = function(idx, limit) {
	if(idx > limit) {
		return limit;
	} else if(idx < 0) {
		return 0;
	}
	return idx;
};

/** @constructor
* @param {Object} ctx
*/
let CellBoundPainter = function (ctx) {
	this._boundLayer = ctx.boundLayer;
	this._layoutX = ctx.layoutX;
	this._layoutY = ctx.layoutY;
	this._hscrollbar = ctx.hscrollbar;
	this._calculateColumnBounds = ctx.calculateColumnBounds;
};

/** @type {Element}
* @private
*/
CellBoundPainter.prototype._boundLayer = null;
/** @type {!TrackLayout}
* @private
*/
CellBoundPainter.prototype._layoutX;
/** @type {!TrackLayout}
* @private
*/
CellBoundPainter.prototype._layoutY;
/** @type {HScrollbar}
* @private
*/
CellBoundPainter.prototype._hscrollbar;
/** @type {Function}
* @private
*/
CellBoundPainter.prototype._calculateColumnBounds;
/** @type {Element}
* @private
*/
CellBoundPainter.prototype._cellBound = null;
/** @type {number}
* @private
*/
CellBoundPainter.prototype._cbLftIdx = 0;
/** @type {number}
* @private
*/
CellBoundPainter.prototype._cbRgtIdx = 0;
/** @type {number}
* @private
*/
CellBoundPainter.prototype._cbTopIdx = 0;
/** @type {number}
* @private
*/
CellBoundPainter.prototype._cbBtmIdx = 0;

/** @public
*/
CellBoundPainter.prototype.dispose = function () {
	let cellBound = this._cellBound;
	if(cellBound) {
		let pn = cellBound.parentNode;
		if(pn) {
			pn.removeChild(cellBound);
		}
		this._cellBound = null;
	}
	this._boundLayer = null;
	this._layoutX = null;
	this._layoutY = null;
	this._hscrollbar = null;
	this._calculateColumnBounds = null;
};

/** @public
* @param {number} colIndex
* @param {number} rowIndex
* @param {number} width
* @param {number} height
* @param {number} colCount
*/
CellBoundPainter.prototype.setCellBounds = function (colIndex, rowIndex, width, height, colCount) {
	let cellBound = this._cellBound;
	if(!cellBound) {
		cellBound = this._cellBound = document.createElement("div");
		cellBound.className = "selection-bound";
	}

	// Validate inputs
	let rgtIndex = colIndex + width; // Exclusive
	let btmIndex = rowIndex + height; // Exclusive

	let rowCount = this._layoutY.getLaneCount();

	this._cbLftIdx = _validateIndex(colIndex, colCount);
	this._cbRgtIdx = _validateIndex(rgtIndex, colCount);
	this._cbTopIdx = _validateIndex(rowIndex, rowCount);
	this._cbBtmIdx = _validateIndex(btmIndex, rowCount);

	this.updateCellBounds();
};

/** @public
*/
CellBoundPainter.prototype.updateCellBounds = function() {
	let cellBound = this._cellBound;
	if(!cellBound) {
		return;
	}

	let layoutX = this._layoutX;
	let layoutY = this._layoutY;
	let lftIdx = this._cbLftIdx;
	let rgtIdx = this._cbRgtIdx; // Exclusive
	let topIdx = this._cbTopIdx;
	let btmIdx = this._cbBtmIdx; // Exclusive
	let lftPx, rgtPx, topPx, btmPx;
	lftPx = rgtPx = topPx = btmPx = 0;
	if(lftIdx < rgtIdx && topIdx < btmIdx) {
		lftPx = layoutX.getLaneStart(lftIdx);
		rgtPx = layoutX.getLaneEnd(rgtIdx - 1);
		topPx = layoutY.getLaneStart(topIdx);
		btmPx = layoutY.getLaneEnd(btmIdx - 1);
	}

	let width = rgtPx - lftPx;
	let height = btmPx - topPx;
	let noBorders = [false, false];
	if(width > 0 && height > 0 && this._hscrollbar) {
		let positions = [0, 0];
		this._calculateColumnBounds(lftIdx, rgtIdx - 1, positions, noBorders);
		lftPx = positions[0];
		rgtPx = positions[1];
		width = rgtPx - lftPx;
	}

	if(width > 0) {
		cellBound.style.left = lftPx + "px";
		cellBound.style.top = topPx + "px";
		cellBound.style.width = width + "px";
		cellBound.style.height = height + "px";

		cellBound.classList.toggle("no-left-bound", noBorders[0]);
		cellBound.classList.toggle("no-right-bound", noBorders[1]);

		if(this._boundLayer) {
			this._boundLayer.appendChild(cellBound);
		}
	} else {
		let pn = cellBound.parentNode;
		if(pn) {
			pn.removeChild(cellBound);
		}
	}
};

CellBoundPainter._proto = CellBoundPainter.prototype;

export default CellBoundPainter;
export { CellBoundPainter };
