
/** @description Conflator is used to stop multiple executions of a function in the specified period of time by merging multiple calls into one single call.
* Conflator guarantees that only one single call will be executed.
* @constructor
* @param {Function|number=} opt_func Bound function that will use this conflator
* @param {number|Function=} opt_ms The delay time in millisecond before executing the function
* @param {*=} opt_thisObj "this" object to be bound with the given function. If the function is already bound, there is no need to provide thisObj parameter
* @example
* let c = new Conflator(function() { console.log("Executed"); }, 1000);
* for(let i = 10; --i >= 0;) {
*   c.conflate(i); // Only one "Executed" text will be logged to console after one second
* }
*/
let Conflator = function (opt_func, opt_ms, opt_thisObj) {
	this._onConflated = this._onConflated.bind(this);

	this._data = [];

	if(typeof opt_func == "function") {
		this.setFunction(opt_func, opt_thisObj);
	} else if (typeof opt_ms == "function") { // Parameter can be swapped places
		this.setFunction(opt_ms, opt_thisObj);
	}

	if(typeof opt_ms == "number") {
		this.setConflatingTime(opt_ms);
	} else if (typeof opt_func == "number") { // Parameter can be swapped places
		this.setConflatingTime(opt_func);
	}
};

/** @type {!Array}
* @private
*/
Conflator.prototype._data; // Array to store all conflated data
/** @type {Function}
* @private
*/
Conflator.prototype._func = null;
/** @type {number}
* @private
*/
Conflator.prototype._delay = 0;
/** @type {number}
* @private
*/
Conflator.prototype._timerId = 0;
/** @type {boolean}
* @private
*/
Conflator.prototype._conflated = false;
/** @type {boolean}
* @private
*/
Conflator.prototype._disabled = false;


/** @description Any existing timer and stored data will be cleared.
* @public
* @return {Array} All stored data. Return null if there is no data
*/
Conflator.prototype.reset = function () {
	if(this._timerId) {
		clearTimeout(this._timerId);
		this._timerId = 0;
	}

	return this.popAllData();
};
/** All stored data will be cleared
* @public
* @return {Array} All stored data. Return null if there is no data
*/
Conflator.prototype.popAllData = function() {
	if(this._data.length) {
		let data = this._data;
		this._data = [];
		return data;
	}
	return null;
};
/** @public
*/
Conflator.prototype.dispose = function () {
	if(this._timerId) {
		clearTimeout(this._timerId);
		this._timerId = 0;
	}
	if(this._data.length) {
		this._data.length = 0;
	}
	this._func = null; // Release user bound function
};

/** @public
* @param {*=} opt_data
* @return {boolean}
*/
Conflator.prototype.conflate = function (opt_data) {
	if(this._disabled) {
		return false;
	}
	if(opt_data != null) {
		this._data.push(opt_data);
	}
	if(this._conflated) {
		return false; // Conflating operation is already completed
	}
	if(!this._timerId) {
		this._timerId = setTimeout(this._onConflated, this._delay);
	}
	return true;
};
/** @public
* @return {boolean}
*/
Conflator.prototype.isConflating = function () {
	return (this._timerId > 0 && !this._conflated && !this._disabled);
};

/** @private
*/
Conflator.prototype._onConflated = function () {
	this._conflated = true;

	try {
		this._func();
	} catch (err) {
		console.error(err.message || err);
	}

	this._timerId = 0;
	this._data.length = 0;
	this._conflated = false;
};

/** @public
* @return {!Array}
*/
Conflator.prototype.getData = function () {
	return this._data;
};
/** @public
* @param {Function} func Bound function that will use this conflator
* @param {*=} opt_thisObj "this" object to be bound with the given function. If the function is already bound, there is no need to provide thisObj parameter
*/
Conflator.prototype.setFunction = function (func, opt_thisObj) {
	this._func = func;
	if(func) {
		if(opt_thisObj) {
			this._func = func.bind(opt_thisObj);
		} else {
			this._func = func;
		}
	} else {
		this._func = null;
	}
};
/** @public
* @param {number} ms The delay time in millisecond before executing the function
*/
Conflator.prototype.setConflatingTime = function (ms) {
	this._delay = ms || 0;
	if(this._delay < 0) {
		this._delay = 0;
	}
};

/** @public
* @param {boolean=} opt_disabled
*/
Conflator.prototype.disable = function (opt_disabled) {
	this._disabled = opt_disabled !== false;
};
/** @public
* @return {boolean}
*/
Conflator.prototype.isEnabled = function () {
	return !this._disabled;
};
/** @public
* @param {boolean=} opt_enabled
*/
Conflator.prototype.enable = function (opt_enabled) {
	this._disabled = opt_enabled === false;
};

Conflator._proto = Conflator.prototype;

export default Conflator;
export { Conflator };
