/**
* @namespace
*/
let Util = {};

// TODO: move to tr-grid-util

/** @const
* @type {boolean}
*/
Util.isIE = (function () {
	let ua = window.navigator.userAgent;
	return (ua.indexOf('MSIE ') > 0) || (ua.indexOf('Trident/') > 0) || (ua.indexOf('Edge/') > 0);
}());
/** @const
* @type {boolean}
*/
Util.isTouchDevice = (function () {
	if ((navigator["maxTouchPoints"] && navigator["maxTouchPoints"] < 256) ||
		(navigator["msMaxTouchPoints"] && navigator["msMaxTouchPoints"] < 256)) {
		return true;
	}
	return false;
}());
/** @const
* @type {boolean}
*/
Util.isSafari = (function () {
	let rejectedExpression = /Chrome|Android|CriOS|FxiOS|EdgiOS/i;
	let safariExpression = /Safari/i;
	let agent = navigator.userAgent;
	if (rejectedExpression.test(agent)) {
		return false;
	}
	return safariExpression.test(agent);
}());

/** @public
 * @type {boolean}
 */
Util.isMobile = (function () {
	return /Android|webOS|iPhone|iPad|iPod|BlackBerry|IEMobile|Opera Mini/i.test(navigator.userAgent);
}());

/** Retrieve an element from Novakit control or any object that has "getElement" method.
* @public
* @ignore
* @param {*} obj If an HTMLElement is given, the same HTMLElement will be returned
* @return {Element} The return value is always element or null
*/
Util.getElement = function (obj) {
	if (Util.isElement(obj)) {
		return /** @type{Element} */(obj);
	} else {
		if (obj && Util.isFunction(obj["getElement"])) {
			return obj["getElement"]();
		}
	}
	return null;
};
/** Some browsers do not support HTMLElement, so we use nodeType to identify the type instead. Works on IE7
* @public
* @ignore
* @param {*} val
* @return {boolean}
*/
Util.isElement = function (val) {
	if (val) {
		return val.nodeType === 1;
	}
	return false;
};
/**
 * Returns `true` if the passed value is empty, `false` otherwise. The value is deemed to be empty if it is either:
 * - `null`
 * - `undefined`
 * - a zero-length array
 * - a zero-length string (Unless the `allowEmptyString` parameter is set to `true`)
 *
 * @param {*} value The value to test.
 * @param {boolean} [allowEmptyString=false] `true` to allow empty strings.
 * @return {boolean}
 */
Util.isEmpty = function (value, allowEmptyString) {
	return (value == null || typeof value === "undefined") ||
		(!allowEmptyString ? value === "" : false) ||
		(Util.isArray(value) && value.length === 0);
};
/**
* @public
* @ignore
* @param {*} val
* @return {boolean}
*/
Util.isNumber = function (val) {
	return typeof val === "number" && isFinite(val);
};
/**
* @public
* @ignore
* @param {*} val
* @return {boolean}
*/
Util.isString = function (val) {
	return typeof val === "string";
};
/** This version of isArray is faster than the one provided by goog.isArray. It works with both literal Array ([]) and Array class (new Array()) for Chrome
* WARNING: It does not work when called across iframe, since Array object is a different object in each iframe
* @public
* @ignore
* @param {*} val
* @return {boolean}
*/
Util.isArray = Array.isArray;
/** This version of isFunction is faster than the one provided by goog.isFunction. However It only works for Modern Browser and not IE
* @public
* @ignore
* @param {*} val
* @return {boolean}
*/
Util.isFunction = function (val) {
	return typeof val === "function";
};
/** Supports only for modern browser (Gecko 1.9+, modern Webkit)
* @public
* @ignore
* @param {!Element|Event|MouseEvent} target
* @param {*=} opt_retObj An object that stores returned values
* @return {!Object}
*/
Util.getClientPosition = function (target, opt_retObj) {
	let pos = opt_retObj || {};
	if (target["getBoundingClientRect"]) {
		let box = target["getBoundingClientRect"]();
		pos["x"] = box.left;
		pos["y"] = box.top;
	} else { // Event
		pos["x"] = target.clientX;
		pos["y"] = target.clientY;
	}
	return /**@type {!Object}*/(pos);
};
/** Top-left of B is used as registration a point (0, 0) for x and y.
 * For example, if A is located to right of B, "x" will be positive, If A is located below B, "y" will be positive.
* @public
* @ignore
* @param {!Element|Event|MouseEvent} A HTMLElement, Browser's event, or Browser's mouse event object
* @param {!Element|Event|MouseEvent} B HTMLElement, Browser's event, or Browser's mouse event object
* @param {*=} opt_retObj An object that stores returned values
* @return {!Object} an object with "x" and "y" values
*/
Util.getRelativePosition = function (A, B, opt_retObj) {
	opt_retObj = Util.getClientPosition(A, opt_retObj);
	let bp = Util.getClientPosition(B);

	opt_retObj["x"] -= bp["x"];
	opt_retObj["y"] -= bp["y"];
	return /**@type {!Object}*/(opt_retObj);
};

/** Move an item in the given array to the specified destination. If the destination equal or greater than array length, the array will be expanded
* @public
* @ignore
* @param {Array} ary
* @param {number} from INCLUSIVE
* @param {number} to INCLUSIVE
* @return {Array} ary
*/
Util.moveArrayItem = function (ary, from, to) {
	if (!ary) { return null; }

	let len = ary.length;
	if (from >= len && to >= len) {
		return /** @type{Array} */(ary); // No item needs to be moved
	}

	let item = ary.splice(from, 1)[0];
	if (to < len) {
		ary.splice(to, 0, item);
	} else {
		ary[to] = item;
	}
	return ary;
};
/** @public
* @function
* @param {Array} ary
* @param {Array} items
* @return {Array}
*/
Util.pushArrayItems = function(ary, items) {
	let len = items ? items.length : 0;
	if(len > 0) {
		if(len < 1e5) { // Maximum number of arguments that the spread operator can take is exactly 125052
			if(len > 1) {
				Array.prototype.push.apply(ary, items);
			} else {
				ary.push(items[0]);
			}
		} else {
			let start = 0;
			let end = 0;
			while(start < len) {
				end += 1e5;
				Array.prototype.push.apply(ary, items.slice(start, end));
				start = end;
			}
		}
	}
	return ary;
};
/** Insert multiple empty items to an array
* @public
* @function
* @param {Array} ary
* @param {number} at Zero or positive number
* @param {number} chunkSize
* @example
* Util.insertEmptyItems(["a", "b"], 1, 2); // ["a", , , "b"]
*/
Util.insertEmptyItems = function (ary, at, chunkSize) {
	if(chunkSize > 0) {
		if(chunkSize > 1) {
			let tailEnd = ary.slice(at);
			ary.length = at; // Remove the tail end part
			ary.length += chunkSize;
			Util.pushArrayItems(ary, tailEnd);
		} else {
			ary.splice(at, 0, null);
		}
	}
};
/** Move all items from the specify index to the front/top of the array
* @public
* @function
* @param {Array} ary
* @param {number} numItems Index of the first item to be moved to the top
* @return {Array}
* @example
* Util.cycleArray(["a", "b", "c"], 2); // ["c", "a", "b"]
*/
Util.cycleArray = function (ary, numItems) {
	if(numItems > 0 && ary) {
		if(numItems < ary.length) {
			let front = ary.splice(0, numItems);
			return ary.concat(front);
		}
	}
	return ary || null;
};

/** STL implementation of lowerBound() of the C++ vector class with no recursion
* @public
* @ignore
* @param {Array.<*>} ary The given array must already be sorted in ascending order
* @param {*} value the value to be searched
* @param {number=} left the begin index of the search range is INCLUSIVE
* @param {number=} right the end index of the search range is EXCLUSIVE
* @param {function(?, ?):boolean=} opt_lessFunc "Less than comparison functor"
* @return {number} Return the lowest index that gives equal or greater than the searched value
*/
Util.lowerBound = function (ary, value, left, right, opt_lessFunc) {
	if (left == null || left < 0) {
		left = 0;
	}
	if (right == null || right < 0 || right > ary.length) {
		right = ary.length;
	}
	if (left > right) {
		left = right;
	}
	if (opt_lessFunc == null) {
		opt_lessFunc = Util._defaultLessComparator;
	}

	let mid = 0;
	while (right > left) {
		mid = (left + right) >> 1;
		if (opt_lessFunc(ary[mid], value)) {
			left = mid + 1;
		} else {
			right = mid;
		}
	}
	return left; // May return ary.length
};

/** Upper bouund is the greatest index that has value that is always greater than the given value
* @public
* @ignore
* @param {Array.<*>} ary The given array must already be sorted in ascending order
* @param {*} value the value to be searched
* @param {number=} left the begin index of the search range is INCLUSIVE
* @param {number=} right the end index of the search range is EXCLUSIVE
* @param {function(?, ?):boolean=} opt_lessFunc "Less than comparison functor"
* @return {number} Return the highest index that gives equal or greater than the searched value
 */
Util.upperBound = function (ary, value, left, right, opt_lessFunc) {
	let at = Util.lowerBound(ary, value, left, right, opt_lessFunc);
	while (ary[at] === value) { ++at; }
	return at;
};

/** This function will replace the string by sequence arguments. (like String.format in JAVA)
 * @param {...string} format - First argument is string to format, the following arguments are the string to substitute
 * @example Util.stringFormat("There is a {0} in a {1}.", "box", "room"); // Result: "There is a box in a room."
 * @return {string}
 */
Util.stringFormat = function (format) {
	let args = Array.prototype.slice.call(arguments, 1);
	return format.replace(/{(\d+)}/g, function (match, number) {
		return (args[number] != null) ? args[number] : match;
	});
};
/**
* @private
* @param {*} left
* @param {*} right
* @return {boolean}
*/
Util._defaultLessComparator = function (left, right) {
	return left < right;
};
/**
* @public
* @function
* @ignore
 */
Util._deprecatedFunction = function () {
	console.warn("Calling deprecated method");
};

/**
 * @public
 * @function
 * @ignore
 * @param {number} lowVal
 * @param {number} lastVal
 * @param {number} highVal
 * @return {number}
 */
Util.calculatePercent = function (lowVal, lastVal, highVal) {
	if ((!lowVal && lowVal !== 0) ||
		(!lastVal && lastVal !== 0) ||
		(!highVal && highVal !== 0)) {
		return NaN;
	}
	if (highVal === lowVal) {
		return 50;
	}
	let percent = (lastVal - lowVal) / (highVal - lowVal);
	if (percent < 0) {
		return 0;
	}
	if (percent > 1) {
		return 100;
	}
	return percent * 100;
};

/** @description The method returns closest ancestor element or the element itself with specified class string. Otherwise, it returns null
 * @public
 * @function
 * @param {Element|Node|undefined} elem
 * @param {string} classStr The class string can only be one single class without spaces
 * @return {Element}
 */
Util.closestElement = function (elem, classStr) {
	let n = elem;
	while (n) {
		if(n.classList) { // DocumentFragment does not have classList property
			if (n.classList.contains(classStr)) {
				return /** @type{Element} */(n);
			}
		}
		n = n.parentNode || n.host; // ShadowRoot has host, but not parentNode
	}
	return null;
};
/** @public
* @function
* @ignore
* @param {Event} e
 */
Util._preventDefault = function (e) {
	e.preventDefault();
	e.stopPropagation();
};

export default Util;
export { Util };
