import { customElement } from "@refinitiv-ui/core/decorators/custom-element.js";
import { BasicElement, html } from "@refinitiv-ui/core";
import "@refinitiv-ui/elements/button";
import "@refinitiv-ui/elements/item";
import "@refinitiv-ui/elements/checkbox";
import translation from "./locale/translation.js";

/**
 * @typedef {Object} CheckboxList~itemState
 * @property {string=} type
 * @property {string=} label
 * @property {boolean=} checked
 * @property {number=} index
 * @property {Element=} elem
 * @property {(string|boolean)=} value
*/
/** @type {number}
* @private
* @constant
*/
const ITEM_LIMIT = 500;

/** @private
* @function
* @param {*=} val
* @return {string}
*/
const _toLabelText = function(val) {
	if(typeof val === "string") {
		return val;
	} else if(val) {
		return val.title || val.label || "";
	}
	return "";
};
/** @private
* @function
* @param {*} val
* @param {number} idx
* @return {CheckboxList~itemState}
*/
const _toItemStates = function(val, idx) {
	let textLabel = _toLabelText(val);
	return {
		type: val.type,
		label: textLabel,
		title: textLabel,
		checked: (val && val.checked) ? true : false,
		index: idx,
		elem: null,
		value: val.value || textLabel
	};
};
/** @private
* @function
* @param {!Object} state
* @param {*} val
*/
const _assignItemState = function(state, val) {
	state.label = _toLabelText(val);
	state.checked = (val && val.checked) ? true : false;
	if(state.elem) {
		_updateElement(state.elem, state.label, state.checked);
	}
};
/** @private
* @function
* @param {!Object} state
* @return {boolean}
*/
const _getCheckedState = function(state) {
	return state.elem ? state.elem._checkbox.checked : state.checked;
};
/** @private
* @function
* @param {!Element} elem
* @param {string} txt
* @param {boolean} checked
*/
const _updateElement = function(elem, txt, checked) {
	if(elem._label.textContent !== txt) {
		elem._label.textContent = txt;
	}
	if(elem._checkbox.checked !== checked) {
		elem._checkbox.checked = checked;
	}
};
/** @private
* @param {string} str
* @param {number} idx
* @param {string} ctx
* @param {string} type
* @returns {boolean}
*/
const _caseInsensitiveFilter = function(str, idx, ctx, type) {
	if(type === "header" || type === "divider") {
		return true;
	}
	return str.toLowerCase().indexOf(ctx) >= 0;
};

class CheckboxList extends BasicElement {
	static get properties() {
		return {
			config: {type: Object},
			data: {type: Array},
			lang: { type: String, reflect: true }
		};
	}

	constructor() {
		super();

		this.selectAll = this.selectAll.bind(this);
		this.deselectAll = this.deselectAll.bind(this);
		this._onItemClicked = this._onItemClicked.bind(this);
		this._toItemElement = this._toItemElement.bind(this);
		this._byLabelText = this._byLabelText.bind(this);

		this.lang = "en";
		this._translation = translation;

		this._config = {};
		this._states = []; // Item states before filtering
		this._cachedItems = []; // To avoid repeatedly creating the same item element
		this._items = []; // Actual existing elements (filtered items)
		this._filterStates = []; // Existing item (after filtered)
		this._inputFilter = null;
	}

	update(changedProperties) {
		if(changedProperties.has("config") && !changedProperties.has("data")) {
			if(Array.isArray(this.config)) {
				this.data = this.config;
				this.config = null;
				this._inputFilter = null;
			} else if(this.config && this.config.data != null) {
				this.data = this.config.data;
				this._inputFilter = null;
			}
		}

		if(changedProperties.has("data")) {
			this._inputFilter = null;
		}

		super.update(changedProperties);
	}

	render() {
		let lang = this._translation[this.lang] ? this.lang : "en";
		let t =  this._translation[lang];
		return html`
			<div class='root'>
				<div>
					<ef-button class="select-btn" id="select_all">${t["Select All"]}</ef-button>
					<ef-button class="select-btn hidden" id="deselect_all">${t["Deselect All"]}</ef-button>
				</div>
				<div id="container"></div>
			</div>
		`;
	}

	firstUpdated(changedProps) {
		// Prepare list item template. TODO: This should be static variable
		let checkBox = document.createElement("ef-checkbox");
		let label = document.createElement("div");
		label.classList.add("item-label");
		let itemContainer = document.createElement("div");
		itemContainer.classList.add("item-container");
		itemContainer.appendChild(checkBox);
		itemContainer.appendChild(label);
		this._templateItemEl = document.createElement("ef-item");
		this._templateItemEl.appendChild(itemContainer);

		// Elements has been rendered. Cache element references for better performance
		let root = this.shadowRoot;
		this._containerEl = root.getElementById("container");

		this._selectAllBtn = root.getElementById("select_all");
		this._deselectAllBtn = root.getElementById("deselect_all");

		this._selectAllBtn.addEventListener("click", this.selectAll);
		this._deselectAllBtn.addEventListener("click", this.deselectAll);
	}

	updated(changedProps) {
		if (changedProps.has("data")) {
			let dataList = this.data;
			let len = dataList ? dataList.length : 0;
			if (this._cachedItems.length < len) {
				this._cachedItems.length = len;
			}

			this._states = dataList.map(_toItemStates);
			this._filterItems();
			this._renderItems();
		}
	}

	/** @private
	* @param {!Object} state
	* @return {!Element}
	*/
	_toItemElement(state) {
		let elem = state.elem;
		if(!elem) {
			let idx = state.index;
			elem = this._cachedItems[idx];
			if(!elem) {
				elem = this._cachedItems[idx] = this._templateItemEl.cloneNode(true);
				elem.addEventListener("click", this._onItemClicked);

				let itemContainer = elem.firstChild; // div.item-container
				elem._checkbox = itemContainer.firstChild; // ef-checkbox
				elem._label = itemContainer.lastChild; // label
			}

			_updateElement(elem, state.label, state.checked);
			state.elem = elem;
		}
		return elem;
	}
	/**
	* @private
	*/
	_renderItems() {
		if(!this._containerEl) {
			return;
		}

		// TODO: Optimize appending and removing items by finding appropriate item number
		let container = this._containerEl;
		while(container.lastChild) {
			container.removeChild(container.lastChild);
		}

		let ary = this._items;
		let filterStates = this._filterStates;
		let len = ary.length;
		for(let i = 0; i < len; ++i) {
			let item = ary[i];
			let state = filterStates[i];
			item.type = "";
			let itemContainer = item.firstChild;
			if(!itemContainer.contains(item._checkbox)){
				itemContainer.insertBefore(item._checkbox, item._label);
			}
			let type = state.type;
			if(type) {
				if(type === "header") {
					itemContainer.removeChild(item._checkbox);
				}
				item.type = type;
			}
			container.appendChild(item);
		}
		this._updateSelectAllState();
	}

	/**
	* @public
	* @param {Array<Object>} list
	* @example
	*   setData([{
	*     title: "Refinitiv"
	*     checked: false
	*   }]);
	*/
	setData(list) {
		this.data = list;
	}

	/**
	* @public
	* @param {Array<Object>=} list
	*/
	updateData(list) {
		if(list && this.data !== list) {
			this.data = list;
			return;
		}

		let states = this._states;
		let len = states.length;
		for(let i = 0; i < len; ++i) {
			_assignItemState(states[i], this.data[i]);
		}

		this._updateSelectAllState();
	}

	/**
	* @public
	* @return {!Array<object>}
	*/
	getSelectedItems() {
		let items = [];
		let states = this._states;
		let len = states.length;
		for(let i = 0; i < len; ++i) {
			let state = states[i];
			if(_getCheckedState(state)) {
				items.push({
					title: state.label,
					label: state.label,
					index: i,
					value: state.value
				});
			}
		}
		return items;
	}

	/**
	* @public
	* @param {boolean=} selected
	*/
	selectAll(selected) {
		selected = selected !== false;
		let ary = this._items;
		let len = ary.length;
		for(let i = 0; i < len; ++i) {
			ary[i]._checkbox.checked = selected;
		}
		this._updateSelectAllState(selected);
		if(selected) {
			this._deselectAllBtn.focus();
		} else {
			this._selectAllBtn.focus();
		}
	}

	/**
	* @public
	*/
	deselectAll() {
		this.selectAll(false);
	}

	/**
	* @public
	* @param {(string|Function)=} inputFilter
	*/
	filter(inputFilter) {
		if(!inputFilter) {
			inputFilter = null;
		}
		let dirty = (typeof inputFilter === "function");
		if(this._inputFilter !== inputFilter) {
			this._inputFilter = inputFilter || null;
			dirty = true;
		}
		if(dirty) {
			this._filterItems();
			this._renderItems();
		}
	}

	/**
	* @private
	*/
	_filterItems() {
		let states = this._states;
		let len = states ? states.length : 0;
		if(!len) {
			this._items.length = 0;
			return; // No new copy is generated
		}

		let inputFilter = this._inputFilter;
		if(inputFilter) {
			if(typeof inputFilter === "function") {
				this._filterMethod = inputFilter;
				this._filterCtx = states;
			} else {
				this._filterMethod = _caseInsensitiveFilter;
				this._filterCtx = (inputFilter + "").toLowerCase(); // Convert to lowercased string for case insensitive comparison
			}
			states = states.filter(this._byLabelText);

			this._filterMethod = null;
			this._filterCtx = null;
		}

		if(states.length > ITEM_LIMIT) {
			states = states.slice(0, ITEM_LIMIT);
		}

		this._filterStates = states;
		this._items = states.map(this._toItemElement);
	}
	/**
	* @private
	* @param {Object} state
	* @param {number} idx
	* @return {boolean}
	*/
	_byLabelText(state, idx) {
		return this._filterMethod(state.label, idx, this._filterCtx, state.type);
	}

	/**
	* @private
	* @param {Event} e
	*/
	_onItemClicked(e) {
		if(e.target.tagName !== "EF-CHECKBOX") { // checkbox may already toggle its state
			let itemEl = e.currentTarget;
			itemEl._checkbox.checked = !itemEl._checkbox.checked;
		}
		this._updateSelectAllState();
	}

	/**
	* @private
	* @param {boolean=} selected=true mean user click "Select All" button, so all check box was checked and button changed to "Deselect All"
	*/
	_updateSelectAllState(selected) {
		if(!this._containerEl) {
			return;
		}

		if(selected == null) {
			let ary = this._items;
			let len = ary.length;
			selected = len ? false : true;
			for (let i = 0; i < len; i++) {
				if (ary[i]._checkbox.checked) {
					selected = true;
					break;
				}
			}
		}

		this._selectAllBtn.classList.toggle("hidden", selected);
		this._deselectAllBtn.classList.toggle("hidden", !selected);
	}
}

customElement("checkbox-list")(CheckboxList);

export { CheckboxList };
export default CheckboxList;
