import FormatterBuilder from "./FormatterBuilder.js";
import ElfUtil from "../../tr-grid-util/es6/ElfUtil.js";

/** @typedef {Object} CoralComboBoxFormatter~Options
* @description CoralComboBoxFormatter options
* @property {Array=} data Item list to be displayed on the coral select. The given object should contains at least value properties. If string is given, it will be treated as the value.
* @property {Array=} entries Alias to `data` property
* @property {Array=} items Alias to `data` property
*/

/** @private
* @param {Object} ctx Context object
* @param {Object} options User options
*/
let onContextCreated = function (ctx, options) {
	let userItems = options.data || options.entries || options.items;
	if(Array.isArray(userItems)) {
		let i;
		let len = userItems.length;
		let userItem;
		for(i = 0; i < len; ++i) {
			userItem = userItems[i];
			if(userItem != null) {
				if (typeof userItem !== "object") {
					userItem = { value: userItem };
				}
			} else {
				userItem = {value: ""};
			}
			let label = userItem.label;
			if(!label && typeof label !== "string") { // 0, false, null, undefined, NaN
				userItem.label = userItem.value + "";
			}
			userItems[i] = userItem;
		}
		ctx.data = userItems;

		let mappedData = ctx._mappedData = {}; // WORKAROUND: Since combobox does not accept raw value for selection, it actually needs item element
		for (i = 0; i < len; ++i) {
			userItem = userItems[i];
			mappedData[userItem.value] = userItem;
		}
	}
};

let onElementCreated = function (element, ctx) {
	element.data = ctx.data;
};

let onElementUpdated = function (element, ctx) {
	if (ctx._mappedData) {
		if(ElfUtil.getElfVersion() > 3) {
			element.value = ctx.value;
		} else {
			element.value = ctx._mappedData[ctx.value];
		}
	}
};

let _changeHandler = function (e, ctx) {
	if (e.detail.value == null) {
		return;
	}
	if(ElfUtil.getElfVersion() > 3) {
		ctx.value = e.detail.value;
	} else {
		ctx.value = e.detail.value.value;
	}
	ctx.setData(ctx.field, ctx.value);
};

/** @constructor
* @param {*=} options
*/
let CoralComboBoxFormatter = function (options) {

};

/** @public
* @param {(FormatterBuilder.Options|CoralComboBoxFormatter.Options|Object)=} options
* @return {Object}
*/
CoralComboBoxFormatter.create = function (options) {
	let defaultOpt = {
		tagName: "ef-combo-box",
		refName: "CoralComboBoxFormatter",
		styles: {
			maxWidth: "100%",
			margin: "0"
		},
		onContextCreated: onContextCreated,
		onElementCreated: onElementCreated,
		onElementUpdated: onElementUpdated,
		changeHook: "value-changed",
		changeHandler: _changeHandler,
		data: null
	};

	return FormatterBuilder.create(options, defaultOpt);
};

export default CoralComboBoxFormatter;
export { CoralComboBoxFormatter };
