/** @typedef {Object} FormatterBuilder~Options
* @description FormatterBuilder options
* @property {string=} field
* @property {Object=} attributes
* @property {Object=} styles
* @property {Object=} events
* @property {Function=} changeHandler
* @property {Array.<Function>=} changeHandlers
* @property {Function=} onElementRendered
*/

/** Dictionary that contain reserved attribute to context element
* @type {!Object}
* @private
* @constant
*/
let ReservedAttributes = {
	"api": true,
	"wrapper": true, // wrapper is generated during runtime binding
	"grid": true,
	"field": true,
	"rowIndex": true,
	"value": true,
	"setData": true,
	"getData": true,
	"disableElement": true,
	"isElementDisabled": true,
	"setError": true,
	"getError": true,
	"styles": true,
	"attributes": true,
	"tagName": true, // for internal use only
	"refName": true, // for internal use only
	"events": true,
	"userEvents": true, // for internal use only
	"defaultEvents": true, // for internal use only
	"onContextCreated": true, // for internal use only
	"onElementCreated": true, // for internal use only
	"onElementUpdated": true, // for internal use only
	"onElementRendered": true,
	"changeHook": true, // for internal use only
	"changeHooks": true, // for internal use only
	"changeHandler": true,
	"changeHandlers": true
};

/** Clone all non-reserved properties from master to base
* @private
* @function
* @param {Object} base Base object
* @param {Object} master Master object. There will be no modification on this object
*/
function _assignProperties(base, master) {
	for (let key in master) {
		if (!ReservedAttributes[key]) {
			base[key] = master[key];
		}
	}
}
/** Clone a property with Object map from master to base
* @private
* @function
* @param {Object} base Base object
* @param {Object} master Master object. There will be no modification on this object
* @returns {boolean} Returns true, if there is any change
*/
function _assignAttributes(base, master) {
	let mObj = master.attributes;
	if(!mObj) {
		return false;
	}

	let dirty = false;
	let bObj = base.attributes;
	if(!bObj) {
		bObj = {};
	}
	for(let key in mObj) {
		dirty = true;
		bObj[key] = mObj[key];
	}
	if(!dirty) {
		return false;
	}

	if(!base.attributes) {
		base.attributes = bObj;
	}
	return true;
}
/** Clone a property with Object map from master to base
* @private
* @function
* @param {Object} base Base object
* @param {Object} master Master object. There will be no modification on this object
* @returns {boolean} Returns true, if there is any change
*/
function _assignStyles(base, master) {
	let mObj = master.styles;
	if(!mObj) {
		return false;
	}

	let dirty = false;
	let bObj = base.styles;
	if(!bObj) {
		bObj = {};
	}
	for(let key in mObj) {
		dirty = true;
		let val = mObj[key];
		if(key === "width" || key === "height") {
			if(typeof val === "number") {
				val = val + "px";
			}
		}
		bObj[key] = val;
	}
	if(!dirty) {
		return false;
	}

	if(!base.styles) {
		base.styles = bObj;
	}
	return true;
}

/** @private
* @function
* @param {Object} ctx
* @param {Function} userFunc
* @param {Object} e
*/
function _positionLocator(ctx, userFunc, e) {
	if (!ctx.grid.isBinding()) {
		ctx.grid.getRelativePosition(e, ctx);
		userFunc(e, ctx);
	}
}
/** Clone event listerners from master to base
* @private
* @function
* @param {Object} base Base object
* @param {Object} master Master object. There will be no modification on this object
* @param {string} optionType
* @returns {boolean} Returns true, if there is any change
*/
function _assignListener(base, master, optionType) {
	let mObj = master.events;
	if(!mObj) {
		return false;
	}

	let dirty = false;
	let bObj = base[optionType];
	if(!bObj) {
		bObj = {};
	}
	for (let key in mObj) {
		let func = mObj[key];
		if (typeof func === "function") {
			dirty = true;
			bObj[key] = _positionLocator.bind(null, base, func);
		}
	}
	if(!dirty) {
		return false;
	}
	if(!base[optionType]) {
		base[optionType] = bObj;
	}
	return true;
}

/** @private
* @param {Object} ctx
* @param {Object} colDef User's column definition object
* @param {tr.Grid} grid Core grid instance
*/
function _onInitContextObject(ctx, colDef, grid) { // Executed once per column
	ctx.grid = grid; // WARNING: using ctx as closure variable
	if (!ctx.field) {
		ctx.field = colDef.field;
		if(ctx.disablingField == null) {
			ctx.disablingField = ctx.field ? ctx.field + "_DISABLED" : null;
		}
		if(ctx.errorField == null) {
			ctx.errorField = ctx.field ? ctx.field + "_ERROR" : null;
		}
	}

	ctx.disableElement = _disableElement;
	ctx.isElementDisabled = _isElementDisabled;
	ctx.setError = _setError;
	ctx.getError = _getError;
}
/** @private
* @this ContextObject
* @param {string} field
* @param {*} value
*/
function _setRtGridData(field, value) {
	let rowDef = this.api.getRowDefinition(this.rowIndex);
	if(rowDef) {
		rowDef.setStaticData(field, value);
	}
}
/** @private
* @this ContextObject
* @param {boolean} field
* @returns {*} Value
*/
function _getRtGridData(field) {
	let rowDef = this.api.getRowDefinition(this.rowIndex);
	return rowDef ? rowDef.getData(field) : null;
}
/** @private
* @this ContextObject
* @param {string} field
* @param {*} value
*/
function _setCoreGridData(field, value) {
	this.api.getDataView().setDataAt(this.rowIndex, field, value);
}
/** @private
* @this ContextObject
* @param {boolean} field
* @returns {*} Value
*/
function _getCoreGridData(field) {
	return this.api.getDataView().getDataAt(this.rowIndex, field);
}
/** @private
* @this ContextObject
* @param {boolean} bool
*/
function _disableElement(bool) {
	if(this.disablingField) {
		this.setData(this.disablingField, bool !== false);
	}
}
/** @private
* @this ContextObject
* @returns {boolean}
*/
function _isElementDisabled() {
	return this.getData(this.disablingField);
}
/** @private
* @this ContextObject
* @param {*} val
*/
function _setError(val) {
	if(this.errorField) {
		this.setData(this.errorField, val);
	}
}
/** @private
* @this ContextObject
* @returns {*}
*/
function _getError() {
	return this.getData(this.errorField);
}

/** @private
* @function
* @param {Object} ctx
* @param {Object} e
*/
function _predefinedBinding(ctx, e) {
	ctx.api = e["grid"];
	if(!ctx.wrapper) {
		ctx.wrapper = "real-time-grid";
		ctx.setData = _setRtGridData;
		ctx.getData = _getRtGridData;
	}
	let rowData = ctx.rowData = e["rowData"];
	if(ctx.field) {
		ctx.value = rowData ? rowData[ctx.field] : null;
	} else {
		ctx.value = e["data"];
	}
	ctx.rowIndex = e["rowIndex"];
	ctx.colIndex = e["colIndex"];
	ctx.cell = e["cell"];

	_commonBinding(ctx);
}
/** @private
* @function
* @param {Object} ctx
* @param {number} rowIndex
* @param {number} columnIndex
* @param {*} value
* @param {Object} cell
* @param {Object} colDef
* @param {Object} dataRow
* @param {Object} dataTable
* @param {Object} dataView
* @param {Object} cgrid
*/
function _predefinedFormatter(ctx, rowIndex, columnIndex, value, cell, colDef, dataRow, dataTable, dataView, cgrid) {
	ctx.api = cgrid;
	if(!ctx.wrapper) {
		ctx.wrapper = "composite-grid";
		ctx.setData = _setCoreGridData;
		ctx.getData = _getCoreGridData;
	}
	let rowData = ctx.rowData = dataRow;
	if (ctx.field) {
		ctx.value = rowData ? rowData[ctx.field] : null;
	} else {
		ctx.value = value;
	}
	ctx.rowIndex = rowIndex;
	ctx.colIndex = columnIndex;
	ctx.cell = cell;

	_commonBinding(ctx);
}
/** @private
* @function
* @param {Object} ctx
*/
function _commonBinding(ctx) {
	let elem = ctx.cell.getContent();
	if(!elem || (elem._formatterName !== ctx.refName)) {
		elem = document.createElement(ctx.tagName);
		elem._formatterName = ctx.refName;
		if (ctx.onElementCreated) {
			ctx.onElementCreated(elem, ctx);
		}

		let attrs = ctx.attributes;
		if(attrs) {
			for(let attrName in attrs) {
				elem.setAttribute(attrName, attrs[attrName] || "");
			}
		}

		let stys = ctx.styles;
		if(stys) {
			for(let styName in stys) {
				elem.style[styName] = stys[styName];
			}
		}

		let evts, evtName;
		evts = ctx.defaultEvents;
		if(evts) {
			for (evtName in evts) {
				elem.addEventListener(
					evtName,
					evts[evtName]
				);
			}
		}
		evts = ctx.userEvents;
		if(evts) {
			for (evtName in evts) {
				elem.addEventListener(
					evtName,
					evts[evtName]
				);
			}
		}
	}

	let rowData = ctx.rowData;
	if(ctx.disablingField) {
		if(rowData && rowData[ctx.disablingField]){
			elem.setAttribute("disabled", "");
		} else {
			elem.removeAttribute("disabled");
		}
	}
	if(ctx.errorField) {
		if(rowData && rowData[ctx.errorField]){
			elem.setAttribute("error", "");
		} else {
			elem.removeAttribute("error");
		}
	}

	if(ctx.onElementUpdated) {
		ctx.onElementUpdated(elem, ctx);
	}

	if (ctx.onElementRendered) {
		ctx.onElementRendered(elem, ctx);
	}

	ctx.cell.setContent(elem);
}

/** @private
* @function
* @param {Object} ctx Context object
* @param {Object} userOpt User options
* @param {Object} defaultOpt Default options
*/
function _buildContextObject(ctx, userOpt, defaultOpt) {
	_assignProperties(ctx, defaultOpt);
	_assignProperties(ctx, userOpt);

	_assignAttributes(ctx, defaultOpt);
	_assignAttributes(ctx, userOpt);

	_assignStyles(ctx, defaultOpt);
	_assignStyles(ctx, userOpt);

	_assignListener(ctx, defaultOpt, "defaultEvents");
	_assignListener(ctx, userOpt, "userEvents");

	ctx.field = userOpt.field || "";
	ctx.refName = defaultOpt.refName || "_defaultFormatter";
	ctx.tagName = defaultOpt.tagName || "div";

	if (defaultOpt.onElementCreated) {
		ctx.onElementCreated = defaultOpt.onElementCreated;
	}
	if (defaultOpt.onElementUpdated) {
		ctx.onElementUpdated = defaultOpt.onElementUpdated;
	}
	if (typeof userOpt.onElementRendered === "function") {
		ctx.onElementRendered = userOpt.onElementRendered;
	}
}

/** @private
*/
function _emptyFunc() {}

/** @private
* @param {Object} ctx
* @return {Object}
*/
function _getContextObject(ctx) {
	return ctx;
}

/** @constructor
* @param {*=} options
*/
let FormatterBuilder = function (options) {
	// TODO: FormatterBuilder should not be initializable
};

/** @public
* @param {FormatterBuilder~Options=} userOptions
* @param {FormatterBuilder~Options=} defaultOpt
* @return {Object}
*/
FormatterBuilder.create = function (userOptions, defaultOpt) {
	userOptions = userOptions || {};
	defaultOpt = defaultOpt || {};

	if (defaultOpt.changeHook) {
		let changeHandler = userOptions.changeHandler || defaultOpt.changeHandler;
		if (!defaultOpt.events) {
			defaultOpt.events = {};
		}
		defaultOpt.events[defaultOpt.changeHook] = changeHandler;
	} else if (defaultOpt.changeHooks) {
		let changeHandlers = userOptions.changeHandlers || defaultOpt.changeHandlers;
		if (!defaultOpt.events) {
			defaultOpt.events = {};
		}
		for (let i = 0; i < changeHandlers.length; i++) {
			defaultOpt.events[defaultOpt.changeHooks[i]] = changeHandlers[i];
		}
	}

	let ctx = {};
	_buildContextObject(ctx, userOptions, defaultOpt);

	if (defaultOpt.onContextCreated) {
		defaultOpt.onContextCreated(ctx, userOptions);
	}

	return {
		render: _emptyFunc,
		init: _onInitContextObject.bind(null, ctx),
		bind: _predefinedFormatter.bind(null, ctx),
		binding: _predefinedBinding.bind(null, ctx),
		getContextObject: _getContextObject.bind(null, ctx)
	};
};

export default FormatterBuilder;
export { FormatterBuilder };
