/** @constructor
*/
let ReferenceCounter = function() {
	this._counter = {};
	this._session = {};
	this._referers = {};
};

/** @type {Object.<string, number>}
* @private
**/
ReferenceCounter.prototype._counter = null;
/** @type {Object.<string, number>}
* @private
**/
ReferenceCounter.prototype._session = null;
/** @type {Object.<string, Object>}
* @private
**/
ReferenceCounter.prototype._referers = null;


/** @public
*/
ReferenceCounter.prototype.reset = function() {
	this._counter = {};
	this._referers = {};
	this.newSession();
};
/** @public
*/
ReferenceCounter.prototype.newSession = function() {
	this._session = {};
};

/** @public
* @return {!Array.<string>}
*/
ReferenceCounter.prototype.getAllReferences = function() {
	return Object.keys(this._counter);
};

/** @public
* @return {boolean}
*/
ReferenceCounter.prototype.hasReference = function() {
	for (const key in this._counter) {
		return true;
	}
	return false;
};

/** @public
* @return {!Array.<string>}
*/
ReferenceCounter.prototype.getAllReferers = function() {
	return Object.keys(this._referers);
};
/** Alias to ReferenceCounter.getAllReferers
* @public
* @function
* @return {!Array.<string>}
*/
ReferenceCounter.prototype.getAllReferrers = ReferenceCounter.prototype.getAllReferers;

/** @public
* @return {Object} Object with newEntries and removedEntries
*/
ReferenceCounter.prototype.getSession = function() {
	let newEntries = [];
	let removedEntries = [];
	for(let key in this._session) {
		let val = this._session[key];
		if(val > 0) {
			newEntries.push(key);
		} else if(val < 0) {
			removedEntries.push(key);
		}
		// else {} // when val 0 do nothing, doesn't change anything
	}
	return {
		newEntries: newEntries.filter(Boolean),
		removedEntries: removedEntries.filter(Boolean)
	};
};

/** @public
* @param {string} key
* @param {string=} referer Referer can only add specific key only once
* @return {boolean} True if new reference is added
*/
ReferenceCounter.prototype.addReference = function(key, referer) {
	if(key) {
		if(referer) {
			let counter = this._referers[referer];
			if(counter) {
				if(counter[key]) { // Single Referer can only add specific key only once
					return false;
				}
			} else {
				counter = this._referers[referer] = {};
			}
			counter[key] = 1;
		}

		if(this._counter[key]) {
			++this._counter[key];
			// The session will not change when a field already exists and a counter is attempted to be added
		} else {
			this._counter[key] = 1;
			if(this._session[key] === -1) {
				this._session[key] = 0;
			} else {
				this._session[key] = 1;
			}

			return true;
		}
	}
	return false;
};
/** @public
* @function
* @param {string} key
* @return {boolean} True if new reference is added
*/
ReferenceCounter.prototype.addRef = ReferenceCounter.prototype.addReference;

/** @public
* @param {string} key
* @param {string=} referer Referer can only add specific key only once
* @param {number=} count Use negative value to remove all counter from the given reference
* @return {boolean} True if existing reference is removed
*/
ReferenceCounter.prototype.removeReference = function(key, referer, count) {
	if(!key) {
		return false;
	}
	let val = this._counter[key];
	if(!val) {
		return false;
	}
	if(count == null) {
		count = 1;
	} else if(count < 0) {
		count = val;
	} else if(!count) {
		return false;
	}

	if(referer) {
		let counter = this._referers[referer];
		if(!counter) {
			return false;
		}
		if(!counter[key]) {
			return false;
		}

		delete counter[key];
	}

	val -= count;
	if(!val || val < 0) {
		delete this._counter[key];
		if(this._session[key] === 1) {
			this._session[key] = 0;
		} else {
			this._session[key] = -1;
		}
		return true;
	}

	this._counter[key] = val;
	return false;
};
/** @public
* @function
* @param {string} key
*/
ReferenceCounter.prototype.release = ReferenceCounter.prototype.removeReference;

/** @public
* @param {(string|Array.<string>|Object)} keys
* @param {string=} referer
* @return {number} Number of newly added entries
*/
ReferenceCounter.prototype.addReferences = function(keys, referer) {
	if(!keys) {
		return 0;
	}
	let count = 0;
	if(Array.isArray(keys)) {
		let len = keys.length;
		for(let i = 0; i < len; ++i) {
			if(this.addReference(keys[i], referer)) {
				++count;
			}
		}
	} else if (typeof keys == "string") {
		if(this.addReference(keys, referer)) {
			++count;
		}
	} else {
		for(let key in keys) {
			if(this.addReference(key, referer)) {
				++count;
			}
		}
	}
	return count;
};

/** Remove all references from specific referer
* @public
* @param {string=} referer
* @return {number} Number of removed entries
*/
ReferenceCounter.prototype.removeReferer = function(referer) {
	let count = 0;
	if(referer) {
		let counter = this._referers[referer];
		if(counter) {
			for(let key in counter) {
				if(this.removeReference(key, referer)) {
					++count;
				}
			}
			delete this._referers[referer];
		}
	}
	return count;
};
/** Alias to ReferenceCounter.removeReferer
* @public
* @function
* @param {string=} referer
* @return {number} Number of removed entries
*/
ReferenceCounter.prototype.removeReferrer = ReferenceCounter.prototype.removeReferer;

/** Log current reference count to dev console
* @public
*/
ReferenceCounter.prototype.log = function() {
	console.log(this._counter);
};


export {ReferenceCounter};
export default ReferenceCounter;
