import { cloneObject } from "../../tr-grid-util/es6/Util.js";

/** @private
* @param {*} a
* @param {*} b
* @param {number} order
* @return {number}
*/
let _defaultCompare = function(a, b, order) {
	if(a == null || a !== a) {
		if(b == null || b !== b) {
			return 0;
		}
		return 1;
	}
	if(b == null || b !== b) {
		return -1;
	}

	if(a < b) {
		return -order;
	}
	if(b < a) {
		return order;
	}
	return 0;
};

/** @constructor
*/
let RowDefSorter = function() {
	this._defaultSorter = this._defaultSorter.bind(this);
	this._dataSorter = this._dataSorter.bind(this);
	this._rowDefSorter = this._rowDefSorter.bind(this);
	this._multiColumnsSorter = this._multiColumnsSorter.bind(this);

	this._globalContext = {};
	this._sortParams = [];
	this._ctxCaches = [];
};

/** @type {!Object}
* @private
*/
RowDefSorter.prototype._globalContext;
/** @type {!Array.<Array>}
* @private
*/
RowDefSorter.prototype._sortParams;
/** @type {!Array.<Object>}
* @private
*/
RowDefSorter.prototype._ctxCaches;
/** @type {Array}
* @private
*/
RowDefSorter.prototype._primaryParams;


/** @public
*/
RowDefSorter.prototype.dispose = function() {
	this._globalContext = {}; // Clear any existing reference
	this._sortParams.length = 0;
	this._ctxCaches.length = 0;
	this._primaryParams = null;
};

/** @public
* @return {Function}
*/
RowDefSorter.prototype.getSorter = function() {
	this._primaryParams = null;
	let sortCount = this._sortParams.length;
	if(sortCount === 1) {
		let params = this._primaryParams = this._sortParams[0];
		let sortLogic = params[1];
		if(sortLogic) {
			let rowSorting = params[3];
			return rowSorting ? this._rowDefSorter : this._dataSorter;
		} else {
			return this._defaultSorter;
		}
	} else if(sortCount > 1) {
		return this._multiColumnsSorter;
	}

	return RowDefSorter._noSorting;
};

/** @public
* @param {string} key
* @param {*} value
*/
RowDefSorter.prototype.reset = function() {
	if(this._sortParams.length) {
		this._sortParams.length = 0;
	}
};
/** @public
* @param {string} key
* @param {*} value
*/
RowDefSorter.prototype.setContext = function(key, value) {
	this._globalContext[key] = value;
};
/** @public
* @param {string} field
* @param {Function} logic
* @param {boolean} rowSorting
* @param {string} order
* @param {number} colIndex
* @param {*} colDef
*/
RowDefSorter.prototype.addColumnContext = function(field, logic, rowSorting, order, colIndex, colDef) {
	if(!field) {
		field = "";
	}
	let sortPriority = this._sortParams.length;
	let ctx = this._ctxCaches[sortPriority];
	if(!ctx) {
		ctx = this._ctxCaches[sortPriority] = cloneObject(this._globalContext);
	}
	let orderNum = 0;
	if(order === "a") {
		orderNum = 1;
	} else if(order === "d") {
		orderNum = -1;
	}

	let params = [
		field, // 0
		(typeof logic === "function") ? logic : null,  // 1
		ctx, // 2
		rowSorting, // 3
		orderNum  // 4
	];

	ctx["colIndex"] = colIndex;
	ctx["field"] = field;
	ctx["formattedField"] = field + "_FORMATTED";
	ctx["colDef"] = colDef;

	this._sortParams.push(params);
};

/** @private
* @param {RowDefinition} rowDefA
* @param {RowDefinition} rowDefB
* @param {number} order
* @return {number}
*/
RowDefSorter._noSorting = function(rowDefA, rowDefB, order) {
	return 0;
};
/** @private
* @param {RowDefinition} rowDefA
* @param {RowDefinition} rowDefB
* @param {number} order
* @return {number}
*/
RowDefSorter.prototype._defaultSorter = function(rowDefA, rowDefB, order) {
	let field = this._primaryParams[0];
	return _defaultCompare(
		rowDefA.getData(field),
		rowDefB.getData(field),
		order
	);
};
/** @private
* @param {RowDefinition} rowDefA
* @param {RowDefinition} rowDefB
* @param {number} order
* @return {number}
*/
RowDefSorter.prototype._dataSorter = function(rowDefA, rowDefB, order) {
	let params = this._primaryParams;
	let field = params[0];
	let sortLogic = params[1];
	return sortLogic(
		rowDefA.getData(field),
		rowDefB.getData(field),
		order,
		params[2]
	);
};
/** @private
* @param {RowDefinition} rowDefA
* @param {RowDefinition} rowDefB
* @param {number} order
* @return {number}
*/
RowDefSorter.prototype._rowDefSorter = function(rowDefA, rowDefB, order) {
	let params = this._primaryParams;
	let sortLogic = params[1];
	return sortLogic(rowDefA, rowDefB, order, params[2]);
};
/** @private
* @param {RowDefinition} rowDefA
* @param {RowDefinition} rowDefB
* @param {number} primaryOrder
* @return {number}
*/
RowDefSorter.prototype._multiColumnsSorter = function(rowDefA, rowDefB, primaryOrder) {
	let sortParams = this._sortParams;
	let sortCount = sortParams.length;
	for(let i = 0; i < sortCount; ++i) {
		let params = sortParams[i];
		let field = params[0];
		let sortLogic = params[1];
		let ctx = params[2];
		let rowSorting = params[3];
		let orderNum = params[4];
		let ret = 0;
		if(sortLogic) {
			if(rowSorting) {
				ret = sortLogic(rowDefA, rowDefB, orderNum, ctx);
			} else {
				ret = sortLogic(rowDefA.getData(field), rowDefB.getData(field), orderNum, ctx);
			}
		} else {
			ret = _defaultCompare(rowDefA.getData(field), rowDefB.getData(field), orderNum);
		}
		if(ret) {
			return ret;
		}
	}
	return 0;
};

export default RowDefSorter;
