import { Ext } from "../../tr-grid-util/es6/Ext.js";
import { GridPlugin } from "../../tr-grid-util/es6/GridPlugin.js";

/** @typedef {Object} StatisticsRowPlugin~Options
* @description Available options describing `statisticsRow` object specified in grid's option
* @property {Array=} rows=null Array of StatisticsRowPlugin~RowOptions
* @property {string=} invalidText="" Text to be shown for invalid statistic (i.e. Not a Number value)
* @property {boolean=} noColoring=false If enabled, all conditional coloring (i.e. tick coloring) will not automatically be applied.
* @property {boolean=} noFormatting=false If enabled, all formatting (i.e. decimal places) will not automatically be applied.
* @property {boolean=} useAllData=true If disabled, filtered out rows will not be included in the calculation.
* @property {Function=} postCalculation Function handler excuted after statistics calculation and before default rendering
* @property {Function=} postRendering Function handler excuted after default rendering
*/

/** @typedef {Object} StatisticsRowPlugin~ColumnOptions
* @description Available column options specified in grid's column configuration
* @property {(boolean|string)=} statistics=false If enabled, the column is qualified for the calculation. If "label" is specified, the column will be used to show statistic label
*/

/** @typedef {Object} StatisticsRowPlugin~RowOptions
* @description Available options describing statistics row specified as item in `rows` array property.
* @property {string=} placement="footer" Available footer are "header", "title", and "footer"
* @property {string=} statistic="" Statistic to be calculated. Predefined values are "average", "min", "max", "count", and "sum"
* @property {string=} label=null By default, label will be statistics name.
* @property {string=} id="" Text Id used for row referencing in extension's APIs
*/

/** @typedef {string|number} StatisticsRowPlugin~RowReference
* @description Statistics Row can be referenced by one of the following ways: <br>
* 1. Number as index counted from top to bottom and only for the statistics rows (excluding normal rows) <br>
* 2. String as statistic (e.g., "average", "min", "max") <br>
* 3. String as id defined in the row configuration object
*/

/** @typedef {Object} StatisticsRowPlugin~Stats
* @description Statistics Object contains calculated statistic values that will be provided in `postCalculation` and `postRendering` events
* @property {string=} field
* @property {boolean=} label Indicating the column is used to show label
* @property {number=} count Count of all valid values in the column. Invalid values such as empty, null, undefined, string values are excluded. For label column, counting method will include only truthy values except number 0, and exclude all other falsy values.
* @property {number=} sum Summation
* @property {number=} average Average
* @property {number=} min Minimum
* @property {number=} max Maximum
*/

/** @event StatisticsRowPlugin#postCalculation
* @description `postCalculation` fired after all basic statistics calculation and before default rendering. This is a place where you should calculate your custom statistics.
* @type {Object}
* @property {number} colCount
* @property {Array.<string>} fields Fields for current columns
* @property {Array.<StatisticsRowPlugin~Stats>} statistics Array of statistic values. You should add your custom statistics to the object
* @property {Array.<Object>} dataRows Data row used for the calculation
* @property {Object} dataView Class object representing current view of grid
* @property {Object} dataTable Class object contains all data
*/

/** @event StatisticsRowPlugin#postRendering
* @description `postRendering` fired after all statistics calculation and default rendering. This is a place where you override default rendering
* @type {Object}
* @property {number} colCount
* @property {Array.<string>} fields Fields for current columns
* @property {Array.<StatisticsRowPlugin~Stats>} statistics Array of statistic values
* @property {Object} headerSection ILayoutGrid UI Section representing header part
* @property {Object} footerSection ILayoutGrid UI Section representing footer part
* @property {Array.<Object>} headerRows Definition objects given at the configuration time
* @property {Array.<Object>} footerRows Definition objects given at the configuration time
*/

/** @constructor
* @param {StatisticsRowPlugin~Options=} options
* @extends {GridPlugin}
*/
var StatisticsRowPlugin = function (options) {
	this._onPostDataSourceChanged = this._onPostDataSourceChanged.bind(this);
	this._onPostSectionDataBinding = this._onPostSectionDataBinding.bind(this);
	this._onColumnAdded = this._onColumnAdded.bind(this);
	this._onExtensionLoaded = this._onExtensionLoaded.bind(this);
	this._onDataChanged = this._onDataChanged.bind(this);
	this._onRowRemoved = this._onRowRemoved.bind(this);

	this._hosts = [];
	this._headerRows = [];
	this._footerRows = [];
	this._rows = [];

	if (options) {
		this.config({ statisticsRow: options });
	}
};
Ext.inherits(StatisticsRowPlugin, GridPlugin);

/** @type {Object.<string, string>}
* @private
*/
StatisticsRowPlugin._aliases = {
	"average": "average",
	"avg": "average",
	"mean": "average",
	"count": "count",
	"max": "max",
	"maximum": "max",
	"min": "min",
	"minimum": "min",
	"sum": "sum",
	"summation": "sum"
};
/** @type {Object.<string, string>}
* @private
*/
StatisticsRowPlugin._defaultLabels = {
	"average": "Average",
	"count": "Count",
	"max": "Maximum",
	"min": "Minimum",
	"sum": "Summation"
};

/** @type {string}
* @private
*/
StatisticsRowPlugin.prototype._invalidText = "";
/** @type {boolean}
* @private
*/
StatisticsRowPlugin.prototype._noColoring = false;
/** @type {boolean}
* @private
*/
StatisticsRowPlugin.prototype._noFormatting = false;

/** @type {Array}
* @private
*/
StatisticsRowPlugin.prototype._headerRows = null;
/** @type {Array}
* @private
*/
StatisticsRowPlugin.prototype._footerRows = null;
/** @type {Array}
* @private
*/
StatisticsRowPlugin.prototype._rows = null;
/** @type {Object}
* @private
* @description Grid section object
*/
StatisticsRowPlugin.prototype._headerSect = null;
/** @type {Object}
* @private
* @description Grid section object
*/
StatisticsRowPlugin.prototype._footerSect = null;
/** @type {Object}
* @private
* @description TextFormattingExtension
*/
StatisticsRowPlugin.prototype._tfExt = null;
/** @type {Object}
* @private
* @description ConditionalColoringExtension
*/
StatisticsRowPlugin.prototype._ccExt = null;
/** @type {boolean}
* @private
*/
StatisticsRowPlugin.prototype._disabledUIs = false;
/** @type {boolean}
* @private
*/
StatisticsRowPlugin.prototype._useAllData = true;
/** @public
* @return {string}
*/
StatisticsRowPlugin.prototype.getName = function () {
	return "StatisticsRowPlugin"; // Read Only
};

/** @public
* @param {Object} host core grid instance
* @param {Object=} options
*/
StatisticsRowPlugin.prototype.initialize = function (host, options) {
	if (this._hosts.indexOf(host) >= 0) { return; }

	this._hosts.push(host);
	setTimeout(this._onExtensionLoaded, 0);

	this._disabledUIs = true; // Prevent unnecessary updates
	this.config(options);
	this._disabledUIs = false;

	host.getDataSource().listen("dataChanged", this._onDataChanged);
	host.listen("rowRemoved", this._onRowRemoved);
	host.listen("postDataSourceChanged", this._onPostDataSourceChanged);
	host.listen("postSectionDataBinding", this._onPostSectionDataBinding);
	host.listen("columnAdded", this._onColumnAdded);

	// In case of lazy loading
	var title = host.getSection("title");
	if(title) {
		this._onPostDataSourceChanged();
	}
};

/** @private
*/
StatisticsRowPlugin.prototype._onExtensionLoaded = function () {
	if(!this._tfExt || !this._ccExt) {
		var host = this._hosts[0];
		if(host) {
			this._tfExt = host.getPlugin("TextFormattingPlugin");
			this._ccExt = host.getPlugin("ConditionalColoringPlugin");
		}
	}
};
/** @public
* @param {Object} host core grid instance
*/
StatisticsRowPlugin.prototype.unload = function (host) {
	var at = this._hosts.indexOf(host);
	if (at < 0) { return; }
	this._hosts.splice(at, 1);

	host.getDataSource().unlisten("dataChanged", this._onDataChanged);
	host.unlisten("rowRemoved", this._onRowRemoved);
	host.unlisten("postDataSourceChanged", this._onPostDataSourceChanged);
	host.unlisten("postSectionDataBinding", this._onPostSectionDataBinding);
	host.unlisten("columnAdded", this._onColumnAdded);

	if (!this._hosts.length) {
		this._headerSect = this._footerSect = null;
		this._tfExt = this._ccExt = null;
	}

	this._dispose();
};

/** @public
* @param {Object=} options Grid configuration object
*/
StatisticsRowPlugin.prototype.config = function (options) {
	if (!options) { return; }

	var extOptions = options.statisticsRow;
	if(extOptions) {
		if(extOptions.invalidText != null) {
			this._invalidText = extOptions.invalidText;
		}
		if(extOptions.noColoring != null) {
			this._noColoring = extOptions.noColoring ? true : false;
		}
		if(extOptions.noFormatting != null) {
			this._noFormatting = extOptions.noFormatting ? true : false;
		}
		if(extOptions.useAllData != null) {
			this._useAllData = extOptions.useAllData ? true : false;
		}
		this.addListener(extOptions, "postCalculation");
		this.addListener(extOptions, "postRendering");

		if(extOptions.rows != null) {
			this.setStatisticsRows(extOptions.rows);
		}
	}

	var columns = options.columns;
	var colCount = columns ? columns.length : 0;
	for(var c = 0; c < colCount; ++c) {
		var colOptions = columns[c];
		this._configColumn(c, colOptions);
	}
};

/** @public
* @param {Object=} gridOptions
* @return {!Object}
*/
StatisticsRowPlugin.prototype.getConfigObject = function (gridOptions) {
	var obj = gridOptions || {};

	var columns = obj.columns;
	if(!columns) {
		columns = obj.columns = [];
	}

	var len = this.getColumnCount();
	for(var i = 0; i < len; ++i) {
		var column = columns[i];
		if(!column) {
			column = columns[i] = {};
		}

		var colOptions = this._getExtColumnConfig(i);
		if(colOptions) {
			if(colOptions.label) {
				column.statistics = "label";
			} else if(colOptions.statistics) {
				column.statistics = true;
			}
		}
	}

	var extOptions = obj.statisticsRow;
	if(!extOptions) {
		extOptions = obj.statisticsRow = {};
	}

	if(this._rows.length > 0) {
		extOptions.rows = this._rows;
	}

	if(this._invalidText) {
		extOptions.invalidText = this._invalidText;
	}

	if(this._noColoring) {
		extOptions.noColoring = this._noColoring;
	}

	if(this._noFormatting) {
		extOptions.noFormatting = this._noFormatting;
	}

	if(!this._useAllData) {
		extOptions.useAllData = this._useAllData;
	}

	return obj;
};

/** Remove all existing statistics rows and create new rows based on the given inputs. Use null to remove all rows.
* @public
* @param {Array.<StatisticsRowPlugin~RowOptions>} rows
*/
StatisticsRowPlugin.prototype.setStatisticsRows = function(rows) {
	var rowCount = rows ? rows.length : 0;
	this._headerRows.length = 0;
	this._footerRows.length = 0;
	for(var r = 0; r < rowCount; ++r) {
		var row = this._prepareStatRow(rows[r]);

		if(row.placement === "header") {
			this._headerRows.push(row);
		} else {
			this._footerRows.push(row);
		}
	}
	this._rows = this._headerRows.concat(this._footerRows);

	this._prepareSections();
	this._updateUIs();
};
/** Get all existng statistics rows
* @public
* @return {Array.<StatisticsRowPlugin~RowOptions>}
*/
StatisticsRowPlugin.prototype.getStatisticsRows = function() {
	return this._rows.slice(); // Clone a copy for users
};

/** Normalize user inputs
* @private
* @param {StatisticsRowPlugin~RowOptions} row
* @return {StatisticsRowPlugin~RowOptions}
*/
StatisticsRowPlugin.prototype._prepareStatRow = function(row) {
	if(!row) {
		return null;
	}
	var placement = row.placement;
	if(placement === "header" || placement === "title") {
		row.placement = "header";
	} else {
		row.placement = "footer";
	}
	var statistic = row.statistic || row.statistics;
	if(row.statistics) { // Handle alias
		row.statistics = null;
	}
	if(statistic) {
		var predefined = StatisticsRowPlugin._aliases[statistic.toLowerCase()];
		if(predefined) {
			statistic = predefined;
		}
	}
	row.statistic = statistic || "custom";

	row.cachedData = {}; // WARNING: This will get returned by getConfigData() method

	return row;
};
/** @private
* @param {StatisticsRowPlugin~RowReference} rowRef
* @return {number}
*/
StatisticsRowPlugin.prototype._getStatRowIndex = function(rowRef) {
	var rowIndex = -1;
	if(typeof rowRef === "string" && rowRef) {
		var rowCount = this._rows.length;
		for(var i = 0; i < rowCount; ++i) {
			var row = this._rows[i];
			if(row.statistic === rowRef) {
				rowIndex = i;
				break;
			} else if(row.id === rowRef) {
				rowIndex = i;
				break;
			}
		}
	} else {
		if(typeof rowRef === "number") {
			rowIndex = rowRef;
		} else if(rowRef) {
			rowIndex = +rowRef;
			if(rowIndex !== rowIndex) {
				rowIndex = -1;
			}
		}

		if(rowIndex < 0 || rowIndex >= this._rows.length) {
			rowIndex = -1;
		}
	}

	return rowIndex;
};

/** Replace and update existing statistics row. Placement and order of the row cannot be changed. Use setStatisticsRows() to update multiple rows at once.
* @public
* @param {StatisticsRowPlugin~RowReference} rowRef Index, stat, or id
* @param {StatisticsRowPlugin~RowOptions} options The given configuration object cannot be null nor the same existing object.
* @return {boolean} Returns true if there is any change, otherwise returns false
*/
StatisticsRowPlugin.prototype.setStatisticsRow = function(rowRef, options) {
	var rowIndex = this._getStatRowIndex(rowRef);
	var prevRow = this._rows[rowIndex];
	if(!prevRow || prevRow === options) {
		return false;
	}

	var row = this._prepareStatRow(options);
	if(!row) {
		return false;
	}

	row.placement = prevRow.placement; // Position cannot be changed
	var subset = (row.placement === "header") ? this._headerRows : this._footerRows;
	var subRowIndex = subset.indexOf(prevRow);

	this._rows[rowIndex] = row;
	subset[subRowIndex] = row;

	this._updateUIs();
	return true;
};
/** Add new statistic row at the bottom of either header or footer section
* @public
* @param {StatisticsRowPlugin~RowOptions} options The given configuration object cannot be null nor the same existing object
*/
StatisticsRowPlugin.prototype.addStatisticsRow = function(options) {
	var row = this._prepareStatRow(options);
	var rowIndex = this._rows.indexOf(row);
	if(!row || rowIndex >= 0) {
		return;
	}
	if(row.placement === "header") {
		this._headerRows.push(row);
	} else {
		this._footerRows.push(row);
	}

	this._rows = this._headerRows.concat(this._footerRows);

	this._prepareSections();
	this._updateUIs();
};
/** Remove existing statistics row
* @public
* @param {StatisticsRowPlugin~RowReference} rowRef Index, stat, or id
*/
StatisticsRowPlugin.prototype.removeStatisticsRow = function(rowRef) {
	var rowIndex = this._getStatRowIndex(rowRef);
	if(rowIndex < 0) {
		return;
	}
	var row = this._rows[rowIndex];
	var subset = (row.placement === "header") ? this._headerRows : this._footerRows;
	var subRowIndex = subset.indexOf(row);

	this._rows.splice(rowIndex, 1);
	subset.splice(subRowIndex, 1);

	this._prepareSections();
	this._updateUIs();
};
// getStatisticsValues(rowRef):Object

/** @private
* @param {number} colIndex
* @param {number} columnConfig
*/
StatisticsRowPlugin.prototype._configColumn = function(colIndex, columnConfig) {
	if(columnConfig) {
		var statistics = columnConfig["statistics"] || columnConfig["statistic"];
		if(!statistics) {
			return;
		}

		var colData = this._newColumnData(colIndex);
		var options = colData["statisticsRow"];
		if(!options) {
			options = colData["statisticsRow"] = {};
		}
		if(statistics === "label") {
			options.label = true;
		} else {
			options.statistics = true;
		}
	}
};
/** @private
* @param {number} colIndex
* @return {Object}
*/
StatisticsRowPlugin.prototype._getExtColumnConfig = function(colIndex) {
	return this._getColumnOption(colIndex, "statisticsRow") || null;
};

/** @private
* @param {Object=} e
*/
StatisticsRowPlugin.prototype._onColumnAdded = function (e) {
	if(e.context && e.colIndex != null) {
		this._configColumn(e.colIndex, e.context);
	}
};

/** @private
*/
StatisticsRowPlugin.prototype._prepareSections = function () {
	var host = this._hosts[0];
	if(!host || this._disabledUIs) {
		return;
	}

	var contentSect = host.getSection("content");
	var sectIndex, section;
	var rowHeight = host.getDefaultRowHeight();
	if(this._headerRows.length) {
		section = this._headerSect;
		if(!section) {
			sectIndex = contentSect ? contentSect.getIndex() : 1;
			section = this._headerSect = host.addSectionAt(sectIndex, "title", "statisticsHeaderRows");

			section.setDefaultRowHeight(rowHeight);
			section.enableClass("statistics-rows");
		}
		section.setRowCount(this._headerRows.length);
	} else if(this._headerSect) {
		host.removeSectionAt(this._headerSect.getIndex());
		this._headerSect = null;
	}

	if(this._footerRows.length) {
		section = this._footerSect;
		if(!section) {
			sectIndex = contentSect ? contentSect.getIndex() + 1 : host.getSectionCount();
			section = this._footerSect = host.addSectionAt(sectIndex, "footer", "statisticsFooterRows");

			section.setDefaultRowHeight(rowHeight);
			section.enableClass("statistics-rows");
		}
		section.setRowCount(this._footerRows.length);
	} else if(this._footerSect) {
		host.removeSectionAt(this._footerSect.getIndex());
		this._footerSect = null;
	}
};
/** @private
*/
StatisticsRowPlugin.prototype._updateUIs = function () {
	var host = this._hosts[0];
	if(!host || this._disabledUIs) {
		return;
	}
	if(host.getRowCount() > 0) {
		host.requestRowRefresh();
	} else {
		this._onPostSectionDataBinding();
	}
};
/** @private
* @param {Object=} e
*/
StatisticsRowPlugin.prototype._onPostDataSourceChanged = function (e) {
	this._prepareSections();
};
/** @private
* @param {Object=} e
*/
StatisticsRowPlugin.prototype._onDataChanged = function (e) {
	this._updateUIs();
};
/** @private
* @param {Object=} e
*/
StatisticsRowPlugin.prototype._onRowRemoved = function (e) {
	if(!e.rowCount) { // Since binding won't be occured for 0 row, we have to force re-rendering on the headers/footers by ourselve
		if(e.sectionType === "content") {
			this._onPostSectionDataBinding();
		}
	}
};

/** @private
* @param {Object=} e
*/
StatisticsRowPlugin.prototype._onPostSectionDataBinding = function (e) {
	if(!this._headerSect && !this._footerSect) {
		return;
	}

	if(e && e.positionChangeOnly) { // To avoid unnecessary calculation and rendering
		return;
	}

	// TODO: Reduce unneccessary calculation due to UI updates
	// Preparing context
	var host = this._hosts[0];
	var colCount = host.getColumnCount();
	var dv = host.getDataSource();
	var dt = dv.getDataSource();
	var rows;
	if(this._useAllData) {
		rows = dt.getAllRowData(); // Slow: A new copy of data is created
	} else {
		rows = dv.getAllRowData();
	}
	var rowCount = rows.length;

	var fields = null;
	if(this._compositeGrid) {
		fields = this._compositeGrid.getColumnFields();
	} else if(this._realTimeGrid) {
		fields = this._realTimeGrid.getColumnFields();
		rows = rows.map(this._rowGetter);
	}
	if(!fields) {
		return;
	}

	// Calculate stats
	var c;
	var statistics = new Array(colCount);
	for(c = 0; c < colCount; ++c) {
		statistics[c] = null;
		var colOptions = this._getExtColumnConfig(c);
		if(!colOptions) {
			continue;
		}
		var field = fields[c];
		var count = 0;
		var stat = null;
		var r, val;
		if(colOptions.statistics) {
			stat = statistics[c] = {};
			var sum = 0;
			var min = NaN;
			var max = NaN;
			for(r = 0; r < rowCount; ++r) {
				val = rows[r][field];
				if(val != null) { // Not undefined or null
					if(typeof val !== "number") { // Slow: Type casting
						val = +val;
					}
					if(val === val) {
						++count;
						if(!(val >= min)) {
							min = val;
						}
						if(!(val <= max)) {
							max = val;
						}
						sum += val;
					}
				}
			}
			stat.field = field;
			stat.count = count;
			stat.sum = (count > 0) ? sum : NaN;
			stat.average = sum / count;
			stat.min = min;
			stat.max = max;
		} else if(colOptions.label) {
			stat = statistics[c] = {"label": true};
			for(r = 0; r < rowCount; ++r) {
				val = rows[r][field];
				if(val || val === 0) { // WARNING: Use different counting criteria from normal statistic.
					++count;
				}
			}
			stat.field = field;
			stat.count = count;
		}
	}

	if(this.hasListener("postCalculation")) {
		this._dispatch("postCalculation", {
			"colCount": colCount,
			"fields": fields,
			"statistics": statistics,
			"dataRows": rows,
			"dataView": dv,
			"dataTable": dt,
			"core": host,
			"extension": this
		});
	}
	// Render stuff
	for(c = 0; c < colCount; ++c) {
		this._renderStatistics(c, statistics[c], e ? e.actualUpdate : false);
	}

	if(this.hasListener("postRendering")) {
		this._dispatch("postRendering", {
			"colCount": colCount,
			"fields": fields,
			"statistics": statistics,
			"headerSection": this._headerSect,
			"footerSection": this._footerSect,
			"headerRows": this._headerRows,
			"footerRows": this._footerRows,
			"core": host,
			"extension": this
		});
	}
};
/** @private
* @param {number} colIndex
* @param {Object} stats
* @param {boolean} actualUpdate
*/
StatisticsRowPlugin.prototype._renderStatistics = function (colIndex, stats, actualUpdate) {
	if(!this._headerSect && !this._footerSect) {
		return;
	}
	if(!stats) {
		return;
	}

	var r, rowCount;
	if(this._headerSect) {
		rowCount = this._headerRows.length;
		for(r = 0; r < rowCount; ++r) {
			this._headerRows[r].cell = this._headerSect.getCell(colIndex, r);
		}
	}
	if(this._footerSect) {
		rowCount = this._footerRows.length;
		for(r = 0; r < rowCount; ++r) {
			this._footerRows[r].cell = this._footerSect.getCell(colIndex, r);
		}
	}

	var rows = this._rows;
	rowCount = rows.length;
	for(r = 0; r < rowCount; ++r) {
		var row = rows[r];
		var txt = null;

		var cell = row.cell;
		if(cell == null) {
			continue;
		}

		var statistic = row.statistic;
		if(stats.label) {
			txt = row.label;
			if(txt == null) {
				txt = StatisticsRowPlugin._defaultLabels[statistic] || statistic;
			}
		} else {
			txt = stats[statistic];
			if(!txt && txt !== 0) {
				txt = this._invalidText;
			} else {
				var num = txt;
				if(this._tfExt && !this._noFormatting) {
					txt = this._tfExt.formatText(colIndex, num);
				}
				if(this._ccExt && !this._noColoring) {
					var rowData = row.cachedData;
					rowData[stats.field] = num;
					this._ccExt.applyColor(colIndex, cell, rowData, actualUpdate);
				}
			}
		}

		cell.setContent(txt);
		row.cell = null;
	}
};



export default StatisticsRowPlugin;
export { StatisticsRowPlugin, StatisticsRowPlugin as StatisticsRow, StatisticsRowPlugin as StatisticsRowExtension };
