

/** @constructor
* @param {Object} menuModel
* @param {Array} columns
* @param {Array} columnGroups
* @param {Array} legacyColumnGroups
*/
let MenuEventAPI = function (menuModel) {
	this._model = menuModel || null; // for mapping string value
	this._items = [];
	this._runningId = 0;
};

/** @public */
MenuEventAPI.prototype.clear = function () {
	this._items.length = 0;
	this._runningId = 0;
};

/** @public
* @param {Array.<string> | Object | string} item
*/
MenuEventAPI.prototype.addItem = function (item) {
	this._addItem(item, 0); // Recursive method
};

/** @public
* @param {Array.<string> | Object | string} item
*/
MenuEventAPI.prototype.addItems = function (item) { // Alias
	this._addItem(item, 0); // Recursive method
};

/** @private
* @param {Array.<string> | Object | string} item
* @param {number} level
*/
MenuEventAPI.prototype._addItem = function (item, level) {
	if (!item) {
		return; // No item to be added
	}
	if (level >= 10) {
		console.warn("Circular reference detected");
		return;
	}

	if (typeof item === "string") { // The item is referencing another item
		if (item === "SEPARATOR" || item.match(/-{3,}/)) { // Special keyword
			this.addSeparator();
		} else if(this._model) {
			if(this._model.items) {
				this._addItem(this._model.items[item], level + 1);
			}
		}
	} else if (Array.isArray(item)) { // The item is an item list
		let len = item.length;
		for (let i = 0; i < len; ++i) {
			this._addItem(item[i], level + 1);
		}
	} else { // Assume the item has an object type
		this._assignId(item);  // The item can have child items
		this._items.push(item);
	}
};

/** @public */
MenuEventAPI.prototype.addSeparator = function () {
	this._items.push({ type: "separator" });
};

/** Get immidiate child menu items
* @public
* @returns {Array}
*/
MenuEventAPI.prototype.getMenuItems = function () {
	return this._items;
};

/** @public
* @param {number} id
* @returns {Array} Array of items (ordered by leaf node to root node)
*/
MenuEventAPI.prototype.findItem = function (id) { // For internal use
	return this._findItem(id, this._items);
};

/** @private
* @param {number} id
* @param {Array.<string> | Object} item
* @returns {Array}
*/
MenuEventAPI.prototype._findItem = function (id, item) {
	if (!item) {
		return null;
	}

	let foundItemPath;
	if (Array.isArray(item)) {
		let len = item.length;
		for (let i = 0; i < len; ++i) {
			foundItemPath = this._findItem(id, item[i]);
			if (foundItemPath) {
				return foundItemPath;
			}
		}
		return null;
	} else if (item.id === id) {
		return [item]; // Return item in path array
	}

	foundItemPath = this._findItem(id, item.items);
	if (foundItemPath) {
		foundItemPath.push(item); // Push parent to the back
	}
	return foundItemPath;
};

/** @private
* @param {Array.<string> | Object} items
*/
MenuEventAPI.prototype._assignId = function (items) {
	if (Array.isArray(items)) {
		for (let i = 0; i < items.length; i++) {
			this._assignId(items[i]);
		}
	} else {
		items.id = "_item_" + this._runningId++;
		if (items.items) {
			this._assignId(items.items);
		}
	}
};



export default MenuEventAPI;
export { MenuEventAPI };
