import EventDispatcher from "../../tr-grid-util/es6/EventDispatcher.js";
import Ext from "../../tr-grid-util/es6/Ext.js";

/** @constructor
* @extends {EventDispatcher}
* @param {Object=} options
*/
let MenuItem = function (options) {
	let t = this;

	t._onItemHovered = t._onItemHovered.bind(t);
	t._onItemClicked = t._onItemClicked.bind(t);

	let li = document.createElement("li");
	li.className = "tr-contextmenu-item";
	li.addEventListener("mouseover", t._onItemHovered);
	li.addEventListener("click", t._onItemClicked);
	t._elem = li;

	t._config(options);
};
Ext.inherits(MenuItem, EventDispatcher);

/** @type {Element}
* @private
*/
MenuItem.prototype._elem;
/** @type {Object}
* @private
*/
MenuItem.prototype._options = null;
/** @type {boolean}
* @private
*/
MenuItem.prototype._isSelectable = true;
/** @type {string}
* @private
*/
MenuItem.prototype._id = "";
/** @type {string}
* @private
*/
MenuItem.prototype._type = "";
/** @type {string}
* @private
*/
MenuItem.prototype._text = "";
/** @type {*}
* @private
*/
MenuItem.prototype._value = null;
/** @type {Array}
* @private
*/
MenuItem.prototype._items = null;

/** @public */
MenuItem.prototype.dispose = function() {
	let t = this;
	t.removeAllEventListeners();
	t._elem.removeEventListener("mouseover", t._onItemHovered);
	t._elem.removeEventListener("click", t._onItemClicked);

	let pn = t._elem.parentNode;
	if(pn) {
		pn.removeChild(t._elem);
	}
	t._elem = t._options = t._value = t._items = null;
};

/** @private
* @param {*=} options
*/
MenuItem.prototype._config = function(options) {
	if (!options) {
		options = {};
	}

	let t = this;
	t._options = options;

	let pref = options["id"];
	if (pref) {
		t.setId(pref);
	}

	pref = options["type"];
	if (pref) {
		t.setType(pref);
	}

	pref = options["text"];
	if (pref) {
		t.setText(pref);
	}

	pref = options["value"];
	if (pref != null) {
		t.setValue(pref);
	}

	pref = options["isDisabled"];
	if (pref != null) {
		if (pref === true) {
			t.setSelectable(false);
		} else {
			t.setSelectable(true);
		}
	}

	pref = options["isSelectable"];
	if (pref != null) {
		if (pref === false) {
			t.setSelectable(false);
		} else {
			t.setSelectable(true);
		}
	}

	pref = options["items"];
	if (Array.isArray(pref)) {
		t.setItems(pref);
	}

};

/** @private
* @param {Object} e
*/
MenuItem.prototype._onItemHovered = function(e) {
	let t = this;

	// Cannot be hover if disabled or has no children
	if (!t.isSelectable()) { return; }

	e["item"] = t._options;
	t._dispatch("itemHovered", e);
};

/** @private
* @param {Object} e
*/
MenuItem.prototype._onItemClicked = function(e) {
	let t = this;

	// Cannot be clicked if disabled or has children
	if (!t.isSelectable() || Array.isArray(t.getItems())) { return; }

	e["item"] = t._options;
	t._dispatch("itemClicked", e);
};

/** @public
* @ignore
* @returns {Object}
*/
MenuItem.prototype._getEventHandlers = function() {
	return {
		mouseover: this._onItemHovered,
		click: this._onItemClicked
	};
};

/** @public
* @param {*} id
*/
MenuItem.prototype.setId = function(id) {
	let t = this;
	t._elem.setAttribute("id", id);
	this._id = id;
};

/** @public
* @returns {*}
*/
MenuItem.prototype.getId = function() {
	return this._id || "";
};

/** @public
* @param {*} text
*/
MenuItem.prototype.setText = function(text) {
	let t = this;
	t._elem.textContent = text + "";
	this._text = text;
};

/** @public
* @returns {string}
*/
MenuItem.prototype.getText = function() {
	return this._text || "";
};

/** @public
* @param {*} value
*/
MenuItem.prototype.setValue = function(value) {
	this._value = value;
};

/** @public
* @returns {*}
*/
MenuItem.prototype.getValue = function() {
	return this._value;
};

/** @public
* @param {boolean=} selectable
*/
MenuItem.prototype.setSelectable = function(selectable) {
	let t = this;
	if (selectable) {
		t._elem.removeAttribute("data-unselectable");
		t._elem.removeAttribute("disabled");
	} else {
		t._elem.setAttribute("data-unselectable", "true");
		t._elem.setAttribute("disabled", "true");
	}
	t._isSelectable = selectable ? true : false;
};

/** @public
* @returns {boolean}
*/
MenuItem.prototype.isSelectable = function() {
	return this._isSelectable;
};

/** @public
* @param {Array} items
*/
MenuItem.prototype.setItems = function(items) {
	let t = this;
	if(Array.isArray(items)) {
		t._items = items;
		t._elem.classList.add("has-child");
	} else {
		t._items = null;
		t._elem.classList.remove("has-child");
	}
};

/** @public
* @returns {Array}
*/
MenuItem.prototype.getItems = function() {
	return this._items;
};

/** @public
* @param {string} type
*/
MenuItem.prototype.setType = function(type) {
	this._elem.setAttribute("data-type", type);
	this._type = type;
};

/** @public
* @returns {string}
*/
MenuItem.prototype.getType = function() {
	return this._type || "";
};

/** @public
* @returns {Element}
*/
MenuItem.prototype.getElement = function() {
	return this._elem;
};

/** @public
* @returns {Object}
*/
MenuItem.prototype.getOptions = function() {
	return this._options;
};



export default MenuItem;
export { MenuItem };
