import MenuItem from "./MenuItem.js";
import Popup from "../../tr-grid-util/es6/Popup.js";
import EventDispatcher from "../../tr-grid-util/es6/EventDispatcher.js";
import Ext from "../../tr-grid-util/es6/Ext.js";

/** @constructor
* @extends {EventDispatcher}
* @param {Element} parentElement
*/
let PopupMenu = function (parentElement) {
	let t = this;

	t.show = t.show.bind(t);
	t.attachTo = t.attachTo.bind(t);
	t.setMenu = t.setMenu.bind(t);
	t.setPosition = t.setPosition.bind(t);
	t._hoveredItem = t._hoveredItem.bind(t);
	t._hoveredChildItem = t._hoveredChildItem.bind(t);
	t._clickedItem = t._clickedItem.bind(t);
	t._clickedChildItem = t._clickedChildItem.bind(t);

	t._init(parentElement);
};
Ext.inherits(PopupMenu, EventDispatcher);

/** @type {Object}
* @private
*/
PopupMenu.prototype._evtArg;
/** @type {Array.<PopupMenu>}
* @private
*/
PopupMenu.prototype._nestedPopups;
/** @type {Array.<MenuItem>}
* @private
*/
PopupMenu.prototype._menuItems;
/** @type {Element}
* @private
*/
PopupMenu.prototype._parentElement;
/** @type {number}
* @private
*/
PopupMenu.prototype._x;
/** @type {number}
* @private
*/
PopupMenu.prototype._y;
/** @type {Popup}
* @private
*/
PopupMenu.prototype._popup;
/** @type {Element}
* @private
*/
PopupMenu.prototype._ulElem;
/** @type {boolean}
* @private
*/
PopupMenu.prototype._isMounted;
/** @type {string}
* @private
*/
PopupMenu.prototype._id = "";

/** @private
* @param {Element} parentElement
*/
PopupMenu.prototype._init = function(parentElement) {
	let t = this;

	let topNode = document.createElement("div");
	topNode.className = "tr-contextmenu-popup";
	let ul = document.createElement("ul");
	ul.className = "tr-contextmenu";
	topNode.appendChild(ul);

	t._popup = new Popup(topNode);
	t._popup.setPositioning("custom");
	t._ulElem = ul;

	t._nestedPopups = [];
	t._menuItems = [];
	t._x = 0;
	t._y = 0;
	t._isMounted = true;
	t._parentElement = parentElement || window.document;

	t._evtArg = {
		"target": t._popup._elem,
		"currentTarget": t._popup._elem
	};
};

/** @public
* @return {boolean}
*/
PopupMenu.prototype.isDisposed = function () {
	return !this._isMounted;
};

/** @public */
PopupMenu.prototype.dispose = function () {
	let t = this;
	if (t._isMounted) {
		t.removeAllEventListeners();
		t._popup.dispose();
		t._disposeChildren();
		t._isMounted = false;
	}
};

/** @private
* @function
* @param {*} obj
*/
let _disposeObject = function(obj) {
	if(obj) {
		obj.dispose();
	}
};

/** @private
*/
PopupMenu.prototype._disposeChildren = function () {
	let t = this;
	t._nestedPopups.forEach(_disposeObject);
	t._menuItems.forEach(_disposeObject);
	t._nestedPopups.length = 0;
	t._menuItems.length = 0;
};

/** @public */
PopupMenu.prototype.show = function () {
	let t = this;
	t._popup.show(true, t._parentElement);
	t.setPosition(t._popup._positioning, t._x, t._y);
};

/** @public */
PopupMenu.prototype.hide = function () {
	this._popup.hide();
};

/** @public
* @param {Element} elem
* @param {Popup~Positioning} positioning
*/
PopupMenu.prototype.attachTo = function (elem, positioning) {
	if (elem) {
		this._popup.attachTo(elem, positioning);
	}
};

/** @public
* @param {Popup~Positioning} positioning
* @param {number} x
* @param {number} y
*/
PopupMenu.prototype.setPosition = function(positioning, x, y) {
	let t = this;

	if (positioning === "custom") {
		t._x = x;
		t._y = y;
		let ww = window.innerWidth;
		let wh = window.innerHeight;
		let sx = (window.scrollX != null) ? window.scrollX : window.pageXOffset;
		let sy = (window.scrollY != null) ? window.scrollY : window.pageYOffset;
		let rb = ww + sx; // view port right bound
		let bb = wh + sy; // view port bottom bound

		let popupElem = t._popup.getElement();
		let aw = popupElem.offsetWidth;
		let ah = popupElem.offsetHeight;

		if (x + aw > rb) { // Do not overflow right bound
			x = rb - aw;
		}
		if (y + ah > bb) { // Do not overflow bottom bound
			y = bb - ah;
		}

		popupElem.style.left = (x - sx) + "px"; // Due to fixed position we have to subtract window scrolling
		popupElem.style.top = (y - sy) + "px";
	}
};

/** @public
* @param {Element} element
* @param {PopupMenu} popupMenu
*/
PopupMenu.prototype.addPopupChild = function(element, popupMenu) {
	if (popupMenu instanceof PopupMenu) {
		popupMenu.attachTo(element, "right");
		this._popup.addFocusElement(popupMenu._popup);
		this._nestedPopups.push(popupMenu);
	}
};

/** @public
* @param {Array.<Object>} menuOptions Array of MenuItem options
*/
PopupMenu.prototype.setMenu = function (menuOptions) {
	let t = this;

	// Clear existing before set a new items
	if (t._menuItems.length) {
		t._disposeChildren();
	}

	if (!menuOptions || !Array.isArray(menuOptions)) {
		return;
	}

	for (let i = 0; i < menuOptions.length; i++) {
		let menuItem = new MenuItem(menuOptions[i]);
		menuItem.addEventListener("itemHovered", t._hoveredItem);
		menuItem.addEventListener("itemClicked", t._clickedItem);
		t._menuItems.push(menuItem);

		if (Array.isArray(menuItem.getItems())) {
			let pm = new PopupMenu(t._parentElement);
			pm._id = menuItem.getId();
			pm.setMenu(menuItem.getItems());
			pm.addEventListener("itemHovered", t._hoveredChildItem);
			pm.addEventListener("itemClicked", t._clickedChildItem);
			t.addPopupChild(menuItem.getElement(), pm);
		}

		t._ulElem.appendChild(menuItem.getElement());
	}
};
/** @public
* @returns {Array.<MenuItem>}
*/
PopupMenu.prototype.getMenuItems = function () {
	return this._menuItems;
};
/** @public
* @returns {Array.<PopupMenu>}
*/
PopupMenu.prototype.getChildPopups = function () {
	return this._nestedPopups;
};

/** @private
* @param {Event} e
*/
PopupMenu.prototype._hoveredChildItem = function (e) {
	this._dispatch("itemHovered", e);
};

/** @private
* @param {Event} e
*/
PopupMenu.prototype._clickedChildItem = function (e) {
	this._dispatch("itemClicked", e);
	this.hide();
};

/** @private
* @param {Event} e
*/
PopupMenu.prototype._hoveredItem = function (e) {
	let t = this;
	let evt = t._evtArg;
	let elem = e.srcElement;

	if (elem) {
		let pms = t._nestedPopups;
		let pmCount = pms.length;
		let elemId = elem.getAttribute("id");
		for (let i = 0; i < pmCount; i++) {
			let pm = pms[i];
			if (pm._id === elemId) {
				pm.show();
			} else {
				pm.hide();
			}
		} // WARNING: evt or _evtArg is not customized to reflect the item that is hovered
	} else {
		evt = e;
	}

	t._dispatch("itemHovered", evt);
};

/** @private
* @param {Event} e
*/
PopupMenu.prototype._clickedItem = function(e) {
	let t = this;
	t._evtArg["item"] = e["item"];
	t._dispatch("itemClicked", t._evtArg);
	t.hide();
};



export default PopupMenu;
export { PopupMenu };
