import { Ext } from "../../tr-grid-util/es6/Ext.js";
import { GridPlugin } from "../../tr-grid-util/es6/GridPlugin.js";
import { ElfUtil } from "../../tr-grid-util/es6/ElfUtil.js";

/** @event PaginationPlugin#pageChanged
* @description Happens after pagination was changed, next, previous, first page, last page.
*/

/** @typedef {Object} PaginationPlugin~Options
* @description The options can be specified by `pagination` property of the main grid's options
* @property {number=} page=1 Initial page to be shown
* @property {number=} pageSize=10 Number of rows per page
* @property {(string|Element)=} element A reference to pagination element. This can be a direct pointer or element's id
* @property {Array=} elements=null Multiple references can be linked to a single grid
* @property {Function=} pageChanged=null Event handler for event after page has been changed
*/

/** @constructor
* @extends GridPlugin
* @param {Object=} options
*/
var PaginationPlugin = function(options) {
	this._onPageChanged = this._onPageChanged.bind(this);
	this._onPageSizeChanged = this._onPageSizeChanged.bind(this);
	this._onDataChanged = this._onDataChanged.bind(this);
	this._resolveUIs = this._resolveUIs.bind(this);
	this._onDispatchEvent = this._onDispatchEvent.bind(this);

	this._hosts = [];
	this._controls = [];

	if (options) {
		this.config({ "pagination": options });
	}
};
Ext.inherits(PaginationPlugin, GridPlugin);

/** @type {!Array.<Element>}
* @private
*/
PaginationPlugin.prototype._controls;
/** @type {*}
* @private
*/
PaginationPlugin.prototype._userUI;
/** @type {number}
* @private
*/
PaginationPlugin.prototype._uiTimer = 0;
/** @type {number}
* @private
*/
PaginationPlugin.prototype._pageChangeTimer = 0;
/** DataView
* @type {Object}
* @private
*/
PaginationPlugin.prototype._dateView = null;
/** @type {boolean}
* @private
*/
PaginationPlugin.prototype._defaultPageSize = true;

/** Called by grid system when initial plugins.
* @public
* @param {Object} host core grid object
* @param {Object} options
*/
PaginationPlugin.prototype.initialize = function(host, options) {
	if (this._hosts.indexOf(host) >= 0) {
		return;
	}

	this._hosts.push(host);
	this._dataView = host.getDataSource(); // TODO: Support multi-grids
	this._dataView.listen("dataChanged", this._onDataChanged);

	this.config(options);

	if(this._defaultPageSize) {
		this._dataView.setPageSize(10); // Default page size is 10
	}

	if(!this._controls.length) {
		this._requestUIUpdate();
	}
};

/** @public
* @param {Object} host core grid object
*/
PaginationPlugin.prototype.unload = function(host) {
	var at = this._hosts.indexOf(host);
	if (at < 0) {
		return;
	}
	this._hosts.splice(at, 1);

	this._dispose();
	if (!this._hosts.length) {
		this._dataView.unlisten("dataChanged", this._onDataChanged);
		this._dataView = null;
		this.setElement(null);
		if(this._uiTimer) {
			clearTimeout(this._uiTimer);
			this._uiTimer = 0;
		}
		if(this._pageChangeTimer) {
			clearTimeout(this._pageChangeTimer);
			this._pageChangeTimer = 0;
		}
	}
};

/** Set pagination element. By default extension will try to find single pagination element within a page.
* @public
* @param {*=} elemRef Reference to pagination element(s). The reference can be a direct pointer, element's id, or an array of the references.
*/
PaginationPlugin.prototype.setElement = function(elemRef) {
	if(this._userUI !== elemRef) {
		this._userUI = elemRef;
		var len = this._controls.length;
		if(len) { // Remove all existing UI resources
			for(var i = 0; i < len; ++i) {
				var control = this._controls[i];
				if (ElfUtil.getElfVersion() < 4) {
					control.removeEventListener("page-changed", this._onPageChanged);
					control.removeEventListener("page-size-changed", this._onPageSizeChanged);
				} else {
					control.removeEventListener("value-changed", this._onPageChanged);
				}
			}
			this._controls.length = 0;
		}
		if(elemRef !== null) {
			this._requestUIUpdate();
		}
	}
};
/** @private
*/
PaginationPlugin.prototype._requestUIUpdate = function() {
	if(!this._uiTimer) {
		this._uiTimer = setTimeout(this._resolveUIs, 100);
	}
};
/** @private
*/
PaginationPlugin.prototype._resolveUIs = function() {
	this._uiTimer = 0;
	if(this._controls.length) {
		return;
	}
	var i, len, elem;
	if(this._userUI) {
		var ary = Array.isArray(this._userUI) ? this._userUI : [this._userUI];

		len = ary.length;
		for(i = 0; i < len; ++i) {
			var ref = ary[i];
			if(typeof ref == "string") {
				elem = document.getElementById(ref);
				if(elem) {
					this._controls.push(elem);
				}
			} else if(ref && ref.nodeType === 1) {
				this._controls.push(ref);
			}
		}
	}

	// Fallback by searching for the first existing pagination element
	if(!this._controls.length && this._userUI !== null) {
		var elems = document.getElementsByTagName("ef-pagination");
		var num = -1;
		if(typeof this._userUI === "number") {
			num = this._userUI;
		} else if (elems.length === 1) {
			num = 0;
		}
		elem = elems[num];
		if(elem) {
			this._controls.push(elem);
		}
	}
	this._userUI = null; // User input can be resolved only once

	for(i = this._controls.length; --i >= 0;) {
		var control = this._controls[i];
		if (ElfUtil.getElfVersion() < 4) {
			control.totalItems = this.getTotalRows();
			control.pageSize = this.getPageSize();
			control.page = this.getPageIndex();
			control.addEventListener("page-changed", this._onPageChanged);
			control.addEventListener("page-size-changed", this._onPageSizeChanged);
		} else {
			control.max = this.getPageCount();
			control.addEventListener("value-changed", this._onPageChanged);
		}
	}
};

/** @override
* @return {string}
*/
PaginationPlugin.prototype.getName = function() {
	return "PaginationPlugin";
};

/** Set Configurations
* @public
* @param {Object=} options
*/
PaginationPlugin.prototype.config = function(options) {
	if (options) {
		this._configPaging(options["pagination"]);
	}
};

/** Get a current state of grid and extension config
 * @public
 * @param {Object=} out_obj
 * @returns {!Object}
 */
PaginationPlugin.prototype.getConfigObject = function (out_obj) {
	var obj = out_obj || {};

	var extOptions = obj.pagination;
	if(!extOptions) {
		extOptions = obj.pagination = {};
	}
	extOptions.page = this.getPageIndex();
	extOptions.pageSize = this.getPageSize();

	return obj;
};

/** Convert built-in config to pagination config
* @public
* @param {string} optionName
* @param {Object} options
* @return {*} The transformed value of the option
*/
PaginationPlugin.prototype.beforeProcessOption = function(optionName, options) {
	if (optionName == "paging" && options) {
		this._configPaging(options);
		return null;
	}
	// eslint-disable-next-line
	return undefined; // undefined is valid value in this function
};

/**
* @private
* @param {Object} options
*/
PaginationPlugin.prototype._configPaging = function(options) {
	if (options) {
		var elem = options.elements || options.element;
		// eslint-disable-next-line
		if(elem !== undefined) { // null is valid input
			this.setElement(elem);
		}
		var num = options.pageSize;
		if(typeof num === "number") {
			this.setPageSize(num);
		}
		num = options.page;
		if(typeof num === "number") {
			this.setPage(num);
		}
		this.addListener(options, "pageChanged");
	}
};

/** @private
* @param {string|number} num
* @return {number}
*/
PaginationPlugin._toNumber = function(num) {
	return (typeof num === "number") ? num : parseInt(num, 10);
};
/** Handle page changed from pagination element
* @private
* @param {Event} e
*/
PaginationPlugin.prototype._onPageChanged = function(e) {
	var pageIndex = PaginationPlugin._toNumber(e.detail.value);
	this.setPage(pageIndex);
	if(!this._pageChangeTimer) {
		this._pageChangeTimer = setTimeout(this._onDispatchEvent, 20);
	}
};
/** @private
*/
PaginationPlugin.prototype._onDispatchEvent = function() {
	this._pageChangeTimer = 0;
	this._dispatch("pageChanged", { page: this.getPageIndex() });
};

/** Handle page size changed from pagination element. Usually, this method is rarely called.
* @private
* @param {Event} e
*/
PaginationPlugin.prototype._onPageSizeChanged = function(e) {
	var size = PaginationPlugin._toNumber(e.detail.value);
	this._setPageSize(size);
};

/** Handle data source changed, interest in insert/remove data's event
* @private
* @param {Event} e
*/
PaginationPlugin.prototype._onDataChanged = function(e) {
	var len = this._controls.length;
	if (len && !e.positionChangeOnly) {
		var pageIndex = this.getPageIndex();
		var i;
		if (ElfUtil.getElfVersion() < 4) {
			var totalRows = this.getTotalRows();
			var pageSize = this.getPageSize();
			for(i = 0; i < len; ++i) {
				this._controls[i].page = pageIndex;
				this._controls[i].totalItems = totalRows;
				this._controls[i].pageSize = pageSize; // In ELF v3 this may fire page-changed and page-size changed
			}
		} else {
			var lastPageIndex = this.getPageCount();
			for(i = 0; i < len; ++i) {
				this._controls[i].value = pageIndex;
				this._controls[i].max = lastPageIndex;
			}
		}
	}
};

/** Get total rows number
* @public
* @return {number}
*/
PaginationPlugin.prototype.getTotalRows = function() {
	return this._dataView ? this._dataView.getSortedRowIds(true).length : 0;
};

/** Get current page index
* @public
* @return {number}
*/
PaginationPlugin.prototype.getPageIndex = function() {
	return this._dataView ? this._dataView.getPageIndex() : 1;
};

/** Get total pages
* @public
* @return {number}
*/
PaginationPlugin.prototype.getPageCount = function() {
	return this._dataView ? this._dataView.getPageCount() : 0;
};

/** Get total rows in a page
* @public
* @return {number}
*/
PaginationPlugin.prototype.getPageSize = function() {
	return this._dataView ? this._dataView.getPageSize() : 0;
};

/** Set page size to grid per page
* @public
* @param {number} size
*/
PaginationPlugin.prototype.setPageSize = function(size) {
	this._defaultPageSize = false; // Page size is specially set
	this._setPageSize(size);
};
/** @private
* @param {number} size
*/
PaginationPlugin.prototype._setPageSize = function(size) {
	if(this._dataView) {
		this._dataView.setPageSize(size);
	}
};

/** Go to specfic page
* @public
* @param {number} page
*/
PaginationPlugin.prototype.setPage = function(page) {
	if (this._dataView) {
		this._dataView.setPage(page);
	}
};
/** Go to specfic page
* @public
* @function
* @param {number} pageIndex
*/
PaginationPlugin.prototype.goto = PaginationPlugin.prototype.setPage;

/** Control pagination go to fisrt page
* @public
*/
PaginationPlugin.prototype.first = function() {
	this._dataView.firstPage();
};

/** Control pagination go to last page
* @public
*/
PaginationPlugin.prototype.last = function() {
	this._dataView.lastPage();
};

/** Control pagination go to next page
* @public
*/
PaginationPlugin.prototype.next = function() {
	this._dataView.nextPage();
};

/** Control pagination go to previous page
* @public
*/
PaginationPlugin.prototype.previous = function() {
	this._dataView.previousPage();
};



export default PaginationPlugin;
export { PaginationPlugin, PaginationPlugin as Pagination, PaginationPlugin as PaginationExtension };
