import { Ext } from "../../tr-grid-util/es6/Ext.js";
import { ElementWrapper } from "../../tr-grid-util/es6/ElementWrapper.js";
import { Table } from "../../tr-grid-util/es6/Table.js";
import { CellWriter } from "./CellWriter.js";

/**
* @constructor
* @extends {ElementWrapper}
*/
let SectionWriter = function () {
	this._cellWriter = new CellWriter();
	this._table = new Table();
	this._columns = [];
	this._elem = this._table.getElement();
};
Ext.inherits(SectionWriter, ElementWrapper);
/** @type {!CellWriter}
* @private
*/
SectionWriter.prototype._cellWriter;
/** @type {!Table}
* @private
*/
SectionWriter.prototype._table;

/** @public
* @return {number}
*/
SectionWriter.prototype.getFirstIndexInView = function () {
	return 0;
};

/** @public
* @return {number}
*/
SectionWriter.prototype.getLastIndexInView = function () {
	return this._table.getRowCount() - 1; // WARNING: This is inclusive
};

/** @public
* @return {number}
*/
SectionWriter.prototype.getColumnCount = function () {
	return this._table.getColumnCount();
};
/** @public
 * @param {number} colIndex
* @return {*}
*/
SectionWriter.prototype.getColumn = function (colIndex) {
	let col = this._columns[colIndex];
	if (!col) {
		col = this._columns[colIndex] = {};
	}
	return col;
};
/** @public
* @param {number} val
*/
SectionWriter.prototype.setColumnCount = function (val) {
	this._table.setColumnCount(val);
};
/** @public
* @return {number}
*/
SectionWriter.prototype.getRowCount = function () {
	return this._table.getRowCount();
};
/** @public
* @param {number} val
*/
SectionWriter.prototype.setRowCount = function (val) {
	this._table.setRowCount(val);
};


/**
* @public
* @param {number} colIndex
* @param {number} rowIndex
* @return {CellWriter}
*/
SectionWriter.prototype.getCell = function (colIndex, rowIndex) {
	let cell = this._table.getCell(colIndex, rowIndex);
	if(cell) {
		this._cellWriter.cloak(cell);
		return this._cellWriter;
	}
	return null;
};
/** @public
* @param {number} colIndex
* @param {number} rowIndex
* @return {Element|Node|NodeList}
*/
SectionWriter.prototype.getCellContent = function (colIndex, rowIndex) {
	let cell = this.getCell(colIndex, rowIndex);
	return cell ? cell.getContent() : null;
};
/** Set data to the specified cell
* @public
* @param {number} colIndex
* @param {number} rowIndex
* @param {string|Element|tr.IElementControl|*} data this can be HTML element, string, number, date, image
*/
SectionWriter.prototype.setCellContent = function (colIndex, rowIndex, data) {
	let cell = this.getCell(colIndex, rowIndex);
	if(cell) {
		cell.setContent(data);
	}
};
/** Apply CSS class to each cell in the row
* @public
* @param {number} rowIndex
* @param {string} className
* @param {boolean} enabled
* @see {@link tr.ILayoutGrid.#setRowStyle}
*/
SectionWriter.prototype.enableRowClass = function (rowIndex, className, enabled) { };

/** This will make the specified cell horizontally span to fit entire row regardless of position of the cell -- cell will shift to the left most of the section.<br>
* If onlyToTheRight is true, the cell will not extend to the left most position<br>
* This is different from cell colSpan, which the cell remain in the same position. tr-stretched class will also be added to the cell
* @public
* @param {*} cellRef Either column index or cell reference is acceptable
* @param {number=} rowIndex If the cell instance is given as the first parameter, this parameter will be ignored.
* @param {boolean=} stretching Default is true
* @param {boolean=} onlyToTheRight=false By default, cell will be stretched to both left and right directions, covering entire row. Set to true to expand only to the right
* @return {*} The cell being stretched
* @example
* grid.stretchCell(1, 0); // The second cell on the first row will be stretched covering the entire row
* grid.stretchCell(cellInstance, 0, true); // The given cell will be stretched covering the entire row
*/
SectionWriter.prototype.stretchCell = function (cellRef, rowIndex, stretching, onlyToTheRight) {
	let cell = this._table.getCell(cellRef, rowIndex);
	if(stretching !== false) {
		cell.setAttribute("colspan", this._table.getColumnCount());
	}
	return cell;
};
/** {@link SectionWriter#stretchCell}
* @public
* @ignore
* @param {number} colIndex
* @param {number} rowIndex
*/
SectionWriter.prototype.unstretchCell = function (colIndex, rowIndex) {};

/** @public
* @return {boolean}
*/
SectionWriter.prototype.hasCellSpan = function () { return false; };
/**
* @public
* @param {*} cellRef
* @param {number=} rowIndex
* @return {*}
*/
SectionWriter.prototype.getStretchedCell = function (cellRef, rowIndex) {
	if(typeof rowIndex === "number") {
		let cell = this._table.getCell(cellRef, rowIndex);
		let colSpan = +cell.getAttribute("colspan");
		if(colSpan > 1) {
			return cell;
		}
	}
	return null;
};
/** @description Returns zero or negative value, if the cell is spanned over (i.e. occupied by the spanning cell). <br>
* Negative value is distance from the spanning cell (i.e. -1 means one cell from the spanning cell and -2 means two cells from the spanning cell) <br>
* Return one, if the cell has no span (normal cell). <br>
* Return value greater than one, if the cell is spanning (getting expanded)
* @public
* @param {number} colIndex
* @param {number} rowIndex
* @return {number}
*/
SectionWriter.prototype.getCellRowSpan = function (colIndex, rowIndex) { return 1; };
/** @public
* @param {number} colIndex
* @param {number} rowIndex
* @param {number} rowSpan
*/
SectionWriter.prototype.setCellRowSpan = function (colIndex, rowIndex, rowSpan) { };
/** @description Returns zero or negative value, if the cell is spanned over (i.e. occupied by the spanning cell). <br>
* Negative value is distance from the spanning cell (i.e. -1 means one cell from the spanning cell and -2 means two cells from the spanning cell) <br>
* Return one, if the cell has no span (normal cell). <br>
* Return value greater than one, if the cell is spanning (getting expanded)
* @public
* @param {number} colIndex
* @param {number} rowIndex
* @return {number}
*/
SectionWriter.prototype.getCellColSpan = function (colIndex, rowIndex) { return 1; };
/** @public
* @param {number} colIndex
* @param {number} rowIndex
* @param {number} colSpan
*/
SectionWriter.prototype.setCellColSpan = function (colIndex, rowIndex, colSpan) { };
/** @public
*/
SectionWriter.prototype.clearCellSpans = function () { };
/** @public
* @param {number} colIndex
*/
SectionWriter.prototype.clearColumnSpans = function (colIndex) { };

/** @public
* @param {number} colIndex
* @return {boolean}
*/
SectionWriter.prototype.isColumnActive = function (colIndex) {
	return true;
};

/**
* @public
* @param {number} colIndex
* @param {number} rowIndex
* @return {Element}
*/
SectionWriter.prototype.getCellElement = function (colIndex, rowIndex) {
	return this._table.getCell(colIndex, rowIndex);
};

/** @public
 * @ignore
 * @param {boolean} enabled
 */
SectionWriter.prototype._startBindingSession = function (enabled) {
	// Prevent error when calling _startBindingSession in LayoutGrid
};



export default SectionWriter;
export { SectionWriter };
