/** @namespace */
let Color = {};

/** @private
* @constant
* @type {RegExp}
*/
const NumRegExp = /\d+/g;

/** Convert CSS rgb or rgba formats to CSS hex color string (with # prefix)
* @public
* @param {string} rgbCode RGB values without # prefix
* @return {string} RGB in hex code (with # prefix)
* @example
* rgb2Hex("rgb(255, 255, 0)"); // "#FFFF00"
* rgb2Hex("rgba(255, 255, 0, 1)"); // "#FFFF00"
* rgb2Hex("255 255.0"); // "#FFFF00"
* rgb2Hex("#FFFF00"); // "#FFFF00"
* rgb2Hex("#1a1a1a"); // "#1a1a1a"
* rgb2Hex("2552550"); // "2552550"
* rgb2Hex("invalid"); // "invalid"
* rgb2Hex(null); // ""
*/
let rgb2Hex = function (rgbCode) {
	if(!rgbCode || typeof rgbCode !== "string") {
		return "";
	}
	if(rgbCode.charAt(0) === "#") {
		return rgbCode;
	}
	let rgb = rgbCode.match(NumRegExp);
	if(!rgb || rgb.length < 3) {
		return rgbCode;
	}

	let hex = "#";
	for(let i = 0; i < 3; i++) {
		let num = +rgb[i];
		if(!(num >= 16)) { // Handle NaN case
			hex += "0";
		}
		hex += (num) ? num.toString(16).toUpperCase() : "0";
	}
	return hex;
};

/** @public
* @function
* @param {Array.<number>} triplet
* @return {string} resultColor
*/
let num2Hex = function (triplet) {
	let rgb = triplet[2] | (triplet[1] << 8) | (triplet[0] << 16);
	return ("#" + (0x1000000 + rgb).toString(16).slice(1));
};
/** Note that Chrome, IE, and Firefox store color in rgb representation.
* @public
* @function
* @param {Array.<number>} triplet
* @return {string} Color string in RGB represetation (e.g. rgb(100, 44, 1))
*/
let num2Rgb = function (triplet) {
	return "rgb(" + triplet[0] + ", " + triplet[1] + ", " + triplet[2] + ")";
};
/** @public
* @function
* @param {string} hex
* @return {Array.<number>} Array of size 3 which contains [red, green, blue]
*/
let hex2Num = function (hex) {
	let hexInt = parseInt(hex.replace(/[^0-9A-F]/gi, ""), 16);
	let r = (hexInt >> 16) & 255;
	let g = (hexInt >> 8) & 255;
	let b = hexInt & 255;
	return [r, g, b];
};
/** @public
* @function
* @param {string} hex Color string with leading # character (e.g. #FFAA00)
* @return {string} Color string in RGB represetation (e.g. rgb(100, 44, 1))
*/
let hex2Rgb = function (hex) {
	if(hex) {
		let hexInt = parseInt(hex.replace(/[^0-9A-F]/gi, ""), 16);
		let r = (hexInt >> 16) & 255;
		let g = (hexInt >> 8) & 255;
		let b = hexInt & 255;
		return "rgb(" + r + ", " + g + ", " + b + ")";
	}
	return "";
};

/** @public
* @function
* @param {number} color A color component (e.g., R, G, or B) with value between 0 and 255 (inclusive)
* @return {number} Normalized luminance value between 0 and 1
*/
let getColorLuminance = function (color) {
	if(!color || color < 0) {
		return 0;
	}
	if(color >= 255) {
		return 1;
	}
	let normalizedColor = color / 255;
	if(normalizedColor <= 0.03928) {
		return normalizedColor / 12.92;
	}
	return Math.pow((normalizedColor + 0.055) / 1.055, 2.4);
};
/** The relative brightness of any point in a colorspace, normalized to 0 for darkest black and 1 for lightest white (https://www.w3.org/TR/WCAG20/#relativeluminancedef)
* @public
* @function
* @param {Array.<number>} triplet
* @return {number} Normalized value between 0 and 1
*/
let getRelativeLuminance = function (triplet) {
	let R = getColorLuminance(triplet[0]);
	let G = getColorLuminance(triplet[1]);
	let B = getColorLuminance(triplet[2]);

	return 0.2126 * R + 0.7152 * G + 0.0722 * B;
};
/** @public
* @function
* @param {number} lumin1
* @param {number} lumin2
* @return {number} Contrast ratios can range from 1 to 21 (commonly written 1:1 to 21:1).
*/
let getContrastRatio = function (lumin1, lumin2) {
	if(lumin1 === lumin2) {
		return 1;
	}
	let darker = lumin1;
	let lighter = lumin2;

	if(lumin1 > lumin2) {
		lighter = lumin1;
		darker = lumin2;
	}
	return (lighter + 0.05) / (darker + 0.05);
};
/** @public
* @function
* @param {Array.<number>} triplet
* @return {string} white or black color in hex code
*/
let getContrastColor = function (triplet) {
	let luminance = getRelativeLuminance(triplet);
	let contrastW = getContrastRatio(1, luminance);
	let contrastB = getContrastRatio(0, luminance);

	if (contrastB >= contrastW) { // Brighter color has more impact to human eye than the darker color
		return "#000000";
	}
	return "#ffffff";
};

/** Blend two colors into single color with the specified ratio
* @public
* @function
* @param {string} baseColor
* @param {string} maxColor
* @param {number} ratio [0, 1]
* @return {Array.<number>} resultColor
*/
let blendColor = function (baseColor, maxColor, ratio) { // This can be optimized further
	if (ratio > 1) {
		ratio = 1;
	} else if(ratio < 0) {
		ratio = 0;
	}

	let baseColorTriplet = hex2Num(baseColor);
	let maxColorTriplet = hex2Num(maxColor);
	let blendResult = [];
	for (let i = 0; i < 3; ++i) {
		let gap = (maxColorTriplet[i] - baseColorTriplet[i]) * ratio;
		blendResult.push(baseColorTriplet[i] + gap);
	}

	return blendResult;
};


Color.rgb2Hex = rgb2Hex;
Color.num2Hex = num2Hex;
Color.num2Rgb = num2Rgb;
Color.hex2Num = hex2Num;
Color.hex2Rgb = hex2Rgb;
Color.getColorLuminance = getColorLuminance;
Color.getRelativeLuminance = getRelativeLuminance;
Color.getContrastRatio = getContrastRatio;
Color.getContrastColor = getContrastColor;
Color.blendColor = blendColor;

export default Color;
export {
	Color,
	rgb2Hex,
	num2Hex,
	num2Rgb,
	hex2Num,
	hex2Rgb,
	getColorLuminance,
	getRelativeLuminance,
	getContrastRatio,
	getContrastColor,
	blendColor
};
