/** Conflator is used to stop multiple executions of a function in the specified period of time by merging multiple calls into one single call. <br>
* Conflator guarantees that only one single call will be executed.
* @constructor
* @param {(Function|number)=} func Bound function that will use this conflator
* @param {(number|Function)=} ms The delay time in millisecond before executing the function
* @param {*=} thisObj "this" object to be bound with the given function. If the function is already bound, there is no need to provide thisObj parameter
* @example
* let c = new Conflator(function() { console.log("Executed"); }, 1000);
* for(let i = 10; --i >= 0;) {
*   c.conflate(i); // Only one "Executed" text will be logged to console after one second
* }
*/
let Conflator = function (func, ms, thisObj) {
	this._onConflated = this._onConflated.bind(this);

	this._data = [];

	if(typeof func == "function") {
		this.setFunction(func, thisObj);
	} else if (typeof ms == "function") { // Parameter can be swapped places
		this.setFunction(ms, thisObj);
	}

	if(typeof ms == "number") {
		this.setConflatingTime(ms);
	} else if (typeof func == "number") { // Parameter can be swapped places
		this.setConflatingTime(func);
	}
};

/** @type {!Array}
* @private
*/
Conflator.prototype._data; // Array to store all conflated data
/** @type {Function}
* @private
*/
Conflator.prototype._func = null;
/** @type {number}
* @private
*/
Conflator.prototype._delay = 0;
/** @type {number}
* @private
*/
Conflator.prototype._timerId = 0;
/** @type {boolean}
* @private
*/
Conflator.prototype._conflated = false;
/** @type {boolean}
* @private
*/
Conflator.prototype._disabled = false;


/** Any existing timer and stored data will be cleared.
* @public
* @return {Array} All stored data. Return null if there is no data
*/
Conflator.prototype.reset = function () {
	if(this._timerId) {
		clearTimeout(this._timerId);
		this._timerId = 0;
	}

	return this.popAllData();
};
/** All stored data will be cleared
* @public
* @return {Array} All stored data. Return null if there is no data
*/
Conflator.prototype.popAllData = function() {
	if(this._data.length) {
		let data = this._data;
		this._data = [];
		return data;
	}
	return null;
};
/** @public */
Conflator.prototype.dispose = function () {
	if(this._timerId) {
		clearTimeout(this._timerId);
		this._timerId = 0;
	}
	if(this._data.length) {
		this._data.length = 0;
	}
	this._func = null; // Release user bound function
};

/** @public
* @param {*=} obj Any data that is needed to be stored for later usage
* @return {boolean}
*/
Conflator.prototype.conflate = function (obj) {
	if(this._disabled) {
		return false;
	}
	if(obj != null) {
		this._data.push(obj);
	}
	if(this._conflated) {
		return false; // Conflating operation is already completed
	}
	if(!this._timerId) {
		this._timerId = setTimeout(this._onConflated, this._delay);
	}
	return true;
};
/** @public
* @return {boolean}
*/
Conflator.prototype.isConflating = function () {
	return (this._timerId > 0 && !this._conflated && !this._disabled);
};

/** @private */
Conflator.prototype._onConflated = function () {
	this._conflated = true;

	try {
		this._func();
	} catch (err) {
		console.error(err.message || err);
	}

	this._timerId = 0;
	this._data.length = 0;
	this._conflated = false;
};

/** @public
* @return {!Array}
*/
Conflator.prototype.getData = function () {
	return this._data;
};
/** @public
* @param {Function} func Bound function that will use this conflator
* @param {*=} thisObj "this" object to be bound with the given function. If the function is already bound, there is no need to provide thisObj parameter
*/
Conflator.prototype.setFunction = function (func, thisObj) {
	this._func = func;
	if(func) {
		if(thisObj) {
			this._func = func.bind(thisObj);
		} else {
			this._func = func;
		}
	} else {
		this._func = null;
	}
};
/** @public
* @param {number} ms The delay time in millisecond before executing the function
*/
Conflator.prototype.setConflatingTime = function (ms) {
	this._delay = ms || 0;
	if(this._delay < 0) {
		this._delay = 0;
	}
};

/** @public
* @return {number} milli second
*/
Conflator.prototype.getConflatingTime = function () {
	return this._delay;
};

/** @public
* @param {boolean=} opt_disabled
*/
Conflator.prototype.disable = function (opt_disabled) {
	this._disabled = opt_disabled !== false;
};
/** @public
* @return {boolean}
*/
Conflator.prototype.isEnabled = function () {
	return !this._disabled;
};
/** @public
* @param {boolean=} opt_enabled
*/
Conflator.prototype.enable = function (opt_enabled) {
	this._disabled = opt_enabled === false;
};

export default Conflator;
export {Conflator};
