/** Deferred promise should be used when asynchronous method does not return promise object (e.g. typical AJAX calls, or native setTimeout()). In other words, deferred promise is used to allow any non-promise logics to be worked in promise pattern.
* @constructor
* @param {Function=} resolveHandler
* @param {Function=} rejectHandler
* @param {*=} ctx Context object that will be provided as the second parameter of the given handlers
* @example
* let def = new Deferred(handler);
* setTimeout(def.resolve, 2000);
* let def2 = new Deferred();
* setTimeout(function() {
*   def2.resolve();
* }, 4000);
*/
let Deferred = function(resolveHandler, rejectHandler, ctx) {
	this.resolve = this.resolve.bind(this);
	this.reject = this.reject.bind(this);

	if(typeof resolveHandler === "function") {
		this._resolveHandler = resolveHandler;
	}
	if(typeof rejectHandler === "function") {
		this._rejectHandler = rejectHandler;
	}
	if(ctx != null) {
		this._ctx = ctx;
	}
	this.promise = new Promise((function(resolve, reject) {
		this._resolve = resolve;
		this._reject = reject;
	}).bind(this));
};
/** @type {Promise<*>}
* @public
*/
Deferred.prototype.promise = null;
/** @type {Function}
* @private
*/
Deferred.prototype._resolve = null; // Always supplied from Promise object
/** @type {Function}
* @private
*/
Deferred.prototype._reject = null; // Always supplied from Promise object
/** @type {Function}
* @private
*/
Deferred.prototype._resolveHandler = null;
/** @type {Function}
* @private
*/
Deferred.prototype._rejectHandler = null;
/** Context object that will be provided as the second parameter of the given handlers
* @type {*}
* @private
*/
Deferred.prototype._ctx = null;
/** @type {boolean}
* @private
*/
Deferred.prototype._fulfilled = false;

/** @public
* @param {*=} result
*/
Deferred.prototype.resolve = function(result) {
	if(this._resolve) { // Promise can be resolved only once
		if(this._resolveHandler) {
			let handlerRes = this._resolveHandler(result, this._ctx || this);
			if(handlerRes != null) {
				result = handlerRes;
			}
		}
		this._resolve(result); // WARNING: Synchronous call. The standard promise behavior requires asynchronous response.
		this._finally();
	}
};
/** @public
* @param {*=} result
*/
Deferred.prototype.reject = function(result) {
	if(this._reject) { // Promise can be resolved only once
		if(this._rejectHandler) {
			let handlerRes = this._rejectHandler(result, this._ctx || this);
			if(handlerRes != null) {
				result = handlerRes;
			}
		}
		this._reject(result); // WARNING: Synchronous call
		this._finally();
	}
};
/** @public
* @return {boolean}
*/
Deferred.prototype.isFulfilled = function() {
	return this._fulfilled;
};

/** @private */
Deferred.prototype._finally = function() {
	// Release all references to avoid memory leak
	this._fulfilled = true;
	this._ctx = this._resolveHandler = this._rejectHandler = null;
	this._resolve = this._reject = null;
};

export default Deferred;
export {Deferred};
