import { injectCss, prettifyCss } from "./Util.js";
import Dom from "./Dom.js";
import ElfUtil from "./ElfUtil.js";

/**
* @constructor
* @param {Object} options
*/
let DragUI = function(options) {
	this._dragBox = options.dragBox;
	this._dragBoxIcon = options.dragBoxIcon;
	this._dragBoxContent = document.createElement("div");
	this._dragBoxContent.className = "drag-box-content";
	this._dragBox.appendChild(this._dragBoxContent);
};

/** @type {string}
* @private
*/
DragUI._styles = ""; // Static variable
/** @type {Element}
* @private
*/
DragUI.prototype._guideline = null;
/** @type {Element}
* @private
*/
DragUI.prototype._dragBox = null;
/** @type {Element}
* @private
*/
DragUI.prototype._dragBoxIcon = null;
/** Core grid instance
* @type {Object}
* @private
*/
DragUI.prototype._grid = null;

/** Apply the theme colors to the specified grid after DragUI.initStyles
* @public
* @param {Object} grid core grid instance
*/
DragUI.applyThemeColor = function(grid) {
	if(!grid || grid._dragUIStyles) {
		return;
	}

	if(DragUI._styles) {
		grid._dragUIStyles = true; // Prevent loading the same style twice
		injectCss(DragUI._styles, grid.getElement());
	}
	// TO_DO : Handle if can not load DragUI._styles try to set new DragUI._styles
	// Warning: This process have to use ElfUtil.getThemeColors() that async process for Backward compatability.
	// When use async process, then it can be load theme multiple times.
};

/** Deprecated in favor of DragUI.initStyles()
* @public
* @param {Object} colors
*/
DragUI.prototype.onThemeLoaded = function(colors) {
	DragUI.initStyles(colors);
};

/** WARNING: This method should be called after the theme has been loaded in order to initialize with the correct colors.
* @public
* @function
* @param {Object=} colors
*/
DragUI.initStyles = function(colors) {
	if(!DragUI._styles) {
		if(!colors) {
			colors = ElfUtil.getColors();
		}

		let ElfVersion = ElfUtil.getElfVersion();
		let cursor = "grabbing";
		if (ElfVersion < 3) {
			cursor = "move";
		}
		let styles = [ // Main Styles without theme
			".tr-row-guideline", [ // Backward compatability of row dragging
				"position: absolute;",
				"left: 0;",
				"top: 0;",
				"pointer-events: none;",
				"text-align: center;",
				"border-top-width: 3px;",
				"border-top-style: solid;"
			],
			".guideline", [
				"position: absolute;",
				"left: 0;",
				"top: 0;",
				"pointer-events: none;",
				"text-align: center;",
				"border-top-style: solid;",
				"width: 3px;",
				"border-top-color: var(--grid-title-icon-color);",
				"color: var(--grid-title-icon-color);",
				"background-color: var(--grid-title-icon-color);"
			],
			".drag-box", [
				"position: absolute;",
				"left: 0;",
				"top: 0;",
				"pointer-events: none;",
				"background-color: var(--grid-dragbox-bgcolor);",
				"border: solid 1px var(--grid-title-sort-icon-color);",
				"width: 134px;",
				"height: 29px;",
				"line-height: 29px;",
				"white-space: nowrap;",
				"overflow-x: clip;",
				"text-overflow: ellipsis;",
				"padding: 0 7px 0 7px;",
				"box-shadow: 0 0 8px var(--grid-dragbox-bordercolor);",
				"z-index: 2" // Cover sort indicator
			],
			".drag-box-disabled", [
				"width: 100px;",
				"left: 0;",
				"top: 0;",
				"position: absolute;"
			],
			".drag-box-content", [
				"overflow-x: clip;",
				"text-overflow: ellipsis;"
			],
			".drag-box-icon", [
				"top: -4px;",
				"left: 12px;",
				"padding: 2px;",
				"border-radius: 100%;",
				"font-size: 11px;",
				"position: absolute;"
			],
			".drag-box-icon[icon='void']", [
				"background-color:var(--grid-void-icon-bgcolor);",
				"color:var(--grid-void-icon-color);"
			],
			".drag-box-icon[icon='add']", [
				"background-color:var(--grid-insertion-icon-bgcolor);",
				"color:var(--grid-insertion-icon-color);"
			],
			":host", [
				"--grid-dragbox-bgcolor: #ffffff;", // If the value cannot be found, default to halo-light
				"--grid-insertion-icon-bgcolor: #39c46e;",
				"--grid-insertion-icon-color: #1A1A1A;",
				"--grid-void-icon-bgcolor: #F5475B;",
				"--grid-void-icon-color: #FFFFFF;",
				"--grid-title-icon-color: var(--grid-guideline-color);" // v3 fallback
			],
			".mouse-dragging *", [ // While dragging cursor should be changing
				"cursor: " + cursor + " !important;"
			],
			".tr-grid .column .cell.drag-indicator", [ // --grid-drag-indicator defualt is "none"
				"border-top: var(--grid-drag-indicator) !important;",
				"border-right: var(--grid-drag-indicator) !important;", // need to set !important to overwrite the priority set by the last column which has the none border
				"border-bottom: var(--grid-drag-indicator);",
				"border-left: var(--grid-drag-indicator);"
			],
			".tr-dragging, .tr-dragging *", [
				"-webkit-touch-callout: none;",
				".user-select(none);"
			],
			".tr-dragbox", [
				"background-color: unset;" // v3 fallback
			]
		];
		let guidelineColor = "#ff9933";
		if(colors.primary) {
			guidelineColor = colors.primary;
		}
		guidelineColor += ";";

		styles.push(".tr-row-guideline", [
			"border-top-color: " + guidelineColor,
			"color: " + guidelineColor
		]);

		if(ElfVersion < 3) { // Backward compatability
			styles.push(".pearl .tr-row-guideline, *[theme=pearl] .tr-row-guideline", [
				"border-top-color: #ee7600;",
				"color: #ee7600;",
				"background-image: linear-gradient(to bottom, rgba(0, 0, 0, 0), rgba(0, 0, 0, 0.1));"
			]);
		}
		DragUI._styles = prettifyCss(styles);
	}
};

/** @private
* @param {!Element|Event|MouseEvent} e
* @param {Object} grid
* @returns {boolean}=true means dragBox can be drop
*/
DragUI.prototype.renderDragBox = function (e, grid) {
	if(!grid) {
		return false;
	}

	let gridElem = grid.getElement();
	let dragBoxHost = gridElem;
	let gridParent = grid.getParent().parentNode;
	if (gridParent.nodeType === 11) {
		dragBoxHost = gridParent;
	}

	let parent = this._dragBox.parentNode;
	if(!parent) {
		dragBoxHost.appendChild(this._dragBox);
	}

	let dragBoxIcon = e.dragBoxIcon || this._dragBox.dragBoxIcon; // The user-supplied icon
	this._dragBoxIcon.style.visibility = "visible";
	parent = this._dragBoxIcon.parentNode;
	let drop = true;
	if(dragBoxIcon === "insertion") {
		drop = false;
		this._dragBoxIcon.icon = "add";
		if(!parent) {
			this._dragBox.appendChild(this._dragBoxIcon);
		}
	} else if (dragBoxIcon === "not-allowed" || dragBoxIcon === "no-drop" || dragBoxIcon === "void" ) {
		drop = false;
		this._dragBoxIcon.icon = "void";
		if(!parent) {
			this._dragBox.appendChild(this._dragBoxIcon);
		}
	} else {
		Dom.removeParent(this._dragBoxIcon);
	}

	let gridRect = gridElem.getBoundingClientRect();

	let x = e.clientX - gridRect.left;
	let y = e.pageY - gridRect.top;

	// space for mouse cursor
	this._dragBox.style.left = x - 5 + "px";
	this._dragBox.style.top = y + 8  + "px";
	return drop;
};

/** @public
* @param {*} content Element, Node, string, number, or everything else.
*/
DragUI.prototype.setContent = function(content) {
	if (content && content["getElement"]) {
		content = content["getElement"]();
	}

	let dragBoxContent = this._dragBoxContent;
	let currentContent = dragBoxContent._content;
	if (content !== currentContent) {
		dragBoxContent._content = content;
		Dom.setContent(dragBoxContent, content);
	}
};

export default DragUI;
export {DragUI};
