
/** @namespace
*/
let ElementObserver = {};

/** @private
* @param {Element} elem
* @param {MutationRecord} mutation
*/
let _onLanguageMutated = function (elem, mutation) {
	if (mutation.type == "attributes" && mutation.attributeName == "lang") {
		elem.lang = mutation.target.lang || "en";
	}
};

/** @private
* @param {Function} listener
* @param {string=} attributeName
* @param {MutationRecord} mutation
*/
let _onAttributeMutated = function (listener, attributeName, mutation) {
	if (mutation.type === "attributes") {
		if (attributeName != null) {
			if (mutation.attributeName === attributeName) {
				listener(mutation.target.getAttribute(attributeName));
			}
		} else {
			listener(mutation.target);
		}
	}
};

/** @private
* @return {string}
*/
ElementObserver._getNewId = function () {
	let id = "observer_" + (++ElementObserver._id);
	return id;
};

/** Observe any element
* @private
* @param {Element} elem
* @param {Function} listener
* @param {Object=} opt_option
*/
ElementObserver._addObserver = function(elem, listener, opt_option) {
	let option = opt_option;
	if (!option) {
		option = {};
	}
	option.attributes = true; //configure it to listen to attribute changes

	ElementObserver._observer.observe(elem, option);
	ElementObserver._addListener(elem, listener);
};

/** @private
* @param {Array.<MutationRecord>} mutations
*/
ElementObserver._onMutationCallback = function(mutations) {
	for (let i = 0; i < mutations.length; i++) {
		let mutation = mutations[i];
		let listeners = ElementObserver._listeners[mutation.target._observeId];
		for (let j = 0; j < listeners.length; j++) {
			listeners[j](mutation);
		}
	}
};

/** @private
* @param {Element} elem
* @param {Function} fn
*/
ElementObserver._addListener = function(elem, fn) {
	let listeners = ElementObserver._listeners;
	if (!listeners) {
		listeners = ElementObserver._listeners = {};
	}
	if (!elem._observeId) {
		let id = ElementObserver._getNewId();
		elem._observeId = id;
	}
	if (!listeners[elem._observeId]) { // Always check to support the using of separated ElfUtil module in testing page
		listeners[elem._observeId] = [];
	}

	let listener = listeners[elem._observeId];
	if (listener.indexOf(fn) < 0) {
		listener.push(fn);
	}
};

/** Add a listener to a html lang attribute
* @public
* @param {Element} element An element within the DOM tree to watch for changes
*/
ElementObserver.addLanguageListener = function(element) {
	if (!element) { return; }

	element.lang = document.documentElement.lang || "en";

	ElementObserver._addObserver(document.documentElement, _onLanguageMutated.bind(null, element));
};

/** Add a listener to a html attribute
* @public
* @param {Element} element An element within the DOM tree to watch for changes
* @param {Function} listener A function which will be called on each attribute change
* @param {string=} attributeName If not specified, listener will be called on every attribute change
*/
ElementObserver.addAttributeListener = function(element, listener, attributeName) {
	if (!element || !listener) { return; }
	ElementObserver._addObserver(element, _onAttributeMutated.bind(null, listener, attributeName));
};

/**
* @type {MutationObserver}
* @private
*/
ElementObserver._observer = new MutationObserver(ElementObserver._onMutationCallback);
/**
* @type {Obect}
* @private
*/
ElementObserver._listeners = {};
/**
* @type {number}
* @private
*/
ElementObserver._id = 0;

export default ElementObserver;
export { ElementObserver };
