/** @namespace */
let ElfDate = {};

/** Check ELF version
* @public
* @function
* @param {*} val String, number, array and any other types are acceptable
* @return {Date} Native Date object or null for Invalid Date
* @example
* ElfDate.from(1000);         // Thu Jan 01 1970 07:00:01 GMT+0700 (Indochina Time)
* ElfDate.from("2000/12/2");  // Sat Dec 02 2000 00:00:00 GMT+0700 (Indochina Time)
* ElfDate.from("2000-12-02"); // Sat Dec 02 2000 07:00:00 GMT+0700 (Indochina Time)
* ElfDate.from("2000-12-2");  // Sat Dec 02 2000 00:00:00 GMT+0700 (Indochina Time)
* ElfDate.from([2000, 11, 2]);// Sat Dec 02 2000 07:00:00 GMT+0700 (Indochina Time)
* ElfDate.from("2/12/2000");  // Sat Feb 12 2000 00:00:00 GMT+0700 (Indochina Time)
* ElfDate.from("20/12/2000"); // null
* ElfDate.from(new Date("Invalid")); // null
*/
ElfDate.from = function (val) {
	if(val) {
		let d = val;
		if(typeof val === "string" || typeof val === "number") {
			d = new Date(val);
		} else if(Array.isArray(val)) {
			d = new Date(Date.UTC(val[0] || 0, val[1] || 0, val[2] || 0));
		}
		if(ElfDate.isValid(d)) {
			return d;
		}
	}
	return null;
};

/** @public
* @function
* @param {Date} d Native Date object
* @return {boolean}
*/
ElfDate.isValid = function (d) {
	if(d && d.getTime) { // Date object must have getTime method
		let ms = d.getTime(); // millisecond since 1 first january 1970
		return ms === ms; // Invalid Date will have NaN as its value
	}
	return false;
};

/** @public
* @function
* @param {number} n
* @return {string}
* @example
* ElfDate.prefixSingleDigit(9); // "09"
* ElfDate.prefixSingleDigit(0); // "00"
* ElfDate.prefixSingleDigit(10); // "10"
*/
ElfDate.prefixSingleDigit = function (n) {
	return (n < 10 && n >= 0) ? "0" + n : "" + n;
};

/** @type {RegExp}
* @private
*/
ElfDate._ymdRe = /^\d{4}\-d{2}\-\d{2}/; // eslint-disable-line

/** Returns yyyy-MM-dd string format for ef-datetime-picker.
* @public
* @function
* @param {string|Date} d
* @return {string} Return string for valid date, otherwise empty string
* @example
* ElfDate.toYMD(new Date("2000/1/1")); // "2000-01-01"
* ElfDate.toYMD("2000/1/1"); // "2000-01-01"
*/
ElfDate.toYMD = function (d) {
	if(typeof d === "string") {
		if(ElfDate._ymdRe.test(d)) {
			return d; // pass through only if we have yyyy-MM-dd
		}
	}
	let dateObj = ElfDate.from(d);
	if(dateObj) {
		return dateObj.getFullYear() + "-" +
			ElfDate.prefixSingleDigit(dateObj.getMonth() + 1) + "-" +
			ElfDate.prefixSingleDigit(dateObj.getDate());
	}
	return "";
};

/** @public
* @function
* @param {Element} edtp Emerald date-time picker element
* @param {*} val
*/
ElfDate.setDate = function (edtp, val) {
	if (edtp) {
		if(edtp.isAttached) { // Only Polymer element has isAttached property
			edtp.value = ElfDate.from(val);
		} else {
			edtp.value = ElfDate.toYMD(val);
		}
	}
};
/** @public
* @function
* @param {Element} edtp Emerald date-time picker element
* @return {Date} Native Date object
*/
ElfDate.getDate = function (edtp) {
	return (edtp) ? ElfDate.from(edtp.value) : null;
};

/** @public
* @function
* @param {*} valA
* @param {*} valB
* @return {number} Return nagative value if A < B, positive value if A > B, and 0 if A === B
*/
ElfDate.compare = function (valA, valB) {
	if(valA !== valB) {
		valA = ElfDate.from(valA);
		valB = ElfDate.from(valB);
		if(valA) {
			if(valB) {
				if(valA < valB) {
					return -1;
				} else if(valA > valB) {
					return 1;
				}
			} else {
				return 1;
			}
		} else if(valB) {
			return -1;
		}
	}
	return 0;
};
/** @public
* @function
* @param {*} valA
* @param {*} valB
* @return {boolean} true if the two has the same value
*/
ElfDate.equal = function (valA, valB) {
	return !ElfDate.compare(valA, valB);
};

/** Extract date part from the given date object to exclude time part and timezone. <br>
* Use ElfDate.from to convert array back to date object.
* @public
* @function
* @param {Date} d
* @return {Array.<number>} Array with 3 numbers. Return null for invalid date
*/
ElfDate.extractDate = function (d) {
	if(ElfDate.isValid(d)) {
		return [d.getFullYear(), d.getMonth(), d.getDate()];
	}
	return null;
};


export default ElfDate;
export { ElfDate };
