/** Abstract base class that provides event management methods for derived class
* @constructor
*/
let EventDispatcher = function () {};

/** @type {Object.<string, Function>}
* @protected
* @ignore
*/
EventDispatcher.prototype._events = null;
/** @type {number}
* @protected
* @ignore
*/
EventDispatcher.prototype._listenerCount = 0;

//#region public function of event
/** Add the function handler into the specific event.
* @public
* @param {string} type Event name
* @param {Function} handler Event handler
*/
EventDispatcher.prototype.addEventListener = function(type, handler) {
	if(!type || !handler) { return; }
	if(!this._events) {
		this._events = {};
	}

	let listeners = this._events[type];
	if(listeners) {
		if(listeners.indexOf(handler) < 0) {
			listeners.push(handler);
			++this._listenerCount;
		}
	} else {
		this._events[type] = [handler];
		++this._listenerCount;
	}
};
/** Remove the function handler from the specific event.
* @public
* @param {string} type Event name
* @param {Function} handler Event handler
*/
EventDispatcher.prototype.removeEventListener = function(type, handler) {
	let listeners = (this._events) ? this._events[type] : null;
	if(listeners) {
		let at = listeners.indexOf(handler);
		if(at >= 0) {
			listeners.splice(at, 1);
			--this._listenerCount;
		}
	}
};
/** Remove all function handlers from all events.
* @public
*/
EventDispatcher.prototype.removeAllEventListeners = function() {
	this._events = null;
	this._listenerCount = 0;
};

/** Check whether the specified event has a function handler.
* @public
* @param {string} type Event name
* @return {boolean}
*/
EventDispatcher.prototype.hasListener = function(type) {
	if(this._listenerCount) {
		if(type) {
			let listeners = this._events ? this._events[type] : null;
			return listeners ? (listeners.length > 0) : false;
		} else {
			return true;
		}
	} else {
		return false;
	}
};
/** Get event listener (function handler) from the specified event.
* @public
* @param {string} type Event name
* @param {number=} idx=0 Index of the listener to be retrieved. This is used when there are multiple event listeners
* @return {Function}
*/
EventDispatcher.prototype.getListener = function(type, idx) {
	if(type) {
		let listeners = this._events ? this._events[type] : null;
		if(listeners) {
			if(!idx) {
				idx = 0;
			}
			return listeners[idx] || null;
		}
	}
	return null;
};
/** A shorthand to retrieve function from an object and add it as an Eventlistener
* @public
* @function
* @param {Object} obj Object that contains a handler with the same name as the given `type`
* @param {string} type Event name
* @example
* let obj = {"mouseUp": function(e) { console.log(e); }};
* plugin.addListener(obj, "mouseUp");
* plugin.addListener(obj, "mouseDown");
*/
EventDispatcher.prototype.addListener = function(obj, type) {
	let func = obj ? obj[type] : null;
	if(typeof func === "function") {
		this.addEventListener(type, func);
	}
};

/** Executed before dispatching, allowing modification of user's event argument object
* @protected
* @ignore
* @param {string} type Event name
* @param {Object} eventArg User's event argument object
* @return {Object} New event argument object
*/
EventDispatcher.prototype._prepareEventArguments = function(type, eventArg) {
	return eventArg;
};

/** Publish the data to the listener.
* @protected
* @ignore
* @param {string} type Event name
* @param {Object} eventArg Event arguments
*/
EventDispatcher.prototype._dispatch = function(type, eventArg) {
	let listeners = this._events ? this._events[type] : null;
	if(listeners) {
		eventArg = this._prepareEventArguments(type, eventArg);
		let len = listeners.length;
		for(let i = 0; i < len; ++i) {
			listeners[i](eventArg);
		}
	}
};

/** @public
* @function
* @param {Event} e
*/
let preventDefault = function(e) {
	if(e && e.preventDefault) {
		e.preventDefault();
		e.stopPropagation();
	}
};

/** @public
* @function
* @param {Event} e
*/
EventDispatcher.preventDefault = preventDefault;

export default EventDispatcher;
export {EventDispatcher, preventDefault};
