/** Provide ability for prototype based Class to inherits another class
* @namespace
* @example
* let BaseClass = function() {};
* BaseClass.prototype.method = function() {};
* BaseClass.prototype._member = 1;
* let DerivedClass = function() {};
* Ext.inherits(DerivedClass, BaseClass); // Derived class will have methods and members of the base class
* // To call base class method
* DeriveClass["base"](this, "methodName", ...params);
*/
let Ext = {};

/** @public
* @function
* @param {Function} childCtor
* @param {Function} parentCtor
*/
Ext.inherits = function (childCtor, parentCtor) {
	/** @constructor
* @ignore
*/
	function tempCtor() { }
	tempCtor.prototype = parentCtor.prototype;
	childCtor.prototype = new tempCtor();
	childCtor.prototype.constructor = childCtor;

	/** Calls superclass constructor/method. <br>
* This function is only available if you use goog.inherits to <br>
* express inheritance relationships between classes. <br>
* NOTE: This is a replacement for goog.base and for superClass_ <br>
* property defined in childCtor.
* @param {!Object} me Should always be "this".
* @param {(string|null)=} methodName The method name to call. Calling superclass constructor can be done with the special string 'constructor'.
* @param {...*} var_args The arguments to pass to superclass <br>
* method/constructor.
* @return {*} The return value of the superclass method/constructor.
*/
	childCtor["base"] = function(me, methodName, var_args) {
		if(!methodName) { methodName = 'constructor'; }
		// Copying using loop to avoid deop due to passing arguments object to function. This is faster in many JS engines as of late 2014.
		let len = arguments.length;
		let args = new Array(len); // http://jsperf.com/creating-an-array
		for (let i = 2; i < len; i++) {
			args[i - 2] = arguments[i];
		}
		return parentCtor.prototype[methodName].apply(me, args);
	};
};

export default Ext;
export {Ext};
