/** @typedef {Object} FilterOperators~Operator
* @property {string} id Short name of the operator
* @property {string} name Full name of the operator
* @property {string} type Data type of the  parameters required by the operator (e.g., number, string)
* @property {number} positive The value indicates that the operator has positive connotation. The negative counterparts for EQ and BEGIN are NEQ and NBEGIN respectively.
* @property {number} caseSensitive The value indicates that the operator requires case sensitive parameters
* @property {string} formula Formula expression used by the operator
*/

/** @typedef {Object.<string, FilterOperators~Operator>} FilterOperators~Operators
* @property {!FilterOperators~Operator} EQ Equal
* @property {!FilterOperators~Operator} NEQ Not equal
* @property {!FilterOperators~Operator} NUMEQ Number equal to
* @property {!FilterOperators~Operator} NUMNEQ Number not equal to
* @property {!FilterOperators~Operator} GT Greater than
* @property {!FilterOperators~Operator} GTE Greater than or equal
* @property {!FilterOperators~Operator} LT Less than
* @property {!FilterOperators~Operator} LTE Less than or equal
* @property {!FilterOperators~Operator} BEGIN Begin
* @property {!FilterOperators~Operator} NBEGIN Not begin
* @property {!FilterOperators~Operator} END End
* @property {!FilterOperators~Operator} NEND Not end
* @property {!FilterOperators~Operator} CONT Contain
* @property {!FilterOperators~Operator} NCONT Not contain
* @property {!FilterOperators~Operator} EQ_BLANK Equal to blank value
* @property {!FilterOperators~Operator} EQ_NBLANK Equal to non blank value
* @property {!FilterOperators~Operator} DT Date is
* @property {!FilterOperators~Operator} DTA Date is after
* @property {!FilterOperators~Operator} DTB Date is before
* @property {!FilterOperators~Operator} TXTEQ Alias to EQ
* @property {!FilterOperators~Operator} BLANK Alias to EQ_BLANK
* @property {!FilterOperators~Operator} NBLANK Alias to EQ_NBLANK
* @property {!FilterOperators~Operator} NEQ_BLANK Alias to EQ_NBLANK
* @property {!FilterOperators~Operator} GREATER_THAN Alias to GT
* @property {!FilterOperators~Operator} GREATER_THAN_OR_EQUAL_TO Alias to GTE
* @property {!FilterOperators~Operator} EQUAL_TO Alias to EQ
* @property {!FilterOperators~Operator} LESS_THAN Alias to LT
* @property {!FilterOperators~Operator} LESS_THAN_OR_EQUAL_TO Alias to LTE
* @property {!FilterOperators~Operator} TEXT_THAT_CONTAINS Alias to CONT
* @property {!FilterOperators~Operator} TEXT_THAT_NOT_CONTAINS Alias to NCONT
* @property {!FilterOperators~Operator} TEXT_IS Alias to EQ
* @property {!FilterOperators~Operator} NOT_BLANK Alias to EQ_NBLANK
*/

/** @type {!FilterOperators~Operators}
* @public
* @const
*/
let FilterOperators = {
	EQ: {
		id: "EQ",
		name: "Equal to",
		type: "string",
		positive: 1,
		caseSensitive: 1,
		formula: "str == input"
	},
	NEQ: {
		id: "NEQ",
		name: "Not equal to",
		type: "string",
		positive: 0,
		caseSensitive: 1,
		formula: "str != input"
	},
	NUMEQ: {
		id: "NUMEQ",
		name: "Number equal to",
		type: "number",
		positive: 1,
		caseSensitive: 1,
		formula: "val == input"
	},
	NUMNEQ: {
		id: "NUMNEQ",
		name: "Number not equal to",
		type: "number",
		positive: 0,
		caseSensitive: 1,
		formula: "val != input"
	},
	GT: {
		id: "GT",
		name: "Greater than",
		type: "number",
		positive: 1,
		caseSensitive: 1,
		formula: "val > input"
	},
	GTE: {
		id: "GTE",
		name: "Greater than or equal to",
		type: "number",
		positive: 1,
		caseSensitive: 1,
		formula: "val >= input"
	},
	LT: {
		id: "LT",
		name: "Less than",
		type: "number",
		positive: 1,
		caseSensitive: 1,
		formula: "val < input"
	},
	LTE: {
		id: "LTE",
		name: "Less than or equal to",
		type: "number",
		positive: 1,
		caseSensitive: 1,
		formula: "val <= input"
	},
	BEGIN: {
		id: "BEGIN",
		name: "Begin with",
		type: "string",
		positive: 1,
		caseSensitive: 0,
		formula: "str.indexOf(input) === 0"
	},
	NBEGIN: {
		id: "NBEGIN",
		name: "Not begin with",
		type: "string",
		positive: 0,
		caseSensitive: 0,
		formula: "str.indexOf(input) !== 0"
	},
	END: {
		id: "END",
		name: "End with",
		type: "string",
		positive: 1,
		caseSensitive: 0,
		formula: "(str.length >= input.length) && str.lastIndexOf(input) === (str.length - input.length)"
	},
	NEND: {
		id: "NEND",
		name: "Not end with",
		type: "string",
		positive: 0,
		caseSensitive: 0,
		formula: "(str.length < input.length) || str.lastIndexOf(input) !== (str.length - input.length)"
	},
	CONT: {
		id: "CONT",
		name: "Contain",
		type: "string",
		positive: 1,
		caseSensitive: 0,
		formula: "str.indexOf(input) >= 0"
	},
	NCONT: {
		id: "NCONT",
		name: "Not contain",
		type: "string",
		positive: 0,
		caseSensitive: 0,
		formula: "str.indexOf(input) < 0"
	},
	EQ_BLANK: {
		id: "EQ_BLANK",
		name: "Equal to blank value",
		type: "blank", // Blank checking
		positive: 1,
		caseSensitive: 1,
		formula: "!val && val !== 0" // false boolean value would not pass
	},
	EQ_NBLANK: {
		id: "EQ_NBLANK",
		name: "Equal to non blank value",
		type: "blank", // Blank checking
		positive: 1,
		caseSensitive: 1,
		formula: "(val || val === 0) ? true : false"
	},
	DT: {
		id: "DT",
		name: "Date is",
		type: "date",
		positive: 1,
		caseSensitive: 1,
		formula: "ts === input"
	},
	DTA: {
		id: "DTA",
		name: "Date is After",
		type: "date",
		positive: 1,
		caseSensitive: 1,
		formula: "ts > input"
	},
	DTB: {
		id: "DTB",
		name: "Date is Before",
		type: "date",
		positive: 1,
		caseSensitive: 1,
		formula: "ts < input"
	}
};

// Aliases
// TXTEQ is equivalent to EQ the type conversion will always be performed
FilterOperators.TXTEQ = FilterOperators.EQ;
FilterOperators.BLANK = FilterOperators.EQ_BLANK;
FilterOperators.NBLANK = FilterOperators.EQ_NBLANK;
FilterOperators.NEQ_BLANK = FilterOperators.EQ_NBLANK;

FilterOperators.GREATER_THAN = FilterOperators.GT;
FilterOperators.GREATER_THAN_OR_EQUAL_TO = FilterOperators.GTE;
FilterOperators.EQUAL_TO = FilterOperators.EQ;
FilterOperators.LESS_THAN = FilterOperators.LT;
FilterOperators.LESS_THAN_OR_EQUAL_TO = FilterOperators.LTE;
// FilterOperators.BETWEEN = // No equivalent operator
FilterOperators.TEXT_THAT_CONTAINS = FilterOperators.CONT;
FilterOperators.TEXT_THAT_NOT_CONTAINS = FilterOperators.NCONT;
FilterOperators.TEXT_IS = FilterOperators.EQ;
// FilterOperators.BLANK = FilterOperators.EQ_BLANK;
FilterOperators.NOT_BLANK = FilterOperators.EQ_NBLANK;


/** @type {!Object.<string, Function>}
* @public
* @const
*/
let OperatorFunctions = {
	EQ: function(str, input) { return str == input; },
	NEQ: function(str, input) { return str != input; },
	NUMEQ: function(num, input) { return num == input; },
	NUMNEQ: function(num, input) { return num != input; },
	GT: function(num, input) { return num > input; },
	GTE: function(num, input) { return num >= input; },
	LT: function(num, input) { return num < input; },
	LTE: function(num, input) { return num <= input; },
	BEGIN: function(str, input) { return str.indexOf(input) === 0; },
	NBEGIN: function(str, input) { return str.indexOf(input) !== 0; },
	END: function(str, input) {
		return (str.length >= input.length) && str.lastIndexOf(input) === (str.length - input.length);
	},
	NEND: function(str, input) {
		return (str.length < input.length) || str.lastIndexOf(input) !== (str.length - input.length);
	},
	CONT: function(str, input) { return str.indexOf(input) >= 0; },
	NCONT: function(str, input) { return str.indexOf(input) < 0; },
	EQ_BLANK: function(val, input) { return !val && val !== 0; },
	EQ_NBLANK: function(val, input) { return (val || val === 0) ? true : false; },
	DT: function(ts, input) { return ts === input; },
	DTA: function(ts, input) { return ts > input; },
	DTB: function(ts, input) { return ts < input; }
};

export default FilterOperators;
export { FilterOperators, OperatorFunctions };
