/** Perf encapsulates window's performance object and is disabled by default for lightweight performance profiling
* @namespace
* @param {Element=} elem
*/
let Perf = {};

/** Enable performance profiling
* @public
* @function
* @param {boolean=} enabled
*/
Perf.enable = function(enabled) {
	if(enabled || enabled == null) {
		let perfObj = window["performance"];
		if(perfObj && typeof perfObj.mark == "function") {
			Perf._perf = perfObj;
		}
	} else { // Disable
		if(Perf.isEnabled()) {
			Perf.clearMarks();
			Perf.clearMeasures();
			Perf._perf = EmptyPerf;
		}
	}
};
/** @public
* @function
* @return {boolean}
*/
Perf.isEnabled = function() {
	return Perf._perf !== EmptyPerf;
};

/** Removes the given mark from the browser's performance entry buffer.
* @public
* @function
*/
Perf.clearMarks = function() {
	Perf._perf.clearMarks();
};
/** Removes the given measure from the browser's performance entry buffer.
* @public
* @function
*/
Perf.clearMeasures = function() {
	Perf._perf.clearMeasures();
};
/** Removes all performance entries with a entryType of "resource" from the browser's performance data buffer. <br>
* ResourceTimings are automatically collected by the browsers for all of the resources (e.g. script, link, css, img, and xmlhttprequest)
* @public
* @function
*/
Perf.clearResourceTimings = function() {
	Perf._perf.clearResourceTimings();
};

/** Returns a list of PerformanceEntry objects based on the given filter. Not supported in Opera
* @public
* @function
* @return {Array.<PerformanceEntry>} An array of PerformanceEntry objects
*/
Perf.getEntries = function() {
	return Perf._perf.getEntries();
};
/** Returns a list of PerformanceEntry objects based on the given name and entry type.
* @public
* @function
* @param {string} entryName
* @param {string=} entryType="mark"
* @return {Array.<PerformanceEntry>} An array of PerformanceEntry objects
*/
Perf.getEntriesByName = function(entryName, entryType) {
	return Perf._perf.getEntriesByName(entryName, entryType);
};
/** Returns a list of PerformanceEntry objects of the given entry type.
* @public
* @function
* @param {string} entryType Available types are mark, measure, resource, frame, navigation, and paint
* @return {Array.<PerformanceEntry>} An array of PerformanceEntry objects
* @see {@link https://developer.mozilla.org/en-US/docs/Web/API/PerformanceEntry/entryType}
*/
Perf.getEntriesByType = function(entryType) {
	return Perf._perf.getEntriesByType(entryType);
};
/** A shorthand for getting single entry by name
* @public
* @function
* @param {string} entryName
* @param {string=} entryType="mark"
* @return {PerformanceEntry}
*/
Perf.getEntry = function(entryName, entryType) {
	let ary = Perf._perf.getEntriesByName(entryName, entryType);
	return ary ? ary[0] : null;
};

/** Creates a timestamp in the browser's performance entry buffer with the given name. Supported in IE10
* @public
* @function
* @param {string} markName
*/
Perf.mark = function(markName) {
	Perf._perf.mark(markName);
};
/** Creates a named timestamp in the browser's performance entry buffer between two specified marks (known as the start mark and end mark, respectively). Supported in IE10
* @public
* @function
* @param {string} measureName
* @param {string=} mark1=navigationStart
* @param {string=} mark2=now
* @example
* Perf.measure("measure a to b", markerNameA, markerNameB);
* Perf.measure("measure a to now", markerNameA);
* Perf.measure("measure from navigation start to b", undefined, markerNameB);
* Perf.measure("measure from the start of navigation to now");
*/
Perf.measure = function(measureName, mark1, mark2) {
	Perf._perf.measure(measureName, mark1, mark2);
};
/** Returns a DOMHighResTimeStamp representing the number of milliseconds elapsed since a reference instant. Supported in IE10 and not supported in Opera for Android
* @public
* @function
* @return {number} In milliseconds
* @see {@link https://developer.mozilla.org/en-US/docs/Web/API/DOMHighResTimeStamp}
*/
Perf.now = function() {
	return Perf._perf.now();
};

/** @private
* @function
* @param {...*} arg
* @return {*}
*/
let emptyMethod = function(arg) {
	return null;
};
/** @type {Object}
* @private
*/
let EmptyPerf = {};
EmptyPerf.clearMarks = emptyMethod;
EmptyPerf.clearMeasures = emptyMethod;
EmptyPerf.clearResourceTimings = emptyMethod;

EmptyPerf.getEntries = emptyMethod;
EmptyPerf.getEntriesByName = emptyMethod;
EmptyPerf.getEntriesByType = emptyMethod;
EmptyPerf.getEntry = emptyMethod;

EmptyPerf.mark = emptyMethod;
EmptyPerf.measure = emptyMethod;
EmptyPerf.now = function() {
	return 0;
};

/** @type {Object}
* @private
*/
Perf._perf = EmptyPerf;

export default Perf;
export {Perf};
