import {Dom} from './Dom.js';
import {Ext} from './Ext.js';
import {ElementWrapper} from './ElementWrapper.js';
import {ElfUtil} from './ElfUtil.js';

/** Provides UI for simple range bar
* @constructor
* @param {Element=} elem
* @param {Object=} options
*/
let RangeBar = function (elem, options) { // TODO: Add type definition for options object showing all available options
	this._elem = elem || Dom.div();
	this._elem.style.lineHeight = "1";
	this._elem.style.position = "relative";
	this._elem.style.textAlign = "center";

	this._slider = Dom.div();
	this._slider.style.width = "3px";
	this._slider.style.height = "100%";
	this._slider.style.position = "absolute";

	this.setStyles(options);
};
Ext.inherits(RangeBar, ElementWrapper);

/** @type {Object}
* @public
* @const
*/
RangeBar.defaults = {
	"trackHeight": "0.8em",
	"thumbColor": "#FFF", // TODO: Make this able to change dynamically
	"naText": "N/A"
};

/** @type {Element}
* @private
*/
RangeBar.prototype._slider = null;
/** @type {number}
* @private
*/
RangeBar.prototype._percent; // Intentionally left undefined
/** @type {boolean}
* @private
*/
RangeBar.prototype._tooltipEnabled = true;
/** @type {boolean}
* @private
*/
RangeBar.prototype._hidden = false;
/** @type {!Object}
* @private
*/
RangeBar.prototype._styles;

/** @public
* @param {number=} lowVal
* @param {number=} lastVal
* @param {number=} highVal
*/
RangeBar.prototype.setValue = function(lowVal, lastVal, highVal) {
	this.setPercent(RangeBar._calculatePercent(lowVal, lastVal, highVal));

	// If all values are not intialized, hide the rangebar instead of showing N/A
	this.hide(lowVal == null && lastVal == null && highVal == null);

	// if(this._tooltipEnabled) {
	// 	this._updateTooltip(lowVal, lastVal, highVal);
	// }
};
/** @public
* @param {boolean=} bool
*/
RangeBar.prototype.hide = function(bool) {
	bool = bool !== false;
	if(this._hidden !== bool) {
		this._hidden = bool;
		this._elem.style.visibility = bool ? "hidden" : "";
	}
};
/** @public
* @return {boolean}
*/
RangeBar.prototype.isHidden = function() {
	return this._hidden;
};

/** @public
* @param {boolean=} enabled
*/
RangeBar.prototype.enableTooltip = function(enabled) {
// 	enabled = enabled != false;
// 	if(this._tooltipEnabled !== enabled) {
// 		if(enabled) {
// 			this._elem.removeAttribute("title");
// 		} else {
// 			this._updateTooltip();
// 		}
// 		this._tooltipEnabled = enabled;
// 	}
// };

// tr.getTooltipLinebreakChar = function () {
// 	return tr.isCoralTooltipLoaded() ? "<br>" : "\n";
};
/** @public
* @param {Object=} obj
*/
RangeBar.prototype.setStyles = function(obj) {
	if(!this._styles) {
		if(!obj) {
			obj = {};
		}
		let defaults = RangeBar.defaults;
		this._styles = {};
		this._styles["trackColor"] = obj["trackColor"] || ElfUtil.themeColors["trackColor"];
		this._styles["fontColor"] = obj["color"] || obj["fontColor"] || defaults["fontColor"] || null;
		this._styles["trackHeight"] = obj["trackHeight"] || defaults["trackHeight"];
		this._styles["thumbColor"] = obj["thumbColor"] || defaults["thumbColor"];
		this._styles["naText"] = obj["naText"] || defaults["naText"];
	} else if(obj) {
		if(obj["trackColor"]) {
			this._styles["trackColor"] = obj["trackColor"];
		}
		if(obj["color"] || obj["fontColor"]) {
			this._styles["fontColor"] = obj["color"] || obj["fontColor"];
		}
		if(obj["trackHeight"]) {
			this._styles["trackHeight"] = obj["trackHeight"];
		}
		if(obj["thumbColor"]) {
			this._styles["thumbColor"] = obj["thumbColor"];
		}
		if(obj["naText"] != null) {
			this._styles["naText"] = obj["naText"] || "";
		}
	}

	if(!this.isValid()) {
		this._elem.style.backgroundColor = this._styles["trackColor"];
	}
	if(this._styles["fontColor"] != null) {
		this._elem.style.color = this._styles["fontColor"];
	}
	this._elem.style.minHeight = this._styles["trackHeight"];
	this._slider.style.backgroundColor = this._styles["thumbColor"];
};

/** @private
* @param {number=} lowVal
* @param {number=} lastVal
* @param {number=} highVal
*/
RangeBar.prototype._updateTooltip = function(lowVal, lastVal, highVal) {
// 	let lineBreak = "\n";
// 	let title = "";
// 	if((typeof lowVal === "number") && (lowVal >= 0)) {
// 		title += ("Low: " + lowVal);
// 	}
// 	if((typeof lastVal === "number") && (lastVal >= 0)) {
// 		if(title) {
// 			title += lineBreak;
// 		}
// 		title += ("Last: " + lastVal);
// 	}
// 	if((typeof highVal === "number") && (highVal >= 0)) {
// 		if(title) {
// 			title += lineBreak;
// 		}
// 		title += ("High: " + highVal);
// 	}

// 	this._elem.setAttribute("title", title);
};
/** @public
* @return {!number}
*/
RangeBar.prototype.getPercent = function() {
	return this._percent;
};
/** @public
* @param {number=} percent [0, 1] or NaN
*/
RangeBar.prototype.setPercent = function(percent) {
	percent = RangeBar._limitPercent(percent);
	if(this._percent !== percent) {
		if(this.isValid() || percent === percent) { // Prevent unnecessary update by checking NaN
			this._percent = percent;
			this._updateUI();
		}
	}
};
/** @private
* @function
* @param {number=} lowVal
* @param {number=} lastVal
* @param {number=} highVal
* @return {number}
*/
RangeBar._calculatePercent = function(lowVal, lastVal, highVal) {
	if((!lowVal && lowVal !== 0) ||
		(!lastVal && lastVal !== 0) ||
		(!highVal && highVal !== 0)) {
		return NaN;
	}
	if(highVal === lowVal) {
		return 0.5;
	}
	return (lastVal - lowVal) / (highVal - lowVal);
};
/** @private
* @function
* @param {number} percent
* @return {number} [0, 1] or NaN
*/
RangeBar._limitPercent = function(percent) {
	if(percent < 0) {
		return 0;
	}
	if(percent > 1) {
		return 1;
	}
	return percent;
};
/** @public
* @return {boolean}
*/
RangeBar.prototype.isValid = function() {
	return this._percent === this._percent;
};

/** @public
* @function
* @return {boolean}
*/
RangeBar.prototype.isInvalid = function() {
	return !this.isValid();
};

/** @private */
RangeBar.prototype._updateUI = function() {
	if(this.isValid()) {
		let percent = ((this._percent * 10000) | 0) / 10000; // rounding number
		this._slider.style.left = "calc(" +
			(percent * 100) +
			"% - " +
			Math.round(percent * 300) / 100 +
			"px)";

		this._elem.style.backgroundColor = this._styles["trackColor"];
		Dom.setContent(this._elem, this._slider);
	} else {
		this._elem.style.backgroundColor = "";
		Dom.setContent(this._elem, this._styles["naText"]);
	}
};

export default RangeBar;
export {RangeBar};
