/** RequestQueue limits number of execution within the specified period of time, while maintain all of the requests in a queue. <br>
* RequestQueue can be used to limit high frequency requests (preventing server overload). <br>
* RequestQueue can be used to merge multiple requests into one single request (reducing network calls). <br>
* RequestQueue can be used to split one large request into several smaller requests (preventing invalid request or long request time). <br>
* RequestQueue can be used to reduce duplicate requests (removing redundancy). <br>
*
* @constructor
* @param {Function} reqBuilder Function that creates a request
* @param {number=} reqLimit Number of requests allowed. Use zero reqLimit to lift up the limit
* @param {number=} timeLimit Time limit in milliseconds
* @example
* let rq = new RequestQueue(function(val, q) { console.log(val); return true; }, 1, 1000);
* for(let i = 0; i < 10; ++i) {
*  rq.add(i); // Number 1 - 10 will be logged to console every one second
* }
*/
let RequestQueue = function (reqBuilder, reqLimit, timeLimit) {
	this._onSuccess = this._onSuccess.bind(this);
	this._onRefreshCounter = this._onRefreshCounter.bind(this);
	this._nextInQueue = this._nextInQueue.bind(this);

	this._reqQueue = [];
	this.setRequestBuilder(reqBuilder);
	this.setFrequencyLimit(reqLimit, timeLimit);
};

/** @type {number}
* @private
*/
RequestQueue.prototype._concurrentLimit = NaN;
/** @type {Object}
* @private
*/
RequestQueue.prototype._freqLimit = null;

/** @type {number}
* @private
*/
RequestQueue.prototype._concurrentReqs = 0;
/** @type {number}
* @private
*/
RequestQueue.prototype._bufferId = 0;

/** @type {!Array}
* @private
*/
RequestQueue.prototype._reqQueue;
/** @type {Function}
* @private
*/
RequestQueue.prototype._reqBuilder = null;


/** Any request left in the queue and unfinished request will be discarded
* @public
*/
RequestQueue.prototype.reset = function() {
	this._reqQueue.length = 0;
	this._concurrentReqs = 0;

	if(this._freqLimit) {
		if(this._freqLimit.timerId) {
			clearTimeout(this._freqLimit.timerId);
			this._freqLimit.timerId = 0;
		}
		this._freqLimit.reqCount = 0;
	}
	if(this._bufferId) {
		clearTimeout(this._bufferId);
		this._bufferId = 0;
	}
};
/** @public */
RequestQueue.prototype.dispose = function() {
	this.reset();

	this._reqBuilder = null; // Free up user bound function
};

/** Limit number of requests that is being processed in the server at one time
* @public
* @param {number=} concurrentLimit Use zero to have no limit at all
*/
RequestQueue.prototype.setConcurrentLimit = function(concurrentLimit) {
	this._concurrentLimit = (concurrentLimit > 0) ? concurrentLimit : NaN;
};
/** Limit number of requests to make within the specified time limit in milliseconds
* @public
* @param {number=} reqLimit Number of requests allowed. Use zero reqLimit to lift up the limit
* @param {number=} timeLimit Time limit in milliseconds
*/
RequestQueue.prototype.setFrequencyLimit = function(reqLimit, timeLimit) {
	if(!reqLimit && !timeLimit) {
		if(this._freqLimit) {
			if(this._freqLimit.timerId) {
				clearTimeout(this._freqLimit.timerId);
			}
			this._freqLimit = null;
		}
		return;
	}
	if(!this._freqLimit) {
		this._freqLimit = {};
		this._freqLimit.reqLimit = 1;
		this._freqLimit.timeLimit = 0; // in milliseconds
		this._freqLimit.reqCount = 0;
		this._freqLimit.timerId = 0;
	}

	if(reqLimit > 0) {
		this._freqLimit.reqLimit = reqLimit;
	}
	if(timeLimit != null) {
		this._freqLimit.timeLimit = timeLimit;
	}
};
/** @public
* @param {Function} reqBuilder Function that creates a request
*/
RequestQueue.prototype.setRequestBuilder = function(reqBuilder) {
	this._reqBuilder = reqBuilder;
};

/** Adding the object to the back of the queue. The back most item will be executed last.
* @public
* @param {*} obj A parameter that will be passed to reqBuilder at the request time. Use native JavaScript object to pass multiple data
*/
RequestQueue.prototype.add = function(obj) {
	this._reqQueue.push(obj);
	if(!this._bufferId) { // Prevent making request repeatedly because items may be added in for loop
		this._bufferId = setTimeout(this._nextInQueue, 0);
	}
};
/** Adding the object to the front of the queue. The front most item will be executed first.
* @public
* @param {*} obj A parameter that will be passed to reqBuilder at the request time. Use native JavaScript object to pass multiple data
*/
RequestQueue.prototype.addFront = function(obj) {
	this._reqQueue.unshift(obj); // Always wait for the queue
	if(!this._bufferId) { // Prevent making request repeatedly because items may be added in for loop
		this._bufferId = setTimeout(this._nextInQueue, 0);
	}
};
/** @public
* @return {!Array}
*/
RequestQueue.prototype.getQueue = function() {
	return this._reqQueue;
};

/** Force making a request without waiting for the queue or limits
* @public
* @suppress {invalidCasts}
* @param {*} obj An object contains anything for constructing at the actual request time
* @return {Promise<Response>}
*/
RequestQueue.prototype.makeRequest = function(obj) {
	let req = this._reqBuilder ? this._reqBuilder(obj, this._reqQueue) : null;
	if(!req) { // 0, null, NaN, undefined, "" (empty string)
		return null; // No request is created at this time
	}

	let prom = null;
	if(typeof req === "string") {
		// WARNING: fetch() is not supported in IE
		prom = fetch(req); // WARNING: Response with Response object.
	} else if(req["then"]) { // thenable object
		prom = /** @type{Promise} */(req);
	} else {
		prom = Promise.resolve(req);
	}

	if(prom) {
		if(this._freqLimit) {
			if(!this._freqLimit.timerId) {
				this._freqLimit.timerId = setTimeout(this._onRefreshCounter, this._freqLimit.timeLimit);
			}
			++this._freqLimit.reqCount;
		}
		++this._concurrentReqs;
		prom["then"](this._onSuccess)["catch"](this._onSuccess);
	}
	return /** @type{Promise<Response>} */(prom);
};

/** @public
* @return {boolean}
*/
RequestQueue.prototype.isLimitReached = function() {
	if(this._freqLimit) {
		if(this._freqLimit.reqCount >= this._freqLimit.reqLimit) {
			return true;
		}
	}
	return (this._concurrentReqs >= this._concurrentLimit);
};

/** @private
* @param {*} res
* @return {Promise}
*/
RequestQueue.prototype._onSuccess = function(res) {
	--this._concurrentReqs;
	this._nextInQueue();
	return Promise.resolve(res);
};
/** @private */
RequestQueue.prototype._onRefreshCounter = function() {
	if(this._freqLimit) {
		this._freqLimit.reqCount = 0; // Any request waiting for response will not be counted toward reqLimit in the next time frame
		this._freqLimit.timerId = 0;
	}
	this._nextInQueue();
};
/** @private */
RequestQueue.prototype._nextInQueue = function() {
	this._bufferId = 0;
	while(this._reqQueue.length) {
		if(this.isLimitReached()) {
			break;
		}
		this.makeRequest(this._reqQueue.shift()); // Shift out the first item in queue
	}
};

export default RequestQueue;
export {RequestQueue};
