import { Dom } from "./Dom.js";
import { Ext } from "./Ext.js";
import { ElementWrapper } from "./ElementWrapper.js";

/** @private
* @function
* @param {Element} parentElem tr element
* @param {number} count Number of elements to be added
* @param {string} childTag Tag name of the cells
*/
let _addElements = function(parentElem, count, childTag) {
	for(let i = count; --i >= 0;) {
		parentElem.appendChild(document.createElement(childTag));
	}
};

/** SubTable represents a single tbody, thead, or tfoot element within table element
* @constructor
* @extends {ElementWrapper}
* @param {(Element|string)=} opt_elem
*/
let SubTable = function(opt_elem) {
	let elem;
	if(opt_elem) {
		if(typeof opt_elem == "string") { // tbody, thead, tfoot
			elem = document.createElement(opt_elem);
		} else {
			elem = /** @type{Element} */(opt_elem);
		}
	} else {
		elem = document.createElement("tbody");
	}
	this._elem = elem;

	this._render = SubTable._defaultCellRenderer;
};
Ext.inherits(SubTable, ElementWrapper);

/** @type {number}
* @private
*/
SubTable.prototype._colCount = 0;

/** @type {number|null}
* @private
*/
SubTable.prototype._defaultRowHeight = null;
/** @type {!Function}
* @private
*/
SubTable.prototype._render;


/** @private
* @return {string} tagName
*/
SubTable.prototype._getCellTagName = function() {
	return this._elem.tagName === "THEAD" ? "TH" : "TD";
};
/** @public
* @ignore
* @param {number=} opt_count
*/
SubTable.prototype.addColumns = function(opt_count) {
	if(opt_count == null) { opt_count = 1; }
	else if(opt_count <= 0) { return; }

	this._colCount += opt_count;

	let cellTag = this._getCellTagName();
	let rows = this._elem.children; // TODO: Must include all suspended rows
	for(let r = rows.length; --r >= 0;) {
		_addElements(rows[r], opt_count, cellTag);
	}
};
/** @public
* @ignore
* @param {number=} opt_count
*/
SubTable.prototype.removeColumns = function(opt_count) {
	if(opt_count == null || opt_count >= this._colCount) {
		this.removeAllColumns();
	}
	if(opt_count <= 0) { return; }

	this._colCount -= opt_count;
	let rows = this._elem.children; // TODO: Must include all suspended rows
	for(let r = rows.length; --r >= 0;) {
		Dom.removeChildren(rows[r], opt_count);
	}
};
/** @public
* @ignore
*/
SubTable.prototype.removeAllColumns = function() {
	if(this._colCount <= 0) { return; }
	this._colCount = 0;

	this.removeAllRows();
};
/** @public
* @ignore
* @param {number} val
*/
SubTable.prototype.setColumnCount = function(val) {
	if(val >= this._colCount) {
		this.addColumns(val - this._colCount);
	} else {
		this.removeColumns(this._colCount - val);
	}
};

/** @public
* @param {number=} at Index for insertion
* @return {Element} tr (HTMLTableRowElement) element
*/
SubTable.prototype.insertRow = function(at) {
	let row = this._elem.insertRow(at);
	_addElements(row, this._colCount, this._getCellTagName());
	return row;
};

/** @public
* @param {number=} count
* @return {!Array.<Element>} Array of tr (HTMLTableRowElement) elements
*/
SubTable.prototype.addRows = function(count) {
	let rows = null;
	if(count == null) {
		count = 1;
	}
	if(count > 0) {
		rows = new Array(count);
		for(let r = 0; r < count; ++r) {
			rows[r] = this.insertRow();
		}

		if(this._defaultRowHeight != null) {
			this._applyDefaultRowHeight(rows, this._defaultRowHeight + "px");
		}
	}
	return rows || [];
};
/** @public
* @param {number=} count
*/
SubTable.prototype.removeRows = function(count) {
	Dom.removeChildren(this._elem, count);
};
/** @public */
SubTable.prototype.removeAllRows = function() {
	Dom.removeChildren(this._elem);
};
/** @public
* @param {number} val
*/
SubTable.prototype.setRowCount = function(val) {
	let rowCount = this.getRowCount();
	if(val >= rowCount) {
		this.addRows(val - rowCount);
	} else {
		this.removeRows(rowCount - val);
	}
};
/** @public
* @param {(number|null)=} val
*/
SubTable.prototype.setDefaultRowHeight = function(val) {
	if(!val) { val = null; }
	if(val === this._defaultRowHeight) { return; }

	this._defaultRowHeight = val;
	let minHeight = (this._defaultRowHeight != null) ? this._defaultRowHeight + "px" : "";

	this._applyDefaultRowHeight(this._elem.children, minHeight);
};

/** @public
* @return {number}
*/
SubTable.prototype.getColumnCount = function() {
	return this._colCount;
};
/** @public
* @return {number}
*/
SubTable.prototype.getRowCount = function() {
	return this._elem.children.length;
};

/** @public
* @param {number} c
* @param {number} r
* @return {Element}
*/
SubTable.prototype.getCell = function(c, r) {
	let tr = this.getRow(r);
	if(tr) {
		return tr.cells[c] || null;
	}
	return null;
};
/** @public
* @param {number} c
* @return {Array.<Element>} Array of td (HTMLTableCellElement) elements
*/
SubTable.prototype.getCellsInColumn = function(c) {
	if(c < 0 || c >= this._colCount) {
		return null;
	}
	let rows = this._elem.children;
	let rowCount = rows.length;
	let ary = new Array(rowCount);
	for(let r = 0; r < rowCount; ++r) {
		ary[r] = rows[r].cells[c];
	}

	return ary;
};
/** @public
* @param {number} r
* @return {Array.<Element>} Array of td (HTMLTableCellElement) elements
*/
SubTable.prototype.getCellsInRow = function(r) {
	let tr = this.getRow(r);
	if(tr) {
		let cells = tr.cells;
		let len = cells.length;
		let ary = new Array(len);
		for(let i = 0; i < len; ++i) {
			ary[i] = cells[i];
		}
		return ary;
	}
	return null;
};
/** @public
* @return {!Array.<Element>} Array of td (HTMLTableCellElement) elements
*/
SubTable.prototype.getAllCells = function() {
	let rows = this._elem.children;
	let rowCount = rows.length;
	let colCount = this._colCount;
	let cellCount = 0;
	let ary = new Array(rowCount * colCount);
	for(let r = 0; r < rowCount; ++r) {
		let tr = rows[r];
		for(let c = 0; c < colCount; ++c) {
			ary[cellCount++] = tr.cells[c];
		}
	}
	return ary;
};
/** @public
* @return {!Array.<Element>} Array of tr (HTMLTableRowElement) elements
*/
SubTable.prototype.getAllRows = function() {
	let chdr = this._elem.children;
	let len = chdr ? chdr.length : 0;
	let ary = new Array(len);
	for(let i = 0; i < len; ++i) {
		ary[i] = chdr[i];
	}
	return ary;
};
/** @public
* @function
* @return {!Array.<Element>} Array of tr (HTMLTableRowElement) elements
*/
SubTable.prototype.getRows = SubTable.prototype.getAllRows;
/** @public
* @param {number} r
* @return {Element} Array of tr (HTMLTableRowElement) elements
*/
SubTable.prototype.getRow = function(r) {
	return this._elem.children[r] || null;
};

/** @public
* @return {!Array.<Array.<string>>}
*/
SubTable.prototype.getTextContents = function() {
	let rows = this._elem.children;
	let rowCount = rows.length;
	let rowContents = new Array(rowCount);
	for(let r = 0; r < rowCount; ++r) {
		let row = rows[r];
		let colContents = new Array(this._colCount);
		rowContents[r] = colContents;
		for(let c = 0; c < this._colCount; ++c) {
			colContents[c] = row.cells[c].textContent;
		}
	}
	return rowContents;
};
/** @public
* @param {number} c
* @param {number} r
* @return {string}
*/
SubTable.prototype.getCellTextContent = function(c, r) {
	let cell = this.getCell(c, r);
	if(cell) {
		return cell.textContent;
	}
	return "";
};
/** @public
* @param {number} c
* @return {string}
*/
SubTable.prototype.getColumnTextContent = function(c) { // New-line delimited
	let rows = this._elem.children;
	let rowCount = rows.length;
	if(c >= 0 && c < this._colCount && rowCount > 0) {
		let str = rows[0].cells[c].textContent;
		for(let r = 1; r < rowCount; ++r) {
			str += "\n" + rows[r].cells[c].textContent;
		}
		return str;
	}
	return "";
};
/** @public
* @param {number} r
* @return {string}
*/
SubTable.prototype.getRowTextContent = function(r) { // Tab delimited
	let tr = this.getRow(r);
	if(tr && this._colCount > 0) {
		let str = tr.cells[0].textContent;
		for(let c = 1; c < this._colCount; ++c) {
			str += "\t" + tr.cells[c].textContent;
		}
		return str;
	}
	return "";
};
/** @public
* @return {string}
*/
SubTable.prototype.getTableTextContent = function() { // Tab delimited
	let rowCount = this._elem.children.length;
	if(rowCount > 0) {
		let str = this.getRowTextContent(0);
		for(let r = 1; r < rowCount; ++r) {
			str += "\n" + this.getRowTextContent(r);
		}
		return str;
	}
	return "";
};
/** @public
* @return {string}
*/
SubTable.prototype.toString = function() {
	let str = this.getElement().outerHTML;
	str = str.replace(/>\s+</g, "><");
	str = str.replace(/><tr/g, ">\n\t<tr");
	str = str.replace(/><td/g, ">\n\t\t<td");
	str = str.replace(/><th/g, ">\n\t\t<th"); // TH can occur in thead
	str = str.replace(/><\/tr/g, ">\n\t</tr");
	let tagName = this.getElement().tagName.toLowerCase();
	str = str.replace("</" + tagName, "\n</" + tagName);
	return str;
};

/** @public
* @param {Event} e
* @return {number}
*/
SubTable.prototype.getColumnIndex = function(e) {
	if(!e) {
		return -1;
	}
	let len, i;
	let tgt = /** @type{Node} */(e.target); // TODO: Support Shadow Root
	if(this._elem.contains(tgt)) {
		let tdElem = Dom.closestTagName(tgt, "TD");
		if(!tdElem) {
			tdElem = Dom.closestTagName(tgt, "TH");
		}
		if(tdElem) {
			let chdr = tdElem.parentElement.children;
			len = chdr.length;
			for(i = 0; i < len; ++i) {
				if(tdElem === chdr[i]) {
					return i;
				}
			}
		}
	} else { // In case of the target is not a child of this element
		let rows = this._elem.children;
		let cells = (rows[0]) ? rows[0].cells : null;
		if(cells) {
			let pos = Dom.getRelativePosition(e, this._elem);
			let x = pos["x"];
			if(x >= 0) {
				len = cells.length;
				for(i = 0; i < len; ++i) {
					x -= cells[i].offsetWidth;
					if(x < 0) { // Not include the right border
						return i;
					}
				}
			}
		}
	}
	return -1;
};
/** @public
* @param {Event} e
* @return {number}
*/
SubTable.prototype.getRowIndex = function(e) {
	if(!e) {
		return -1;
	}
	let rows = this._elem.children; // This return HTML collection
	let rowCount = rows ? rows.length : 0;
	if(!rowCount) {
		return -1;
	}
	let i;
	let tgt = /** @type{Node} */(e.target); // TODO: Support Shadow Root
	if(this._elem.contains(tgt)) {
		let trElem = Dom.closestTagName(tgt, "TR");
		if(trElem) {
			for(i = 0; i < rowCount; ++i) {
				let row = rows[i];
				if(row === trElem) {
					return i;
				}
			}
		}
	}

	// In case of the target is not a child of this element
	let pos = Dom.getRelativePosition(e, this._elem);
	let y = pos["y"];
	if(y >= 0) {
		for(i = 0; i < rowCount; ++i) {
			y -= rows[i].offsetHeight;
			if(y < 0) { // Not include the right border
				return i;
			}
		}
	}
	return -1;
};
/** @public
* @param {number} colIndex
* @param {number=} rowIndex
* @return {number}
*/
SubTable.prototype.getColumnLeft = function(colIndex, rowIndex) {
	let cell = this.getCell(colIndex, rowIndex ? rowIndex : 0);
	if(cell) {
		return cell.offsetLeft; // WARNING: Hidden cells or cell spanning can cause unexpected results
	}
	return NaN;
};
/** @public
* @param {number|Element} rowRef
* @return {number}
*/
SubTable.prototype.getRowTop = function(rowRef) {
	let row = null;
	if(typeof rowRef === "number") {
		row = this.getRow(rowRef);
	} else {
		row = Dom.closestTagName(rowRef, "TR");
	}
	return (row) ? row.offsetTop : NaN;
};
/** @public
* @param {Element} cell
* @param {Object=} ret_obj
* @return {!Object} Object with x and y property
*/
SubTable.prototype.getCellPosition = function(cell, ret_obj) {
	if(!ret_obj) { ret_obj = {}; }
	ret_obj["x"] = cell.offsetLeft;
	let pn = cell.parentNode; // Expecting tr element
	ret_obj["y"] = (pn) ? pn.offsetTop : NaN;
	return ret_obj;
};

/** @public
* @param {Function} func
*/
SubTable.prototype.setCellRenderer = function(func) {
	this._render = func || SubTable._defaultCellRenderer;
};
/** @public
* @return {Function}
*/
SubTable.prototype.getCellRenderer = function() {
	return this._render || SubTable._defaultCellRenderer;
};
/** @public
* @override
* @param {Element} elem
* @param {string=} opt_elementType
*/
SubTable.prototype.cloak = function(elem, opt_elementType) {
	if(elem) {
		this._elem = elem;
		let rows = elem.getElementsByTagName("TR");
		this._colCount = rows[0] ? rows[0].children.length : 0;
	}
};

/** @public
* @param {number} c1 Starting column index
* @param {number} c2 Destination column index
* @param {number} r1 Starting row index
* @param {number} r2 Destination row index
* @return {Element} Top left cell element
*/
SubTable.prototype.spanBlock = function (c1, c2, r1, r2) { // WARNING: It's c c r r
	let cell = this.getCell(c1, r1);
	if(!cell) {
		return null;
	}
	if(c2 < c1) {
		c2 = c1;
	}
	if(r2 < r1) {
		r2 = r1;
	}
	let curColSpan = +cell.getAttribute("colspan");
	let curRowSpan = +cell.getAttribute("rowspan");
	let c3 = (curColSpan) ? c1 + curColSpan - 1 : c1;
	let r3 = (curRowSpan) ? r1 + curRowSpan - 1 : r1;

	// TODO: Optimize below logics
	let c, r;
	for(c = c3; c >= c1; --c) {
		for(r = r3; r >= r1; --r) {
			cell = this.getCell(c, r);
			if(cell) {
				cell.style.display = "";
			}
		}
	}

	for(c = c2; c >= c1; --c) {
		for(r = r2; r >= r1; --r) {
			cell = this.getCell(c, r);
			if(cell) {
				cell.style.display = "none";
			}
		}
	}

	if(cell) { // The last cell from the loop is the top left cell
		let colSpan = (c2 - c1 + 1);
		let rowSpan = (r2 - r1 + 1);
		if(colSpan > 1) {
			cell.setAttribute("colspan", colSpan);
		} else {
			cell.removeAttribute("colspan");
		}
		if(colSpan > 1) {
			cell.setAttribute("rowspan", rowSpan);
		} else {
			cell.removeAttribute("rowspan");
		}
		cell.style.display = "";
	}
	return cell;
};

/** Parse table content into 2d array for table input
* @public
* @function
* @param {Element} tbl
* @return {Array.<Array.<string>>}
*/
SubTable.parseTableContent = function(tbl) {
	if(!tbl) {
		return null;
	}

	let tbody = (tbl.tagName == "TBODY") ? tbl : tbl.getElementsByTagName("TBODY")[0];
	if(!tbody) {
		if(tbl.tagName == "TABLE") { // Table element can directly host tr element without tbody element
			tbody = tbl;
		}
	}
	if(!tbody) {
		return null;
	}

	let trs = tbody.getElementsByTagName("TR");
	let rowCount = trs.length;
	let rows = new Array(rowCount);
	for(let i = 0; i < rowCount; ++i) {
		let tds = trs[i].getElementsByTagName("TD"); // TODO: Handle TH cell
		let colCount = tds.length;
		let columns = new Array(colCount);

		rows[i] = columns;
		for(let j = 0; j < colCount; ++j) {
			columns[j] = tds[j].textContent;
		}
	}
	return rows;
};

/** @private
* @param {!(NodeList<!Element>|Array.<Element>)} rows
* @param {string} minHeight
*/
SubTable.prototype._applyDefaultRowHeight = function(rows, minHeight) {
	for(let r = rows.length; --r >= 0;) {
		rows[r].style.height = minHeight;
	}
};

/** @private
* @function
* @param {Element} tdCell
* @param {*} val
* @param {number=} c
* @param {number=} r
*/
SubTable._defaultCellRenderer = function(tdCell, val, c, r) {
	if(val != null) {
		if(val["nodeType"]) {
			tdCell.appendChild(/** @type{Node} */(val));
		} else {
			tdCell.textContent = val;
		}
	} else {
		tdCell.textContent = "";
	}
};

export default SubTable;
export { SubTable };
