import {Dom} from './Dom.js';

/** @namespace */
let TextHighlighter = {};
/** /[^\w-_ .]+/g
* @type {RegExp}
* @public
*/
TextHighlighter.invalidChars = /[^\w-_ .]+/g;
/** /\s+/
* @type {RegExp}
* @public
*/
TextHighlighter.spaces = /\s+/;
/** Default is &lt;mark> tag
* @type {string}
* @public
*/
TextHighlighter.defaultTag = "mark";

/** Highlight all descendants from the given element. Only leaf node, that has no children, will be highlighted.
* @function
* @param {Element} elem
* @param {string|RegExp} keyword If omit, the content, that has been highlighted, will be restored to its original form
* @param {string=} tagName Default is &lt;mark> tag
* @return {Element} the same given elem
* @example
* let parent = Dom.div();
* let child1 = Dom.text("Lorem ipsum dolor sit amet");
* let child2 = Dom.text("Quisque volutpat, neque et pretium accumsan");
* Dom.appendChild(parent, child1, child2);
* TextHighlighter.highlightAll(parent, "e", "mark"); // only child1 and child2 are modified
*/
TextHighlighter.highlightAll = function(elem, keyword, tagName) {
	let rule = TextHighlighter.buildRule(keyword);
	if(!tagName) {
		tagName = TextHighlighter.defaultTag;
	}

	let nodes = [elem];
	for(let i = 0; i < nodes.length; ++i) {
		let curNode = nodes[i];
		let chdr = curNode.children;
		let nonLeafNode = !curNode._highlightData && chdr && chdr.length;
		if(nonLeafNode) {
			let len = chdr.length;
			for(let j = 0; j < len; ++j) {
				nodes.push(chdr[j]);
			}
		} else { // leaf node
			TextHighlighter.highlightContent(curNode, rule, tagName);
		}
	}
	return elem;
};

/** Highlight text content of the given element. This method does not preserve internal structure of the element. Thus, this should be used on a leaf node where content is just text without layout.
* @function
* @param {Element} elem
* @param {string|RegExp} keyword If omit, the content, that has been highlighted, will be restored to its original form
* @param {string=} tagName Default is &lt;mark> tag
* @return {Element} Element with children of &lt;span>&lt;/span> and &lt;mark>&lt;/mark> tags
* @example
* let el = Dom.text("Lorem ipsum dolor sit amet");
* let highlightedEl = TextHighlighter.highlightContent(el, "e", "mark");
* highlightedEl === el; // true
*/
TextHighlighter.highlightContent = function(elem, keyword, tagName) {
	let reset = true;
	if(keyword) {
		let text = elem.textContent;
		if(text) {
			let highlightData = elem._highlightData;
			let splitInfo = TextHighlighter._split(text, keyword, null, highlightData);
			if(splitInfo) {
				let i, len;
				if(!highlightData) {
					elem._highlightData = splitInfo;

					// Save original child nodes
					let cns = elem.childNodes;
					len = cns.length;
					let originalNodes = new Array(len);
					for(i = 0; i < len; ++i) {
						originalNodes[i] = cns[i];
					}
					splitInfo["originalNodes"] = originalNodes;
				}
				Dom.removeChildren(elem);

				if(!tagName) {
					tagName = TextHighlighter.defaultTag;
				}

				let chunks = splitInfo["chunks"];
				len = chunks.length;
				for(i = 0; i < len; ++i) {
					let chunk = chunks[i];
					if(chunk) {
						let tag = Dom.create((i & 1) ? tagName : "span");
						tag.textContent = chunk;
						elem.appendChild(tag);
					}
				}
				reset = false;
			}
		}
	}
	if(reset) {
		TextHighlighter.restore(elem);
	}
	return elem;
};

/** Clear all highlights and reset back to its original state.
* @function
* @param {Element} elem
* @return {boolean} Returns true if there is any change
*/
TextHighlighter.restore = function(elem) {
	if(elem && elem._highlightData) {
		Dom.removeChildren(elem);
		Dom.appendChild(elem, elem._highlightData["originalNodes"]);
		elem._highlightData = null;
		return true;
	}
	return false;
};

/** Get matching score. The score is calculated by the following criteria: <br>
* - Number of the matched keywords <br>
* - Length of the matched keywords <br>
* - Position of the matched keywords <br>
* - Length of the original text <br>
*
* @function
* @param {Element} elem
* @return {number}
*/
TextHighlighter.getScore = function(elem) {
	if(elem) {
		let obj = /** @type{Object} */(elem._highlightData);
		if(obj) {
			let score = obj["score"];
			if(score == null) {
				let positions = /** @type{Array.<number>} */(obj["positions"]);
				let matches = /** @type{Array.<string>} */(obj["matches"]);
				let matchCount = matches.length;

				score = matchCount * 1000; // Number of the matched keywords
				for(let i = 0; i < matchCount; ++i) {
					let curTxt = matches[i];
					let pos = positions[i];
					score += curTxt.length * 100; // Length of the matched keywords
					score += pos ? (400 / pos) | 0 : 500;
				}
				if(matchCount) {
					let origLen = obj["original"] ? obj["original"].length : 0;
					score += origLen ? (1000 / origLen) | 0 : 0;
				}
				obj["score"] = score; // Cache for later usage
			}
			return /** @type{number} */(score);
		}
	}
	return 0;
};

/** Split the given haystack while preserving the matched texts (i.e. the results will contain the given needle). The matched texts will always be in the odd index (e.g. 1, 3, 5, ...) of the result array.
* @function
* @param {string} haystack
* @param {string|RegExp} needle
* @param {(boolean|number)=} caseSensitive
* @return {!Array.<string>}
* @example
* TextHighlighter.split("aabbccbb ddbbee", "bb"); // ["aa", "bb", "cc", "bb", " dd", "bb", "ee"]
* TextHighlighter.split("entire_word", "entire_word"); // ["", "entire_word"]
* TextHighlighter.split("not_found", "aaa"); // ["not_found"]
*/
TextHighlighter.split = function (haystack, needle, caseSensitive) {
	let obj = TextHighlighter._split(haystack, needle, caseSensitive);
	if(obj) {
		return /** @type{!Array.<string>} */(obj["chunks"]);
	}
	return [haystack];
};
/** @private
* @function
* @param {string} haystack
* @param {string|RegExp} needle
* @param {(boolean|number)=} caseSensitive
* @param {Object=} out_obj
* @return {Object}
*/
TextHighlighter._split = function (haystack, needle, caseSensitive, out_obj) {
	if(!haystack || !needle) {
		return null;
	}
	let regExp = TextHighlighter.buildRule(needle, caseSensitive);

	let matches = haystack.match(regExp);
	if(!matches) {
		return null;
	}

	if(!out_obj) {
		out_obj = {};
	}

	let matchCount = matches.length;
	let positions = new Array(matchCount);
	let start = 0;
	let curIdx = 0;
	let chunks = [];
	do {
		let curMatch = matches[curIdx];
		let matchLen = curMatch.length;
		let at = positions[curIdx] = haystack.indexOf(curMatch, start);
		chunks.push(
			haystack.slice(start, at),
			haystack.slice(at, at + matchLen)
		);
		start = at + matchLen;
	} while(++curIdx < matchCount);

	let end = haystack.length;
	if(start < end) {
		chunks.push(haystack.slice(start, end));
	}

	out_obj["chunks"] = chunks;
	out_obj["matches"] = matches;
	out_obj["positions"] = positions;
	out_obj["original"] = haystack;
	out_obj["score"] = null;
	return out_obj;
};

/** Build RegExp for highlighting. This is useful to speed up the process, as the same rule will be used on multiple elements.
* @function
* @param {string|RegExp} txt
* @param {(boolean|number)=} caseSensitive If true, the rule is produced with case sensitive characters
* @return {RegExp}
* @example
* let regExp = TextHighlighter.buildRule("A + B"); // Equivalent to new RegExp(/A|B/g);
*/
TextHighlighter.buildRule = function(txt, caseSensitive) {
	if(typeof txt != "string") {
		return /** @type{RegExp} */(txt) || null;
	}

	if(txt) {
		txt = txt.replace(TextHighlighter.invalidChars, " ");
		txt = txt.trim();
	}
	if(!txt) {
		return null;
	}

	let option = caseSensitive ? "g" : "gi";
	let keywords = txt.split(TextHighlighter.spaces);
	let rule = (keywords.length > 1) ? keywords.join("|") : txt;

	return new RegExp(rule, option);
};

export default TextHighlighter;
export {TextHighlighter};
