/** Provide a way to execute a function with context object. The timer can be started, extended, stopped at any time. Use {@link Delay} for promisified version of Timer class
* @constructor
* @param {number} duration Delay before executing the given function
* @param {Function=} func Function to be executed
* @param {*=} ctx Context object that will be supplied to the function as the first parameter
* @example
* let timer = new Timer(2000, (ctx)=>{ console.log(ctx)}, "AAA");
* timer.start(); // AAA will be logged after 2 seconds
* timer.start(); // Timer is reset. AAA will be logged only once
*/
let Timer = function(duration, func, ctx) {
	this._duration = (typeof duration == "number") ? duration : 0;
	this._onTimeout = this._onTimeout.bind(this);
	this._method = func || null;
	this._ctx = ctx != null ? ctx : null;
};
/** If calling start before the timer end, the timer will be reset, and effectively extended the duration.
* @public
*/
Timer.prototype.start = function() {
	if(this._timerId) {
		clearTimeout(this._timerId); // Extend duration
	}
	this._timerId = setTimeout(this._onTimeout, this._duration);
};
/** @public */
Timer.prototype.stop = function() {
	if(this._timerId) {
		clearTimeout(this._timerId);
		this._timerId = 0;
	}
};
/** @public
* @return {boolean}
*/
Timer.prototype.isActive = function() {
	return this._timerId ? true : false;
};

/** @private */
Timer.prototype._onTimeout = function() {
	this._timerId = 0;
	if(this._method) {
		this._method(this._ctx || this);
	}
};

export default Timer;
export {Timer};
