import {Ext} from "./Ext.js";
import {EventDispatcher} from "./EventDispatcher.js";

/** @typedef {Object} TouchProxy~Options
* @description TouchProxy options
* @property {boolean=} useTouchMove=false If enabled, touch events are added to the element
* @property {boolean=} preventDefault=false If enabled, event default behaviors are prevented
*/

/** @event TouchProxy#touchStart
* @description Alias to native touchstart
* @type {Event}
*/
/** @event TouchProxy#touchEnd
* @description Alias to native touchend and touchcancel
* @type {Event}
*/
/** @event TouchProxy#touchMove
* @description Alias to native touchmove
* @type {Event}
*/

/** @constructor
* @extends {EventDispatcher}
* @param {Element=} elem
* @param {(boolean|Object)=} options Boolean is treated as useTouchMove
*/
let TouchProxy = function (elem, options) {
	this._onTouchStart = this._dispatch.bind(this, "touchStart");
	this._onTouchEnd = this._dispatch.bind(this, "touchEnd");
	this._onTouchMove = this._dispatch.bind(this, "touchMove");

	if(options) {
		if(typeof options == "boolean") {
			this._touchMove = true;
		} else {
			if(options["useTouchMove"]) {
				this._touchMove = true;
			}
			if(options["preventDefault"]) {
				this._defaultPrevented = true;
			}
		}
	}
	if(!this._defaultPrevented) {
		this._evtOptions = TouchProxy.passive;
	}
	this.setElement(elem);
};
Ext.inherits(TouchProxy, EventDispatcher);

/** @type {Element}
* @private
*/
TouchProxy.prototype._elem;
/** @type {!Function}
* @private
*/
TouchProxy.prototype._onTouchStart;
/** @type {!Function}
* @private
*/
TouchProxy.prototype._onTouchEnd;
/** @type {!Function}
* @private
*/
TouchProxy.prototype._onTouchMove;
/** @type {boolean}
* @private
*/
TouchProxy.prototype._touchMove = false;
/** @type {boolean}
* @private
*/
TouchProxy.prototype._defaultPrevented = false;
/** @type {!EventListenerOptions}
* @private
*/
TouchProxy.prototype._evtOptions;
/** @type {!EventListenerOptions}
* @public
*/
TouchProxy.passive = {"passive": true};

/** @public
* @return {Element}
*/
TouchProxy.prototype.getElement = function() {
	return this._elem;
};
/** @public
* @param {Element=} elem
*/
TouchProxy.prototype.setElement = function(elem) {
	this.detachTouchEvents();
	this._elem = elem || null;
	this.attachTouchEvents(this._touchMove);
};
/** @public
* @function
* @param {Element} elem
*/
TouchProxy.prototype.applyTrait = TouchProxy.prototype.setElement;
/** @public */
TouchProxy.prototype.dispose = function() {
	this.setElement(null);
};

/** @public
* @param {Function=} touchStartFunc
* @param {Function=} touchEndFunc
* @param {Function=} touchMoveFunc
*/
TouchProxy.prototype.addTouchListeners = function(touchStartFunc, touchEndFunc, touchMoveFunc) {
	if(touchStartFunc) {
		this.addEventListener("touchStart", touchStartFunc);
	}
	if(touchEndFunc) {
		this.addEventListener("touchEnd", touchEndFunc);
	}
	if(touchMoveFunc) {
		this.addEventListener("touchMove", touchMoveFunc);
	}
};
/** @public
* @function
*/
TouchProxy.prototype.removeTouchListeners = TouchProxy.prototype.unlistenAll;

/** @public */
TouchProxy.prototype.attachTouchMoveEvent = function() {
	let elem = this._elem;
	if(elem) {
		if(this._defaultPrevented) {
			elem.addEventListener("touchmove", EventDispatcher.preventDefault);
		}
		elem.addEventListener("touchmove",  this._onTouchMove, this._evtOptions);
	}
};
/** @public */
TouchProxy.prototype.detachTouchMoveEvent = function() {
	let elem = this._elem;
	if(elem) {
		if(this._defaultPrevented) {
			elem.removeEventListener("touchmove", EventDispatcher.preventDefault);
		}
		elem.removeEventListener("touchmove",  this._onTouchMove, this._evtOptions);
	}
};
/** @public
* @param {boolean=} touchMove
*/
TouchProxy.prototype.attachTouchEvents = function(touchMove) {
	let elem = this._elem;
	if(elem) {
		elem.addEventListener("touchstart", this._onTouchStart, this._evtOptions);
		elem.addEventListener("touchend", this._onTouchEnd, this._evtOptions);
		elem.addEventListener("touchcancel",  this._onTouchEnd, this._evtOptions);
		if(touchMove) {
			this.attachTouchMoveEvent();
		}
	}
};
/** @public */
TouchProxy.prototype.detachTouchEvents = function() {
	let elem = this._elem;
	if(elem) {
		elem.removeEventListener("touchstart", this._onTouchStart, this._evtOptions);
		elem.removeEventListener("touchend", this._onTouchEnd, this._evtOptions);
		elem.removeEventListener("touchcancel",  this._onTouchEnd, this._evtOptions);
		this.detachTouchMoveEvent();
	}
};

/** @public
* @function
* @param {Event} e
* @return {number} 0 is Left mouse button
*/
TouchProxy.getButtonType = function(e) {
	if(e.changedTouches) {
		return 0;
	} else {
		return e.button || 0; // Support from IE9+
	}
};
/** @public
* @function
* @param {Event|Object} e
* @param {Object=} obj
* @return {!Object}
* @see {@link https://developer.mozilla.org/en-US/docs/Web/API/MouseEvent}
* @see {@link https://developer.mozilla.org/en-US/docs/Web/API/TouchEvent}
*/
TouchProxy.getPosition = function(e, obj) {
	let touches = e.changedTouches;
	let mPos = (touches) ? touches[touches.length - 1] : e;

	if(!obj) {
		obj = {};
	}
	obj["x"] = obj["pageX"] = mPos.pageX; // window.pageXOffset
	obj["y"] = obj["pageY"] = mPos.pageY; // window.pageYOffset
	obj["clientX"] = mPos.clientX; // Allow relative position calculation and distance comparison
	obj["clientY"] = mPos.clientY;
	obj["timeStamp"] = e.timeStamp;
	return obj;
};


export {TouchProxy};
export default TouchProxy;
