import {Ext} from "../Ext.js";
import {Util} from "../Util.js";
import {EventDispatcher} from "../EventDispatcher.js";

/** @constructor
* @extends {EventDispatcher}
*/
let ADCSubscription = function() {
	this._adcFieldMap = {};
	this._queue = {};
	this._onADCUpdate = this._onADCUpdate.bind(this);
	this._refreshData = this._refreshData.bind(this);
	this._processQueue = this._processQueue.bind(this);
};
Ext.inherits(ADCSubscription, EventDispatcher);


/** @type {Object}
* @private
*/
ADCSubscription.prototype._adcFieldMap = null;
/** @type {Object}
* @private
*/
ADCSubscription.prototype._queue = null;
/** @type {string}
* @private
*/
ADCSubscription.prototype._state = "stop";
/** @type {string}
* @private
*/
ADCSubscription.prototype._userId = "GUEST";
/** @type {string}
* @private
*/
ADCSubscription.prototype._productId = "ANONYMOUS";
/** @type {string}
* @private
*/
ADCSubscription.prototype._adcUrl = "/datacloud/rest/select";
/** @type {number|null}
* @private
*/
ADCSubscription.prototype._intervalId = null;
/** @type {number}
* @private
*/
ADCSubscription.prototype._intervalTime = 3600000; // 1 Hour
/** @type {number|null}
* @private
*/
ADCSubscription.prototype._timerId = null;
/** @type {string}
* @private
*/
ADCSubscription.prototype._lang = "";
/** @static
* @constant {string}
*/
ADCSubscription.PENDING_DATA = "_PENDING♪_";


//#region public function of ADC service
/** Set the settings of ADC
* @public
* @param {string} userId UUID
* @param {string} productId Application name
* @param {string} url Url of ADC server
* @param {string=} opt_lang Language of Eikon UI
*/
ADCSubscription.prototype.setDataCloudSettings = function (userId, productId, url, opt_lang) {
	if(userId != null) {
		this._userId = userId;
	}
	if(productId != null) {
		this._productId = productId;
	}
	if(url != null) {
		this._adcUrl = url;
	}
	if(opt_lang) {
		this._lang = opt_lang;
	}
};
/** Set the interval time for retrieving data
* @public
* @param {number} time Millisecond
*/
ADCSubscription.prototype.setIntervalTime = function(time) {
	time = +time;
	if(!time) { // NaN or 0
		return;
	}
	this._intervalTime = time;
	this._stop();
	this._start();
};
/** Subscribe symbol to the spcified field. This method will start the interval for retrieving data automatically.
* @public
* @param {string} symbol RIC
* @param {string} field ADC field name
* @return {*} Return the value if there is existing. Return undefined when requesting the invalid field.
*/
ADCSubscription.prototype.addSymbolToField = function (symbol, field) {
	if(field.indexOf("undefined") >= 0) { // Invalid field name
		return;
	}

	let fieldMap = this._adcFieldMap[field];
	if(!fieldMap) {
		fieldMap = {};
		this._adcFieldMap[field] = fieldMap;
	}

	let value = fieldMap[symbol];
	if(value != null) {
		return value; // eslint-disable-line
	}

	fieldMap[symbol] = ADCSubscription.PENDING_DATA; // First value to cache
	this._addQueue(field);
	return ADCSubscription.PENDING_DATA; // eslint-disable-line
};
//#endregion public function of ADC service


//#region public function of event
/** Add the function handler into the specific event. This method will start interval automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
ADCSubscription.prototype.addEventListener = function(type, handler) {
	ADCSubscription.base(this, "addEventListener", type, handler);
	this._start();
};
/** Remove the function handler from the specific event. If there is not any event handlers, the service will stop automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
ADCSubscription.prototype.removeEventListener = function(type, handler) {
	ADCSubscription.base(this, "removeEventListener", type, handler);
	if(!this.hasListener()) {
		this._stop();
		this._adcFieldMap = {};
		this._queue = {};
	}
};
/** Remove all function handlers of all events and stop the interval of retrieving data of service.
* @override
*/
ADCSubscription.prototype.removeAllEventListeners = function() {
	ADCSubscription.base(this, "removeAllEventListeners");
	this._stop();
	this._adcFieldMap = {};
	this._queue = {};
};
//#endregion public function of event


//#region private function of ADC service
/** Start the interval of retrieving data
* @private
*/
ADCSubscription.prototype._start = function() {
	if(this._state === "start") {
		return;
	}
	this._intervalId = window.setInterval(this._refreshData, this._intervalTime);
	this._state = "start";
};
/** Stop the interval of retrieving data
* @private
*/
ADCSubscription.prototype._stop = function() {
	if(this._timerId !== null) {
		window.clearTimeout(this._timerId);
	}
	if(this._intervalId !== null) {
		window.clearInterval(this._intervalId);
	}
	this._timerId = null;
	this._intervalId = null;
	this._state = "stop";
};
/** Refresh all the ADC data that had been subscribed to the service.
* @private
*/
ADCSubscription.prototype._refreshData = function() {
	this._addQueue(Object.keys(/** @type{!Object} */(this._adcFieldMap)));
};
/** Manage the queue to retrieve the ADC data. Prevent to call the same ADC field multiple time.
* @private
* @param {string|Array.<string>} fields
*/
ADCSubscription.prototype._addQueue = function(fields) {
	if(this._timerId !== null) {
		window.clearTimeout(this._timerId);
	}

	if(!Array.isArray(fields)) {
		fields = [fields];
	}
	if(!this._queue) {
		this._queue = {};
	}
	for(let i = fields.length; --i >= 0;) {
		this._queue[fields[i]] = true;
	}
	this._timerId = window.setTimeout(this._processQueue, 0);
};
/** Process the data queue to retrieve ADC data.
* @private
*/
ADCSubscription.prototype._processQueue = function() {
	let fields = Object.keys(/** @type{!Object} */(this._queue));
	for(let j = fields.length; --j >= 0;) {
		this._requestDataCloud(fields[j]);
	}
	this._queue = {}; // Clear queue
	this._timerId = null; // Clear timeout id
};
/** Publish the ADC request to the ADC server.
* @private
* @param {string} field ADC field name
* @return {Promise}
*/
ADCSubscription.prototype._requestDataCloud = function(field) {
	let RICs = Object.keys(this._adcFieldMap[field]);
	let payload = "formula=" + field + "&identifiers=" + RICs;
	if(this._productId) {
		payload += '&productid=' + this._productId;
	}
	if(this._userId) {
		payload += '&uuid=' + this._userId;
	}
	if(this._lang) {
		payload += '&lang=' + this._lang;
	}

	return Util.requestByUrlEncoded(this._adcUrl, payload)
		.then(function(json) {
			json["requestField"] = field; // Attached requested field
			return json;
		})
		.then(this._onADCUpdate)
		.catch(function(err) {
			console.log("ADC: ", err);
		});
};
/** Validate type of value that are number, string, boolean or not.
* @private
* @param {*} value
* @return {boolean}
*/
ADCSubscription.prototype._isValidValueType = function(value) {
	let t = typeof value;
	return (value === null || t == "string" || t == "number" || t == "boolean");
};
/** Function for processing the ADC response and publish the event to the listener.
* @private
* @param {Object} json JSON object of ADC response
*/
ADCSubscription.prototype._onADCUpdate = function(json) {
	if(json["status"] && json["status"].toLowerCase() !== "ok") {
		console.warn("Response status is not ready");
		return;
	}

	let rows = json["rows"];
	let reqField = json["requestField"];
	let header = rows[0];
	let ricIndex = -1, fieldIndex = -1;

	let type, field;
	for(let idx in header) {
		type = header[idx]["i"];
		field = header[idx]["r"];
		if(type === "instrument") {
			ricIndex = idx;
		}
		if (type === "value" && reqField === field) {
			fieldIndex = idx;
		}
	}

	let outputHash = {};
	let rowsLen = rows.length;
	let symbol, value, rowData;
	for(let i = 1; i < rowsLen; i++) {
		symbol = rows[i][ricIndex];
		value = rows[i][fieldIndex];

		rowData = outputHash[symbol];
		if (!rowData) {
			rowData = {};
			outputHash[symbol] = rowData;
		}

		if(rowData[reqField] == null && this._isValidValueType(value)) {
			rowData[reqField] = value;
		}
	}

	// Publish data
	for(symbol in outputHash) {
		rowData = outputHash[symbol];
		if(rowData[reqField]) { // Have the data to be updated
			this._adcFieldMap[reqField][symbol] = rowData[reqField]; // Cache the data in service
			this._dispatch("dataChanged", { "ric": symbol, "values": rowData, "sender": this });
		}
	}
};

export {ADCSubscription};
