import {Ext} from "../Ext.js";
import {Util} from "../Util.js";
import {EventDispatcher} from "../EventDispatcher.js";

/** @constructor
* @extends {EventDispatcher}
*/
let AdFinSubscription = function() {
	this._cache = {};
	this._queue = [];
	this.onAdFinUpdate = this.onAdFinUpdate.bind(this);
	this._clearCache = this._clearCache.bind(this);
	this._processQueue = this._processQueue.bind(this);
};
Ext.inherits(AdFinSubscription, EventDispatcher);

/** @typedef {{Function:string, Parameters:Array.<Array.<string|number|null>>} } */
AdFinSubscription.Request;

/** call signature -> value
* @type {Object}
* @private
*/
AdFinSubscription.prototype._cache = null;
/** @type {Array.<AdFinSubscription.Request>}
* @private
*/
AdFinSubscription.prototype._queue = null;
/** @type {string}
* @private
*/
AdFinSubscription.prototype._state = "stop";
/** @type {string}
* @private
*/
AdFinSubscription.prototype._userId = "GUEST";
/** @type {string}
* @private
*/
AdFinSubscription.prototype._productId = "SPREADSHEET";
/** @type {string}
* @private
*/
AdFinSubscription.prototype._adfinUrl = ((window && window.location && window.location.origin)
	? window.location.origin
	: "" ) + "/tracsweb/plugin.svc/adf/CompositeMulti/json2";
/** @type {number|null}
* @private
*/
AdFinSubscription.prototype._intervalId = null;
/** @type {number}
* @private
*/
AdFinSubscription.prototype._intervalTime = 3600000; // 1 Hour
/** @type {number|null}
* @private
*/
AdFinSubscription.prototype._timerId = null;
/** @type {number}
* @private
*/
AdFinSubscription._chunckSize = 50;
/** Conflator delay (in ms)
* @type {number}
* @public
*/
AdFinSubscription.ConflatorDelay = 1000;

//#region public function of Adfin service
/** Set the settings of Adfin
* @public
* @param {string} userId UUID
* @param {string} productId Application name
* @param {string} url Url of Adfin plugin
*/
AdFinSubscription.prototype.setAdFinSettings = function (userId, productId, url) {
	if(userId) {
		this._userId = userId;
	}
	if(productId) {
		this._productId = productId;
	}
	if(url) {
		this._adfinUrl = url;
	}
};
/** Set the interval time for retrieving data
* @public
* @param {number} time Millisecond
*/
AdFinSubscription.prototype.setIntervalTime = function(time) {
	time = +time;
	if(!time) { // NaN or 0
		return;
	}
	this._intervalTime = time;
	this._stop();
	this._start();
};
/** Subscribe symbol to the specified field. This method will start the interval for retrieving data automatically.
* @public
* @param {string} name the name of the function
* @param {Array.<Array.<string|number|null>>} parameters the parameters of the function
* @return {*} Return the value if there is existing. Return undefined when requesting the invalid field.
*/
AdFinSubscription.prototype.call = function (name, parameters) {
	if(!name || !parameters) {
		throw "Invalid parameters";
	}

	for(let i = 0, length = parameters.length; i < length; ++i) {
		if(parameters[i][1] === "...") {
			return "...";
		}
	}

	/** @type AdFinSubscription.Request */let request = {"Function": name, "Parameters": parameters};

	let signature = JSON.stringify(request);
	let entry = this._cache[signature];
	if(entry && (entry.pending === false)) {
		return entry.value;
	}
	if(!entry) {
		this._cache[signature] = {
			"pending": true,
			"value": "..."
		};
		this._addQueue([request]);
	}

	return "...";
};
//#endregion public function of Adfin service


//#region public function of event
/** Add the function handler into the specific event. This method will start interval automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
AdFinSubscription.prototype.addEventListener = function(type, handler) {
	AdFinSubscription.base(this, "addEventListener", type, handler);
	this._start();
};
/** Remove the function handler from the specific event. If there is not any event handlers, the service will stop automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
AdFinSubscription.prototype.removeEventListener = function(type, handler) {
	AdFinSubscription.base(this, "removeEventListener", type, handler);
	if(!this.hasListener()) {
		this._stop();
		this._cache = {};
		this._queue = [];
	}
};
/** Remove all function handlers of all events and stop the interval of retrieving data of service.
* @override
*/
AdFinSubscription.prototype.removeAllEventListeners = function() {
	AdFinSubscription.base(this, "removeAllEventListeners");
	this._stop();
	this._cache = {};
	this._queue = [];
};
//#endregion public function of event


//#region private function of Adfin service
/** Start the interval of retrieving data
* @private
*/
AdFinSubscription.prototype._start = function() {
	if(this._state === "start") {
		return;
	}
	this._intervalId = window.setInterval(this._clearCache, this._intervalTime);
	this._state = "start";
};
/** Stop the interval of retrieving data
* @private
*/
AdFinSubscription.prototype._stop = function() {
	if(this._timerId !== null) {
		window.clearTimeout(this._timerId);
	}
	if(this._intervalId !== null) {
		window.clearInterval(this._intervalId);
	}
	this._timerId = null;
	this._intervalId = null;
	this._state = "stop";
};
/** Clear the cache
* @private
*/
AdFinSubscription.prototype._clearCache = function() {
	this._cache = {};
};
/** Manage the queue to retrieve the Adfin results. Prevent to call the same Adfin function multiple time with same parameters.
* @private
* @param {Array.<AdFinSubscription.Request>} requests
*/
AdFinSubscription.prototype._addQueue = function(requests) {
	if(this._timerId !== null) {
		window.clearTimeout(this._timerId);
	}

	if(!Array.isArray(requests)) {
		requests = [requests];
	}
	if(!this._queue) {
		this._queue = [];
	}
	this._queue = this._queue.concat(requests);
	this._timerId = window.setTimeout(this._processQueue, AdFinSubscription.ConflatorDelay);
};
/** Process the data queue to retrieve the Adfin results.
* @private
*/
AdFinSubscription.prototype._processQueue = function() {
	/** @type {Array.<AdFinSubscription.Request>} */ let chunckQueue = [];
	for(let i = 0, length = this._queue.length; i < length; ++i) {
		chunckQueue.push(this._queue[i]);
		if(chunckQueue.length === AdFinSubscription._chunckSize) {
			this._requestAdFin(chunckQueue);
			chunckQueue = [];
		}
	}
	if(chunckQueue.length) {
		this._requestAdFin(chunckQueue);
		chunckQueue = [];
	}
	this._queue = []; // Clear queue
	this._timerId = null; // Clear timeout id
};
/** Publish the Adfin request to the Adfin plugin.
* @private
* @param {Array.<AdFinSubscription.Request>} requests AdFin functions to be called (with parameters)
* @return {Promise}
*/
AdFinSubscription.prototype._requestAdFin = function(requests) {
	return fetch(this._adfinUrl, {
		method: "POST",
		headers: {
			"X-TRACS-ApplicationId": this._productId, // Application ID for the Thin Monitor
			"X-TRACS-UserId": this._userId // UUID
		},
		body: JSON.stringify(requests) // body data type must match "Content-Type" header
	})
		.then(Util.toJSON)
		.then(function(json) {
			json["signatures"] = [];
			json["functions"] = [];
			for(let i = 0, length = requests.length; i < length; ++i) {
				json["signatures"].push(JSON.stringify(requests[i])); // Attach the signature of the requests
				json["functions"].push(requests[i]["Function"]); // Attach the function of the requests
			}
			return json;
		})
		.then(this.onAdFinUpdate)
		.catch(function(err) {
			//		console.error(err);
		});
};
/** Function for processing the Adfin response and publish the event to the listener.
* @private
* @param {Object} json JSON object of Adfin response
*/
AdFinSubscription.prototype.onAdFinUpdate = function(json) {
	let signatures = json["signatures"];
	let functions = json["functions"];
	let toDispatch = {};
	for(let i = 0, length = signatures.length; i < length; ++i) {
		let value;
		let signature = signatures[i];
		let func = functions[i];
		let entry = json[i];
		if(!entry) {
			continue;
		}
		for(let j = 0, length2 = entry.length; j < length2; ++j) {
			if((entry[j].length >= 1) && (typeof (entry[j][0]) === "number")) {
				value = entry[j][0];
				break;
			} else if((entry[j].length === 2) && (typeof(entry[j][0]) === "string")) {
				value = entry[j][1];
				break;
			}
		}
		if(value == null) {
			delete this._cache[signature];
		} else {
			this._cache[signature]["pending"] = false;
			this._cache[signature]["value"] = value;
		}
		toDispatch[func] = true;
	}

	for(let functionName in toDispatch) {
		let rowData = {};
		//		rowData[functionName] = value;
		this._dispatch("dataChanged", { "function": functionName, "values": rowData, "sender": this }); // Publish data
	}
};

export {AdFinSubscription};
