import {Ext} from "../Ext.js";
import {EventDispatcher} from "../EventDispatcher.js";

/** IntervalSubscription is intended for updating data on the specified interval. It is useful when there is no data update triggered from the data source (e.g. market close)
* @constructor
* @extends {EventDispatcher}
* @param {string=} funcName Field name
* @param {Function=} func
* @param {number=} time=3600000 Interval duration
*/
let IntervalSubscription = function(funcName, func, time) {
	if(typeof func !== "function") {
		console.warn("No function is provided");
		return;
	}
	if(!funcName || typeof funcName !== "string") {
		funcName = func.name;
	}
	if(!funcName) {
		console.warn("No function name is provided");
		return;
	}

	try {
		this._value = func() || null; // First execution
	} catch(err) {
		console.warn(err);
		return;
	}

	time = +time;
	if(time) { // Not NaN or 0
		this._intervalTime = time;
	}

	this._onValueUpdate = this._onValueUpdate.bind(this);

	this._func = func;
	this._funcName = funcName;
};
Ext.inherits(IntervalSubscription, EventDispatcher);

/** @type {?Function}
* @private
*/
IntervalSubscription.prototype._func = null;
/** @type {string}
* @private
*/
IntervalSubscription.prototype._funcName = "";
/** @type {*}
* @private
*/
IntervalSubscription.prototype._value = null;
/** @type {number}
* @private
*/
IntervalSubscription.prototype._intervalId = 0;
/** @type {number}
* @private
*/
IntervalSubscription.prototype._intervalTime = 3600000; // 1 Hour


/** Set the interval time for retrieving data
* @public
* @param {number} time Millisecond
*/
IntervalSubscription.prototype.setIntervalTime = function(time) {
	time = +time;
	if(!time) { // NaN or 0
		return;
	}
	this._intervalTime = time;
	this._stop();
	this._start();
};
/** Get the value of this interval service
* @public
* @return {*}
*/
IntervalSubscription.prototype.getValue = function() {
	if(this._value === null) {
		this._value = this._func() || null;
	}
	return this._value;
};
/** Get the value of this interval service name
* @public
* @return {string}
*/
IntervalSubscription.prototype.getFunctionName = function() {
	return this._funcName;
};


/** Add the function handler into the specific event. This method will start interval automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
IntervalSubscription.prototype.addEventListener = function(type, handler) {
	IntervalSubscription.base(this, "addEventListener", type, handler);
	this._start();
};
/** Remove the function handler from the specific event. If there is not any event handlers, the service will stop automatically.
* @override
* @param {string} type Event name
* @param {Function} handler Event handler
*/
IntervalSubscription.prototype.removeEventListener = function(type, handler) {
	IntervalSubscription.base(this, "removeEventListener", type, handler);
	if(!this.hasListener()) {
		this._stop();
		this._value = null;
	}
};
/** Remove all function handlers of all events and stop the interval of retrieving data of service.
* @override
*/
IntervalSubscription.prototype.removeAllEventListeners = function() {
	IntervalSubscription.base(this, "removeAllEventListeners");
	this._stop();
	this._value = null;
};


/** Start the interval of retrieving data
* @private
*/
IntervalSubscription.prototype._start = function() {
	if(!this._intervalId) {
		let calTime = this._intervalTime - (Date.now() % this._intervalTime);
		this._intervalId = setTimeout(this._onValueUpdate, calTime);
	}
};
/** Stop the interval of retrieving data
* @private
*/
IntervalSubscription.prototype._stop = function() {
	if(this._intervalId) {
		window.clearTimeout(this._intervalId);
		this._intervalId = 0;
	}
};
/** Function for executing the function and publish the event to the listener.
* @private
*/
IntervalSubscription.prototype._onValueUpdate = function() {
	this._value = this._func() || null;

	let rowData = {};
	rowData[this._funcName] = this._value;
	this._dispatch("dataChanged", {
		"function": this._funcName,
		"values": rowData,
		"sender": this }
	); // Publish data

	// Stop then start service for calculating the precious timer to update value (Especially DateTime function)
	this._stop();
	this._start();
};

export {IntervalSubscription};
