
import {RealtimeService} from "./RealtimeService.js";
import {TSIService} from "./TSIService.js";
import {ADCService} from "./ADCService.js";
import {ADCSubscription} from "./ADCSubscription.js";
import {TSISubscription} from "./TSISubscription.js";
// import AdFinSubscription from "./AdFinSubscription.js";

/**  {lang=en} Realtime data access function <br>
* {lang=ja} リアルタイム データ アクセス関数 <br>
* {lang=zh} 实时数据访问函数 <br>
* {lang=ko} 실시간 데이터 액세스 함수 <br>
* {lang=ru} Realtime data access function
* @namespace Realtime
*/
let Realtime = {};

/**  {lang=en} The TR function is used for retrieving the data from Analytic Data Cloud service (ADC) or Snapshot Service (SnS). To use with 1 parameter (field), the data of corresponding RIC of that row will be returned. To use with 2 parameters (RIC and field), the data of specified RIC and field will be returned. Sns fields are available for upward compatibility. It is recommended that only ADC fields are used. <br>
* {lang=ja} TR 関数は、Analytic Data Cloud サービス (ADC) または Snapshot サービス (SnS) からデータを読み込んでいます。1つのパラメータ (フィールド) を使用するには、その行の対応する RIC のデータが返されます。2つのパラメータ (RIC およびフィールド) を使用するには、指定された RIC およびフィールドのデータが返されます。 <br>
* {lang=zh} TR 函数用于从分析数据云服务 (ADC) 或or 快照服务 (SnS) 中获取数据。要与 1 个参数 (字段) 一起使用，该行对应 RIC 的数据将被返回。要与 2 个参数 (RIC 和字段) 一起使用，指定 RIC 和字段的数据将被返回。 <br>
* {lang=ko} TR 함수는 Analytic Data Cloud 서비스(ADC) 또는 Snapshot Service(SnS)에서 데이터를 검색하는 데 사용됩니다. 매개변수 1개(필드)와 함께 사용하면 그 행의 해당 RIC 데이터가 반환됩니다. 매개변수 2개(RIC와 필드)와 함께 사용하면 지정된 RIC와 필드의 데이터가 반환됩니다. Snapshot Service(SnS) 필드는 상위 호환성을 위해 사용할 수 있습니다. ADC 필드만 사용하는 것이 좋습니다. <br>
* {lang=ru} The TR function is used for retrieving the data from Analytic Data Cloud service (ADC) or Snapshot Service (SnS). To use with 1 parameter (field), the data of corresponding RIC of that row will be returned. To use with 2 parameters (RIC and field), the data of specified RIC and field will be returned. Sns fields are available for upward compatibility. It is recommended that only ADC fields are used.
* @public
* @function TR
* @suppress {globalThis}
* @this tr.fn.Formula
* @memberof Realtime
* @param {string} param1 <br>
* {lang=en} ADC field. RIC if param2 is specified <br>
* {lang=ja} ADC フィールド。param2 が指定されているときは、RIC <br>
* {lang=zh} ADC 字段。如果指定了 param2，则为RIC <br>
* {lang=ko} ADC 필드. param2가 지정되면 RIC <br>
* {lang=ru} ADC field. RIC if param2 is specified
* @param {string=} param2 <br>
* {lang=en} ADC field <br>
* {lang=ja} ADC フィールド <br>
* {lang=zh} ADC 字段 <br>
* {lang=ko} ADC 필드 <br>
* {lang=ru} ADC field
* @return {*} {lang=en} Returned value can be string, number, or boolean <br>
* {lang=ja} 返される値は文字列、数値またはブール型です <br>
* {lang=zh} 返回值可以是字符串、数字或布尔数据 <br>
* {lang=ko} 반환된 값은 문자열, 숫자 또는 불리언 <br>
* {lang=ru} Returned value can be string, number, or boolean
* @example
* =TR("TR.PriceClose")
* =TR("TRI.TO", "TR.Volume")
* =TR("TR.PRICECLOSE(SDATE=0D,CURN=USD)")
* =((TR("TR.PRICECLOSE(SDATE=0D,CURN=USD)")/TR("TR.PRICECLOSE(SDATE=CY0,CURN=USD)")))-1
*/
Realtime.TR = function(param1, param2) {
	let argsLen = arguments.length;
	let field, res;
	if(argsLen === 1) { // TR("field")
		field = arguments[0];
		if(typeof field !== "string" || field === "") {
			this.setDataError("Pending data");
			return; // eslint-disable-line
		}

		if(this.addField(field)) { // Dispatch event to subscribe field
			this.setDataError("Pending data");
			return; // eslint-disable-line
		}
		// Get the data from the cache. Return as it has.
		return this["__D"](field);// eslint-disable-line
	} else if(argsLen === 2) { // TR("RIC", "field")
		let symbol = arguments[0];
		field = arguments[1];

		if (typeof symbol !== "string" || symbol === "" || typeof field !== "string" || field === "") {
			this.setDataError("Pending data");
			return; // eslint-disable-line
		}

		// Init ADC service if not exist
		let adcService = this.getIntervalService("ADC");
		if (!adcService) {
			adcService = ADCService.create();
			if (!adcService) {
				this.setDataError("Pending data");
				return; // eslint-disable-line
			}
			this.setIntervalService("ADC", adcService);
		}

		res = adcService["addSymbolToField"](symbol, field); // Get the cache data from service.
		if (res === ADCSubscription.PENDING_DATA) { // Invalid field or Pending data from requesting
			this.setDataError("Pending data");
		}
		return res; // eslint-disable-line
	} else { // Invalid used of TR function
		return; // eslint-disable-line
	}

};

/**  {lang=en} Return realtime value of the specified symbol and field <br>
* {lang=ja} 指定の RIC およびフィールドのリアルタイム値を返します <br>
* {lang=zh} 返回指定代码与字段的实时值 <br>
* {lang=ko} 지정된 기호 및 필드의 실시간 값을 반환 <br>
* {lang=ru} Return realtime value of the specified symbol and field
* @public
* @function RECORD
* @suppress {globalThis}
* @this tr.fn.Formula
* @memberof Realtime
* @param {string} symbol <br>
* {lang=en} RIC <br>
* {lang=ja} RIC <br>
* {lang=zh} RIC <br>
* {lang=ko} RIC <br>
* {lang=ru} RIC
* @param {string|number} field <br>
* {lang=en} Realtime field <br>
* {lang=ja} リアルタイム フィールド <br>
* {lang=zh} 实时字段 <br>
* {lang=ko} 실시간 필드 <br>
* {lang=ru} Realtime field
* @return {*} {lang=en} Returned value can be string, or number <br>
* {lang=ja} 返される値は文字列または数値です <br>
* {lang=zh} 返回值可以是字符串或数字 <br>
* {lang=ko} 반환된 값은 문자열 또는 숫자 <br>
* {lang=ru} Returned value can be string, or number
* @example
* =RECORD("FRF=","BID")
*/
Realtime.RECORD = function(symbol, field) {
	let fieldType = typeof field;
	if (typeof symbol !== "string" ||
		symbol === "" ||
		(fieldType !== "string" && fieldType !== "number") ||
		field === "") {
		this.setDataError("Pending data");
		return; // eslint-disable-line
	}

	// Init realtime service if not exist
	let rtService = this.getRealtimeService(); // Return Quotes2's subscription object
	if (!rtService) {
		rtService = RealtimeService.create();
		if (!rtService) {
			this.setDataError("Pending data"); // TODO: No Realtime service shold not cause pending
			return; // eslint-disable-line
		}
		rtService._addedFields = {}; // For preventing duplication
		rtService['start']();
		this.setRealtimeService(rtService);
	}

	// Try casting string field to number (numeric field id)
	let fid = (fieldType === "number") ? field : +field;

	// If the field can be converted to number, the given field is numeric field id
	if(fid === fid) { // Resolve numeric field id to text field name
		let resolvedFieldName = RealtimeService.getResolvedFieldName(fid);
		if(resolvedFieldName != null) {
			field = resolvedFieldName;
		} else {
			field = ""; // fid cannot be subscribed directly
			RealtimeService.resolveFieldId(fid, this);
		}
	}

	if(field && !rtService._addedFields[field]) {
		rtService['addFields'](field);
		rtService._addedFields[field] = field;
	}

	let res;
	let data = this.getCachedData(symbol);
	if (!data) {
		rtService['addRic'](symbol);
		this.setCachedData(symbol, {});
	} else {
		res = data[field];
		if(res == null && rtService._addedFields[field]) {
			// Symbol and field has been successfully registered. We are waiting for the response.
			// Real-time that does not have the value will also fall into this case.
			res = NaN;  // NaN is not available to proceed data that may cause error calculation
			data[field] = res; // Cache for next time
		} // else { The field is waiting to be resolved from field id
	}

	if (res == null) {
		this.setDataError("Pending data");
	}
	return res; // eslint-disable-line
};

/**  {lang=en} The MTXSERIES function is used for retrieving the data from Time Series service (MTXSERIES). To use with 2 parameters (date and field), the data of corresponding RIC of that row will be returned. To use with 3 parameters (RIC, date and field), the data of specified RIC will be returned. <br>
* {lang=ja} MTXSERIES 関数は、タイムシリーズ サービス (MTXSERIES) からデータを読み込んでいます。2つのパラメータ (日付、フィールド) を使用すると、その行の対応する RIC のデータが返されます。3つのパラメータ (RIC、日付、フィールド) を使用すると、指定された RIC のデータが返されます。 <br>
* {lang=zh} MTXSERIES 函数用于从时间序列服务 (MTXSERIES) 获取数据。要与 2 个参数 (日期与字段) 一起使用，该行相应 RIC 的数据将被返回。要与 3 个参数 (RIC、日期与字段) 一起使用，指定 RIC 的数据将被返回。 <br>
* {lang=ko} MTXSERIES 함수는 시계열 서비스(MTXSERIES)에서 데이터를 검색하는 데 사용됩니다. 매개변수 2개(데이터 및 필드)와 함께 사용하면 그 행의 해당 RIC 데이터가 반환됩니다. 매개변수 3개(RIC와 데이터, 필드)와 함께 사용하면 지정된 RIC의 데이터가 반환됩니다. <br>
* {lang=ru} The MTXSERIES function is used for retrieving the data from Time Series service (MTXSERIES). To use with 2 parameters (date and field), the data of corresponding RIC of that row will be returned. To use with 3 parameters (RIC, date and field), the data of specified RIC will be returned.
* @public
* @function MTXSERIES
* @suppress {globalThis}
* @memberof Realtime
* @param {string|number} param1 <br>
* {lang=en} Relative date or absolute date. RIC if param3 is specified <br>
* {lang=ja} 相対日時または絶対日時。param3 が指定されているときは、RIC <br>
* {lang=zh} 相对日期或绝对日期。如果指定了 param3，则为RIC <br>
* {lang=ko} 상대 날짜나 절대 날짜. param3이 지정되면 RIC <br>
* {lang=ru} Relative date or absolute date. RIC if param3 is specified
* @param {string|number} param2 <br>
* {lang=en} Time Series field. Date if param3 is specified <br>
* {lang=ja} タイムシリーズ フィールドparam3 が指定されているときは、日付 <br>
* {lang=zh} 时间序列字段。如果指定了 param3，则为日期 <br>
* {lang=ko} 시계열 필드. param3이 지정되면 날짜 <br>
* {lang=ru} Time Series field. Date if param3 is specified
* @param {string=} param3 <br>
* {lang=en} Time Series field <br>
* {lang=ja} タイムシリーズ フィールド <br>
* {lang=zh} 时间序列字段 <br>
* {lang=ko} 시계열 필드 <br>
* {lang=ru} Time Series field
* @return {*} {lang=en} Returned value will be number <br>
* {lang=ja} 返される値は数値です <br>
* {lang=zh} 返回值将是一个数字 <br>
* {lang=ko} 반환된 값은 숫자 <br>
* {lang=ru} Returned value will be number
* @example
* =MTXSERIES("1D", "BID.CLOSE") // Get BID.CLOSE of 1 day ago
* =MTXSERIES("0D", "BID.CLOSE") // Get BID.CLOSE of today (update every 1 hour)
* =MTXSERIES("1D2W3M4Q5Y", "BID.CLOSE") // D, W, M, Q and Y can be used for Day, Week, Month, Quarter and Year respectively.
* =MTXSERIES("2017-02-22", "ASK.CLOSE")
* =MTXSERIES(TODAY()-1, "ASK.CLOSE") // Get ASK.CLOSE of 1 day ago
*
* =MTXSERIES("TRI.TO", "1Q", "BID.CLOSE")
* =MTXSERIES("TRI.TO","2016-11-22", "ASK.CLOSE")
*/
Realtime.MTXSERIES = function(param1, param2, param3) {
	let argsLen = arguments.length;
	let symbol, date, field;
	if(argsLen === 2) { // TSI("Date", "Field")
		this.addField("X_RIC_NAME", "idn"); // Add required field for triggering when adding new RIC
		date = arguments[0];
		field = arguments[1];
		symbol = this["__D"]("X_RIC_NAME");
	} else if(argsLen === 3) { // TSI("RIC", Date", "Field")
		symbol = arguments[0];
		date = arguments[1];
		field = arguments[2];
	} else { // Invalid used of TSI function
		return;
	}

	if(typeof date === "number") { // To support date number since 1 Jan, 1900 (value is 1) that normally used in DateTime function
		// WARNING: Use external method from tr-grid
		date = tr.fn.DateTime["DATETEXT"](date);
	}

	if (typeof symbol !== "string" || symbol === ""
			|| typeof date !== "string" || date === ""
			|| typeof field !== "string" || field === "") {
		this.setDataError("Pending data");
		return;
	}

	// Init tsi service if not exist
	let tsiService = this.getIntervalService("TSI");
	if (!tsiService) {
		tsiService = TSIService.create();
		if (!tsiService) {
			this.setDataError("Pending data");
			return;
		}
		this.setIntervalService("TSI", tsiService);
	}

	let res = tsiService["addSymbolAndFieldToPeriod"](symbol, field, date); // Get the cache data from service.;
	if (res === TSISubscription.PENDING_DATA) { // Pending data from requesting, Invalid field will return null
		this.setDataError("Pending data");
	}
	return res; // eslint-disable-line
};

export {Realtime};
