import {Util} from "../Util.js";
import {Formula} from "./Formula.js";

/** @namespace */
let RealtimeService = {};

/** @type {Object}
* @private
*/
RealtimeService._quotes2 = null;
/** @type {Object.<string, Object>}
* @private
*/
RealtimeService._cache = {};
/** @type {string}
* @private
*/
RealtimeService._udipServiceUrl = "/Apps/UdipApi/GetFields";
/** @type {string}
* @public
*/
RealtimeService.uuid = "";

/** @public
* @return {Object}
*/
RealtimeService.create = function() {
	let quotes2 = RealtimeService._quotes2;
	if (!quotes2) {
		let j = window ? window["JET"] : null;
		if (j && j["Quotes2"]) {
			quotes2 = j["Quotes2"];
			RealtimeService._quotes2 = quotes2;
		} else {
			return null;
		}
	}
	return quotes2['create']();
};

/** @public
* @param {Object} quotes2
*/
RealtimeService.setQuote2 = function(quotes2) {
	RealtimeService._quotes2 = quotes2;
};

/** @public
* @param {string} fid
* @return {string|null}
*/
RealtimeService.getResolvedFieldName = function(fid) {
	let cache = RealtimeService._cache[fid];
	if(cache) {
		if(cache.value != null) {
			return /** @type{string} */(cache.value);
		}
	}
	return null;
};

// @param {Formula} formula External formula object
/** Resolve numeric field id to text field name
* @public
* @param {string} fid String of numeric field id
* @param {Object=} obj Listener object. Usually, this is External Formula formula object
* @return {!Promise}
*/
RealtimeService.resolveFieldId = function(fid, obj) {
	fid = fid + ""; // Cast to string
	let cache = RealtimeService._cache[fid];
	if(cache) {
		if(cache.value != null) {
			return Promise.resolve(cache.value);
		}
	} else {
		cache = RealtimeService._cache[fid] = {"fid": fid}; // Prevent duplicate request
	}

	// The request has not been resolved. Put any listener to the queue
	if(obj) {
		let q = cache.queue;
		if(!q) {
			q = cache.queue = [];
		}
		if(q.indexOf(obj) < 0) { // WANRING: this will be called multiple times
			q.push(obj);
		}
	}

	if(cache.promise) { // The request has been sent. we will wait for the response
		return cache.promise; // Original promise
	}

	// At this point there is no pending request (no promise) for the given fid
	let payload = {
		"universe": [
			"Public",
			"PublicPrivate"
		],
		"UUID": RealtimeService.uuid,
		"sources": [
			"RT"
		],
		"language": "en-US",
		"sortBy": "relevance",
		"searchText": fid
	};

	let prom = cache.promise = Util.post(RealtimeService._udipServiceUrl, payload)
		.then(RealtimeService._onSuccessfulResolving.bind(null, cache))
		.catch(RealtimeService._onFailedResolving.bind(null, cache));

	return prom;
};

/** @private
* @function
* @param {Object} cache Bound cache object from global cache
* @param {Object} json
* @return {string}
*/
RealtimeService._onSuccessfulResolving = function(cache, json) {
	cache.value = ""; // value can be empty due to the given invalid json structure
	let data = json["data"];
	if(data) {
		let fields = data["fields"];
		if(fields) {
			let info = fields[0];
			if(info) {
				if(info["oid"] === cache.fid) {
					cache.value = info["path"]; // Cache the value
				}
			}
		}
	}

	RealtimeService._onResponseEnd(cache);
	return cache.value;
};
/** @private
* @function
* @param {Object} cache Bound cache object from global cache
* @param {Object} err
* @return {string}
*/
RealtimeService._onFailedResolving = function(cache, err) {
	if(err.response) {
		if(err.response.status === 200) {
			cache.value = "";
		} else { // Server error retry another round
			cache.retry = cache.retry ? cache.retry + 1 : 1;

			if(cache.retry >= 3) {
				if(!cache.timerId) {
					cache.timerId = setTimeout(
						RealtimeService._onCacheClearing.bind(null, cache),
						300000 // Clear cache after 5 minutes
					);
				}
			}
			return Promise.reject("Server error");
		}
	} else {  // Successful response but no data due to internal error
		cache.value = "";
	}
	RealtimeService._onResponseEnd(cache);
	return ""; // This is resolved as success
};
/** Clean up cache once the promise has been fulfilled
* @private
* @function
* @param {Object} cache Bound cache object from global cache
*/
RealtimeService._onResponseEnd = function(cache) {
	cache.promise = null; // Promise has been fulfilled
	let q = cache.queue;
	if(q) { // Notify any object that is waiting for the response
		let len = q.length;
		for(let i = 0; i < len; ++i) {
			let obj = q[i];
			if(obj instanceof Formula) { // WARNING: Referencing external object
				obj.dispatchDataChanged(); // Force data re-calculation
			}
		}
		cache.queue = null;
	}
};
/** Delete cache
* @private
* @function
* @param {Object} cache Bound cache object from global cache
*/
RealtimeService._onCacheClearing = function(cache) {
	if (cache.fid) {
		delete RealtimeService._cache[cache.fid];
	}
	cache.value = null;
	cache.timerId = 0;
	RealtimeService._onResponseEnd(cache);
};

export {RealtimeService};
