
/** @constructor
* @param {string} str
* @param {number=} index
*/
let VariableToken = function(str, index) {
	str = str.trim();

	if(this._inputConverter[str]) {
		str = this._inputConverter[str];
	}

	this._input = str;
	if(index != null) {
		this._index = index;
	}

	if(this._constants[str] != null) {
		this._value = this._constants[str];
		this._type = typeof this._value;
		return;
	}

	let firstChar = str.charAt(0);
	let lastChar = (str.length > 1) ? str.charAt(str.length - 1) : "";
	if(firstChar === '"' && lastChar === '"') {
		this._value = str.substring(1, str.length - 1);
		this._type = "string";
		return;
	}
	if(firstChar === "'" && lastChar === "'") {
		this._value = str.substring(1, str.length - 1);
		this._type = "string";
		return;
	}
	if(firstChar === "(" && lastChar === ")") {
		this._resolveVariableGroup(str);
		return;
	}
	if(firstChar === "[" && lastChar === "]") {
		this._resolveArray(str); // Array has undefined value (this._value == null)
		return;
	}

	let adc = str.match(/TR\.\w+\(/);
	if(adc) { // ADC field with parameters
		this._type = "dynamic"; // ADC field with params always a dynamic token field
		return;
	}

	let funcs = str.match(/[\w._]+\(/g);
	if(funcs && funcs.length === 1) { // There must be only one method in the given str
		this._resolveFunction(str, funcs[0]); // Function has undefined value (this._value == null)
		return;
	}

	let words = str.match(/[\w._♦]+/g);
	if(words && words.length > 1) { // A group of expression without parentheses
		this._resolveVariableGroup(str);
		return;
	}

	if(firstChar === "♦") {
		this._type = "token";
		return;
	}

	let num = Number(str);
	if(num === num) { // If not NaN
		this._value = num;
		this._type = "number";
		return;
	}

	this._type = "dynamic";
};

/** @private
* @type {string}
*/
VariableToken.prototype._input = "";
/** @private
* @type {number}
*/
VariableToken.prototype._index = -1;
/** @private
* @type {string}
*/
VariableToken.prototype._type = "";
/** @private
* @type {Array.<VariableToken>}
*/
VariableToken.prototype._subVars = null;
/** @private
* @type {string}
*/
VariableToken.prototype._funcName = "";

/** @private
* @type {*}
*/
VariableToken.prototype._value;

/** @private
* @type {string}
*/
VariableToken.prototype._valueWithTokenField = "";

/** Static variables for input conversion
* @private
* @type {Object.<string, string>}
*/
VariableToken.prototype._inputConverter = {
	"TRUE": "true",
	"FALSE": "false",
	"PI": "Math.PI",
	"TRUE()": "true",
	"FALSE()": "false",
	"PI()": "Math.PI"
};
/** Static variables for defining constant
* @private
* @type {Object.<string, number>}
*/
VariableToken.prototype._constants = {
	"true": true,
	"false": false,
	"null": null,
	"NaN": NaN,
	"Math.E": Math.E, // returns Euler's number
	"Math.PI": Math.PI, // returns PI
	"Math.SQRT2": Math.SQRT2, // returns the square root of 2
	"Math.SQRT1_2": Math.SQRT1_2, // returns the square root of 1/2
	"Math.LN2": Math.LN2, // returns the natural logarithm of 2
	"Math.LN10": Math.LN10, // returns the natural logarithm of 10
	"Math.LOG2E": Math.LOG2E, // returns base 2 logarithm of E
	"Math.LOG10E": Math.LOG10E
};

/** @public
* @return {string}
*/
VariableToken.prototype.getExpression = function() {
	if(this._type === "dynamic") {
		return 'this.__D("' + this._input + '", __R)';
	}
	if(this._type === "group") {
		return /** @type{string} */(this._value);
	}
	if(this._type === "function") {
		if(this._subVars) {
			return "this.__F[" + this._index + "](" + this._subVars.join(", ") + ")";
		} else {
			return "this.__F[" + this._index + "]()";
		}
	}
	if(this._type === "array") {
		if(this._subVars) {
			return "[" + this._subVars.join(", ") + "]";
		} else {
			return "[]";
		}
	}
	return this._input;
};

/** @public
 * @return {string}
 */
VariableToken.prototype.getInputWithTokenField = function() {
	if(this._type === "dynamic") {
		return '♠' + this._input + '♥';
	}
	if(this._type === "group") {
		return /** @type{string} */(this._valueWithTokenField);
	}
	if(this._type === "function") {
		return this._funcName + "(" + this.getChildInputWithToken() + ")";
	}
	if(this._type === "array") {
		return "[" + this.getChildInputWithToken() + "]";
	}
	return this._input;
};

/** @public
 * @return {string}
 */
VariableToken.prototype.getChildInputWithToken = function() {
	let len = this._subVars ? this._subVars.length : 0;
	if(len <= 0) {
		return "";
	}
	let inputs = new Array(len);
	for(let i = 0; i < len; ++i) {
		inputs[i] = this._subVars[i].getInputWithTokenField();
	}
	return inputs.join(", ");
};

/** @public
* @function
* @return {string}
*/
VariableToken.prototype.toString = VariableToken.prototype.getExpression;

/** @public
* @return {*}
*/
VariableToken.prototype.getValue = function() { return this._value; };
/** @public
* @return {string}
*/
VariableToken.prototype.getInput = function() { return this._input; };
/** @public
* @return {number}
*/
VariableToken.prototype.getIndex = function() { return this._index; };
/** @public
* @param {number} num
*/
VariableToken.prototype.setIndex = function(num) { this._index = num; };

/** @public
* @return {string}
*/
VariableToken.prototype.getFunctionName = function() { return this._funcName; };

/** @public
* @param {Object=} opt_obj
* @return {string}
*/
VariableToken.prototype.getFieldNames = function(opt_obj) {
	if(this._subVars) {
		if(opt_obj) {
			let len = this._subVars.length;
			for(let i = 0; i < len; ++i) {
				this._subVars[i].getFieldNames(opt_obj);
			}
		}
		return this._subVars[0].getFieldNames();
	} else if(this._type === "dynamic") {
		if(opt_obj) {
			opt_obj[this._input] = true;
		}
		return this._input;
	}

	return "";
};

/** @private
* @param {string} str
* @returns {string}
*/
VariableToken.prototype._resolveVariableGroup = function(str) {
	let i = this._subVars ? this._subVars.length : 0;
	this._type = "group";

	this._value = str.replace(/[\w._♦]+/g, function(varStr) {
		if(!this._subVars) {
			this._subVars = [];
		}
		let varToken = new VariableToken(varStr, this._subVars.length);
		this._subVars.push(varToken);
		return varToken.getExpression();
	}.bind(this));

	this._valueWithTokenField = str.replace(/[\w._♦]+/g, function(varStr) {
		return this._subVars[i++].getInputWithTokenField();
	}.bind(this));

	return this._value;
};

/** @private
* @param {string} str
* @param {string} matchedName
*/
VariableToken.prototype._resolveFunction = function(str, matchedName) {
	this._type = "function";
	this._funcName = matchedName.substr(0, matchedName.length - 1); // Extract function name

	// str = str.replace(/ /g, ""); // Replace all white spaces
	str = str.replace(matchedName, ""); // Strip function name and "("
	str = str.substr(0, str.length - 1); // Strip ")"
	if(str) {
		this._subVars = [];
		let params = str.split(",");
		let len = params.length;
		for(let i = 0; i < len; ++i) {
			this._subVars[i] = new VariableToken(params[i], i);
		}
	}
};
/** @private
* @param {string} str
*/
VariableToken.prototype._resolveArray = function(str) {
	this._type = "array";

	str = str.substr(1, str.length - 1); // Strip "[" and "]"
	if(str) {
		this._subVars = [];
		let params = str.split(",");
		let len = params.length;
		for(let i = 0; i < len; ++i) {
			this._subVars[i] = new VariableToken(params[i], i);
		}
	}
};

export default VariableToken;
export { VariableToken };
