
/**
 * @namespace Info
 * @description
 *    {lang=en} Valid/Invalid value checking functions
 *    {lang=ja} 有効/無効 の確認関数
 *    {lang=zh} 有效/无效值检查函数
 *    {lang=hant} 有效/無效值檢查函數
 *    {lang=ko} 유효/무효 값 확인 함수
 *    {lang=ru} Valid/Invalid value checking functions
 *    {lang=de} Prüffunktionen für gültige/ungültige Werte
 */
let Info = {};

/**
 * @function ISEMPTY
 * @memberof Info
 * @description
 *    {lang=en} The ISEMPTY function tests if an initial supplied expression (or value) contains a value.
 *    {lang=ja} ISEMPTY 関数は、最初に指定した式 (または値) に空白があるかどうかをテストします。
 *    {lang=zh} ISEMPTY 函数测试初始提供的表达式 (或值) 是否有空值。
 *    {lang=hant} ISEMPTY 函數測試初始提供的表達式 (或值) 是否有空值。
 *    {lang=ko} ISEMPTY 함수는 초기 입력한 식(또는 값)에 값이 들어 있는지 테스트합니다.
 *    {lang=ru} The ISEMPTY function tests if an initial supplied expression (or value) contains a value.
 *    {lang=de} Die ISEMPTY-Funktion prüft, ob ein ursprünglich gelieferter Ausdruck (oder Wert) einen Wert enthält.
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISEMPTY(COL3)
 */
Info["ISEMPTY"] = function(value) { // undefined, null, and empty string return 1. 0 and NaN return 0
	return (value === value && value !== 0 && !value) ? 1 : 0;
};
/**
 * @function ISERR
 * @memberof Info
 * @description
 *    {lang=en} The ISERR function tests if an initial supplied expression (or value) returns an Error, except #N/A
 *    {lang=ja} ISERR 関数は、最初に指定した式 (または値) が #N/A 以外のエラーを返すかどうかをテストします。
 *    {lang=zh} ISERR 函数测试初始提供的表达式 (或值) 是否返回任何错误，#N/A 错误除外
 *    {lang=hant} ISERR 函數測試初始提供的表達式 (或值) 是否返回任何錯誤，#N/A 錯誤除外
 *    {lang=ko} ISERR 함수는 초기 입력한 식(또는 값)이 #N/A를 제외한 오류를 반환하는지 테스트합니다.
 *    {lang=ru} The ISERR function tests if an initial supplied expression (or value) returns an Error, except #N/A
 *    {lang=de} Die ISERR-Funktion prüft, ob ein anfänglich gelieferter Ausdruck (oder Wert) einen Fehler ausgibt, außer #N/A
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISERR(COL3)
 */
Info["ISERR"] = function(value) { // undefined, null, and NaN return 1.  0 and empty string return 0
	return (value !== value || value == null) ? 1 : 0;
};
/**
 * @function ISNA
 * @memberof Info
 * @description
 *    {lang=en} The ISNA function tests if an initial supplied expression (or value) returns an #N/A Error, and if so, returns true; Otherwise the function returns false.
 *    {lang=ja} ISNA 関数は、最初に指定された式 (または値) が #N/A エラーを返すかどうかをテストします。#N/A エラーの場合、TRUE を返し、それ以外は FALSE を返します。
 *    {lang=zh} ISNA 函数测试初始提供的表达式 (或值) 是否返回 #N/A 错误，如果是，返回 true; 否则函数返回 false。
 *    {lang=hant} ISNA 函數測試初始提供的表達式 (或值) 是否返回 #N/A 錯誤，如果是，返回 true; 否則函數返回 false。
 *    {lang=ko} ISNA 함수는 초기 입력한 식(또는 값)이 #N/A 오류를 반환하는지 테스트하고 #N/A Error를 반환하면 TRUE를 반환하지 않으면 FALSE를 반환합니다.
 *    {lang=ru} The ISNA function tests if an initial supplied expression (or value) returns an #N/A Error, and if so, returns true; Otherwise the function returns false.
 *    {lang=de} Die ISNA-Funktion prüft, ob ein anfänglich gelieferter Ausdruck (oder Wert) einen #N/A-Fehler ausgibt, und wenn ja, wird „Wahr“ ausgegeben; andernfalls gibt die Funktion „Falsch“ aus.
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISNA(COL3)
 */
Info["ISNA"] = function(value) {
	return (value === "") ? 1 : 0;
};
/**
 * @function ISNUMBER
 * @memberof Info
 * @description
 *    {lang=en} The ISNUMBER function tests if a supplied value is a number. If so, the function returns TRUE; Otherwise it returns FALSE.
 *    {lang=ja} ISNUMBER 関数は、指定された値が数値かどうかをテストします。文字列の場合 TRUE を返し、それ以外は FALSE を返します。
 *    {lang=zh} ISNUMBER 函数测试提供的值是否是数字。如果是，函数返回 TRUE; 否则返回 FALSE。
 *    {lang=hant} ISNUMBER 函數測試提供的值是否是數字。如果是，函數返回 TRUE; 否則返回 FALSE。
 *    {lang=ko} ISNUMBER 함수는 입력한 값이 숫자인지 테스트합니다. 숫자라면 이 함수는 TRUE를, 숫자가 아니면 FALSE를 반환합니다.
 *    {lang=ru} The ISNUMBER function tests if a supplied value is a number. If so, the function returns TRUE; Otherwise it returns FALSE.
 *    {lang=de} Die ISNUMBER-Funktion prüft, ob ein gelieferter Wert eine Zahl ist. Wenn dies der Fall ist, gibt die Funktion WAHR aus; andernfalls gibt sie FALSCH aus.
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISNUMBER(CF_NETCHNG)
 */
Info["ISNUMBER"] = function(value) {
	return (typeof value === "number") ? 1 : 0;
};
/**
 * @function ISSTRING
 * @memberof Info
 * @description
 *    {lang=en} The ISSTRING function tests if a supplied value is a string. If so, the function returns TRUE; Otherwise it returns FALSE.
 *    {lang=ja} ISSTRING 関数は、指定された値が文字列かどうかをテストします。文字列の場合 TRUE を返し、それ以外は FALSE を返します。
 *    {lang=zh} ISSTRING 函数测试提供的值是否是字符串。如果是，函数返回 TRUE; 否则返回 FALSE。
 *    {lang=hant} ISSTRING 函數測試提供的值是否是字符串。如果是，函數返回 TRUE; 否則返回 FALSE。
 *    {lang=ko} ISSTRING 함수는 입력한 값이 문자열인지 테스트합니다. 문자열이라면 이 함수는 TRUE를, 문자가 아니면 FALSE를 반환합니다.
 *    {lang=ru} The ISSTRING function tests if a supplied value is a string. If so, the function returns TRUE; Otherwise it returns FALSE.
 *    {lang=de} Die ISSTRING-Funktion prüft, ob ein gelieferter Wert eine Zeichenfolge ist. Wenn dies der Fall ist, gibt die Funktion WAHR aus; andernfalls gibt sie FALSCH aus.
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISSTRING(CF_NAME)
 */
Info["ISSTRING"] = function(value) {
	return (typeof value === "string") ? 1 : 0;
};
/**
 * @function ISNASTRING
 * @memberof Info
 * @description
 *    {lang=en} The ISNASTRING  function tests if a supplied value is a none-empty string. The function returns TRUE for an empty string, Otherwise it returns FALSE
 *    			Zero, null, undefined, number, array with item and object with key will return 0
 *    {lang=ja} ISNASTRING 関数は、指定された値が空白でない文字列かどうかをテストします。
 *    			ゼロ, null, 未定義, 数値, 項目のある配列およびキー付オブジェクトは 0 を返します。
 *    {lang=zh} ISNASTRING  函数测试提供的值是否是非空字符串。该函数对空字符串返回 TRUE，否则返回 FALSE。
 *    			零、空、未定义、数字、数组与项目，以及对象与关键字返回 0。
 *    {lang=hant} ISNASTRING 函數測試提供的值是否是非空字符串。該函數對空字符串返回 TRUE，否則返回 FALSE。
 *    			零、空、未定義、數字、數組與項目，以及對象與關鍵字返回 0。
 *    {lang=ko} ISNASTRING 함수는 입력한 값이 비어 있지 않은 문자열인지 테스트합니다. 이 함수는 빈 문자열의 경우 TRUE를, 그렇지 않으면 FALSE를 반환합니다.
 *    {lang=ru} The ISNASTRING  function tests if a supplied value is a none-empty string. The function returns TRUE for an empty string, Otherwise it returns FALSE
 *    			Zero, null, undefined, number, array with item and object with key will return 0
 *    {lang=de} Die ISNASTRING-Funktion prüft, ob ein gelieferter Wert eine nicht leere Zeichenfolge ist. Die Funktion gibt für eine leere Zeichenkette WAHR aus, andernfalls gibt sie FALSCH aus.
 *    			Zero, null, undefined, number, array with item and object with key will return 0
 * @param  {*} value
 *    {lang=en} expression (or value)
 *    {lang=ja} 式 (または値)
 *    {lang=zh} 表达式 (或值)
 *    {lang=hant} 表達式 (或值)
 *    {lang=ko} 식(또는 값)
 *    {lang=ru} expression (or value)
 *    {lang=de} Ausdruck (oder Wert)
 * @return {number}
 * @example
 *    =ISNASTRING(CF_SOURCE)
 */
Info["ISNASTRING"] = function(value) {
	if(value) {
		if(value.length === 0) {
			return 1; // []
		}
		if(value.constructor === Object) {
			for(let key in value) {
				return 0; // {"a": 1}
			}
			return 1; // {}
		}
	} else if(value === "") {
		return 1; // ""
	}
	return  0; // "abc", [1, 2, 3], 0, 1, null, undefined, NaN
};

export default Info;
export { Info };
