/**
 * @namespace Logic
 * @description
 *    {lang=en} Functions that handle logical conditions of data
 *    {lang=ja} データの理論関数
 *    {lang=zh} 处理数据的逻辑条件的函数
 *    {lang=hant} 處理數據的邏輯條件的函數
 *    {lang=ko} 데이터의 논리적 조건을 처리하는 함수
 *    {lang=ru} Functions that handle logical conditions of data
 *    {lang=de} Funktionen, die logische Bedingungen von Daten verarbeiten
 */
let Logic = {};

/**
 * @function AND
 * @memberof Logic
 * @description
 *    {lang=en} The AND function returns 1 (TRUE), if all arguments in List are true. It returns 0 (FALSE) if one or more arguments in List are false. Use operand && for a shorthand.
 *    {lang=ja} AND 関数は、リストのすべての引数が TRUEの 場合、1 (TRUE) を返します。リストの1つ以上の引数が FALSEの 場合、0 (FALSE) を返します。省略形は、オペランド [ && ] を使用します。
 *    {lang=zh} 如果列表中的全部参数为真，则 AND 函数返回 1 (TRUE)。如果一个或多个列表中的参数为假，则返回 0 (FALSE)。 速记使用操作符 &&。
 *    {lang=hant} 如果列表中的全部參數為真，則 AND 函數返回 1 (TRUE)。如果一個或多個列表中的參數為假，則返回 0 (FALSE)。速記使用操作符 &&。
 *    {lang=ko} AND 함수는 목록의 모든 인수가 참이면 1(TRUE)을 반환합니다. 목록에 있는 하나 이상의 인수가 거짓이면 0(FALSE)을 반환합니다. 약칭으로 피연산자 &&를 사용합니다.
 *    {lang=ru} The AND function returns 1 (TRUE), if all arguments in List are true. It returns 0 (FALSE) if one or more arguments in List are false. Use operand && for a shorthand.
 *    {lang=de} Die UND-Funktion gibt 1 (WAHR) aus, wenn alle Argumente in der Liste wahr sind. Sie gibt 0 (FALSCH) aus, wenn eines oder mehrere Argumente in der Liste falsch sind. Verwenden Sie Operand && für eine Kurzschrift.
 * @return {number}
 * @example
 *    =AND(5>2, 1+1=2)
 *    =5>2 && 1+1=2
 */
Logic["AND"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	for(let i = args.length; --i >= 0;) {
		if(!args[i]) {
			return 0;
		}
	}
	return 1;
};

/**
 * @function FALSE
 * @memberof Logic
 * @description
 *    {lang=en} The FALSE function returns the value 0 (FALSE).
 *    {lang=ja} FALSE 関数は、値 0 (FALSE) を返します。
 *    {lang=zh} FALSE 函数返回值 0 (FALSE)。
 *    {lang=hant} FALSE 函數返回值 0 (FALSE)。
 *    {lang=ko} FALSE 함수는 값 0(FALSE)을 반환합니다.
 *    {lang=ru} The FALSE function returns the value 0 (FALSE).
 *    {lang=de} Die FALSCH-Funktion gibt den Wert 0 (FALSCH) aus.
 * @return {number}
 * @example
 *    =FALSE()
 */
Logic["FALSE"] = function() {
	return 0;
};

/**
 * @function IF
 * @memberof Logic
 * @description
 *    {lang=en} The IF function returns Expression 1 if Cond is TRUE, (a numeric value  0). Otherwise, it returns Expression 2.
 *    {lang=ja} IF 関数は、指定された式が TRUE の場合、理論式 1を返します (数値 0)。 それ以外は、理論式 2を返します。
 *    {lang=zh} 如果条件为 TRUE，则 IF 函数返回表达式 1, (数值为  0)。否则，返回表达式 2。
 *    {lang=hant} 如果條件為 TRUE，則 IF 函數返回表達式 1, (數值為 0)。否則，返回表達式 2。
 *    {lang=ko} IF 함수는 Cond가 TRUE인 경우 Expression 1을 반환합니다(숫자 값 0). 그렇지 않으면 Expression 2를 반환합니다.
 *    {lang=ru} The IF function returns Expression 1 if Cond is TRUE, (a numeric value  0). Otherwise, it returns Expression 2.
 *    {lang=de} Die IF-Funktion gibt Ausdruck 1 aus, wenn die Bedingung WAHR ist (ein numerischer Wert 0). Andernfalls gibt sie Ausdruck 2 aus.
 * @param  {boolean|string} cond
 *    {lang=en} Boolean condition
 *    {lang=ja} ブール条件
 *    {lang=zh} 布尔条件
 *    {lang=hant} 布爾條件
 *    {lang=ko} 불리언 조건
 *    {lang=ru} Boolean condition
 *    {lang=de} Boolesche Bedingung
 * @param  {number|string} exp1
 *    {lang=en} Expression 1
 *    {lang=ja} 理論式 1
 *    {lang=zh} 表达式 1
 *    {lang=hant} 表達式 1
 *    {lang=ko} Expression 1
 *    {lang=ru} Expression 1
 *    {lang=de} Ausdruck 1
 * @param  {number|string} exp2
 *    {lang=en} Expression 2
 *    {lang=ja} 理論式 2
 *    {lang=zh} 表達式 2
 *    {lang=hant} Expression 2
 *    {lang=ko} Expression 2
 *    {lang=ru} Expression 2
 *    {lang=de} Ausdruck 2
 * @return {number|string}
 * @example
 *    =IF(CF_LAST > 0, "Positive", "Negative")
 */
Logic["IF"] = function(cond, exp1, exp2) {
	return cond ? exp1 : exp2;
};

/**
 * @function NOT
 * @memberof Logic
 * @description
 *    {lang=en} The NOT function negate the value of the condition. Use operand ! for a shorthand.
 *    {lang=ja} NOT 関数は、FALSE の場合は TRUEを、TRUE の場合は FALSE を返します。省略形は、オペランド [ !]を使用します。
 *    {lang=zh} NOT 函数否定条件值。速记使用操作符 !。
 *    {lang=hant} NOT 函數否定條件值。速記使用操作符 !。
 *    {lang=ko} NOT 함수는 조건의 값을 무효화합니다. 약칭으로 피연산자 !를 사용하십시오.
 *    {lang=ru} The NOT function negate the value of the condition. Use operand ! for a shorthand.
 *    {lang=de} Die NICHT-Funktion negiert den Wert der Bedingung. Verwenden Sie Operand ! für eine Kurzschrift.
 * @param  {number|string} a
 *    {lang=en} Cond
 *    {lang=ja} 条件
 *    {lang=zh} 条件
 *    {lang=hant} 條件
 *    {lang=ko} 조건
 *    {lang=ru} Cond
 *    {lang=de} Beding.
 * @return {number}
 * @example
 *    =NOT(CF_NETCHNG > 0)
 *    =!(CF_NETCHNG > 0)
 */
Logic["NOT"] = function(a) {
	return !a ? 1 : 0;
};

/**
 * @function OR
 * @memberof Logic
 * @description
 *    {lang=en} The OR function returns 1 (TRUE), if any argument in List is true. It returns 0 (FALSE), if all arguments in List are FALSE. Use operand || for a shorthand.
 *    {lang=ja} OR 関数は、リストのいずれかの引数が TRUE の場合、1 (TRUE) を返します。リストのすべての引数が FALSE の場合、0 (FALSE) を返します。省略形は、オペランド [ || ] を使用します。
 *    {lang=zh} 如果列表中的任意参数为真，则 OR 函数返回 1 (TRUE)。如果列表中的全部参数为假，则返回 0 (FALSE)。速记使用操作符 ||。
 *    {lang=hant} 如果列表中的任意參數為真，則 OR 函數返回 1 (TRUE)。如果列表中的全部參數為假，則返回 0 (FALSE)。速記使用操作符 ||。
 *    {lang=ko} OR 함수는 목록의 어떤 인수가 참이면 1(TRUE)을 반환합니다. 목록에 있는 모든 인수가 거짓이면 0(FALSE)을 반환합니다. 약칭으로 피연산자 ||를 사용합니다.
 *    {lang=ru} The OR function returns 1 (TRUE), if any argument in List is true. It returns 0 (FALSE), if all arguments in List are FALSE. Use operand || for a shorthand.
 *    {lang=de} Die ODER-Funktion gibt 1 (WAHR) aus, wenn eines der Argumente in der Liste wahr ist. Sie gibt 0 (FALSCH) aus, wenn alle Argumente in der Liste FALSCH sind. Verwenden Sie Operand || für eine Kurzschrift.
 * @return {number}
 * @example
 *    =OR(5>10, 4+2=8, 10<1)
 *    =5>10 || 4+2=8 || 10<1
 */
Logic["OR"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	for(let i = args.length; --i >= 0;) {
		if(args[i]) {
			return 1;
		}
	}
	return 0;
};

/**
 * @function TRUE
 * @memberof Logic
 * @description
 *    {lang=en} The TRUE function returns the value 1 (TRUE).
 *    {lang=ja} TRUE 関数は、理論値 1 (TRUE) を返します。
 *    {lang=zh} TRUE 函数返回值 1 (TRUE)。
 *    {lang=hant} TRUE 函數返回值 1 (TRUE)。
 *    {lang=ko} TRUE 함수는 값 1(TRUE)을 반환합니다.
 *    {lang=ru} The TRUE function returns the value 1 (TRUE).
 *    {lang=de} Die WAHR-Funktion gibt den Wert 0 (WAHR) aus.
 * @return {number}
 * @example
 *    =TRUE()
 */
Logic["TRUE"] = function() {
	return 1;
};

export default Logic;
export { Logic };
