/**
 * @namespace Statistic
 * @description
 *    {lang=en} Statistic functions
 *    {lang=ja} 統計関数
 *    {lang=zh} 统计函数
 *    {lang=hant} 統計函數
 *    {lang=ko} 통계 함수
 *    {lang=ru} Statistic functions
 *    {lang=de} Statistische Funktionen
 */
let Stat = {};

/**
* @function AVERAGE
* @memberof Statistic
* @description
*    {lang=en} Returns the average (arithmetic mean) of the arguments. This function supports any number of arguments AVERAGE(number1, [number2], [number3], ...)
*    {lang=ja} 引数の平均値 (算術平均) を返します。この関数は、任意の値の引数 AVERAGE をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 返回参数的平均值 (算数平均数)。这个函数支持任意数量的参数 AVERAGE(number1, [number2], [number3], ...)
*    {lang=hant} 返回參數的平均值 (算數平均數)。這個函數支持任意數量的參數 AVERAGE(number1, [number2], [number3], ...)
*    {lang=ko} 인수의 평균(산술 평균)을 반환합니다. 이 함수는 인수를 제한 없이 지원합니다. AVERAGE(number1, [number2], [number3], ...)
*    {lang=ru} Returns the average (arithmetic mean) of the arguments. This function supports any number of arguments AVERAGE(number1, [number2], [number3], ...)
*    {lang=de} Gibt den Durchschnitt (arithmetische Mitte) der Argumente aus. Diese Funktion unterstützt den DURCHSCHNITT einer beliebigen Anzahl von Argumenten (Zahl1, [Zahl2], [Zahl3], ...)
* @return {number}
* @example
*    =AVERAGE(6.10,13.20,15.40)
*/
Stat["AVERAGE"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	let argLen = args.length;
	if(argLen <= 0) { return NaN; }

	let sum = 0;
	for(let i = argLen; --i >= 0;) {
		sum += args[i];
	}
	return sum / argLen;
};

/**
* @function AVG
* @memberof Statistic
* @description
*    {lang=en} Returns the average (arithmetic mean) of the arguments. This function supports any number of arguments AVG(number1, [number2], [number3], ...)
*    {lang=ja} 引数の平均値 (算術平均) を返します。この関数は、任意の値の引数 AVG をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 返回参数的平均值 (算数平均数)。这个函数支持任意数量的参数 AVG(number1, [number2], [number3], ...)
*    {lang=hant} 返回參數的平均值 (算數平均數)。這個函數支持任意數量的參數 AVG(number1, [number2], [number3], ...)
*    {lang=ko} 인수의 평균(산술 평균)을 반환합니다. 이 함수는 인수를 제한 없이 지원합니다. AVG(number1, [number2], [number3], ...)
*    {lang=ru} Returns the average (arithmetic mean) of the arguments. This function supports any number of arguments AVG(number1, [number2], [number3], ...)
*    {lang=de} Gibt den Durchschnitt (arithmetische Mitte) der Argumente aus. Diese Funktion unterstützt den DURCHSCHN. einer beliebigen Anzahl von Argumenten (Zahl1, [Zahl2], [Zahl3], ...)
* @return {number}
* @example
*    =AVG(6.10,13.20,15.40)
*/
Stat["AVG"] = Stat["AVERAGE"];

/**
* @function COUNT
* @memberof Statistic
* @description
*    {lang=en} Returns the number of entries in a number field that is in a range or array of numbers.
*              This function supports any number of arguments COUNT(number1, [number2], [number3], ...)
*    {lang=ja} 数値列のフィールドに入力された数値の数を返します。
*              この関数は、任意の値の引数 COUNT をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 返回在某数字范围或某数组中的数字项输入的次数。
*              这个函数支持任意数量的参数 COUNT(number1, [number2], [number3], ...)
*    {lang=hant} 返回在某數字範圍或某數組中的數字項輸入的次數。
*              這個函數支持任意數量的參數 COUNT(number1, [number2], [number3], ...)
*    {lang=ko} 범위 또는 숫자 배열에 있는 숫자 필드의 항목 수를 반환합니다.
*    {lang=ru} Returns the number of entries in a number field that is in a range or array of numbers.
*              This function supports any number of arguments COUNT(number1, [number2], [number3], ...)
*    {lang=de} Gibt die Anzahl der Einträge in einem Zahlenfeld aus, das sich in einem Bereich oder einer Anordnung von Zahlen befindet.
*              This function supports any number of arguments COUNT(number1, [number2], [number3], ...)
* @return {number}
* @example
*    =COUNT(6.10,13.20,15.40)
*/
Stat["COUNT"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	let count = 0;
	for(let i = args.length; --i >= 0;) {
		let arg = args[i];
		if(arg != null && arg == arg) {
			++count;
		}
	}
	return count;
};

/**
* @function MAX
* @memberof Statistic
* @description
*    {lang=en} Returns the largest value from a supplied set of numeric values.
*              This function supports any number of arguments MAX(number1, [number2], [number3], ...)
*    {lang=ja} 指定された数値の中から最大値を返します。
*              この関数は、任意の値の引数 MAX をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 返回所提供的数值集中的最大值。
*              这个函数支持任意数量的参数 MAX(number1, [number2], [number3], ...)
*    {lang=hant} 返回所提供的數值集中的最大值。
*              這個函數支持任意數量的參數 MAX(number1, [number2], [number3], ...)
*    {lang=ko} 제공된 숫자 값의 집합에서 가장 큰 값을 반환합니다.
*    {lang=ru} Returns the largest value from a supplied set of numeric values.
*              This function supports any number of arguments MAX(number1, [number2], [number3], ...)
*    {lang=de} Gibt den größten Wert aus einem gelieferten Satz numerischer Werte aus.
*              This function supports any number of arguments MAX(number1, [number2], [number3], ...)
* @return {number}
* @example
*    =MAX(6.10,13.20,15.40)
*/
Stat["MAX"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	if(args.length <= 0) { return NaN; }

	let max = args[0];
	for(let i = args.length; --i >= 1;) {
		if(args[i] > max) {
			max = args[i];
		}
	}
	return max;
};

/**
* @function MIN
* @memberof Statistic
* @description
*    {lang=en} Returns the smallest value from a supplied set of numeric values.
*              This function supports any number of arguments MIN(number1, [number2], [number3], ...)
*    {lang=ja} 指定された数値の中から最小値を返します。
*              この関数は、任意の値の引数 MIN をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 返回所提供的数值集中的最小值。
*              这个函数支持任意数量的参数 MIN(number1, [number2], [number3], ...)
*    {lang=hant} 返回所提供的數值集中的最小值。
*              這個函數支持任意數量的參數 MIN(number1, [number2], [number3], ...)
*    {lang=ko} 제공된 숫자 값의 집합에서 가장 작은 값을 반환합니다.
*    {lang=ru} Returns the smallest value from a supplied set of numeric values.
*              This function supports any number of arguments MIN(number1, [number2], [number3], ...)
*    {lang=de} Gibt den kleinsten Wert aus einem gelieferten Satz numerischer Werte aus.
*              This function supports any number of arguments MIN(number1, [number2], [number3], ...)
* @return {number}
* @example
*    =MIN(6.10,13.20,15.40)
*/
Stat["MIN"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : arguments;
	if(args.length <= 0) { return NaN; }

	let min = args[0];
	for(let i = args.length; --i >= 1;) {
		if(args[i] < min) {
			min = args[i];
		}
	}
	return min;
};

/**
* @function STDEV
* @memberof Statistic
* @description
*    {lang=en} Estimates standard deviation based on a sample (unbias standard deviation for small sample size)
*              This function supports any number of arguments STDEV(number1, [number2], [number3], ...)
*    {lang=ja} 標本に基づいて予想標準偏差を返します (小さいサンプル サイズの偏りのない標準偏差)
*              この関数は、任意の値の引数 STDEV をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 基于样本的预估标准差 (小样本规模的无偏标准差)
*              这个函数支持任意数量的参数 STDEV(number1, [number2], [number3], ...)
*    {lang=hant} 基於樣本的預估標準差 (小樣本規模的無偏標準差)
*              這個函數支持任意數量的參數 STDEV(number1, [number2], [number3], ...)
*    {lang=ko} 표본을 기준으로 표준편차를 추정(표본 크기가 작으면 편의가 없는 표준편차)
*    {lang=ru} Estimates standard deviation based on a sample (unbias standard deviation for small sample size)
*              This function supports any number of arguments STDEV(number1, [number2], [number3], ...)
*    {lang=de} Schätzt die Standardabweichung basierend auf einer Stichprobe (objektive Standardabweichung für kleine Stichprobengrößen)
*              This function supports any number of arguments STDEV(number1, [number2], [number3], ...)
* @return {number}
* @example
*    =STDEV(6.10,13.20,15.40)
*/
Stat["STDEV"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : Array.prototype.slice.call(arguments);
	let argLen = args.length;
	if(argLen <= 0) { return NaN; }
	if(argLen === 1) { return 0; }

	let mean = Stat["AVERAGE"](args);

	let sum = 0;
	for(let i = argLen; --i >= 0;) {
		let diff = args[i] - mean;
		sum += diff * diff;
	}
	return Math.sqrt(sum / (argLen - 1));
};

/**
* @function STDEVP
* @memberof Statistic
* @description
*    {lang=en} Calculates standard deviation based on the entire population (bias standard deviation for small sample size)
*              This function supports any number of arguments STDEVP(number1, [number2], [number3], ...)
*    {lang=ja} 母集団の標準偏差を返します (小さいサンプル サイズのバイアス標準偏差)
*              この関数は、任意の値の引数 STDEVP をサポートしています (数値 1, 数値 2, 数値 3, ...)
*    {lang=zh} 基于整个总体计算标准差 (小样本规模的无偏标准差)
*              这个函数支持任意数量的参数 STDEVP(number1, [number2], [number3], ...)
*    {lang=hant} 基於整個總體計算標準差 (小樣本規模的無偏標準差)
*              這個函數支持任意數量的參數 STDEVP(number1, [number2], [number3], ...)
*    {lang=ko} 전체 모집단을 기준으로 표준편차를 계산(표본 크기가 작은 경우 편의가 있는 표준편차)
*    {lang=ru} Calculates standard deviation based on the entire population (bias standard deviation for small sample size)
*              This function supports any number of arguments STDEVP(number1, [number2], [number3], ...)
*    {lang=de} Berechnet die Standardabweichung basierend auf der Gesamtpopulation (Tendez-Standardabweichung für kleine Stichprobengrößen)
*              This function supports any number of arguments STDEVP(number1, [number2], [number3], ...)
* @return {number}
* @example
*    =STDEVP(6.10,13.20,15.40)
*/
Stat["STDEVP"] = function(/* ... args*/) {
	let args = (arguments[0] instanceof Array) ? arguments[0] : Array.prototype.slice.call(arguments);
	let argLen = args.length;
	if(argLen <= 0) { return NaN; }
	if(argLen === 1) { return 0; }

	let mean = Stat["AVERAGE"](args);

	let sum = 0;
	for(let i = argLen; --i >= 0;) {
		let diff = args[i] - mean;
		sum += diff * diff;
	}
	return Math.sqrt(sum / argLen);
};

export default Stat;
export { Stat };
