/**
 * @namespace Text
 * @description
 *    {lang=en} Text manipulation functions
 *    {lang=ja} 文字列に関する関数
 *    {lang=zh} 文本控制函数
 *    {lang=ko} 텍스트 조작 함수
 *    {lang=ru} Text manipulation functions
 */
let Text = {};

let _toString = function(text) {
	if(!text && text !== 0) {
		return "";
	}

	return text + "";
};

/**
 * @function EXACT
 * @memberof Text
 * @description
 *   {lang=en} The EXACT function returns 1(TRUE) if text string 1 is same as text string 2. Otherwise returns 0(FALSE). Use operand = for a shorthand.
 *   {lang=ja} EXACT 関数は、2つの文字列を比較して同じ場合は 1(TRUE) を返します。異なる場合は 0(FALSE) を返します。オペランド [ = ] を使用します。
 *   {lang=zh} 如果文本字符串 1 与文本字符串 2 相同，则 EXACT 函数返回 1(TRUE)。否则返回 0(FALSE)。快速记录请使用操作符 =
 *   {lang=ko} EXACT 함수는 텍스트 문자열 1이 텍스트 문자열 2와 같으면 1(TRUE)을 반환합니다. 다르면 0(FALSE)을 반환합니다. 약칭으로 피연산자 =를 사용합니다.
 *   {lang=ru} The EXACT function returns 1(TRUE) if text string 1 is same as text string 2. Otherwise returns 0(FALSE). Use operand = for a shorthand.
 * @param  {string} text1
 *   {lang=en} The text string 1.
 *   {lang=ja} 文字列 1
 *   {lang=zh} 文本字符串 1。
 *   {lang=ko} 텍스트 문자열 1.
 *   {lang=ru} The text string 1.
 * @param  {string} text2
 *   {lang=en} The text string 2.
 *   {lang=ja} 文字列 2
 *   {lang=zh} 文本字符串 2。
 *   {lang=ko} 텍스트 문자열 2.
 *   {lang=ru} The text string 2.
 * @return {number}
 * @example
 *   =EXACT(CF_CURR,"USD")
 *   =CF_CURR="USD"
 */
Text["EXACT"] = function(text1, text2) {
	text1 = text1 + "";
	text2 = text2 + "";
	return (text1 === text2) ? 1 : 0;
};
/**
 * @function FIND
 * @memberof Text
 * @description
 *   {lang=en} The FIND function returns the position of a specified character or sub-string within a supplied text string.
 *             REMARK: The function is case-sensitive. The first character is counted as 0.
 *   {lang=ja} FIND 関数は、指定された文字列内で指定された文字の位置を返します。
 *             注記: この関数は、大文字と小文字を区別します。最初の文字は [0] としてカウントされます。
 *   {lang=zh} FIND 函数返回指定的字符或子字符串在所提供的文本字符串中的位置。
 *   {lang=ko} FIND 함수는 제공된 텍스트 문자열 내에서 지정된 문자 또는 하위 문자열의 위치를 반환합니다.
 *   {lang=ru} The FIND function returns the position of a specified character or sub-string within a supplied text string.
 *             注意: 该函数区分大小写。首个字符计为 0。
 * @param  {string} findText
 *   {lang=en} The character or sub-string that you wish to find.
 *   {lang=ja} 検索文字列
 *   {lang=zh} 您希望查找的字符或子字符串。
 *   {lang=ko} 찾으려는 문자나 하위 문자열.
 *   {lang=ru} The character or sub-string that you wish to find.
 * @param  {string} withinText
 *   {lang=en} The text string that is to be searched.
 *   {lang=ja} 検索する文字列
 *   {lang=zh} 要搜索的文本字符串。
 *   {lang=ko} 검색할 텍스트 문자열.
 *   {lang=ru} The text string that is to be searched.
 * @param  {number=} opt_startNum
 *   {lang=en} An optional argument that specifies the position in the within_text string, from which the search should begin.
 *   {lang=ja} 検索文字列内で検索を開始する位置を指定するオプショナル引数。
 *   {lang=zh} 可选参数，指定搜索开始于 within_text 字符串中的位置。
 *   {lang=ko} 검색을 시작해야 하는 within_text 문자열의 위치를 지정하는 선택적 인수.
 *   {lang=ru} An optional argument that specifies the position in the within_text string, from which the search should begin.
 * @return {number}
 * @example
 *    =FIND("r","surfthewave",5)
 */
Text["FIND"] = function(findText, withinText, opt_startNum) {
	findText = _toString(findText);
	if(!findText) { // Invalid findText
		return 0;
	}
	withinText = _toString(withinText);
	if(!withinText) { // Invalid base text
		return 0;
	}
	opt_startNum = +opt_startNum;
	if (!opt_startNum || opt_startNum < 0) {
		opt_startNum = 0;
	}
	let pos = withinText.indexOf(findText, opt_startNum);
	return pos;
};
/**
 * @function LEFT
 * @memberof Text
 * @description
 *   {lang=en} The LEFT function returns a specified number of characters from the start of a supplied text string.
 *             REMARK: If opt_numChars, (the number of returned characters), is more than the length of text string, return the whole string.
 *   {lang=ja} LEFT 関数は、指定された文字列の先頭から指定された数の文字を返します。
 *             注記: opt_numChars (返される文字数) がテキスト文字列の長さを超えている場合、文字列全体を返します。
 *   {lang=zh} LEFT 函数返回从所提供的文本字符串起始处的指定数量的字符。
 *             注意: 如果 opt_numChars, (返回的字符数), 多于文本字符串的长度，则返回整个字符串。
 *   {lang=hant} LEFT 函數返回從所提供的文本字符串起始處的指定數量的字符。
 *             注意: 如果 opt_numChars, (返回的字符數), 多於文本字符串的長度，則返回整個字符串。
 *   {lang=ko} LEFT 함수는 제공된 텍스트 문자열의 시작 부분에서 지정된 수의 문자를 반환합니다.
 *   {lang=ru} The LEFT function returns a specified number of characters from the start of a supplied text string.
 *   {lang=de} Die Funktion LINKS gibt eine bestimmte Anzahl von Zeichen ab dem Beginn einer bereitgestellten Textzeichenfolge zurück.
 *             REMARK: If opt_numChars, (the number of returned characters), is more than the length of text string, return the whole string.
 * @param {string} text
 *   {lang=en} The original text string.
 *   {lang=ja} 元の文字列
 *   {lang=zh} 初始文本字符串。
 *   {lang=hant} 初始文本字符串。
 *   {lang=ko} 원본 텍스트 문자열.
 *   {lang=ru} The original text string.
 *   {lang=de} Die ursprüngliche Textzeichenfolge.
 * @param {number} opt_numChars
 *   {lang=en} An optional argument that specifies the number of characters to be returned from the start of the supplied text.
 *   {lang=ja} 指定された文字列の先頭から指定された文字数を返すオプショナル引数です。
 *   {lang=zh} 可选参数，指定从所提供的文本起始处要返回的字符数。
 *   {lang=hant} 可選參數，指定從所提供的文本起始處要返回的字符數。
 *   {lang=ko} 제공된 텍스트의 시작 부분에서 반환할 문자 수를 지정하는 선택적 인수.
 *   {lang=ru} An optional argument that specifies the number of characters to be returned from the start of the supplied text.
 *   {lang=de} Ein optionales Argument, das die Anzahl der Zeichen spezifiziert, die vom Anfang des bereitgestellten Textes zurückgegeben werden sollen.
 * @return {string}
 * @example
 *    =LEFT("Surf the Wave",4)
 */
Text["LEFT"] = function(text, opt_numChars) {
	opt_numChars = +opt_numChars;
	if(!opt_numChars) {
		opt_numChars = 0;
	}
	if(opt_numChars < 0) {
		opt_numChars = 1;
	}
	return _toString(text).substr(0, opt_numChars);
};
/**
 * @function LENGTH
 * @memberof Text
 * @description
 *   {lang=en} The LENGTH function returns a number of characters of a text string.
 *   {lang=ja} LENGTH 関数は、文字列の文字数を返します。
 *   {lang=zh} LENGTH 函数返回文本字符串的一定数量的字符。
 *   {lang=hant} LENGTH 函數返回文本字符串的一定數量的字符。
 *   {lang=ko} LENGTH 함수는 텍스트 문자열의 문자 수를 반환합니다.
 *   {lang=ru} The LENGTH function returns a number of characters of a text string.
 *   {lang=de} Die Funktion LÄNGE gibt eine Anzahl von Zeichen einer Textzeichenfolge zurück.
 * @param  {string} text
 *   {lang=en} Text string.
 *   {lang=ja} 文字列
 *   {lang=zh} 文本字符串。
 *   {lang=hant} 文本字符串。
 *   {lang=ko} 텍스트 문자열.
 *   {lang=ru} Text string.
 *   {lang=de} Textzeichenfolge.
 * @return {number}
 * @example
 *    =LENGTH("Markets")
 */
Text["LENGTH"] = function(text) {
	return _toString(text).length;
};
/**
 * @function LEN
 * @memberof Text
 * @description
 *   {lang=en} The LEN function returns a number of characters of a text string.
 *   {lang=ja} LEN 関数は、文字列の文字数を返します。
 *   {lang=zh} LEN 函数返回文本字符串的一定数量的字符。
 *   {lang=hant} LEN 函數返回文本字符串的一定數量的字符。
 *   {lang=ko} LEN 함수는 텍스트 문자열의 문자 수를 반환합니다.
 *   {lang=ru} The LEN function returns a number of characters of a text string.
 *   {lang=de} Die Funktion LEN gibt eine Anzahl von Zeichen einer Textzeichenfolge zurück.
 * @param  {string} text
 *   {lang=en} Text string.
 *   {lang=ja} 文字列
 *   {lang=zh} 文本字符串。
 *   {lang=hant} 文本字符串。
 *   {lang=ko} 텍스트 문자열.
 *   {lang=ru} Text string.
 *   {lang=de} Textzeichenfolge.
 * @return {number}
 * @example
 *    =LEN("Markets")
 */
Text["LEN"] = Text["LENGTH"];

/**
 * @function LOWER
 * @memberof Text
 * @description
 *   {lang=en} The LOWER function converts all uppercase letters in text string to lowercase.
 *   {lang=ja} LOWER 関数は、文字列のすべての文字を小文字に変換します。
 *   {lang=zh} LOWER 函数将文本字符串中的全部大写字母转换成小写。
 *   {lang=hant} LOWER 函數將文本字符串中的全部大寫字母轉換成小寫。
 *   {lang=ko} LOWER 함수는 텍스트 문자열의 모든 대문자를 소문자로 변환합니다.
 *   {lang=ru} The LOWER function converts all uppercase letters in text string to lowercase.
 *   {lang=de} Die Funktion KLEINBUCHSTABEN konvertiert alle Großbuchstaben in der Textzeichenfolge in Kleinbuchstaben.
 * @param  {string} text
 *   {lang=en} Text string.
 *   {lang=ja} 文字列
 *   {lang=zh} 文本字符串。
 *   {lang=hant} 文本字符串。
 *   {lang=ko} 텍스트 문자열.
 *   {lang=ru} Text string.
 *   {lang=de} Textzeichenfolge.
 * @return {string}
 * @example
 *    =LOWER("HELLO WORLD")
 */
Text["LOWER"] = function(text) {
	return _toString(text).toLowerCase();
};

/**
 * @function MID
 * @memberof Text
 * @description
 *   {lang=en} The MID function returns a specified number of characters from the middle of a supplied text string.
 *             REMARK: The first character is counted as 0.
 *   {lang=ja} MID 関数は、文字列の指定された位置から指定された数の文字を返します。
 *             注記: 最初の文字は [0] としてカウントされます。
 *   {lang=zh} MID 函数从所提供的文本字符串的中间返回指定数量的字符。
  *             注意: 首字符计为 0。
 *   {lang=hant} MID 函數從所提供的文本字符串的中間返回指定數量的字符。
 *             注意: 首字符計為 0。
 *   {lang=ko} MID 함수는 제공된 텍스트 문자열의 중간 부분에서 지정된 수의 문자를 반환합니다.
 *   {lang=ru} The MID function returns a specified number of characters from the middle of a supplied text string.
 *             REMARK: The first character is counted as 0.
 *   {lang=de} Die Funktion MITTE gibt eine bestimmte Anzahl von Zeichen aus der Mitte einer bereitgestellten Textzeichenfolge zurück.
 *             REMARK: The first character is counted as 0.
 * @param  {string} text
 *   {lang=en} The original text string.
 *   {lang=ja} 元の文字列
 *   {lang=zh} 初始文本字符串。
 *   {lang=hant} 初始文本字符串。
 *   {lang=ko} 원본 텍스트 문자열.
 *   {lang=ru} The original text string.
 *   {lang=de} Die ursprüngliche Textzeichenfolge.
 * @param  {number} startNum
 *   {lang=en} An integer that specifies the position of the first character that you want to be returned.
 *   {lang=ja} 返される最初の文字の位置を指定する整数。
 *   {lang=zh} 一个整数，指定您想要返回的首字符的位置。
 *   {lang=hant} 一個整數，指定您想要返回的首字符的位置。
 *   {lang=ko} 반환할 첫 번째 문자의 위치를 지정하는 정수입니다.
 *   {lang=ru} An integer that specifies the position of the first character that you want to be returned.
 *   {lang=de} Eine Ganzzahl, die die Position des ersten Zeichens angibt, das zurückgegeben werden soll.
 * @param  {number} numChars
 *   {lang=en} An integer that specifies the number of characters (beginning with start_num), to be returned from the supplied text.
 *   {lang=ja} 指定された文字列から返される文字数を指定する整数。
 *   {lang=zh} 一个整数，指定从所提供的文本要返回的字符数 (以 start_num 开头)。
 *   {lang=hant} 一個整數，指定從所提供的文本要返回的字符數 (以 start_num 開頭)。
 *   {lang=ko} 제공된 텍스트에서 반환할 문자 수(start_num으로 시작하는)를 지정하는 정수입니다.
 *   {lang=ru} An integer that specifies the number of characters (beginning with start_num), to be returned from the supplied text.
 *   {lang=de} Eine Ganzzahl, die die Anzahl der Zeichen (beginnend mit start_num) spezifiziert, die vom bereitgestellten Text zurückgegeben werden soll.
 * @return {string}
 * @example
 *    =MID("antidisestablishmentarianism",7,13)
 */
Text["MID"] = function(text, startNum, numChars) {
	return _toString(text).substr(startNum, numChars);
};
/**
 * @function REPLACE
 * @memberof Text
 * @description
 *   {lang=en} The REPLACE function replaces all or part of a text string with another string.
 *   {lang=ja} REPLACE 関数は、指定された数の文字を新しい文字で置き換えた文字列を返します。
 *   {lang=zh} REPLACE 函数将全部或部分文本字符串替代为另一字符串。
 *   {lang=hant} REPLACE 函數將全部或部分文本字符串替代為另一字符串。
 *   {lang=ko} REPLACE 함수는 텍스트 문자열 전부 또는 일부를 다른 문자열로 바꿉니다.
 *   {lang=ru} The REPLACE function replaces all or part of a text string with another string.
 *   {lang=de} Die Funktion ERSETZEN tauscht eine Textzeichenfolge ganz oder teilweise mit einer anderen Zeichenfolge aus.
 * @param  {string} oldText
 *   {lang=en} The original text string, that you want to replace a part of.
 *   {lang=ja} 置き換えをする元の文字列。
 *   {lang=zh} 您想要部分替代的初始文本字符串。
 *   {lang=hant} 您想要部分替代的初始文本字符串。
 *   {lang=ko} 일부를 바꾸려는 원본 텍스트 문자열.
 *   {lang=ru} The original text string, that you want to replace a part of.
 *   {lang=de} Die ursprüngliche Textzeichenfolge, von der Sie einen Teil ersetzen möchten.
 * @param  {number} numChars
 *   {lang=en} The number of characters to replace.
 *   {lang=ja} 置き換えをする文字数。
 *   {lang=zh} 要替代的字符数。
 *   {lang=hant} 要替代的字符數。
 *   {lang=ko} 바꿀 문자 수.
 *   {lang=ru} The number of characters to replace.
 *   {lang=de} Die Anzahl der zu ersetzenden Zeichen.
 * @param  {number} startNum
 *   {lang=en} The position, within old_text, of the first character that you want to replace.
 *   {lang=ja} 置き換えをする文字列中で、置き換えを開始する文字の位置。
 *   {lang=zh} 您想要替代的旧文本中首字符的位置。
 *   {lang=hant} 您想要替代的舊文本中首字符的位置。
 *   {lang=ko} old_text 내에서 바꾸려는 첫 번째 문자의 위치.
 *   {lang=ru} The position, within old_text, of the first character that you want to replace.
 *   {lang=de} Die Position, innerhalb von old_text, des ersten Zeichens, das Sie ersetzen möchten.
 * @param  {string} newText
 *   {lang=en} The replacement text.
 *   {lang=ja} 置き換える文字列。
 *   {lang=zh} 替代文本。
 *   {lang=hant} 替代文本。
 *   {lang=ko} 대체 텍스트.
 *   {lang=ru} The replacement text.
 *   {lang=de} Der Ersatztext.
 * @return {string}
 * @example
 *    =REPLACE("Rollerblading",7,6,"skates")
 */
Text["REPLACE"] = function(oldText, numChars, startNum, newText) {
	newText = newText + "";
	oldText = _toString(oldText);
	let arrOldText = oldText.split("");
	arrOldText.splice(startNum, numChars, newText);
	return arrOldText.join("");
};
/**
 * @function RIGHT
 * @memberof Text
 * @description
 *   {lang=en} The RIGHT function returns a specified number of characters from the end of a supplied text string.
 *   {lang=ja} RIGHT 関数は、指定された文字列の末尾から指定された文字数を返します。
 *   {lang=zh} RIGHT 函数从所提供的文本字符串末尾返回指定数量的字符。
 *   {lang=hant} RIGHT 函數從所提供的文本字符串末尾返回指定數量的字符。
 *   {lang=ko} RIGHT 함수는 제공된 텍스트 문자열의 끝부분에서 지정된 수의 문자를 반환합니다.
 *   {lang=ru} The RIGHT function returns a specified number of characters from the end of a supplied text string.
 *   {lang=de} Die Funktion RECHTS gibt eine bestimmte Anzahl von Zeichen vom Ende einer bereitgestellten Textzeichenfolge zurück.
 * @param  {string} text
 *   {lang=en} The original text string.
 *   {lang=ja} 元の文字列。
 *   {lang=zh} 初始文本字符串。
 *   {lang=hant} 初始文本字符串。
 *   {lang=ko} 원본 텍스트 문자열.
 *   {lang=ru} The original text string.
 *   {lang=de} Die ursprüngliche Textzeichenfolge.
 * @param  {number} opt_numChars
 *   {lang=en} An optional argument that specifies the number of characters to be returned from the end of the supplied text.
 *   {lang=ja} 指定された文字列の末尾から指定された文字数を返すオプショナル引数です。
 *   {lang=zh} 可选参数，指定从所提供文本的末尾返回的字符数。
 *   {lang=hant} 可選參數，指定從所提供文本的末尾返回的字符數。
 *   {lang=ko} 제공된 텍스트의 끝부분에서 반환할 문자 수를 지정하는 선택적 인수.
 *   {lang=ru} An optional argument that specifies the number of characters to be returned from the end of the supplied text.
 *   {lang=de} Ein optionales Argument, das die Anzahl der Zeichen spezifiziert, die vom Ende des bereitgestellten Textes zurückgegeben werden sollen.
 * @return {string}
 * @example
 *    =RIGHT("Portishead",4)
 */
Text["RIGHT"] = function(text, opt_numChars) {
	text = _toString(text);
	opt_numChars = +opt_numChars;
	if(!opt_numChars) {
		opt_numChars = 0;
	}
	if(opt_numChars < 0) {
		opt_numChars = 1;
	}
	let startNum = text.length - opt_numChars;
	return text.substr(startNum, opt_numChars);
};
/**
 * @function TRIM
 * @memberof Text
 * @description
 *   {lang=en} The TRIM function removes leading, trailing, and consecutive spaces from a string.
 *   {lang=ja} TRIM 関数は、指定された文字列の余分な空白を削除します。
 *   {lang=zh} TRIM 函数从一个字符串中删除开头、结尾与连续的空格。
 *   {lang=hant} TRIM 函數從一個字符串中刪除開頭、結尾與連續的空格。
 *   {lang=ko} TRIM 함수는 문자열에서 앞과 뒤, 연속된 공백을 제거합니다.
 *   {lang=ru} The TRIM function removes leading, trailing, and consecutive spaces from a string.
 *   {lang=de} Die Funktion KÜRZEN entfernt führende, nachfolgende und aufeinanderfolgende Leerzeichen aus einer Zeichenfolge.
 * @param  {string} text
 *   {lang=en} Text string.
 *   {lang=ja} 文字列
 *   {lang=zh} 文本字符串。
 *   {lang=hant} 文本字符串。
 *   {lang=ko} 텍스트 문자열.
 *   {lang=ru} Text string.
 *   {lang=de} Textzeichenfolge.
 * @return {string}
 * @example
 *    =TRIM(" Information Management Systems ")
 */
Text["TRIM"] = function(text) {
	return _toString(text).trim();
};
/**
 * @function UPPER
 * @memberof Text
 * @description
 *   {lang=en} The UPPER function converts all lowercase letters in text string to uppercase.
 *   {lang=ja} UPPER 関数は、文字列のすべての小文字を大文字に変換します。
 *   {lang=zh} UPPER 函数将文本字符串中的全部小写字母转换成大写。
 *   {lang=hant} UPPER 函數將文本字符串中的全部小寫字母轉換成大寫。
 *   {lang=ko} UPPER 함수는 텍스트 문자열의 모든 소문자를 대문자로 변환합니다.
 *   {lang=ru} The UPPER function converts all lowercase letters in text string to uppercase.
 *   {lang=de} Die Funktion GROSSBUCHSTABEN konvertiert alle Kleinbuchstaben in der Textzeichenfolge in Großbuchstaben.
 * @param  {string} text
 *   {lang=en} Text string.
 *   {lang=ja} 文字列
 *   {lang=zh} 文本字符串。
 *   {lang=hant} 文本字符串。
 *   {lang=ko} 텍스트 문자열.
 *   {lang=ru} Text string.
 *   {lang=de} Textzeichenfolge.
 * @return {string}
 * @example
 *    =UPPER("Hello World")
 */
Text["UPPER"] = function(text) {
	return _toString(text).toUpperCase();
};
/**
 * @function VALUE
 * @memberof Text
 * @description
 *   {lang=en} The VALUE function converts a string to a number.
 *   {lang=ja} VALUE 関数は、文字列を数値に変換します。
 *   {lang=zh} VALUE 函数将字符串转换成数字。
 *   {lang=hant} VALUE 函數將字符串轉換成數字。
 *   {lang=ko} VALUE 함수는 문자열을 숫자로 변환합니다.
 *   {lang=ru} The VALUE function converts a string to a number.
 *   {lang=de} Die Funktion WERT konvertiert eine Zeichenfolge in eine Zahl.
 * @param {string} text
 *   {lang=en} text
 *   {lang=ja} 文字列
 *   {lang=zh} 文本
 *   {lang=hant} 文本
 *   {lang=ko} 텍스트
 *   {lang=ru} text
 *   {lang=de} Text
 * @return {number}
 * @example
 *    =VALUE("987.1")
 */
Text["VALUE"] = function(text) {
	return +text;
};
/**
 * @function CHAR
 * @memberof Text
 * @description
 *   {lang=en} The CHAR function converts a number to an ASCII character.
 *   {lang=ja} CHAR 関数は、数値を ASCII 文字に変換します。
 *   {lang=zh} CHAR 函数将数字转换成 ASCII 字符。
 *   {lang=hant} CHAR 函數將數字轉換成 ASCII 字符。
 *   {lang=ko} CHAR 함수는 숫자를 ASCII 문자로 변환합니다.
 *   {lang=ru} The CHAR function converts a number to an ASCII character.
 *   {lang=de} Die Funktion CHAR konvertiert eine Zahl in ein ASCII-Zeichen.
 * @param {number} code
 *   {lang=en} character code
 *   {lang=ja} 文字コード
 *   {lang=zh} 字符代码
 *   {lang=hant} 字符代碼
 *   {lang=ko} 문자 코드
 *   {lang=ru} character code
 *   {lang=de} Zeichencode
 * @return {string}
 * @example
 *    =CHAR(1)
 */
Text["CHAR"] = function(code) {
	let num = +code;
	return String.fromCharCode(num);
};
/**
 * @function CLEAN
 * @memberof Text
 * @description
 *   {lang=en} The CLEAN function removes non-ASCII characters, (characters not in the range 32-127 inclusive), from the string.
 *   {lang=ja} CLEAN 関数は、ASCII 以外の文字 (32-127 の範囲にない文字) を文字列から削除します。
 *   {lang=zh} CLEAN 函数从字符串中删除非 ASCII 字符 (不在 32-127 范围中的字符)。
 *   {lang=hant} CLEAN 函數從字符串中刪除非 ASCII 字符 (不在 32-127 範圍中的字符)。
 *   {lang=ko} CLEAN 함수는 문자열에서 비 ASCII 문자(32-127 범위에 없는 문자)를 제거합니다.
 *   {lang=ru} The CLEAN function removes non-ASCII characters, (characters not in the range 32-127 inclusive), from the string.
 *   {lang=de} Die Funktion SÄUBERN entfernt Nicht-ASCII-Zeichen (Zeichen, die nicht im Bereich 32-127 einschließlich liegen) aus der Zeichenfolge.
 * @param {string} text
 *   {lang=en} Text
 *   {lang=ja} 文字列
 *   {lang=zh} 文本
 *   {lang=hant} 文本
 *   {lang=ko} 텍스트
 *   {lang=ru} Text
 *   {lang=de} Text
 * @return {string}
 * @example
 *    =CLEAN("Kobérña")
 */
Text["CLEAN"] = function(text) {
	let arrText = _toString(text).split("");
	for(let i = arrText.length; --i >= 0;) {
		let ascii = arrText[i].charCodeAt(0);
		if(ascii < 32 || ascii > 127) {
			arrText[i] = "";
		}
	}
	return arrText.join("");
};
/**
 * @function CODE
 * @memberof Text
 * @description
 *   {lang=en} The CODE function returns the ASCII code that corresponds to the first character in a string.
 *   {lang=ja} CODE 関数は、文字列の先頭文字の ASCII コードを返します。
 *   {lang=zh} CODE 函数返回对应字符串中首字符的 ASCII 代码。
 *   {lang=hant} CODE 函數返回對應字符串中首字符的 ASCII 代碼。
 *   {lang=ko} CODE 함수는 문자열의 첫 번째 문자에 해당하는 ASCII 코드를 반환합니다.
 *   {lang=ru} The CODE function returns the ASCII code that corresponds to the first character in a string.
 *   {lang=de} Die Funktion CODE gibt den ASCII-Code zurück, welcher dem ersten Zeichen in einer Zeichenfolge entspricht.
 * @param  {string} text
 *   {lang=en} Text
 *   {lang=ja} 文字列
 *   {lang=zh} 文本
 *   {lang=hant} 文本
 *   {lang=ko} 텍스트
 *   {lang=ru} Text
 *   {lang=de} Text
 * @return {number}
 * @example
 *    =CODE("A")
 */
Text["CODE"] = function(text) {
	return _toString(text).charCodeAt(0);
};
/**
 * @function PROPER
 * @memberof Text
 * @description
 *   {lang=en} The PROPER function converts the first letter of all words in sentence to uppercase, and converts all remaining letters to lowercase.
 *   {lang=ja} ROPER 関数は、文中の各単語の先頭文字を大文字に変換し、残りの文字をすべて小文字に変換します。
 *   {lang=zh} PROPER 函数将句子中全部单词的首字母转换成大写，其他字母全部保持小写。
 *   {lang=hant} PROPER 函數將句子中全部單詞的首字母轉換成大寫，其他字母全部保持小寫。
 *   {lang=ko} PROPER 함수는 문장의 모든 단어의 첫 글자를 대문자로 변환하고 나머지 문자를 모두 소문자로 변환합니다.
 *   {lang=ru} The PROPER function converts the first letter of all words in sentence to uppercase, and converts all remaining letters to lowercase.
 *   {lang=de} Die Funktion RICHTIG konvertiert den ersten Buchstaben aller Wörter im Satz in Großbuchstaben und alle verbleibenden Buchstaben in Kleinbuchstaben.
 * @param  {string} text
 *   {lang=en} Text
 *   {lang=ja} 文字列
 *   {lang=zh} 文本
 *   {lang=hant} 文本
 *   {lang=ko} 텍스트
 *   {lang=ru} Text
 *   {lang=de} Text
 * @return {string}
 * @example
 *    =PROPER("A STRANGE BLACK CAT ON YOUR PORCH BRINGS PROSPERITY.")
 */
Text["PROPER"] = function(text) {
	text = _toString(text);
	let arrTokens = text.split(/[\W_]/g);
	let arrSyms = text.match(/[\W_]/g);
	let arrLen = arrTokens.length;
	let result = "";
	for(let i = 0; i < arrLen; i++) {
		let token = arrTokens[i];
		let len = token.length;
		if(len === 1) {
			result += token.toUpperCase();
		} else if(len !== 0) {
			result += token.charAt(0).toUpperCase() + token.substr(1).toLowerCase();
		}

		let sym = arrSyms[i];
		if(sym) {
			result += sym;
		}
	}
	return result;
};
/**
 * @function REPEAT
 * @memberof Text
 * @description
 *   {lang=en} The REPEAT function duplicates the text string in specific times.
 *   {lang=ja} REPEAT 関数は、特定の時刻にテキスト文字を複製します。
 *   {lang=zh} REPEAT 函数将文本字符串重复指定次数。
 *   {lang=hant} REPEAT 函數將文本字符串重複指定次數。
 *   {lang=ko} REPEAT 함수는 텍스트 문자열을 지정된 횟수만큼 복제합니다.
 *   {lang=ru} The REPEAT function duplicates the text string in specific times.
 *   {lang=de} Die Funktion WIEDERHOLEN dupliziert die Textzeichenfolge zu bestimmten Zeiten.
 * @param  {string} text
 *   {lang=en} Text
 *   {lang=ja} 文字列
 *   {lang=zh} 文本
 *   {lang=hant} 文本
 *   {lang=ko} 텍스트
 *   {lang=ru} Text
 *   {lang=de} Text
 * @param  {number} numRep
 *   {lang=en} Number of times
 *   {lang=ja} 回数
 *   {lang=zh} 次数
 *   {lang=hant} 次數
 *   {lang=ko} 횟수
 *   {lang=ru} Number of times
 *   {lang=de} Mehrfach
 * @return {string}
 * @example
 *    =REPEAT("Ho! ",3)
 */
Text["REPEAT"] = function(text, numRep) {
	numRep = +numRep;
	if(!numRep || numRep < 0) {
		return "";
	}

	text = _toString(text);
	if(text === "") {
		return "";
	}

	let newText = text;
	for(let i = 1; i < numRep; i++) {
		newText += text;
	}
	return newText;
};
/**
 * @function STRING
 * @memberof Text
 * @description
 *   {lang=en} The STRING function rounds the number to a specified precision, and then converts it into a string.
 *   {lang=ja} STRING 関数は、数値を指定された少数に処理した後、文字列に返します。
 *   {lang=zh} STRING 函数将数字四舍五入到指定精度，然后将其转换成字符串。
 *   {lang=hant} STRING 函數將數字四捨五入到指定精度，然後將其轉換成字符串。
 *   {lang=ko} STRING 함수는 숫자를 지정된 정밀도로 반올림한 다음 문자열로 변환합니다.
 *   {lang=ru} The STRING function rounds the number to a specified precision, and then converts it into a string.
 *   {lang=de} Die Funktion STRING rundet die Zahl auf eine bestimmte Genauigkeit und konvertiert sie dann in eine Zeichenfolge.
 * @param {number} num1
 *   {lang=en} Number 1
 *   {lang=ja} 数値 1
 *   {lang=zh} 数字 1
 *   {lang=hant} 數字 1
 *   {lang=ko} 숫자 1
 *   {lang=ru} Number 1
 *   {lang=de} Nummer 1
 * @param {number} num2
 *   {lang=en} Number 2
 *   {lang=ja} 数値 2
 *   {lang=zh} 数字 2
 *   {lang=hant} 數字 2
 *   {lang=ko} 숫자 2
 *   {lang=ru} Number 2
 *   {lang=de} Nummer 2
 * @return {string}
 * @example
 *    =STRING(12.129,2)
 */
Text["STRING"] = function(num1, num2) {
	num1 = +num1;
	num2 = +num2;
	if(!num1) {
		return "";
	}
	if(!num2 || num2 < 0) {
		num2 = 0;
	} else if(num2 > 9) {
		num2 = 9;
	}

	return (num1).toFixed(num2);
};

export default Text;
export {
	Text
};
