/** @description CollectionDict stores a collection (Array) of any value using a text (string) as a key for accessing the collection.
* @constructor
*/
let CollectionDict = function() {
	this._dict = {};
};
/** @type {!Object.<string, Array>}
* @private
*/
CollectionDict.prototype._dict;
/** @type {number}
* @private
*/
CollectionDict.prototype._count = 0;

/** @public
* @param {string} key
* @param {*} item Null or undefined item is not allowed
* @return {boolean} Return true if the item has been added
*/
CollectionDict.prototype.addItem = function(key, item) {
	if(key && item != null) {
		let ary = this._dict[key];
		if(ary) {
			ary.push(item);
		} else {
			this._dict[key] = [item];
		}

		++this._count;
		return true;
	}
	return false;
};
/** @public
* @param {string} key
* @param {Array} items The given item list should not contain null or undefined
* @return {boolean} Return true if an item has been added
*/
CollectionDict.prototype.addItems = function(key, items) {
	if(key) {
		if(!Array.isArray(items)) {
			return this.addItem(key, items);
		}
		let len = items ? items.length : 0;
		if(len) {
			let ary = this._dict[key];
			if(!ary) {
				ary = this._dict[key] = [];
			}

			// Fastest way to add multiple items at once to an array. WARNING: There is a 10,000 item limit.
			Array.prototype.push.apply(ary, items);
			this._count += len;
			return true;
		}
	}
	return false;
};

/** @public
* @param {string} key
* @param {*} item
* @return {boolean} Return true if the item has been removed
*/
CollectionDict.prototype.removeItem = function(key, item) {
	let ary = this._dict[key];
	if(ary) {
		let at = ary.indexOf(item);
		if(at >= 0) {
			if(ary.length > 1) {
				ary.splice(at, 1);
			} else {
				delete this._dict[key];
			}
			--this._count;
			return true;
		}
	}
	return false;
};
/** @public
* @param {string} key
* @return {boolean} Return true if any item has been removed
*/
CollectionDict.prototype.removeItemsByKey = function(key) {
	let ary = this._dict[key];
	if(ary) {
		delete this._dict[key];
		this._count -= ary.length;
		return true;
	}
	return false;
};
/** @public
* @return {boolean} Return true if there is any change
*/
CollectionDict.prototype.removeAllItems = function() {
	if(this._count) {
		this._dict = {};
		this._count = 0;
		return true;
	}
	return false;
};

/** @public
* @param {string} key
* @return {Array} Return null if no item exist
*/
CollectionDict.prototype.getItems = function(key) {
	return this._dict[key] || null;
};
/** @public
* @return {Array} Return null if no item exist
*/
CollectionDict.prototype.getAllItems = function() {
	if(this._count > 0) {
		let ary = [];
		let dict = this._dict;
		for(let key in dict) {
			// Fastest way to add multiple items at once to an array. WARNING: There is a 10,000 item limit.
			Array.prototype.push.apply(ary, dict[key]);
		}
		return ary;
	}
	return null;
};
/** @public
* @param {string=} key
* @return {number}
*/
CollectionDict.prototype.getItemCount = function(key) {
	if(key) {
		let items = this._dict[key] || null;
		if(items) {
			return items.length;
		}
		return 0;
	}
	return this._count;
};
/** @public
* @return {!Array.<string>}
*/
CollectionDict.prototype.getAllKeys = function() {
	return Object.keys(this._dict);
};

export default CollectionDict;
export { CollectionDict };
